/* $Modified: Wed Nov  3 15:27:38 2010 by uwer $ */
// vim: set ft=cpp:
#ifndef CURRENT_H_
#define CURRENT_H_

#include <complex>

#include "mytypes.h"

#define NMAX 32

class Current {
 public:
  Current(const int ngluon_, DOUBLE (*k_)[4], int *hel_ ) 
    : hel(hel_), k(k_), ngluon(ngluon_), zero(0.), one(1.), two(2.), I(zero,one) {
    DOUBLE ref_[4] = {DOUBLE(3.),DOUBLE(2.),DOUBLE(2.),DOUBLE(1.)};
    setRefMom(ref_);
  }
    
    int *hel;
    DOUBLE (*k)[4];
        
    void momentumUpdate();
    void helicityUpdate();
    void rescaleMomenta(const DOUBLE x);

    void printMomenta(const int N);
    void printCurrents(const int N);
    void fillPartialCurrents();
    void setRefMom(const DOUBLE* ref_);
 
    
    std::complex<DOUBLE> gluonAmplitude();
    DOUBLE checkAmplitude(); 
   
    std::complex<DOUBLE> partialGluonAmplitude(const int i, const int j,
					       const int hel_i,const int hel_j,
					       const std::complex<DOUBLE> l_i[4],
					       const std::complex<DOUBLE> l_j[4]);
    
    std::complex<DOUBLE> partialGluonScalarAmplitude(const int i, const int j,
						     const std::complex<DOUBLE> l_i[4],
						     const std::complex<DOUBLE> l_j[4]);
    
    int checkHelsum(const int i, const int j,
		    const int hel_i, const int hel_j);
 
   
    inline DOUBLE* posX(const int i, const int j){ 
      return KKK[(ngluon+j-i) % ngluon][i % ngluon]; 
    }

    inline DOUBLE posS(const int i, const int j){ 
      return S_ij[(ngluon+j-i) % ngluon][i % ngluon]; 
    }

    static std::string helicity2string(const int n, int hel[]);
    static std::string int2string(int number);
    static void sethel(unsigned int n, int htype, int *helicity);
    static void sethel(unsigned int n, std::string hstr, int *helicity);

 private:
    const int ngluon;
    int Helsum[NMAX][NMAX];
    const DOUBLE zero, one, two;
    const std::complex<DOUBLE> I;
    DOUBLE refMom[4];
    DOUBLE KKK[NMAX][NMAX][4];
    DOUBLE S_ij[NMAX][NMAX];
    std::complex<DOUBLE> JJJ[NMAX][NMAX][4];

    template<typename T> void get1PtCurrents(const int helicity, 
					     std::complex<DOUBLE> JJ[4],
					     const T KK[4]);

    // vertices
    void vertex3GGG(const int len, const int k, const int m, const int l);
    void vertex4GGGG(const int len, const int k, const int  m, 
		     const int j, const int l);
    
    template<typename T>
      inline void vertex3GGG_c(const int dest, const int k, const int y,
			       const int l, 
			       T KK[][4], std::complex<DOUBLE> JJ[][4]);
    void vertex4GGGG_c(const int len, const int k, const int x, const int y,
		       const int l, std::complex<DOUBLE> JJ[][4]);
    
    
    template<typename T>
      inline void vertex3GSS_c(const int dest, const int k, const int x,
			       const  int l, T KK[][4],
			       std::complex<DOUBLE> JJ[]);
    void vertex4GGSS_c(const int len, const int k, const int x, 
		       const int y, const int l,
		       std::complex<DOUBLE> JJ[]);

   template<typename T>
     inline T* posX(const int i, const int j,
		    T Matrix[NMAX][NMAX][4]){
     return &Matrix[(ngluon+j-i) % ngluon][i % ngluon][0];
   }
 
  
   inline int posX(const int i, const int j, int Matrix[NMAX][NMAX]){
      return Matrix[(ngluon+j-i) % ngluon][i % ngluon];
    }

   void cyclicMomSum();
   void cyclicSum( int Matrix[NMAX][NMAX] );

};

#endif
