/* $Modified: Thu Jul  8 19:22:50 2010 by uwer $ */
#include "LoopIntegrals.h"
#include <iostream>

using namespace std;

/*
 * For the scalar one-loop integrals we use....
 *
 * Refs: [1] van Oldenborgh: FF: A Package To Evaluate One Loop Feynman Diagrams
 *           Comput.Phys.Commun.66:1-15,1991
 *       [2] R.Keith Ellis, Giulia Zanderighi,  "Scalar one-loop integrals for QCD",
 *           JHEP 0802:002,2008.
 *
 */

void info(string s){
  cout << "# INTEGRALS: " << s << endl;
}
#ifdef QCDLOOP
extern "C" {
  void qlinit_();
  complex<double> qli4_(const double &p1,const double &p2,const double &p3,const double &p4,
			const double &s,const double& t,
			const double& m1,const double &m2,const double &m3,const double &m4,
			const double &mu2,const int &ep);
  
  complex<double> qli3_(const double &p1,const double &p2,const double &p3,const double& m1,
			const double &m2,const double&m3,const double &mu2,const int &ep);
  
  complex<double> qli2_(const double & p1,const double & m1,
			const double & m2, const double & mu2,
			const int & ep);
}
#endif

class InitLoopIntegrals {
public:
  InitLoopIntegrals(){
#ifdef QCDLOOP
    qlinit_();
    info("FF [1] and QCDLoop [2] are used to calculate the");
    info("scalar one-loop integrals");
    info("[1] van Oldenborgh: FF: A Package To Evaluate One Loop Feynman Diagrams");
    info("    Comput.Phys.Commun.66:1-15,1991");
    info("[2] R.Keith Ellis, Giulia Zanderighi,  Scalar one-loop integrals for QCD, ");
    info("    JHEP 0802:002,2008");
#else
    info("Note: All scalar loop integrals are set to 1."); 
    info("Recompile the package with option QCDLOOP"); 
    info("if QCDLOOP/FF should be used."); 
#endif
  }
};

static InitLoopIntegrals dummy;

complex<DOUBLE> I4(DOUBLE s, DOUBLE t,DOUBLE M1,DOUBLE M2,
		   DOUBLE M3,DOUBLE M4,
		   DOUBLE m1,DOUBLE m2,DOUBLE m3,DOUBLE m4,
		   DOUBLE MUR2, int eps){
#ifdef QCDLOOP
#ifdef EXTENDED_PREC
  return( qli4_(to_double(M1),to_double(M2),to_double(M3),to_double(M4),
	        to_double(s),to_double(t),to_double(m1),to_double(m2),
		to_double(m3),to_double(m4),
		to_double(MUR2),eps) );
#else
  return( qli4_(M1,M2,M3,M4,s,t,m1,m2,m3,m4,MUR2,eps) );
#endif
#else
  return (DOUBLE(1.));
#endif
}

complex<DOUBLE> I3(DOUBLE M1,DOUBLE M2,DOUBLE M3,DOUBLE m1,
		   DOUBLE m2,DOUBLE m3,DOUBLE MUR2, int eps){
#ifdef QCDLOOP
#ifdef EXTENDED_PREC
  return( qli3_(to_double(M1),to_double(M2),to_double(M3),to_double(m1),
		to_double(m2),to_double(m3),to_double(MUR2),eps) );
#else
  return( qli3_(M1,M2,M3,m1,m2,m3,MUR2,eps) );
#endif
#else
  return(DOUBLE(1.));
#endif
}

complex<DOUBLE> I2(DOUBLE s, DOUBLE m1, DOUBLE m2, DOUBLE  MUR2, int eps){
#ifdef QCDLOOP
#ifdef EXTENDED_PREC
  return qli2_(to_double(s),to_double(m1),to_double(m2),to_double(MUR2),eps);
#else
  return qli2_(s,m1,m2,MUR2,eps);
#endif
#else
  return (DOUBLE(1.));
#endif
}
