// $Modified: Tue Oct 12 13:00:46 2010 by uwer $
// vim: set foldmethod=marker:
#include "NGluon.h"
#include <cfloat>

#define DOTP(X,Y) ( X[0]*Y[0] - ( X[1]*Y[1] + X[2]*Y[2] + X[3]*Y[3] ) )

using namespace std;

//{{{ top level and initalisation functions


template<typename T>
static DOUBLE compare(const T & a, const T& b){
  return( abs( (a-b)/ (a+b)) * DOUBLE(2.) );
}

void NGluon::setScaleTest(bool scaleTest_){
  scaleTest = scaleTest_;
  if(scaleTest)
    info("Scaling test to check accuracy is used");
  else {
    info("No scaling test to check accuracy is used");
  }
}

void NGluon::setVerbosity(VERBOSITY verbosity_){

  if (verbosity_ == QUIET){
    verbosity = FULL;
    info("Verbosity is set to NGluon::QUIET");
    verbosity = QUIET;
    return;
  }
  
  if (verbosity_ == FULL){
    verbosity = FULL;
    info("output on screen is turned on!");
    return;
  }

  info("output on screen not selected correctly");
  info("  -> use either NGluon::FULL or NGluon::QUIET to switch on or off"); 

}

void NGluon::info(string s){
  if (verbosity == FULL)
    cout << "# ONELOOP: " << s << endl;
}

void NGluon::info(string s, DOUBLE tmp){
  if (verbosity == FULL)
    cout << "# ONELOOP: " << s << " " << tmp << endl;
}

void NGluon::info(string s, int tmp){
  if (verbosity == FULL)
    cout << "# ONELOOP: " << s << " " << tmp << endl;
}

void NGluon::error(string s) {
  cout << "# ONELOOP: ERROR: " << s << endl;
  exit(1);
}

void NGluon::error(string s, int tmp){
    cout << "# ONELOOP: ERROR: "   << s << " " << tmp << endl;
    exit(1);
}

void NGluon::setMuR(DOUBLE muR_){
  muR=muR_;
  info("Renormalization scale set to mu =", muR);
}

void NGluon::setMassiveFModes(int box, int tri, int bub){
  FourierNumRBox = box;
  FourierNumRTri = tri;
  FourierNumRBub = bub;

  if ( FourierNumRBox > 3 )
    error("maximum number of massive box modes = ",3); 

  if( FourierNumRTri > 3 )
    error("maximum number of massive triangle modes = ",3); 

  if( FourierNumRBub > 3 )
    error("maximum number of massive bubble modes = ",3);

  const int fModes = 2*FourierNumRBox+1;
  const DOUBLE omega = two*PI/fModes;
  mFModesBox = two * fModes;

  for (int t = 0; t < fModes; t++){
    sign[2*t+1]  =  one;
    sign[2*t]    = -one;
    sMBox[2*t+1] =  exp(I*two*omega*static_cast<DOUBLE>(t));
    sMBox[2*t]   =  sMBox[2*t+1];
    fBaseBoxRat[2*t+1][0] =  one/ mFModesBox;
    fBaseBoxRat[2*t][0]   =  fBaseBoxRat[2*t+1][0];
    fBaseBoxRat[2*t+1][1] =  one/(sMBox[2*t+1] * mFModesBox);
    fBaseBoxRat[2*t][1]   =  fBaseBoxRat[2*t+1][1];
    fBaseBoxRat[2*t+1][2] =  one/(sMBox[2*t+1]*sMBox[2*t+1]*mFModesBox);
    fBaseBoxRat[2*t][2]   =  fBaseBoxRat[2*t+1][2];
  }

  setScalarMass(FourierNumRTri,sMTri);
  mFModesTri = FourierNumRTri + 1;

  setScalarMass(FourierNumRBub,sMBub);
  mFModesBub = FourierNumRBub + 1;
}

void NGluon::setScalarMass(const int p, complex<DOUBLE>* scalarMass){

  const DOUBLE omegaM = two*PI / (p+1);
  if ( p == 2 ){ 
    scalarMass[0] = one;
    scalarMass[1] = complex<DOUBLE>(-one/two, -sqrt(three)/two );
    scalarMass[2] = complex<DOUBLE>(-one/two,  sqrt(three)/two );
    return;
  }  
  if ( p == 1 ){
    scalarMass[0] =  one;
    scalarMass[1] = -one;
    return;
  }
  for (int t = 0; t < p+1; t++){
    scalarMass[t] = exp(I*two*omegaM*DOUBLE(t));
  }
}


void NGluon::updateCMEnergy(){
  cmEnergy = sqrt( two * DOT( BGC.posX(0,0), BGC.posX(1,1) ) );
}

complex<DOUBLE> NGluon::evalAmp(){

  BGC.momentumUpdate();
  BGC.helicityUpdate();
  BGC.fillPartialCurrents();
  updateCMEnergy();

  Atree = BGC.gluonAmplitude();

  Acc = Arat = Aeps1 = Aeps2 = zero;

  boxes();
  triangles();
  bubbles();

  pentagonsR();
  boxesR();
  trianglesR();
  bubblesR();

  Afinite = Acc + Arat;

  IRCheckm2 = compare(Aeps2/Atree,IRpoles(-2));
  IRCheckm1 = compare(Aeps1/Atree,IRpoles(-1));

  /****************
   * scaling test *
   ****************/

  if (scaleTest == true){
    DOUBLE x = 1000;
    DOUBLE fac = pow(x,ngluon-4);
    DOUBLE muRX = muR;
    complex<DOUBLE> AccX, AratX, AfiniteX, Aeps1X, Aeps2X, AtreeX;
    
    AccX  = Acc;
    AratX = Arat;
    AfiniteX  = Afinite;
    Aeps1X    = Aeps1;
    Aeps2X    = Aeps2;
    AtreeX    = Atree;
    
    BGC.rescaleMomenta(x);
    BGC.fillPartialCurrents();
    cmEnergy *= x;
    muR *= x*x;

    Atree = BGC.gluonAmplitude();

    Acc = Arat = Aeps1 = Aeps2 = zero;

    boxes();
    triangles();
    bubbles();
    
    pentagonsR();
    boxesR();
    trianglesR();
    bubblesR();

    Afinite = Acc + Arat;

    absError = (AfiniteX - Afinite*fac);
    Afinite  = (AfiniteX + Afinite*fac)/two; 
    relError =  abs(absError/Afinite);

    absErrorCC = (AccX - Acc*fac);
    Acc        = (AccX + Acc*fac)/two;
    relErrorCC = abs(absErrorCC/Acc);

    absErrorRR = (AratX - Arat*fac);
    Arat       = (AratX + Arat*fac)/two;
    relErrorRR = abs(absErrorRR/Arat);

    absErrorEps1 = (Aeps1X - Aeps1*fac);
    Aeps1        = (Aeps1X + Aeps1*fac)/two;
    relErrorEps1 = abs(absErrorEps1/Aeps1);

    absErrorEps2 = (Aeps2X - Aeps2*fac);
    Aeps2        = (Aeps2X + Aeps2*fac)/two;
    relErrorEps2 = abs(absErrorEps2/Aeps2);
 
    Atree    = (AtreeX + Atree*fac)/two;

    muR = muRX;

  }

  return ( Afinite );

}

complex<DOUBLE> NGluon::getAtree(){ return (Atree); }
complex<DOUBLE> NGluon::getAcc(){ return (Acc); }
complex<DOUBLE> NGluon::getArat(){ return (Arat); }
complex<DOUBLE> NGluon::getAeps2(){ return (Aeps2); }
complex<DOUBLE> NGluon::getAeps1(){ return (Aeps1); }
complex<DOUBLE> NGluon::getAfinite(){ return (Afinite); }

complex<DOUBLE> NGluon::getAbsError(){ return (absError); }
complex<DOUBLE> NGluon::getAbsErrorEps1(){ return (absErrorEps1); }
complex<DOUBLE> NGluon::getAbsErrorEps2(){ return (absErrorEps2); }
complex<DOUBLE> NGluon::getAbsErrorRR(){ return (absErrorRR); }


DOUBLE NGluon::getRelError(){ return (relError); }
DOUBLE NGluon::getRelErrorCC(){ return (relErrorCC); }
DOUBLE NGluon::getRelErrorRR(){ return (relErrorRR); }
DOUBLE NGluon::getRelErrorEps1(){ return (relErrorEps1); }
DOUBLE NGluon::getRelErrorEps2(){ return (relErrorEps2); }

DOUBLE NGluon::getIRCheck(int i){ 
  if ( -2 == i ) return (IRCheckm2); 
  if ( -1 == i ) return (IRCheckm1); 
  error("Wrong argument to getIRcheck");
  return(DOUBLE(1.e300));
}


complex<DOUBLE> NGluon::IRpoles(const int eps){

  if ( eps == -2 ) 
    return(-static_cast<complex<DOUBLE> >(ngluon));

  if(eps == -1){
    DOUBLE MUR2 = muR*muR;
    complex<DOUBLE> pole = -complex<DOUBLE>(11.0)/DOUBLE(3.0);
    for(int i=0; i < ngluon-1; i++) {
      DOUBLE logarg = -MUR2 / DOUBLE(2.)/ DOTP(BGC.k[i],BGC.k[i+1]);
      pole -= log( abs( logarg ) );
      if(logarg < 0.0){
	pole -= I*PI;
      }
    }
    DOUBLE logarg = -MUR2 / DOUBLE(2.)/ DOTP(BGC.k[ngluon-1],BGC.k[0]);
    pole -= log( abs( logarg) );
    if(logarg < 0.0){
      pole -= I*PI;
    }
    return (pole);
  }
  error("Wrong argument to IRpoles");
  return(DOUBLE(1.e300));
  
}


int NGluon::CheckPhaseSpacePoint(int n, DOUBLE p[][4]){

  DOUBLE acc = 1.e-13;
  DOUBLE ircut = 1.e-5;
  DOUBLE smin = 1e300, s = 2.0 * DOTP(p[0], p[1]);

  for(int i=0; i<n; i++){
    for(int j = i+1; j < n; j++){
      DOUBLE sij = 2. * DOTP(p[i],p[j]);
      if (abs(sij) < smin)
    smin = abs(sij);
    }
  }

  if (smin/s < ircut)
    return(1);
 
  for(int i=0; i<n; i++){
    if ( abs(DOTP(p[i],p[i])/p[i][0]) > acc)
      return(1);
  }

  for (int mu=0; mu<4; mu++){
    DOUBLE psum=0;
    for(int i=0; i<n; i++){
      psum += p[i][mu];
    }
    if ( abs(psum)/sqrt(s) > acc)
      return(1);
  }
  return(0);
}

//}}}

string NGluon::helicity2string(const int n, int hel[]){
  return Current::helicity2string(n,hel);
}
string NGluon::int2string(const int number){
  return Current::int2string(number);
}
void NGluon::sethel(int n, int htype, int *helicity){
  Current::sethel(n,htype,helicity);
}
void NGluon::sethel(int n, std::string hstr, int *helicity){
  Current::sethel(n, hstr, helicity);
}

//{{{ integrands

complex<DOUBLE> NGluon::boxProduct(const Topology &topo){
  
  complex<DOUBLE> Boxcoefficient = zero;
  for (int w =-1; w<2; w+=2){ 
    for (int x =-1; x<2; x+=2){
      for (int y=-1; y<2; y+=2){
	for (int z=-1; z<2; z+=2){
	  // search for vanishing helicity configurations
	  if ( ! ( BGC.checkHelsum(topo.h0, topo.h1-1, -w, x) && 
		   BGC.checkHelsum(topo.h1, topo.h2-1, -x, y) && 
		   BGC.checkHelsum(topo.h2, topo.h3-1, -y, z) && 
		   BGC.checkHelsum(topo.h3, topo.h0-1, -z, w) ) )
	    continue; 
	  Boxcoefficient +=  
	    BGC.partialGluonAmplitude(topo.h0, topo.h1-1, -w, x, topo.l0, topo.l1) *
	    BGC.partialGluonAmplitude(topo.h1, topo.h2-1, -x, y, topo.l1, topo.l2) *
	    BGC.partialGluonAmplitude(topo.h2, topo.h3-1, -y, z, topo.l2, topo.l3) *
	    BGC.partialGluonAmplitude(topo.h3, topo.h0-1, -z, w, topo.l3, topo.l0);
	}
      }
    }
  }  
  return Boxcoefficient;
}

complex<DOUBLE> NGluon::triangleProduct(const Topology &topo){
  
  complex<DOUBLE> Trianglecoefficient = zero;
  // sum over internal helicities - multiply tree amplitudes
  for (int x =-1; x<2; x+=2) {
    for (int y=-1; y<2; y+=2) {
      for (int z=-1; z<2; z+=2) {
	// search for vanishing helicity configurations
	if ( ! ( BGC.checkHelsum(topo.h0, topo.h1-1, -x, y) &&
		 BGC.checkHelsum(topo.h1, topo.h2-1, -y, z) &&
		 BGC.checkHelsum(topo.h2, topo.h0-1, -z, x) ) ) 
	  continue; 
	
	Trianglecoefficient +=  
	  BGC.partialGluonAmplitude(topo.h0, topo.h1-1, -x, y, topo.l0, topo.l1) *
	  BGC.partialGluonAmplitude(topo.h1, topo.h2-1, -y, z, topo.l1, topo.l2) *
	  BGC.partialGluonAmplitude(topo.h2, topo.h0-1, -z, x, topo.l2, topo.l0);
      }
    }
  }  
  return ( Trianglecoefficient );
}

complex<DOUBLE> NGluon::bubbleProduct(const Topology &topo){
  
  complex<DOUBLE> Bubblecoefficient= zero;
  // sum over internal helicities - multiply tree amplitudes
  for (int x =-1; x<2; x+=2){
    for (int y=-1; y<2; y+=2){
      // search for vanishing helicity configurations
      if (! ( BGC.checkHelsum(topo.h0, topo.h1-1, -x, y) && 
	      BGC.checkHelsum(topo.h1, topo.h0-1, -y, x) ) )  
	continue; 
      
      Bubblecoefficient +=  
	BGC.partialGluonAmplitude(topo.h0, topo.h1-1,-x,y, topo.l0, topo.l1) *
	BGC.partialGluonAmplitude(topo.h1, topo.h0-1,-y,x, topo.l1, topo.l0);
    }
  }
  return ( Bubblecoefficient );
}

complex<DOUBLE> NGluon::pentagonProductS(const Topology &topo){

  return ( BGC.partialGluonScalarAmplitude(topo.h0, topo.h1-1, topo.l0, topo.l1)*
	   BGC.partialGluonScalarAmplitude(topo.h1, topo.h2-1, topo.l1, topo.l2)*
	   BGC.partialGluonScalarAmplitude(topo.h2, topo.h3-1, topo.l2, topo.l3)*
	   BGC.partialGluonScalarAmplitude(topo.h3, topo.h4-1, topo.l3, topo.l4)*
	   BGC.partialGluonScalarAmplitude(topo.h4, topo.h0-1, topo.l4, topo.l0) );
}

complex<DOUBLE> NGluon::boxProductS(const Topology &topo){

  return ( BGC.partialGluonScalarAmplitude(topo.h0, topo.h1-1, topo.l0, topo.l1) *
	   BGC.partialGluonScalarAmplitude(topo.h1, topo.h2-1, topo.l1, topo.l2) *
	   BGC.partialGluonScalarAmplitude(topo.h2, topo.h3-1, topo.l2, topo.l3) *
	   BGC.partialGluonScalarAmplitude(topo.h3, topo.h0-1, topo.l3, topo.l0) );
}

complex<DOUBLE> NGluon::triangleProductS(const Topology &topo){
  
  return ( BGC.partialGluonScalarAmplitude(topo.h0, topo.h1-1, topo.l0, topo.l1)*
	   BGC.partialGluonScalarAmplitude(topo.h1, topo.h2-1, topo.l1, topo.l2)*
	   BGC.partialGluonScalarAmplitude(topo.h2, topo.h0-1, topo.l2, topo.l0));
}

complex<DOUBLE> NGluon::bubbleProductS(const Topology &topo){

  return ( BGC.partialGluonScalarAmplitude(topo.h0, topo.h1-1, topo.l0, topo.l1)*
	   BGC.partialGluonScalarAmplitude(topo.h1, topo.h0-1, topo.l1, topo.l0));
}

//}}}

//{{{ boxes
void NGluon::boxes(){

  Topology topo;

  const int R = 4;
  DOUBLE q[7];
  DOUBLE *PP[4];

  complex<DOUBLE> VR[4] = { zero };
  complex<DOUBLE> nR[4][4] = { { zero } };

  complex<DOUBLE> root;
  complex<DOUBLE> integrand[2], C4;
  DOUBLE sgnhlp[2]={one, DOUBLE(-1.)};
  int boxkey;

  for (topo.h3=3;  topo.h3 < ngluon;  topo.h3++){
    for (topo.h2=2;  topo.h2 < topo.h3;  topo.h2++){
      for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
	for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){
	  
	  PP[0] = BGC.posX(topo.h0,topo.h1-1);
	  PP[1] = BGC.posX(topo.h1,topo.h2-1);
	  PP[2] = BGC.posX(topo.h2,topo.h3-1);
	  PP[3] = BGC.posX(topo.h1,topo.h3-1); 
	  
	  q[0] = BGC.posS(topo.h0,topo.h1-1);
	  q[1] = zero;
	  q[2] = BGC.posS(topo.h1,topo.h2-1);
	  q[3] = BGC.posS(topo.h1,topo.h3-1);
	  
	  q[4] = BGC.posS(topo.h2,topo.h3-1);
	  q[5] = BGC.posS(topo.h0,topo.h2-1);
	  q[6] = BGC.posS(topo.h0,topo.h3-1);

	  VNVB(R, PP, q, VR, nR);

	  boxkey = C.boxkey(topo.h0, topo.h1, topo.h2, topo.h3);
	  C.setboxmom(boxkey, 0, nR[0] );
	  C.setboxmom(boxkey, 1, VR );

	  root = sqrt(-S(VR));

	  for (int count = 0; count < 2; count++){
	    for (int mu=0; mu<4; mu++){
	      topo.l1[mu] = VR[mu] + sgnhlp[count] * root * nR[0][mu];
	      topo.l2[mu] = topo.l1[mu] - PP[1][mu];
	      topo.l3[mu] = topo.l1[mu] - PP[3][mu];
	      topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	    }	  
	    integrand[count] = boxProduct(topo);
	  }

	  C4 = ( integrand[0] + integrand[1] )/two;
	  C.setbox(boxkey,0,C4);
	  C.setbox(boxkey,1, (integrand[0]-integrand[1]) / (two*root) );

	  Acc += I * C4 *I4(q[3], q[5], q[2], q[4], q[6], q[0],
			    zero, zero, zero, zero, muR, 0);
	  
	  Aeps1 += I * C4 *I4(q[3], q[5],q[2], q[4], q[6], q[0],
			      zero, zero, zero, zero, muR, -1);
	  
	  Aeps2 += I * C4 *I4(q[3], q[5], q[2], q[4], q[6], q[0],
			      zero, zero, zero, zero, muR, -2);

	  //printBoxCoef(topo,"fullbox");
	}
      }
    }
  }
}
//}}}

//{{{ triangles
void NGluon::triangles(){

  Topology topo;

  const int R = 3;
  DOUBLE q[4];
  DOUBLE *PP[2];

  complex<DOUBLE> VR[4]={zero};
  complex<DOUBLE> nR[4][4]={{zero}};

  complex<DOUBLE> kappaSq;
  complex<DOUBLE> kappa;
  complex<DOUBLE> integrand;

  int maxMode;
  complex<DOUBLE> (*alpha)[8][2];
  complex<DOUBLE> (*fBaseTri)[8][7];
  bool massless;
  int key;

  for (topo.h2=2;  topo.h2 < ngluon;  topo.h2++){
    for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
      for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){

	PP[0] = BGC.posX(topo.h0, topo.h1-1);
	PP[1] = BGC.posX(topo.h1, topo.h2-1);

	q[0] = BGC.posS(topo.h0, topo.h1-1);
	q[1] = zero;
	q[2] = BGC.posS(topo.h1, topo.h2-1);
	q[3] = BGC.posS(topo.h0, topo.h2-1);

	VNVB(R, PP, q, VR, nR);
	
	key = C.trikey(topo.h0, topo.h1, topo.h2);
	C.settrimom(key, 0, nR[0]);
	C.settrimom(key, 1, nR[1]);
	C.settrimom(key, 2, VR);

	massless = false;
	if ( (topo.h1-(topo.h0+1)==0) || (topo.h2-(topo.h1+1)==0)
	     || (topo.h0-(topo.h2+1)%ngluon==0) ){ 
	  massless = true;
	  alpha = &alphaTriC;
	  fBaseTri = &fBaseTriC;
	  maxMode = fmTri+1;
	  kappaSq = cmEnergy*cmEnergy;
	  kappa   = sqrt(kappaSq);
	} else {
	  alpha = &alphaTriR;
	  fBaseTri = &fBaseTriR;
	  maxMode = fmTri;
	  kappaSq = -S(VR);
	  kappa = sqrt(kappaSq);
	}
	
	complex<DOUBLE> result[7] = {zero};
	for ( int fm = 0; fm < maxMode; fm++){     
	  for (int mu=0; mu<4; mu++){
	    topo.l1[mu] = VR[mu] + kappa * ( + (*alpha)[fm][0] * nR[0][mu] 
					     + (*alpha)[fm][1] * nR[1][mu] );
	    topo.l2[mu] = topo.l1[mu] - PP[1][mu];
	    topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	  }
	  
	  integrand = triangleProduct(topo) 
	    - boxSub(topo.h0, topo.h1, topo.h2, topo.h1, topo.l1);

	  for (int b = 0; b < 7; b++){
	    result[b] += (*fBaseTri)[fm][b] * integrand;
	  }
	}

	if ( massless ) {
	  result[1] /= kappa;
	  result[2] /= kappa;
	  result[3] /= (two*kappaSq);
	  result[4] /= kappaSq;
	  result[5] /= (kappa*kappaSq);
	  result[6] /= -(kappa*kappaSq);
	} else {
	  result[1]  = (result[1]-result[6])/kappa;
	  result[2]  = (result[2]+result[5])/kappa;
	  result[3] *= (-one)/kappaSq;
	  result[4] *= two/kappaSq;
	  result[5] *= (-four)/(kappaSq*kappa);
	  result[6] *= ( four)/(kappaSq*kappa);
	}
	for (int b=0; b<7; b++){
	  C.settri(key, b, result[b] );
	}

	Acc   += result[0] * I3(q[2], q[0], q[3], zero, zero, zero, muR,  0);
	Aeps1 += result[0] * I3(q[2], q[0], q[3], zero, zero, zero, muR, -1);
	Aeps2 += result[0] * I3(q[2], q[0], q[3], zero, zero, zero, muR, -2);
	//printTriCoef(topo,"fulltri");
      }
    }
  }
}
//}}}

//{{{ bubbles
void NGluon::bubbles(){

  Topology topo;
  
  const int R = 2;
  DOUBLE q[2];
  DOUBLE *PP[1];
  
  complex<DOUBLE> VR[4]={zero};
  complex<DOUBLE> nR[4][4]={{zero}};

  complex<DOUBLE> kappa, integrand, C2;
  int key;

  for (topo.h0 = 0;  topo.h0 < ngluon-2;  topo.h0++){
    for (topo.h1 = topo.h0+2;  topo.h1 < ngluon;  topo.h1++){
 
      // no on-shell bubbles
      if ( topo.h1-topo.h0 == ngluon-1 ) 
	continue;

      PP[0] = BGC.posX(topo.h0,topo.h1-1);

      q[0] = BGC.posS(topo.h0,topo.h1-1);
      q[1] = zero;

      VNVB(R, PP, q, VR, nR);
      
      key = C.bubkey(topo.h0, topo.h1);
      
      C.setbubmom(key, 0, nR[0]);
      C.setbubmom(key, 1, nR[1]);
      C.setbubmom(key, 2, nR[2]);
      C.setbubmom(key, 3, VR);
      
      kappa = sqrt( -S(VR) );

      complex<DOUBLE> result[2] = {zero};
      
      for (int fm = 0; fm < fmBub; fm++){
	for (int mu=0; mu<4; mu++){
	  topo.l1[mu] = VR[mu] + kappa * (  alphaBub[fm][0] * nR[0][mu] 
					   + alphaBub[fm][1] * nR[1][mu] 
					   + alphaBub[fm][2] * nR[2][mu] );
	  topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	}
	
	integrand = bubbleProduct(topo)
	  - boxTriSub(topo.h0, topo.h1, topo.h1, topo.l1);

	result[0] += integrand;
	result[1] += fBaseBub[fm] * integrand;

      }
      
      C2 = (result[0]-two/three*result[1]) / DOUBLE(fmBub);
      C.setbub(key,0,C2);

      Acc    += - I * C2 * I2(q[0], zero, zero, muR, 0);
      Aeps1  += - I * C2 * I2(q[0], zero, zero, muR, -1);
      Aeps2  += - I * C2 * I2(q[0], zero, zero, muR, -2);

      // printBubCoef(topo,"fullbub"); 
    }
  }
}
//}}}

//{{{ subtraction terms
complex<DOUBLE> NGluon::boxSub(const int i, const int j, const int k, 
				const int ref, complex<DOUBLE> *loop1){

  int key;
  complex<DOUBLE> sT = zero;

  for (int l=0; l<i; l++){  
    key = C.boxkey(l,i,j,k); 
    sT += ( C.box(key,0) + C.box(key,1) * DOT(loop1,C.boxmom(key,0)) ) 
      / ( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
  }
  for (int l=i+1; l<j; l++){
    key = C.boxkey(i,l,j,k);
    sT += ( C.box(key,0) + C.box(key,1) * DOT(loop1,C.boxmom(key,0)))
      / ( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
  }
  for (int l=j+1; l<k; l++){  
    key = C.boxkey(i,j,l,k);
    sT += ( C.box(key,0) + C.box(key,1) * DOT(loop1,C.boxmom(key,0)))
      / ( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
  }
  for (int l=k+1; l<ngluon; l++){
    key = C.boxkey(i,j,k,l);
    sT += ( C.box(key,0) + C.box(key,1) * DOT(loop1,C.boxmom(key,0)))
      / ( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
  }
  return (I * sT);
}


complex<DOUBLE> NGluon::cBar(int i, int j, int k, const complex<DOUBLE>* loop1){
  
  int key = C.trikey(i,j,k);
  complex<DOUBLE> s1 = DOT(C.trimom(key,0),loop1);
  complex<DOUBLE> s2 = DOT(C.trimom(key,1),loop1);

  return ( C.tri(key,0) + s1 * C.tri(key,1) + s2*C.tri(key,2) 
	   + (s1*s1-s2*s2) * C.tri(key,3) + s1 * s2 * C.tri(key,4) 
	   + s1*s2*s1*C.tri(key,5) + s1*s2*s2*C.tri(key,6) );
}


complex<DOUBLE> NGluon::boxTriSub(const int i, const int j, const int ref,
					complex<DOUBLE> *loop1){
  complex<DOUBLE> Prop;  
  complex<DOUBLE> sT3 = zero;
  complex<DOUBLE> sT4 = zero;
 
  for (int k=0; k<i; k++){
    Prop = one/( -two*DOT(loop1,BGC.posX(ref,k-1) ) + BGC.posS(ref,k-1) );
    sT3 += cBar(k,i,j, loop1) * Prop;
    sT4 += boxSub(k,i,j,ref,loop1) * Prop;
  }
  for (int k=i+1; k<j; k++){ 
    Prop = one/( -two*DOT(loop1,BGC.posX(ref,k-1) ) + BGC.posS(ref,k-1) );
    sT3 += cBar(i,k,j, loop1) * Prop;
    sT4 += boxSub(i,k,j,ref,loop1) * Prop;
  }
  for (int k=j+1; k<ngluon; k++){
    Prop = one/( -two*DOT(loop1,BGC.posX(ref,k-1) ) + BGC.posS(ref,k-1) );
    sT3 += cBar(i,j,k, loop1) * Prop;
    sT4 += boxSub(i,j,k,ref,loop1) * Prop;
  }
  return ( I * ( sT3 + sT4 / two ) );
}
//}}}

//{{{ Rational: pentagons
void NGluon::pentagonsR(){

  Topology topo;
  DOUBLE *PP[4];
  
  for (topo.h4=4; topo.h4<ngluon; topo.h4++){
    for (topo.h3=3; topo.h3 < topo.h4; topo.h3++){
      for (topo.h2=2; topo.h2 < topo.h3; topo.h2++){
	for (topo.h1=1; topo.h1 < topo.h2; topo.h1++){
	  for (topo.h0=0; topo.h0 < topo.h1; topo.h0++){
	    
	  PP[0] = BGC.posX(topo.h0,topo.h1-1);
	  PP[1] = BGC.posX(topo.h1,topo.h2-1);
	  PP[2] = BGC.posX(topo.h2,topo.h3-1);
	  PP[3] = BGC.posX(topo.h3,topo.h4-1);

	  pentBasis(PP, topo);

	  C.setpentagonR(C.penkey(topo.h0,topo.h1,topo.h2,topo.h3,topo.h4),
			 pentagonProductS(topo));
	  C.setpentagonM(C.penkey(topo.h0,topo.h1,topo.h2,topo.h3,topo.h4),
			 abs(topo.scalarMSq) );
	  }
	}
      }
    }
  }
}
//}}}

//{{{ Rational: boxes
void NGluon::boxesR(){

  Topology topo;
  int key;
  DOUBLE r;
  complex<DOUBLE> kappa, kappaSq, integrand;

  DOUBLE *PP[3];
  complex<DOUBLE>* VR;
  complex<DOUBLE>* nR;
 
  for (topo.h3=3;  topo.h3 < ngluon;  topo.h3++){
    for (topo.h2=2;  topo.h2 < topo.h3;  topo.h2++){
      for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
	for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){

	  PP[0] = BGC.posX(topo.h0,topo.h1-1);
	  PP[1] = BGC.posX(topo.h1,topo.h2-1);
	  PP[2] = BGC.posX(topo.h2,topo.h3-1);

	  key = C.boxkey(topo.h0, topo.h1, topo.h2, topo.h3);
	  nR  = C.boxmom(key, 0 );
	  VR  = C.boxmom(key, 1 );

	  complex<DOUBLE> result[6]={zero};

	  if ( ngluon > 4){
	    r = pentMaxM(topo.h0, topo.h1, topo.h2, topo.h3);
	  }
	  else
	    r = cmEnergy*cmEnergy;

	  for (int t = 0; t < mFModesBox; t++){
	    topo.scalarMSq = r * sMBox[t];
	    kappaSq = -(S(VR)-topo.scalarMSq);
	    kappa = sqrt(kappaSq);

	    for (int mu=0; mu<4; mu++){
	      topo.l1[mu] = VR[mu] + sign[t] * kappa * nR[mu];
	      topo.l2[mu] = topo.l1[mu] - PP[1][mu];
	      topo.l3[mu] = topo.l1[mu] - BGC.posX(topo.h1,topo.h3-1)[mu];
	      topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	    }
	    if ( ngluon > 4 ){
	      integrand = boxProductS(topo) 
		-pentSubR(topo.h0,topo.h1,topo.h2,topo.h3,topo.h1,topo.l1,topo.scalarMSq);
	    }
	    else 
	      integrand = boxProductS(topo);

	    result[0] +=           integrand * fBaseBoxRat[t][0];
	    result[1] += sign[t] * integrand * fBaseBoxRat[t][0]/kappa;
	    result[2] +=           integrand * fBaseBoxRat[t][1]/r;
	    result[3] += sign[t] * integrand * fBaseBoxRat[t][1]/(r*kappa);
	    result[4] +=           integrand * fBaseBoxRat[t][2]/(r*r);
	  }
	  for (int x=0; x<5; x++){
	    C.setboxR( key, x, result[x]);
	  }
	  Arat += -I * result[4] / three;
	  // printBoxCoef(topo,"rat");
	}
      }
    }
  }
}
//}}}

//{{{ Rational: triangles
void NGluon::trianglesR(){
  
  Topology topo;
  DOUBLE r;
  int key;
  DOUBLE inv_mFModesTri = one/mFModesTri;
  complex<DOUBLE> kappa, kappaSq, sT4, sT5, integrand;

  DOUBLE *PP[2];
  complex<DOUBLE>* VR;
  complex<DOUBLE>* nR[2];

  for (topo.h2=2;  topo.h2 < ngluon;  topo.h2++){
    for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
      for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){

	PP[0] = BGC.posX(topo.h0, topo.h1-1);
	PP[1] = BGC.posX(topo.h1, topo.h2-1);

	key   = C.trikey(topo.h0,topo.h1,topo.h2);
	nR[0] = C.trimom(key,0);
	nR[1] = C.trimom(key,1);
	VR    = C.trimom(key,2);

	complex<DOUBLE> result[4][7] = {{zero}};
	complex<DOUBLE> final[10] = {zero};

	if ( ngluon > 4 ){
	  r = boxPentMaxM(topo.h0, topo.h1, topo.h2);
	}
	else
	  r = cmEnergy*cmEnergy;

	for (int t = 0; t < mFModesTri; t++){
	  topo.scalarMSq = r * sMTri[t];
	  kappaSq        = -(S(VR)-topo.scalarMSq);
	  kappa          = sqrt(kappaSq);
	  
	  for (int fm = 0; fm < fmTri; fm++){      
	    for (int mu=0; mu<4; mu++){
	      topo.l1[mu] = VR[mu] + kappa * ( + alphaTriR[fm][0]*nR[0][mu] 
					       + alphaTriR[fm][1]*nR[1][mu] );
	      topo.l2[mu] = topo.l1[mu] - PP[1][mu];
	      topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	    }
	    boxPentSubR(topo.h0,topo.h1,topo.h2, topo.h1,topo.l1,topo.scalarMSq,
			sT4,sT5);
	    integrand = triangleProductS(topo) - sT4 - sT5 / two;

	    for (int f = 0; f < 7; f++){
	      result[t][f] += fBaseTriR[fm][f] * integrand;
	    }
	  } 
	  result[t][1]  = (result[t][1]-result[t][6]) / kappa;
	  result[t][2]  = (result[t][2]+result[t][5]) / kappa;
	  result[t][3] *= (-one) / kappaSq;
	  result[t][4] *=  two/kappaSq;
	  result[t][5] *= (-four)/(kappaSq*kappa);
	  result[t][6] *= ( four)/(kappaSq*kappa);
	  
	  final[0]  +=  result[t][0];
	  final[1]  +=  result[t][1];
	  final[2]  +=  result[t][2];
	  final[3]  +=  result[t][3];
	  final[4]  +=  result[t][4];
	  final[5]  +=  result[t][5];
	  final[6]  +=  result[t][6];
	  final[7]  +=  result[t][1]/topo.scalarMSq ;
	  final[8]  +=  result[t][2]/topo.scalarMSq;
	  final[9]  +=  result[t][0]/topo.scalarMSq; 

	}
	for (int z=0; z<10; z++){
	  final[z] *= inv_mFModesTri; 
	  C.settriR(key,z,final[z]);
	}
	Arat  += - final[9];
	
      }
    }
  }
}
//}}}

//{{{ Rational: bubbles
void NGluon::bubblesR(){
  
  Topology topo;
  DOUBLE r;
  int key;
  complex<DOUBLE> kappa, sT3, sT4, sT5, integrand;

  DOUBLE *PP[1];
  complex<DOUBLE>* VR;
  complex<DOUBLE>* nR[3];
  
  for (topo.h0 = 0;  topo.h0 < ngluon-2;  topo.h0++){
    for (topo.h1 = topo.h0+2;  topo.h1 < ngluon;  topo.h1++){
 
      // no on-shell bubbles
      if(topo.h1-topo.h0==ngluon-1) 
	continue;

      PP[0] = BGC.posX(topo.h0,topo.h1-1);

      key   = C.bubkey(topo.h0,topo.h1);
      nR[0] = C.bubmom(key,0);
      nR[1] = C.bubmom(key,1);
      nR[2] = C.bubmom(key,2);
      VR    = C.bubmom(key,3);
      
      // determine correct radius
      if ( ngluon > 4 )
	r = triBoxPentMaxM(topo.h0, topo.h1)*cmEnergy*cmEnergy/abs(S(VR));
      else
	r = cmEnergy*cmEnergy;

      complex<DOUBLE> final = zero;
      for (int t = 0; t < mFModesBub; t++){
	topo.scalarMSq = r * sMBub[t];
	kappa = sqrt(-(S(VR)-topo.scalarMSq) );
	
	complex<DOUBLE> result[2] = {zero};

	for (int fm = 0; fm < fmBub; fm++){
	  for (int mu=0; mu<4; mu++){
	    topo.l1[mu] = VR[mu] + kappa * (   alphaBub[fm][0] * nR[0][mu] 
					     + alphaBub[fm][1] * nR[1][mu] 
					     + alphaBub[fm][2] * nR[2][mu] );
	    topo.l0[mu] = topo.l1[mu] + PP[0][mu];
	  }
	  triBoxPentSubR(topo.h0,topo.h1, topo.h1,topo.l1,topo.scalarMSq,
			 sT3,sT4,sT5);	    
	  integrand = bubbleProductS(topo) - sT3 - sT4/two - sT5/six;

	  result[0] += integrand;
	  result[1] += fBaseBub[fm] * integrand;
	}
	
	final += (result[0] - two/three * result[1])
	  /( DOUBLE(fmBub) * mFModesBub * topo.scalarMSq );
      }
      C.setbubR(key,0,final);
      Arat  += I * final  * S(PP[0]) / three;
      //printBubCoef(topo,"rat");
    }
  }
}
//}}}

//{{{ Rational : subtractions
complex<DOUBLE> NGluon::pentSubR(const int i, const int j, const int k,
				  const int l, const int ref,
				  const complex<DOUBLE> *loop1,
				  const complex<DOUBLE> scalarMSq){
  
  complex<DOUBLE> sT5 = zero;
  
  for (int m=0; m<i; m++){
    sT5 += C.pentagonR(C.penkey(m,i,j,k,l)) 
      /( - two * DOT(loop1,BGC.posX(ref,m-1) ) + BGC.posS(ref,m-1) );
  }
  for (int m=i+1; m<j; m++){
    sT5+=C.pentagonR(C.penkey(i,m,j,k,l))
      /( - two * DOT(loop1,BGC.posX(ref,m-1) ) + BGC.posS(ref,m-1) );
  }
  for (int m=j+1; m<k; m++){
    sT5+=C.pentagonR(C.penkey(i,j,m,k,l))
      /( - two * DOT(loop1,BGC.posX(ref,m-1) ) + BGC.posS(ref,m-1) );
  }
  for (int m=k+1; m<l; m++){
    sT5+=C.pentagonR(C.penkey(i,j,k,m,l))
      /( - two * DOT(loop1,BGC.posX(ref,m-1) ) + BGC.posS(ref,m-1) );
  }
  for (int m=l+1; m<ngluon; m++){
    sT5+=C.pentagonR(C.penkey(i,j,k,l,m))
      /( - two * DOT(loop1,BGC.posX(ref,m-1) ) + BGC.posS(ref,m-1) );
  }
  return ( - I * sT5 );  // -I from propagators
}


void NGluon::boxPentSubR(const int i, const int j, const int k,
			  const int ref,
			  const complex<DOUBLE> *loop1,
			  const complex<DOUBLE> scalarMSq,
			  complex<DOUBLE> &sT4, complex<DOUBLE> &sT5){
  
  complex<DOUBLE> Prop;
  complex<DOUBLE> sT5tmp;
  sT4 = zero;
  sT5 = zero;
  for (int l=0; l<i; l++){
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT4 += dBarR(l,i,j,k,loop1,scalarMSq) * Prop;
    sT5 += pentSubR(l,i,j,k, ref, loop1,scalarMSq) * Prop;
  }
  for (int l=i+1; l<j; l++){     
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT4 +=  dBarR(i,l,j,k,loop1,scalarMSq) * Prop;
    sT5 += pentSubR(i,l,j,k, ref, loop1,scalarMSq) * Prop;
  }
  for (int l=j+1; l<k; l++){ 
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT4 += dBarR(i,j,l,k,loop1,scalarMSq) * Prop;
    sT5 += pentSubR(i,j,l,k, ref,loop1,scalarMSq) * Prop;
  }
  for (int l=k+1; l<ngluon; l++){ 
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT4 += dBarR(i,j,k,l,loop1,scalarMSq) * Prop;
    sT5 += pentSubR(i,j,k,l, ref,loop1,scalarMSq) * Prop;
  }
  sT4 *= -I;  // -I from propagators
  sT5 *= -I;
}

complex<DOUBLE> NGluon::dBarR(const int i, const int j, const int k, 
			       const int l, const complex<DOUBLE> *loop1,
			       const complex<DOUBLE> scalarMSq){
  
  int bkey = C.boxkey(i,j,k,l);
  complex<DOUBLE> sXl  = DOT(loop1,C.boxmom(bkey,0));
  return ( + C.boxR(bkey,0) + C.boxR(bkey,1) * sXl 
	   + C.boxR(bkey,2) * scalarMSq 
	   + C.boxR(bkey,3) * sXl*scalarMSq 
	   + C.boxR(bkey,4) * scalarMSq*scalarMSq );
}

void NGluon::triBoxPentSubR(const int i, const int j, const int ref,
			     const complex<DOUBLE> *loop1,
			     const complex<DOUBLE> scalarMSq,
			     complex<DOUBLE> &sT3, 
			     complex<DOUBLE> &sT4,
			     complex<DOUBLE> &sT5){

  complex<DOUBLE> Prop;
  complex<DOUBLE> sT4tmp, sT5tmp;
  sT3 = sT4 = sT5 = zero;
  
  for (int l=0; l<i; l++){
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT3  += cBarR(l,i,j,loop1,scalarMSq) * Prop;
    boxPentSubR(l,i,j, ref, loop1,scalarMSq, sT4tmp, sT5tmp);
    sT4    += sT4tmp*Prop;
    sT5    += sT5tmp*Prop;
  }
  for (int l=i+1; l<j; l++){
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT3  += cBarR(i,l,j,loop1,scalarMSq) * Prop;
    boxPentSubR(i,l,j, ref, loop1,scalarMSq, sT4tmp, sT5tmp);
    sT4    += sT4tmp*Prop; 
    sT5    += sT5tmp*Prop;
  }
  for (int l=j+1; l<ngluon; l++){
    Prop  = one/( -two*DOT(loop1,BGC.posX(ref,l-1) ) + BGC.posS(ref,l-1) );
    sT3  += cBarR(i,j,l,loop1,scalarMSq) * Prop;
    boxPentSubR(i,j,l, ref, loop1,scalarMSq, sT4tmp, sT5tmp);
    sT4    += sT4tmp*Prop;
    sT5    += sT5tmp*Prop;
  }
  sT3 *= -I;    // -I from the propagator
  sT4 *= -I;
  sT5 *= -I;
}

complex<DOUBLE> NGluon::cBarR(const int i, const int j, const int k,
			       const complex<DOUBLE> *loop1,
			       const complex<DOUBLE> mSqScalar ){

  int key = C.trikey(i,j,k);
  complex<DOUBLE> s1 = DOT( C.trimom(key,0), loop1 );
  complex<DOUBLE> s2 = DOT( C.trimom(key,1), loop1 );

  return ( + C.triR(key,0) + s1*C.triR(key,1) + s2*C.triR(key,2) 
	   + (s1*s1 - s2*s2) * C.triR(key,3) + s1*s2 * C.triR(key,4) 
	   + s1*s2*s1 * C.triR(key,5) + s1*s2*s2 * C.triR(key,6)
	   + s1*mSqScalar * C.triR(key,7)
	   + s2*mSqScalar * C.triR(key,8) 
	   + mSqScalar * C.triR(key,9) );
}



DOUBLE NGluon::pentMaxM(const int i, const int j, const int k, const int l){
  
  DOUBLE pentMaxM = zero;
  DOUBLE pen;

  for (int m=0; m<i; m++){
    pen = C.pentagonM(C.penkey(m,i,j,k,l));
    if( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int m=i+1; m<j; m++){
    pen = C.pentagonM(C.penkey(i,m,j,k,l));
    if( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int m=j+1; m<k; m++){
    pen = C.pentagonM(C.penkey(i,j,m,k,l));
    if( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int m=k+1; m<l; m++){
    pen = C.pentagonM(C.penkey(i,j,k,m,l));
    if( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int m=l+1; m<ngluon; m++){
    pen = C.pentagonM(C.penkey(i,j,k,l,m));
    if( pen > pentMaxM )
      pentMaxM = pen;
  }
  return ( pentMaxM ); 
}


DOUBLE NGluon::boxPentMaxM(const int i, const int j, const int k){
  
  DOUBLE boxPentMaxM = zero;
  DOUBLE pen;

  for (int l=0; l<i; l++){
    pen = pentMaxM(l,i,j,k);
    if ( pen > boxPentMaxM )
      boxPentMaxM = pen;
  }
  for (int l=i+1; l<j; l++){     
    pen = pentMaxM(i,l,j,k);
    if ( pen > boxPentMaxM )
      boxPentMaxM = pen;
  }
  for (int l=j+1; l<k; l++){ 
    pen = pentMaxM(i,j,l,k);
    if ( pen > boxPentMaxM )
      boxPentMaxM = pen;
  }
  for (int l=k+1; l<ngluon; l++){ 
    pen = pentMaxM(i,j,k,l);
    if ( pen > boxPentMaxM )
      boxPentMaxM = pen;
  }
  return ( boxPentMaxM );
}



DOUBLE NGluon::triBoxPentMaxM(const int i, const int j){
  
  DOUBLE pentMaxM = zero;
  DOUBLE pen;

  for (int l=0; l<i; l++){
    pen =  boxPentMaxM(l,i,j);
    if ( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int l=i+1; l<j; l++){
    pen =  boxPentMaxM(i,l,j);
    if ( pen > pentMaxM )
      pentMaxM = pen;
  }
  for (int l=j+1; l<ngluon; l++){ 
    pen =  boxPentMaxM(i,j,l);
    if ( pen > pentMaxM )
      pentMaxM = pen;
  }
  return ( pentMaxM );
}
//}}}

//{{{ Cut-constructible Fourier Modes
void NGluon::initFourier4D(){

  const int p=3;
  const int theta=2;
  const int phi=1;
  // triangle case, Fourier modes set to 2p+1=7;
  //  info("# Fourier modes in the 4D-triangle part set to p =",p);
  if ( (p < 3) || (p > 7))
    error("# Fourier modes p must be: 2 < p < 7");

  fmTri = 2*p+1;

  const complex<DOUBLE> omegaC = two*PI*I/(p+one);
  const DOUBLE NC = one/(two*(p+one));

  int FM = 0;
  for (int sigma = 0; sigma < 2; sigma++){
    DOUBLE signum = (sigma & 1) ? one : -one;
    for (DOUBLE fm = 0; fm < p+1; fm += one){
      alphaTriC[FM][0]=   exp(omegaC*fm);
      alphaTriC[FM][1]= I * exp(omegaC*fm)*signum;

      fBaseTriC[FM][0] = NC;
      fBaseTriC[FM][1] = NC*exp(-omegaC*fm);
      fBaseTriC[FM][2] = NC*exp(-omegaC*fm) * (-I*signum);
      fBaseTriC[FM][3] = NC*exp(-two*omegaC*fm);
      fBaseTriC[FM][4] = NC*exp(-two*omegaC*fm) * (-I*signum);
      fBaseTriC[FM][5] = NC*exp(-three*omegaC*fm) * (-I*signum);
      fBaseTriC[FM][6] = NC*exp(-three*omegaC*fm);
      FM++;
    }
  }

  const DOUBLE omegaR = two*PI/fmTri;
  const DOUBLE NR = one/fmTri;
  const DOUBLE NR2 = NR*two;

  for (int fm = 0; fm < fmTri; fm++){
    alphaTriR[fm][0] = sin( omegaR * fm );
    alphaTriR[fm][1] = cos( omegaR * fm ); 

    fBaseTriR[fm][0] = NR;
    fBaseTriR[fm][1] = alphaTriR[fm][0] * NR2;
    fBaseTriR[fm][2] = alphaTriR[fm][1] * NR2;
    fBaseTriR[fm][3] = cos(two * omegaR * fm ) * NR2;
    fBaseTriR[fm][4] = sin(two * omegaR * fm ) * NR2;
    fBaseTriR[fm][5] = cos(three * omegaR * fm ) * NR2;
    fBaseTriR[fm][6] = sin(three * omegaR * fm ) * NR2;

  }

  // bubble case:
  //  info("# Fourier modes in the 4D-bubble part set to (2*theta+1)(2*phi+1) =",
  //   (2*theta+1)*(2*phi+1));
  if ((2*theta+1)*(2*phi+1) < 15 || (2*theta+1)*(2*phi+1) > 49)
    error("# Fourier modes theta*phi for the 4D-part must be: 15 <= t*p <= 49");
  
  const int FourierModesTheta = 2*theta + 1;
  const int FourierModesPhi   = 2*phi + 1;
  const DOUBLE omegaT = two*PI/FourierModesTheta;
  const DOUBLE omegaP = two*PI/FourierModesPhi;
  int fm = 0;
  for (int l = -theta; l <= theta; l++){
    for (int m = -phi; m <= phi; m++){
      alphaBub[fm][0] = sin(omegaT * l) * cos(omegaP * m);
      alphaBub[fm][1] = sin(omegaT * l) * sin(omegaP * m);
      alphaBub[fm][2] = cos(omegaT * l);

      fBaseBub[fm] = cos(omegaT * two * l);

      fm++;      
    }
  }
  fmBub = FourierModesTheta * FourierModesPhi;
}
//}}}

//{{{ loop momentum parameterisations
void NGluon::pentBasis(DOUBLE* PP[4], Topology& topo){

  DOUBLE a, b, S1, S2, P1P2, root, gg, gdet, detsign, kf[2][4];
  complex<DOUBLE> c, d, b3, b4, trace5, cvec[2][4];

  S1 = BGC.posS(topo.h0,topo.h1-1);
  S2 = BGC.posS(topo.h1,topo.h2-1);

  P1P2 = DOT(PP[0], PP[1]);
  root = sqrt( P1P2 * P1P2 - S1 * S2);
  detsign = (P1P2 < zero) ? -one : one;
  gg = P1P2 + detsign*root;
  gdet = one / ( gg * gg - S1 * S2 );
  for (int mu = 0; mu < 4; mu++){
    kf[0][mu] = gg * gdet * ( gg * PP[0][mu] - S1 * PP[1][mu] );
    kf[1][mu] = gg * gdet * ( gg * PP[1][mu] - S2 * PP[0][mu] );
  }

  a =  ( S1 * S2 + S2 * gg ) * gdet;
  b = -( S1 * S2 + S1 * gg ) * gdet;

  trace5 = AB3(kf[0],PP[2],kf[1],PP[3],kf[0]) - AB3(kf[1],PP[2],kf[0],PP[3],kf[1]);

  b3 = - two * a *DOT(kf[0],PP[2]) - two * b * DOT(kf[1],PP[2]) 
    + two * DOT(PP[1],PP[2]) + BGC.posS(topo.h2,topo.h3-1);
  b4 = - two * a *DOT(kf[0],PP[3]) - two * b * DOT(kf[1],PP[3]) 
    + two * DOT(PP[2],PP[3]) + two * DOT(PP[1],PP[3]) 
    + BGC.posS(topo.h3,topo.h4-1);

  c = one/trace5*(   b3 * AB(kf[1],PP[3],kf[0]) - b4 * AB(kf[1],PP[2],kf[0]) );
  d = one/trace5*( - b3 * AB(kf[0],PP[3],kf[1]) + b4 * AB(kf[0],PP[2],kf[1]) );

  topo.scalarMSq = gg * ( a*b - c*d );

  eta_mu(cvec[0],kf[0],kf[1]);
  eta_mu(cvec[1],kf[1],kf[0]);
  for (int mu = 0; mu < 4; mu++){
    topo.l1[mu] = a*kf[0][mu]+b*kf[1][mu]+c*cvec[0][mu]/two+d*cvec[1][mu]/two;
    topo.l2[mu] = topo.l1[mu] - PP[1][mu];
    topo.l3[mu] = topo.l1[mu] - PP[1][mu] - PP[2][mu];
    topo.l4[mu] = topo.l1[mu] - PP[1][mu] - PP[2][mu] - PP[3][mu];
    topo.l0[mu] = topo.l1[mu] + PP[0][mu];
  }

}

complex<DOUBLE> NGluon::AB(const DOUBLE p1[4], const DOUBLE p2[4],
			    const DOUBLE p3[4]){
  
  complex<DOUBLE> t1,t11,t12,t16,t19,t2,t22,t3,t5,t6,t9;
  t1 = pp(p2);
  t2 = pTb(p2);
  t3 = pp(p3);
  t5 = pTb(p3);
  t6 = one/t5;
  t9 = pm(p1);
  t11 = pT(p2);
  t12 = pm(p2);
  t16 = pTb(p1);
  t19 = SpNorm(p3);
  t22 = SpNorm(p1);
  return(((t1-t2*t3*t6)*t9-(t11-t12*t3*t6)*t16)/t19*t22);
}

complex<DOUBLE> NGluon::AB3(const DOUBLE p1[4], const DOUBLE p2[4],
			     const DOUBLE p3[4], const DOUBLE p4[4],
			     const DOUBLE p5[4]){
  
  complex<DOUBLE> t1,t10,t12,t13,t14,t17,t19,t2,t21,t22,t24,t26,t29,
    t3,t31,t33,t36,t39,t4,t42,t5,t7,t8;
  
  t1 = pp(p2);
  t2 = pm(p3);
  t3 = pp(p4);
  t4 = pTb(p4);
  t5 = pp(p5);
  t7 = pTb(p5);
  t8 = one/t7;
  t10 = t3-t4*t5*t8;
  t12 = pTb(p3);
  t13 = pT(p4);
  t14 = pm(p4);
  t17 = t13-t14*t5*t8;
  t19 = t2*t10-t12*t17;
  t21 = pTb(p2);
  t22 = pT(p3);
  t24 = pp(p3);
  t26 = -t22*t10+t24*t17;
  t29 = pm(p1);
  t31 = pT(p2);
  t33 = pm(p2);
  t36 = pTb(p1);
  t39 = SpNorm(p5);
  t42 = SpNorm(p1);
  return(((t1*t19+t21*t26)*t29-(t31*t19+t33*t26)*t36)/t39*t42);
}



void NGluon::epsilon4DComponent(std::complex<DOUBLE> nt[4], DOUBLE* P[]){

  nt[0]= I * (  P[0][3] * P[1][2] * P[2][1] - P[0][2] * P[1][3] * P[2][1] 
	       - P[0][3] * P[1][1] * P[2][2] + P[0][1] * P[1][3] * P[2][2] 
	       + P[0][2] * P[1][1] * P[2][3] - P[0][1] * P[1][2] * P[2][3]);
  
  nt[1]= -I * ( - P[0][3] * P[1][2] * P[2][0] + P[0][2] * P[1][3] * P[2][0] 
		+ P[0][3] * P[1][0] * P[2][2] - P[0][0] * P[1][3] * P[2][2] 
		- P[0][2] * P[1][0] * P[2][3] + P[0][0] * P[1][2] * P[2][3] );
  
  nt[2]= -I * (  P[0][3] * P[1][1] * P[2][0] - P[0][1] * P[1][3] * P[2][0] 
		- P[0][3] * P[1][0] * P[2][1] + P[0][0] * P[1][3] * P[2][1] 
		+ P[0][1] * P[1][0] * P[2][3] - P[0][0] * P[1][1] * P[2][3] );
  
  nt[3]= -I * ( - P[0][2] * P[1][1] * P[2][0] + P[0][1] * P[1][2] * P[2][0] 
		+ P[0][2] * P[1][0] * P[2][1] - P[0][0] * P[1][2] * P[2][1] 
		- P[0][1] * P[1][0] * P[2][2] + P[0][0] * P[1][1] * P[2][2] );

}

template<typename T> T NGluon::det(int n, T b[4][4]){
  T sum=0, c[4][4];
  assert( n > 0 );
  if ( n == 1 )
    return(b[0][0]);
  if ( n == 2 )
    return( b[0][0]*b[1][1]-b[0][1]*b[1][0] );
  for(int p=0;p<n;p++) {
    int h=0; 
    int k=0;
    for(int i=1; i<n; i++) {
      for(int j=0; j<n; j++) {
	if(j==p) continue;
	c[h][k]=b[i][j];
	k++;
	if(k==n-1) {
	  h++;
	  k=0;
	}
      }
    }
    DOUBLE sgn = ( p & 1 ) ? -one : one;
    // ( p & 1 ) checks if the last digit of the binary representation 
    // of p is 0 or 1:  1 = odd, 0 = even	
    sum += b[0][p] * sgn * det(n-1,c);
  }
  return (sum);
}
template DOUBLE NGluon::det(int n, DOUBLE b[4][4]);

void NGluon::solveEqn(DOUBLE A[4][4], DOUBLE B[4][4], const int n, DOUBLE& gramdet){

  DOUBLE sigma;
  DOUBLE uT[n];
  DOUBLE H;
  DOUBLE temp[n];
  DOUBLE tempB[4];

  for (int k=0; k<n-1; k++){

    sigma = 0.;
    for (int i=k; i<n; i++){
      sigma += A[i][k] * A[i][k];
    }
    sigma = sqrt(sigma);
    if ( A[k][k] < 0 ) sigma = -sigma;

    // construct reflection matrix
    uT[k] = A[k][k] + sigma;
    for (int i=k+1; i<n; i++){
      uT[i] = A[i][k];
    }
    H = sigma * uT[k];
    A[k][k] = -sigma;
    for (int i=k+1; i<n; i++){
      A[i][k] = 0.;
    }
    
    // Householder reduction left hand side
    for (int j=k+1; j<n; j++){
      temp[j] = 0.;
      for (int i=k; i<n; i++){
	temp[j] += uT[i] * A[i][j];
      }
      temp[j] /= H;
    }
    for (int i=k+1; i<n; i++){
      for (int l=k; l<n; l++){
	A[l][i] = A[l][i] - uT[l] * temp[i];
      }
    }

    // Householder reduction right hand side
    for (int mu = 0; mu < 4; mu++){
      tempB[mu] = 0.;
      for (int j=k; j<n; j++){
	tempB[mu] += uT[j] * B[j][mu];
      }
      tempB[mu] /= H;
      for (int j=k; j<n; j++){
	B[j][mu] = B[j][mu] - tempB[mu] * uT[j];
      }
    }
  }

  // backsubstitution
  for (int mu = 0; mu < 4; mu++){
    for (int k = n-1 ; k >= 0; k--){
      temp[mu] = 0.;
      for (int j = k ; j < n-1; j++) {
	temp[mu] +=  A[k][j+1] * B[j+1][mu];
      }
      B[k][mu] -= temp[mu];
      B[k][mu] /= A[k][k];
    }
  }

  // compute determinant
  if ( n == 0 ) 
    cerr << "QRdet: n must be positive!"<<endl;
  gramdet = 1.0;
  for (int i = 0; i < n; i++){
    gramdet *= A[i][i];
  }
  DOUBLE signum = (n & 1) ? one : -one;
  gramdet *= signum;
}


#define DIM 4
void NGluon::VNVB(const int R, DOUBLE *PP[], const DOUBLE q[],
		   complex<DOUBLE> Vr[DIM], complex<DOUBLE> nt[DIM][DIM] ) {

  int n;
  if ( R-1 >= DIM ) 
    n = DIM;
  else 
    n = R-1;  

  DOUBLE gramMatrix[DIM][DIM];
  DOUBLE vPhys[DIM][DIM];
  DOUBLE gramCopy[DIM][DIM];

  for(int i = 0; i < n; i++) {
    gramMatrix[i][i] = S( PP[i] );
    gramCopy[i][i] = gramMatrix[i][i];
    for(int j = 0; j < i; j++) {
      gramMatrix[i][j] = DOT(PP[i],PP[j]);
      gramMatrix[j][i] = gramMatrix[i][j];
      gramCopy[i][j] = gramMatrix[i][j];
      gramCopy[j][i] = gramMatrix[i][j];
    }
  }
  
  for (int i = 0; i < n; i++){
    for (int mu = 0; mu < DIM; mu++){
      vPhys[i][mu] = PP[i][mu];
    }
  }

  if ( n == 0 ) 
    gramDet = one;
  else if ( n == 1 ){
    gramDet = q[0];
    for(int mu=0;mu<4;mu++){
      vPhys[0][mu] /= q[0];
    }
  } 
  else {
    solveEqn(gramCopy, vPhys, n, gramDet);
  }
  
  for(int mu = 0; mu < DIM; mu++){
    Vr[mu] = zero;
  }
  for (int i=0; i<n; i++){
    for (int mu = 0; mu < DIM; mu++){
      Vr[mu] += (q[i+1]-q[i])*vPhys[i][mu]/two;
    }
  }
  
  if (R > DIM) return;

  // choose three random vectors to span the trivial spaces
  DOUBLE ref[DIM-1][DIM] = { {4.,1.,2.,3.},
			     {7.,0.,3.,4.},
			     {5.,2.,1.,1.} };
  
 DOUBLE* Mom[DIM-1];
 for(int i = 0; i < n; i++) {
   Mom[i] = PP[i];
 }

 for(int i = n; i < DIM-1; i++){
   Mom[i] = ref[i];
 }
 
  for(int i = n; i < DIM-1; i++){
    gramMatrix[i][i] = S(Mom[i]);
    for(int j = 0; j < i; j++) {
      gramMatrix[i][j] = DOT(Mom[i],Mom[j]);
      gramMatrix[j][i] = gramMatrix[i][j];
    }
  }
  
  complex<DOUBLE> rootGramDet[DIM-n];
  int count = 0;
  rootGramDet[count] = gramDet;  
  rootGramDet[count] = sqrt( rootGramDet[count] );

  for (int t = n+1; t < DIM; t++){
    count++;
    rootGramDet[count] = det(t,gramMatrix);
    rootGramDet[count] = sqrt( rootGramDet[count] );
    for (int mu = 0; mu < DIM; mu++) {
      for (int i = 0; i < t; i++) {
	for (int j = 0; j < t; j++) {
	  gramCopy[i][j] = gramMatrix[i][j];
	}
      }
      for (int i = 0; i < t; i++) {
	gramCopy[t-1][i]=Mom[i][mu];
      }
      nt[count-1][mu] = 
	det(t,gramCopy) / ( rootGramDet[count] * rootGramDet[count-1] );
    }
  }
  
  epsilon4DComponent(nt[count], Mom); 
  for (int mu = 0; mu < DIM; mu++){
    nt[count][mu] /= rootGramDet[count];
  }
} 
#undef DIM
//}}}

//{{{ output control
template<typename T> void NGluon::printMomentum(const T *mom){
  cout << "mom = ";
  for (int mu=0; mu<4; mu++){
    cout << mom[mu] << ",  ";
  }
  cout << endl;
}
template void NGluon::printMomentum(const DOUBLE *mom);
template void NGluon::printMomentum(const complex<DOUBLE> *mom);

void NGluon::printIntegralCoef(){
  Topology topo;
  for (topo.h3=3;  topo.h3 < ngluon;  topo.h3++){
    for (topo.h2=2;  topo.h2 < topo.h3;  topo.h2++){
      for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
	for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){
	  printBoxCoef(topo);
	}
      }
    }
  }
  for (topo.h2=2;  topo.h2 < ngluon;  topo.h2++){
    for (topo.h1=1;  topo.h1 < topo.h2;  topo.h1++){
      for (topo.h0=0;  topo.h0 < topo.h1;  topo.h0++){
	printTriCoef(topo);
      }
    }
  }
  for (topo.h0 = 0;  topo.h0 < ngluon-2;  topo.h0++){
    for (topo.h1 = topo.h0+2;  topo.h1 < ngluon;  topo.h1++){
      if( topo.h1-topo.h0 != ngluon-1){
	printBubCoef(topo);
      }
    }
  }
}

void NGluon::printBoxCoef(const Topology& topo, const string s){
  int key = C.boxkey(topo.h0,topo.h1,topo.h2,topo.h3);
  if (s == "fullbox" ){
    cout << "box"<<topo.h0<<topo.h1<<topo.h2<<topo.h3<<endl;

    for (int i =0; i<2; i++){
      cout <<"  C4["<<i<<"] = "<< C.box(key,i)<<endl;
    }
    return;
  }
  if (s == "rat"){
    cout << "boxR"<<topo.h0<<topo.h1<<topo.h2<<topo.h3<<endl;
    for (int i =0; i<5; i++){
      cout <<"  C4R["<<i<<"] = "<< C.boxR(key,i)<<endl;
    }
    return;
  }
  if (s == "cof" ){
    cout << "  C4["<<topo.h0<<topo.h1<<topo.h2<<topo.h3<<"] = "
	 <<  C.box(key,0)<<endl;
  }
}
void NGluon::printTriCoef(const Topology& topo, const string s){
  int key = C.trikey(topo.h0,topo.h1,topo.h2);
  if (s == "fulltri" ){
    cout << "tri"<<topo.h0<<"|"<<topo.h1<<"|"<<topo.h2<<endl;
    for (int i =0; i<7; i++){
      cout <<"  C3["<<i<<"] = "<< C.tri(key,i)<<endl;
    }
    return;
  }
  if (s == "rat"){
    cout << "triR"<<topo.h0<<"|"<<topo.h1<<"|"<<topo.h2<<endl;
    for (int i =0; i<10; i++){
      cout <<"  C3R["<<i<<"] = "<< C.triR(key,i)<<endl;
    }
    return;
  }
  if (s == "cof" ){
    cout << "  C3["<<topo.h0<<"|"<<topo.h1<<"|"<<topo.h2<<"]  = "
	 <<  C.tri(key,0)<<endl;
  }
}

void NGluon::printBubCoef(const Topology& topo, const string s){
  int key = C.bubkey(topo.h0,topo.h1);
  if (s == "rat"){
    cout << "bubR"<<topo.h0<<topo.h1 << " = "<< C.bubR(key,0)<<endl;
    return;
  }
  cout << "  C2["<<topo.h0<<topo.h1<<"]   = "
       <<  C.bub(key,0)<<endl;
}
//}}}
