/* $Modified: Wed Nov  3 15:20:59 2010 by uwer $ */
// vim: set ft=cpp:
// vim: set foldmethod=marker:
#ifndef ONELOOP_H_
#define ONELOOP_H_
#define NDEBUG 
#include <assert.h>
#include <iostream>
#include <complex>
#include <cstdlib>
#include <cmath>

#include "Current.h"
#include "Coefficients.h"
#include "LoopIntegrals.h"

#include "mytypes.h"

class NGluon {
 public:
  enum VERBOSITY {QUIET=0, FULL=1};
  
  NGluon( int ngluon_, DOUBLE k_[][4], int hel_[]) : 
    ngluon(ngluon_),  
    zero(0.),one(1.),two(2.),three(3.),four(4.),five(5.),six(6.),seven(7.),
    eight(8.),nine(9.), elevenThird(11./3.),
    PI(SETVAL(3.14159265358979323846264338327950288419716939937510582097494459230)),
    I(zero,one), 
    BGC(ngluon_,k_,hel_), C(ngluon_) {
    verbosity = FULL;
    if ( ngluon > NMAX ) 
      error("Recompile package with NMAX in mytypes.h set to ", ngluon);
#ifdef EXTENDED_PREC
    info("Extended precision (QD) is used!");
    if (one._ndigits == 31 ){
      info("Arithmetics is done using dd_real providing 31");
      info("decimal digits.");
    }
    if (one._ndigits == 62 ){
      info("Arithmetics is done using qd_real providing 62");
      info("decimal digits.");
    }
#endif
    /* Defaults as described in the documentation */
    scaleTest = false;
    muR = one;
    initFourier4D();
    setMassiveFModes(1,2,1);
  }
  
  std::complex<DOUBLE> evalAmp();
  std::complex<DOUBLE> getAtree();
  std::complex<DOUBLE> getAeps2();
  std::complex<DOUBLE> getAeps1();
  std::complex<DOUBLE> getAcc();
  std::complex<DOUBLE> getArat();
  std::complex<DOUBLE> getAfinite();
  std::complex<DOUBLE> getAbsError();
  std::complex<DOUBLE> getAbsErrorEps1();
  std::complex<DOUBLE> getAbsErrorEps2();
  std::complex<DOUBLE> getAbsErrorRR();
  DOUBLE getRelError();
  DOUBLE getRelErrorCC();
  DOUBLE getRelErrorRR();
  DOUBLE getRelErrorEps1();
  DOUBLE getRelErrorEps2();
  DOUBLE getIRCheck(int i);

  void setScaleTest(bool scaleTest_);
  void setMuR(DOUBLE muR_);

  void printIntegralCoef();

  void setVerbosity(VERBOSITY output_);
  std::complex<DOUBLE> IRpoles(const int eps);
  static int CheckPhaseSpacePoint(int n, DOUBLE p[][4]);

  static std::string int2string(const int number);
  static std::string helicity2string(const int n, int hel[]);
  static void sethel(int n, int htype, int *helicity);
  static void sethel(int n, std::string hstr, int *helicity);

 private: 
  const int ngluon;
  int fmTri, fmBub;
  DOUBLE cmEnergy, gramDet, relError, relErrorCC, relErrorRR, relErrorEps1,
    relErrorEps2,IRCheckm1, IRCheckm2;
  DOUBLE mFModesBox, mFModesTri, mFModesBub;
  const DOUBLE zero,one,two,three,four,five,six,seven,eight,nine,elevenThird,PI;
  const std::complex<DOUBLE> I;
  std::complex<DOUBLE> Atree, Acc, Arat, Afinite, Aeps2, Aeps1, absError,
    absErrorCC, absErrorRR, absErrorEps1, absErrorEps2;

  std::complex<DOUBLE> alphaTriR[8][2], alphaTriC[8][2],
                       fBaseTriR[8][7], fBaseTriC[8][7];
  DOUBLE alphaBub[15][3], fBaseBub[15];

  std::complex<DOUBLE> sMBox[14], sMTri[4], sMBub[4];
  DOUBLE sign[14];
  std::complex<DOUBLE> fBaseBoxRat[14][3];

  Current BGC;
  Coefficients C;

  VERBOSITY verbosity;
  bool scaleTest;
  DOUBLE muR;

  int FourierNumRBox;
  int FourierNumRTri;
  int FourierNumRBub;

  NGluon(const NGluon&);
  void initFourier4D();
  void setMassiveFModes(int box, int tri, int bub);

  void setScalarMass(const int p, std::complex<DOUBLE>* scalarMass);
  void updateCMEnergy();

  void info(std::string s);
  void info(std::string s, DOUBLE tmp);
  void info(std::string s, int tmp);
  void error(std::string s);
  void error(std::string s, int tmp);

  template<typename T> void printMomentum(const T *mom);

  void printBoxCoef(const Topology& topo, const std::string s = "cof");
  void printTriCoef(const Topology& topo, const std::string s = "cof");
  void printBubCoef(const Topology& topo, const std::string s = "cof");

  std::complex<DOUBLE> boxProduct(const Topology &topo);
  std::complex<DOUBLE> triangleProduct(const Topology &topo);
  std::complex<DOUBLE> bubbleProduct(const Topology &topo);
  
  std::complex<DOUBLE> pentagonProductS(const Topology &topo);
  std::complex<DOUBLE> boxProductS(const Topology &topo);
  std::complex<DOUBLE> triangleProductS(const Topology &topo);
  std::complex<DOUBLE> bubbleProductS(const Topology &topo);
  
  void boxes();
  void triangles();
  void bubbles();
  
  void pentagonsR();
  void boxesR();
  void trianglesR();
  void bubblesR();

  std::complex<DOUBLE> boxSub(const int i, const int j,
			      const int k, const int ref,
			      std::complex<DOUBLE> *loop1);
  std::complex<DOUBLE> boxTriSub(const int i, const int j, const int ref,
				 std::complex<DOUBLE> *loop1);
  std::complex<DOUBLE> cBar(int i, int j, int k, 
			    const std::complex<DOUBLE> *loop1);
  std::complex<DOUBLE> pentSubR(const int l, const int i, const int j,
				const int k, const int ref,
				const std::complex<DOUBLE> *loop1,
				const std::complex<DOUBLE> scalarMSq);
  void boxPentSubR(const int i, const int j, const int k, const int ref,
		   const std::complex<DOUBLE> *loop1,
		   const std::complex<DOUBLE> scalarMassSquared,
		   std::complex<DOUBLE> &ST4, std::complex<DOUBLE> &ST5);
  void triBoxPentSubR(const int i, const int j, const int ref,
		      const std::complex<DOUBLE> *loop1,
		      const std::complex<DOUBLE> scalarMSq,
		      std::complex<DOUBLE> &ST3, std::complex<DOUBLE> &ST4,
		      std::complex<DOUBLE> &ST5);
  std::complex<DOUBLE> dBarR(const int i, const int j, const int k, 
			     const int l, const std::complex<DOUBLE> *loop1,
			     const std::complex<DOUBLE> scalarMSq );
  std::complex<DOUBLE> cBarR(const int i, const int j, const int k,
			     const std::complex<DOUBLE> *loop1,
			     const std::complex<DOUBLE> scalarMSq );
  
  DOUBLE pentMaxM(const int i, const int j, const int k, const int l);
  DOUBLE boxPentMaxM(const int i, const int j, const int k);
  DOUBLE triBoxPentMaxM(const int i, const int j);
  
  template<typename T> T S(const T p[4]){
    return(p[0]*p[0] - ( p[1]*p[1] + p[2]*p[2] + p[3]*p[3] ));
  }
  template<typename T> T S3(const T p[4]){
    return( p[1]*p[1] + p[2]*p[2] + p[3]*p[3]);
  }
  template<typename T, typename U> 
    std::complex<DOUBLE> DOT(const T p1[4], const U p2[4]){
    return  ( p1[0]*p2[0] - ( p1[1]*p2[1] + p1[2]*p2[2] + p1[3]*p2[3] ) );
  }
  template<typename T> 
    T DOT(const T p1[4], const T p2[4]){
    return  ( p1[0]*p2[0] - ( p1[1]*p2[1] + p1[2]*p2[2] + p1[3]*p2[3] ) );
  }
  template<typename T>
    std::complex<DOUBLE> pp(const T p[4]){ return (p[0]+p[2]);}
  template<typename T>
    std::complex<DOUBLE> pm(const T p[4]){ return (p[0]-p[2]);}
  template<typename T>
    std::complex<DOUBLE> pT(const T p[4]){ return (p[3]+I*p[1]);}
  template<typename T>
    std::complex<DOUBLE> pTb(const T p[4]){ return (p[3]-I*p[1]);}
  std::complex<DOUBLE> SpNorm(const DOUBLE p[4]){
    return sqrt(abs(pp(p)))/pTb(p)*exp(I*arg(pp(p))/two);
  }
  std::complex<DOUBLE> SpNorm(const std::complex<DOUBLE> p[4]){
    return sqrt(abs(pp(p)))/pTb(p)*exp(I*arg(pp(p))/two);
  }
  std::complex<DOUBLE> AB(const DOUBLE p1[4], const DOUBLE p2[4],
			  const DOUBLE p3[4]);
  std::complex<DOUBLE> AB3(const DOUBLE p1[4], const DOUBLE p2[4],
			   const DOUBLE p3[4], const DOUBLE p4[4],
			   const DOUBLE p5[4]);
  template<typename T,typename U>
    void eta_mu(std::complex<DOUBLE> eta[4], const T a[4], const U b[4]){
    std::complex<DOUBLE> t1, t2, t3, t4, t5, norm;
    t1 = pp(a);
    t2 = pm(a);
    t3 = pTb(a);
    t4 = pp(b);
    t5 = pTb(b);
    norm = SpNorm(a)/SpNorm(b)/t5;
    eta[0] =   (t2*t5+t4*t3)  * norm;
    eta[3] =   (t4*t2+t3*t5)  * norm;
    eta[2] =   (-t2*t5+t4*t3) * norm;
    eta[1] = I*(-t4*t2+t3*t5) * norm;
  }
  
  void solveEqn(DOUBLE A[4][4],DOUBLE B[4][4],const int n, DOUBLE& gramdet);
  void epsilon4DComponent(std::complex<DOUBLE> nt[4], DOUBLE* P[]);
  template<typename T> T det(int n, T b[4][4]);
  void pentBasis(DOUBLE* PP[4], Topology& topo);
  void VNVB(const int R, DOUBLE *PP[], const DOUBLE q[],
	    std::complex<DOUBLE> Vr[4],
	    std::complex<DOUBLE> nt[4][4] );

};

#endif // ONELOOP_H_
