// $Modified: Mon Oct 11 23:45:04 2010 by uwer $
#include "analytic.h"

#define ZERO DOUBLE(0.)
#define ONE DOUBLE(1.)
#define TWO DOUBLE(2.)
#define THREE DOUBLE(3.)
#define FOUR DOUBLE(4.)
#define FIVE DOUBLE(5.)
#define SIX DOUBLE(6.)
#define SEVEN DOUBLE(7.)
#define EIGHT DOUBLE(8.)
#define NINE DOUBLE(9.)
#define TEN DOUBLE(10.)
#define ELEVEN DOUBLE(11.)
#define PI DOUBLE(SETVAL(3.14159265358979323846264338327950288419716939937510582097494459230))
#define I std::complex<DOUBLE>(ZERO,ONE)

static FourMomentum<DOUBLE> KK(int i,int j,FourMomentum<DOUBLE> *P){
 FourMomentum<DOUBLE> ans(ZERO,ZERO,ZERO,ZERO);
 for(int k=i-1;k<j;k++){
   ans=ans+P[k];
 }
 return ans;
}

static std::complex<DOUBLE> Ch(int n,FourMomentum<DOUBLE> *P,std::complex<DOUBLE> tree){

  std::complex<DOUBLE> ans=ZERO;
  std::complex<DOUBLE> tr1;
  std::complex<DOUBLE> tr2;
  std::complex<DOUBLE> tr3;
  DOUBLE s1;
  DOUBLE s2;
  DOUBLE r;
  FourMomentum<DOUBLE> P1; 
  FourMomentum<DOUBLE> P2;
  
  for(int m=4;m<n;m++){
    
    P1=KK(2,m-1,P);
    P2=KK(2,m,P);
    
    tr1=-B(P[0],P[1])*A(P[1],P[m-1])*BA(P[m-1],P1,P[0]);
    tr2=-B(P[0],P[1])*AB(P[1],P1,P[m-1])*A(P[m-1],P[0]);
    tr3=tr1-tr2;
    s1=S(P1);
    s2=S(P2);
    r=s1/s2;
    
    ans+=-(r-1/r)/TWO/(1-r)/(1-r)/(1-r)/s2/s2/s2*tr1*tr2*tr3;
  }
  
  return -pow(S(P[0]+P[1]),-THREE)/THREE*tree*ans;
  
}

static std::complex<DOUBLE> f1(int n,int l,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  FourMomentum<DOUBLE> P6;
  FourMomentum<DOUBLE> P7;

  P1=KK(i1+2,i2,P);
  P2=KK(2,i1+1,P);
  P3=KK(2,i2,P);
  P4=KK(i1+2,i2-1,P);
  P5=KK(2,i2-1,P);
  P6=KK(l,l+1,P);
  P7=KK(i1+2,l,P);
	
  if(l==i2){
    return -pow(S(P1),TWO)*AA(P[0],P1,P2,P[0])*AA(P[0],P3,P4,P[i2-1])*BA(P[i2-1],P5,P[0])*IAA(P[0],P2,P4,P[i2-1]);
  }
  else{
    return AA(P[0],P3,P1,P[l])*AA6(P[0],P3,P1,P6,P7,P1,P3,P[0])*IA(P[l-1],P[l]);
  }
}

static std::complex<DOUBLE> f2(int n,int l,int p,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  FourMomentum<DOUBLE> P6;

  P1=KK(i1+2,i2,P);
  P2=KK(2,i1+1,P);
  P3=KK(i2+1,n,P);
  P4=KK(l,i2,P);
  P5=KK(i1+2,l-1,P);
  P6=KK(l,p,P);

  if(p==i2){
    return AA(P[i2-1],P1,P2,P[0])/S(P1)/AA4(P[0],P3,P4,P5,P2,P[0]);
  }
  else{
    return A(P[p-1],P[p])/AA4(P[0],P3,P1,P5,P6,P[p]);
  }

}

static std::complex<DOUBLE> f3(int n,int l,int p,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  FourMomentum<DOUBLE> P6;

  P1=KK(i1+2,i2,P);
  P2=KK(2,i1+1,P);
  P3=KK(i2+1,n,P);
  P4=KK(i2+1,p,P);
  P5=KK(p+1,n,P);
  P6=KK(l,p,P);

  if(l==i2+1){
    return AA(P[0],P2,P1,P[i2])/AA4(P[0],P5,P4,P1,P2,P[0]);
  }
  else{
    return A(P[l-2],P[l-1])/AA(P[0],P5,P6,P[l-2]);
  }

}

static std::complex<DOUBLE> C1(int n,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;

  P1=KK(i2+1,n,P);
  P2=KK(i1+2,i2,P);
  P3=KK(i1+3,i2,P);

  return A(P[i1],P[i1+1])/AA(P[0],P1,P2,P[i1])/AA(P[0],P1,P3,P[i1+1]);
}

static std::complex<DOUBLE> C2(int n,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;

  P1=KK(2,i1+1,P);
  P2=KK(i1+2,i2,P);
  P3=KK(i1+2,i2-1,P);

  return A(P[i2-1],P[i2])*C1(n,i1,i2,P)/S(P2)/AA(P[0],P1,P2,P[i2])/AA(P[0],P1,P3,P[i2-1]);
}

static std::complex<DOUBLE> C3(int n,int i1,int i2,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  
  P1=KK(i1+2,i2,P);

  return pow(S(P1),TWO)*pow(S(P1),TWO)*C2(n,i1,i2,P);
}

static std::complex<DOUBLE> T1(int n,int i1,int i2,int j,FourMomentum<DOUBLE> *P){

  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  FourMomentum<DOUBLE> P6;
  std::complex<DOUBLE> ans;


  P1=KK(i2+1,n,P);
  P2=KK(i1+2,i2,P);
  P3=KK(2,i2,P);
  P4=KK(2,j-1,P);
  P5=KK(i1+2,j-1,P);
  P6=KK(2,i1+1,P);

  ans=S(P2)*A(P[0],P[j-1])*AA(P[0],P1,P2,P[j-1])
    *( AA(P[0],P3,P2,P[j-1])*BA(P[j-1],P4,P[0])-AB3(P[0],P3,P2,P5,P[j-1])*A(P[j-1],P[0]) )
    *IAA(P[0],P6,P2,P[j-1])*IAA(P[0],P6,P2,P[j-1])/TWO;
  return ans;
}

static std::complex<DOUBLE> T2a(int n,int i1,int i2,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  std::complex<DOUBLE> ans=ZERO;
  
  P1=KK(i2+1,n,P);
  P2=KK(i1+2,i2,P);

  for(int l=i1+2;l<i2+1;l++){
	ans+=AA(P[0],P1,P2,P[l-1])*f1(n,l,i1,i2,P);
  }
  return ans;
}

static std::complex<DOUBLE> T2b(int n,int i1,int i2,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  FourMomentum<DOUBLE> P6;
  FourMomentum<DOUBLE> P7;
  std::complex<DOUBLE> ABF=ZERO;
  std::complex<DOUBLE> ans=ZERO;

  P1=KK(i2+1,n,P);
  P2=KK(i1+2,i2,P);
  P5=KK(2,i2,P);


  for(int l=i1+3;l<i2;l++){
    for(int p=l+1;p<i2+1;p++){
      P3=KK(l,p,P);
      P4=KK(i1+2,p,P);
      P6=KK(i1+2,l-1,P);
      P7=KK(l+1,p,P);

      ABF=ZERO;
      for(int m1=l;m1<p;m1++){
         for(int m2=m1+1;m2<p+1;m2++){
           for(int m3=l;m3<p;m3++){
	     for(int m4=m3+1;m4<p+1;m4++){
	       ABF+=AB3(P[0],P5,P2,P6,P[m1-1])*A(P[m1-1],P[m2-1])*B(P[m2-1],P[m3-1])*A(P[m3-1],P[m4-1])*BA3(P[m4-1],P4,P2,P5,P[0]);
	     }
	   }
	 }
      }
      std::complex<DOUBLE> tmp = ONE;
      tmp *= AA6(P[0],P1,P2,P3,P4,P2,P5,P[0])/AA4(P[0],P1,P2,P6,P3,P[p-1]);
      tmp *= AA6(P[0],P1,P2,P3,P4,P2,P5,P[0])/AA4(P[0],P1,P2,P4,P3,P[l-2]);
      tmp *= AA6(P[0],P1,P2,P3,P4,P2,P5,P[0])/AA4(P[0],P1,P2,P4,P7,P[l-1]);
      tmp *= f2(n,l,p,i1,i2,P);
      tmp /= S(P3);
      tmp *= A(P[l-2],P[l-1]);
      tmp *= ABF;
      ans += tmp;
    }
  }
  return -ans;
}

static std::complex<DOUBLE> T3a(int n,int i1,int i2,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  std::complex<DOUBLE> ans=ZERO;

  for(int l=i2+1;l<n;l++){
    P1=KK(l,l+1,P);
    P2=KK(l+1,n,P);
    
    ans+=A(P[0],P[l-1])*A(P[0],P[l])*AA(P[0],P1,P2,P[0])/A(P[l-1],P[l]);
  }

  return ans;
}

static std::complex<DOUBLE> T3b(int n,int i1,int i2,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  std::complex<DOUBLE> ans=ZERO;

  P1=KK(2,i2,P);
  P2=KK(2,i1+1,P);
  P3=KK(i1+2,i2,P);

  ans=AB(P[0],P1,P[i2])*AA(P[0],P2,P1,P[0])*A(P[0],P[i2])*A(P[0],P[i2])/AA(P[0],P2,P3,P[i2]);
  return ans;

}

static std::complex<DOUBLE> T3c(int n,int i1,int i2,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  FourMomentum<DOUBLE> P2;
  FourMomentum<DOUBLE> P3;
  FourMomentum<DOUBLE> P4;
  FourMomentum<DOUBLE> P5;
  std::complex<DOUBLE> ABF=ZERO;
  std::complex<DOUBLE> ans=ZERO;

  for(int l=i2+1;l<n-1;l++){
    for(int p=l+1;p<n;p++){

  P1=KK(l,p,P);
  P2=KK(p+1,n,P);
  P3=KK(l+1,p,P);
  P4=KK(2,l-1,P);
  P5=KK(l,p-1,P);

  ABF=ZERO;
  for(int m1=l;m1<p;m1++){
    for(int m2=m1+1;m2<p+1;m2++){
      for(int m3=l;m3<p;m3++){
	for(int m4=m3+1;m4<p+1;m4++){
	  ABF+=AB(P[0],P4,P[m1-1])*A(P[m1-1],P[m2-1])*B(P[m2-1],P[m3-1])*A(P[m3-1],P[m4-1])*BA(P[m4-1],P2,P[0]);
	}
      }
    }
  }

    ans+=AA(P[0],P1,P2,P[0])*AA(P[0],P1,P2,P[0])*AA(P[0],P1,P2,P[0])*IAA(P[0],P2,P3,P[l-1])*
    A(P[p-1],P[p])*ABF*f3(n,l,p,i1,i2,P)
    /S(P1)/AA(P[0],P4,P5,P[p-1])/AA(P[0],P4,P1,P[p]);
    }
  }

  return ans;
}

static std::complex<DOUBLE> T4(int n,int i1,FourMomentum<DOUBLE> *P){
  FourMomentum<DOUBLE> P1;
  std::complex<DOUBLE> ans=ZERO;

  P1=KK(2,i1+1,P);

  ans=-B(P[i1+1],P[i1+2])*A(P[i1+2],P[0])/AB(P[0],P1,P[i1+1])/TWO;
  return ans;
}

static std::complex<DOUBLE> Rh(int n,FourMomentum<DOUBLE> *P,std::complex<DOUBLE> tree){

std::complex<DOUBLE> ans=ZERO;

for(int i1=1;i1<n-3;i1++){
  for(int i2=i1+3;i2<n;i2++){
    ans+=
      +C1(n,i1,i2,P)*( T1(n,i1,i2,i2,P)+T1(n,i1,i2,i2+1,P) )
      +C2(n,i1,i2,P)*( T2a(n,i1,i2,P)+T2b(n,i1,i2,P) )
      +C3(n,i1,i2,P)*( T3a(n,i1,i2,P)+T3b(n,i1,i2,P)+T3c(n,i1,i2,P) );
  }
  ans+=T4(n,i1,P);
}

return ONE/THREE*tree*ans;

}


std::complex<DOUBLE> RMHV(int n,FourMomentum<DOUBLE> *P,std::complex<DOUBLE> tree){

  return Ch(n,P,tree)+Rh(n,P,tree)+TWO/NINE*tree;

}


std::complex<DOUBLE> R6pppppp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,
			      FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,
			      FourMomentum<DOUBLE> p5,FourMomentum<DOUBLE> p6)
{
  {
    return(-A(p1,p3)*A(p4,p6)*B(p1,p6)*B(p3,p4)*IA(p1,p2)*IA(p1,p6)*IA(p2,p3)*
IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*I/THREE-A(p1,p3)*B(p1,p5)*B(p3,p4)*IA(p1,p2)*IA(p1
,p6)*IA(p2,p3)*IA(p3,p4)*IA(p5,p6)*I/THREE-A(p1,p3)*B(p1,p6)*B(p3,p5)*IA(p1,p2)*
IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*I/THREE-A(p1,p4)*B(p1,p6)*B(p4,p5)*IA(p1
,p2)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*I/THREE-A(p2,p4)*B(p2,p6)*B(p4,p5)*
IA(p1,p2)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*I/THREE-A(p3,p5)*B(p1,p5)*B(p2
,p3)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*I/THREE-A(p3,p6)*B(p1,p6)
*B(p2,p3)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*I/THREE-A(p4,p6)*B(
p1,p6)*B(p2,p4)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*I/THREE-A(p4,
p6)*B(p2,p6)*B(p3,p4)*IA(p1,p2)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*I/THREE-
B(p1,p4)*B(p2,p3)*IA(p1,p6)*IA(p2,p3)*IA(p4,p5)*IA(p5,p6)*I/THREE-B(p1,p5)*B(p2,
p4)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p5,p6)*I/THREE-B(p1,p6)*B(p2,p5)*IA(p1,p6)*
IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*I/THREE-B(p2,p5)*B(p3,p4)*IA(p1,p2)*IA(p1,p6)*IA(
p3,p4)*IA(p5,p6)*I/THREE-B(p2,p6)*B(p3,p5)*IA(p1,p2)*IA(p1,p6)*IA(p3,p4)*IA(p4,
p5)*I/THREE-B(p3,p6)*B(p4,p5)*IA(p1,p2)*IA(p1,p6)*IA(p2,p3)*IA(p4,p5)*I/THREE);
  }
}


std::complex<DOUBLE> R6mppppp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,
			      FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,
			      FourMomentum<DOUBLE> p5,FourMomentum<DOUBLE> p6)
{
  {
    return(I*(pow(AB(p1,p2+p3,p6),THREE)*IA(p1,p2)*IA(p2,p3)*pow(IA(p4,p5),TWO)*
IAB(p3,p1+p2,p6)*IS(p1+p2+p3)+pow(AB(p1,p3+p4,p2),THREE)*pow(IA(p3,p4),TWO)*IA(p5
,p6)*IA(p6,p1)*IAB(p5,p3+p4,p2)*IS(p2+p3+p4)+pow(B(p2,p6),THREE)*IB(p1,p2)*IB(p6,
p1)*IS(p3+p4+p5)*(B(p2,p3)*B(p3,p4)*IA(p4,p5)*IAB(p5,p3+p4,p2)-B(p4,p5)*B(p5,p6
)*IA(p3,p4)*IAB(p3,p1+p2,p6)+B(p3,p5)*IA(p3,p4)*IA(p4,p5))-pow(A(p1,p3),THREE)*B(
p2,p3)*A(p2,p4)*pow(IA(p2,p3),TWO)*pow(IA(p3,p4),TWO)*IA(p4,p5)*IA(p5,p6)*IA(p6
,p1)+pow(A(p1,p5),THREE)*A(p4,p6)*B(p5,p6)*IA(p1,p2)*IA(p2,p3)*IA(p3,p4)*pow(IA(
p4,p5),TWO)*pow(IA(p5,p6),TWO)-pow(A(p1,p4),THREE)*A(p3,p5)*AB(p1,p2+p3,p4)*IA(p1
,p2)*IA(p2,p3)*pow(IA(p3,p4),TWO)*pow(IA(p4,p5),TWO)*IA(p5,p6)*IA(p6,p1))/THREE);
  }
}


std::complex<DOUBLE> R6mmpppp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,
			      FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,
			      FourMomentum<DOUBLE> p5,FourMomentum<DOUBLE> p6)
{
  {
    return(-pow(B(p3,p6),THREE)*pow(IA(p4,p5),TWO)*IB(p1,p2)*IB(p1,p6)*IB(p2,p3)*
I/THREE-TWO/NINE*pow(A(p1,p2),THREE)*IA(p1,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,
p6)*I+pow(A(p1,p2),THREE)*pow(A(p1,p4),TWO)*A(p2,p3)*A(p2,p4)*pow(B(p1,p2),THREE)*
pow(B(p3,p4),TWO)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*AB(p1,p2+p3,p4)*pow(
IAB(p4,p2+p3,p4),TWO)*pow(IS(p1+p2),THREE)*IS(p2+p3)*I/SIX+pow(A(p1,p2),THREE)*pow
(A(p1,p4),TWO)*A(p2,p3)*A(p2,p4)*pow(B(p1,p2),THREE)*pow(B(p3,p4),TWO)*IA(p1,p6)*
IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*AB(p1,p2+p3,p4)*pow(IAB(p4,p2+p3,p4),TWO)*pow(IS(
p1+p2),THREE)*IS(p2+p3+p4)*I/SIX+pow(A(p1,p2),THREE)*A(p1,p4)*pow(A(p2,p4),TWO)*
pow(B(p1,p2),THREE)*B(p3,p4)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*pow(AB(p1,p2
+p3,p4),TWO)*pow(IAB(p4,p2+p3,p4),TWO)*pow(IS(p1+p2),THREE)*IS(p2+p3)*I/SIX+pow(
A(p1,p2),THREE)*A(p1,p4)*pow(A(p2,p4),TWO)*pow(B(p1,p2),THREE)*B(p3,p4)*IA(p1,p6)*
IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*pow(AB(p1,p2+p3,p4),TWO)*pow(IAB(p4,p2+p3,p4),TWO
)*pow(IS(p1+p2),THREE)*IS(p2+p3+p4)*I/SIX+pow(A(p1,p2),THREE)*pow(A(p1,p5),TWO)*A(
p2,p5)*pow(B(p1,p2),THREE)*B(p5,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*pow(
AB(p2,p1+p6,p5),TWO)*pow(IAB(p5,p1+p6,p5),TWO)*pow(IS(p1+p2),THREE)*IS(p1+p6)*I/
SIX+pow(A(p1,p2),THREE)*pow(A(p1,p5),TWO)*A(p2,p5)*pow(B(p1,p2),THREE)*B(p5,p6)*IA(
p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*pow(AB(p2,p1+p6,p5),TWO)*pow(IAB(p5,p1+p6,
p5),TWO)*pow(IS(p1+p2),THREE)*IS(p2+p3+p4)*I/SIX-pow(A(p1,p2),THREE)*A(p1,p5)*A(p1
,p6)*pow(A(p2,p5),TWO)*pow(B(p1,p2),THREE)*pow(B(p5,p6),TWO)*IA(p2,p3)*IA(p3,p4)*
IA(p4,p5)*IA(p5,p6)*AB(p2,p1+p6,p5)*pow(IAB(p5,p1+p6,p5),TWO)*pow(IS(p1+p2),THREE
)*IS(p1+p6)*I/SIX-pow(A(p1,p2),THREE)*A(p1,p5)*A(p1,p6)*pow(A(p2,p5),TWO)*pow(B(
p1,p2),THREE)*pow(B(p5,p6),TWO)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*AB(p2,p1+
p6,p5)*pow(IAB(p5,p1+p6,p5),TWO)*pow(IS(p1+p2),THREE)*IS(p2+p3+p4)*I/SIX-pow(A(
p1,p2),THREE)*A(p3,p5)*B(p4,p6)*B(p5,p6)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*
IAB(p1,p2+p3,p4)*IAB(p3,p1+p2,p6)*I/SIX-pow(A(p1,p2),THREE)*pow(B(p3,p5),TWO)*IA
(p1,p6)*IA(p3,p4)*IA(p4,p5)*AB(p4,p3+p5,p4)*IAB(p2,p1+p6,p5)*IAB(p6,p1+p2,p3)*
IS(p1+p2+p6)*I/THREE+pow(A(p1,p2),THREE)*pow(B(p3,p5),TWO)*IA(p1,p6)*IA(p3,p4)*IA(
p4,p5)*IAB(p2,p1+p6,p5)*IAB(p6,p1+p2,p3)*I/THREE+pow(A(p1,p2),THREE)*pow(B(p4,p6),
TWO)*IA(p2,p3)*IA(p4,p5)*IA(p5,p6)*AB(p5,p4+p6,p5)*IAB(p1,p2+p3,p4)*IAB(p3,p1+
p2,p6)*IS(p1+p2+p3)*I/THREE-pow(A(p1,p2),TWO)*A(p1,p4)*B(p3,p4)*IA(p1,p6)*IA(p2,
p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*IB(p2,p3)*I/SIX-pow(A(p1,p2),TWO)*A(p1,p5)*A
(p3,p5)*pow(B(p3,p5),THREE)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IAB(p2,p1+p6,p5)*IAB(p5
,p3+p4,p2)*IAB(p6,p1+p2,p3)*I/THREE-pow(A(p1,p2),TWO)*A(p2,p5)*B(p5,p6)*IA(p1,p6
)*IA(p2,p3)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*IB(p1,p6)*I/SIX-A(p1,p2)*A(p1,p4)*B(
p3,p4)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*IB(p2,p3)*AB(p1,p2+p4,p3)*IS(p2+
p3+p4)*I/SIX-A(p1,p2)*A(p1,p5)*B(p3,p4)*B(p5,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)
*AA(p1,p5+p6,p3+p4,p5)*IAB(p1,p2+p3,p4)*IAB(p5,p3+p4,p2)*IS(p2+p3+p4)*I/SIX-A(
p1,p2)*A(p2,p4)*A(p3,p5)*pow(B(p3,p5),TWO)*B(p5,p6)*pow(IA(p3,p4),TWO)*IA(p4,p5
)*IB(p1,p6)*AB(p5,p1+p2,p6)*IAB(p2,p1+p6,p5)*IAB(p5,p3+p4,p2)*IAB(p6,p1+p2,p3)*
I/THREE+A(p1,p2)*B(p5,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*IB(p1,p6)*pow(AB(p5,p1+
p2,p6),TWO)*IAB(p3,p1+p2,p6)*IAB(p5,p3+p4,p2)*I/SIX+A(p1,p2)*IA(p1,p6)*IA(p3,
p4)*IA(p4,p5)*IA(p5,p6)*IB(p2,p3)*pow(AB(p1,p2+p4,p3),TWO)*IS(p2+p3+p4)*I/THREE+
pow(A(p1,p5),TWO)*A(p1,p6)*pow(B(p3,p4),THREE)*B(p5,p6)*pow(IA(p5,p6),TWO)*IB(p2,
p3)*IAB(p1,p2+p3,p4)*IAB(p5,p3+p4,p2)*IS(p2+p3+p4)*I/THREE-pow(A(p1,p5),TWO)*pow
(B(p3,p4),TWO)*B(p5,p6)*IA(p4,p5)*IA(p5,p6)*IB(p2,p3)*AB(p1,p2+p4,p3)*IAB(p1,p2
+p3,p4)*IAB(p5,p3+p4,p2)*IS(p2+p3+p4)*I/THREE+A(p3,p5)*B(p3,p5)*pow(IA(p3,p4),
TWO)*pow(IA(p4,p5),TWO)*IB(p1,p2)*IB(p1,p6)*AB(p4,p1+p2,p3)*AB(p4,p1+p2,p6)*AB(
p5,p1+p2,p6)*IAB(p5,p3+p4,p2)*IAB(p6,p1+p2,p3)*I/THREE+A(p3,p5)*B(p3,p6)*pow(IA(
p3,p4),TWO)*pow(IA(p4,p5),TWO)*IB(p1,p2)*IB(p1,p6)*pow(AB(p4,p1+p2,p6),TWO)*IAB
(p5,p3+p4,p2)*I/THREE-A(p3,p5)*IA(p1,p6)*IA(p3,p4)*IA(p4,p5)*IA(p5,p6)*IB(p2,p3)
*pow(AB(p1,p2+p4,p3),THREE)*IAB(p5,p3+p4,p2)*IS(p2+p3+p4)*I/THREE+B(p5,p6)*IA(p3,
p4)*pow(IA(p4,p5),TWO)*IA(p5,p6)*IB(p1,p2)*IB(p1,p6)*pow(AB(p5,p1+p2,p6),TWO)*
AA(p4,p3+p5,p1+p2,p5)*IAB(p3,p1+p2,p6)*IAB(p5,p3+p4,p2)*I/THREE);
  }
}


std::complex<DOUBLE> inv(std::complex<DOUBLE> x){
return ONE/x;
}

std::complex<DOUBLE> num(std::complex<DOUBLE> x){
return x;
}


std::complex<DOUBLE> FbtAB(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  DOUBLE t2;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t8;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    t2 = S1*t1;
    K1f = K1-t2*XX;
    t4 = YY0*YY0;
    t6 = A(px,XX);
    t8 = B(K1f,py);
    t11 = AB(px,K1f,py);
    t12 = AB(px,XX,py);
    t16 = A(px,K1f);
    t17 = B(XX,py);
    return(-t4*S1*t1*t6*t8+YY0*(t11-t2*t12)+t16*t17);
  }
}


std::complex<DOUBLE> FbtAA(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t8;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    K1f = K1-S1*t1*XX;
    t4 = YY0*YY0;
    t6 = A(px,XX);
    t8 = BA(K1f,py,pz);
    t11 = AA(px,K1f,py,pz);
    t12 = IAB(XX,K1f,XX);
    t14 = AA(px,XX,py,pz);
    t18 = A(px,K1f);
    t19 = BA(XX,py,pz);
    return(-t4*S1*t1*t6*t8+YY0*(t11-S1*t12*t14)+t18*t19);
  }
}


std::complex<DOUBLE> FbtBB(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t8;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    K1f = K1-S1*t1*XX;
    t4 = YY0*YY0;
    t6 = BA(pz,py,XX);
    t8 = B(K1f,px);
    t11 = BB(pz,py,K1f,px);
    t12 = t1;
    t14 = BB(pz,py,XX,px);
    t18 = BA(pz,py,K1f);
    t19 = B(XX,px);
    return(
	t4*S1*t1*t6*t8
	-YY0*(t11-S1*t12*t14)
	-t18*t19
	);
  }
}


std::complex<DOUBLE> FbtAB2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0,std::complex<DOUBLE> YY2
)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t9;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    K1f = K1-S1*t1*XX;
    t4 = A(px,K1f);
    t5 = B(K1f,py);
    t9 = A(px,XX);
    t11 = B(XX,py);
    return(t4*t5*YY2-S1*YY2*t1*t9*t11+(-t1-TWO*S1*YY0*YY2*t1)*t9*t5);
  }
}


std::complex<DOUBLE> FbtAA2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,
std::complex<DOUBLE> YY0,std::complex<DOUBLE> YY2)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t7;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    K1f = K1-S1*t1*XX;
    t4 = AA(px,K1f,py,pz);
    t7 = AA(px,XX,py,pz);
    t15 = A(px,XX);
    t17 = BA(K1f,py,pz);
    return(t4*YY2-S1*YY2*t1*t7+(-t1-TWO*S1*YY0*YY2*t1)*t15*t17);
  }
}


std::complex<DOUBLE> FbtBB2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,
std::complex<DOUBLE> YY0,std::complex<DOUBLE> YY2)
{
  FourMomentum<DOUBLE> K1f;
  DOUBLE S1;
  DOUBLE gg;
  DOUBLE t1;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t7;
  {
    S1 = S(K1);
    gg = TWO*dot(XX,K1);
    t1 = ONE/gg;
    K1f = K1-S1*t1*XX;
    t4 = BB(px,K1f,py,pz);
    t7 = BB(px,XX,py,pz);
    t15 = B(px,K1f);
    t17 = AB(XX,py,pz);
    return(t4*YY2-S1*YY2*t1*t7+(-t1-TWO*S1*YY0*YY2*t1)*t15*t17);
  }
}


std::complex<DOUBLE> IFbtAB(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtAB(px,py,K1,XX,YY0);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> IFbtAA(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0
)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtAA(px,py,pz,K1,XX,YY0);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> IFbtBB(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0
)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtBB(px,py,pz,K1,XX,YY0);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> IFbtAB2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,std::complex<DOUBLE> YY0,
std::complex<DOUBLE> YY2)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtAB2(px,py,K1,XX,YY0,YY2);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> IFbtAA2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,
std::complex<DOUBLE> YY0,std::complex<DOUBLE> YY2)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtAA2(px,py,pz,K1,XX,YY0,YY2);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> IFbtBB2(FourMomentum<DOUBLE> px,FourMomentum<DOUBLE> py,FourMomentum<DOUBLE> pz,FourMomentum<DOUBLE> K1,FourMomentum<DOUBLE> XX,
std::complex<DOUBLE> YY0,std::complex<DOUBLE> YY2)
{
  std::complex<DOUBLE> t1;
  {
    t1 = FbtBB2(px,py,pz,K1,XX,YY0,YY2);
    return(ONE/t1);
  }
}


std::complex<DOUBLE> Delta(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  {
    t1 = dot(p1,p2);
    t2 = t1*t1;
    t3 = dot(p1,p1);
    t4 = dot(p2,p2);
    return(-t2+t3*t4);
  }
}


std::complex<DOUBLE> IDelta(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2)
{
  std::complex<DOUBLE> t1;
  {
    t1 = Delta(p1,p2);
    return(ONE/t1);
  }
}



std::complex<DOUBLE> tr5(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4 ){
return AB3(p1,p2,p3,p4,p1)-AB3(p1,p4,p3,p2,p1);
}


std::complex<DOUBLE> C4123mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t8;
  {
    t1 = p2+p1;
    t2 = AA4(p1,p5,p4,p6,t1,p3);
    t3 = AA4(p1,p6,p4,p5,t1,p3);
    t5 = ONE/(t2-t3);
    t6 = A(p1,p3);
    t7 = t6*t6;
    t8 = t7*t7;
    t10 = IA(p1,p2);
    t12 = IA(p2,p3);
    t13 = IA(p4,p5);
    t14 = t13*t13;
    t16 = IA(p4,p6);
    t17 = IA(p5,p6);
    t18 = t17*t17;
    t22 = IB(p4,p5);
    t23 = IB(p4,p6);
    t25 = IB(p5,p6);
    t26 = t25*t25;
    t29 = IS(p3+p2+p1);
    t30 = AB3(p5,p5,p4,p6,p4);
    t31 = AB3(p5,p6,p4,p5,p4);
    t33 = num(t30-t31);
    t35 = AB3(p5,p5,p4,p6,p6);
    t36 = AB3(p5,p6,p4,p5,p6);
    t38 = num(t35-t36);
    t39 = t38*t38;
    t40 = t39*I;
    t44 = tr5(p5,p1,p6,p4);
    t46 = tr5(p5,p3,p6,p4);
    t52 = t16*t16;
    t55 = IB(p1,p2);
    t56 = IB(p2,p3);
    t60 = t22*t22;
    t61 = t23*t23;
    t64 = AB3(p1,p5,p4,p6,p2);
    t65 = AB3(p1,p6,p4,p5,p2);
    t67 = num(t64-t65);
    t68 = t67*t67;
    t71 = AB3(p3,p5,p4,p6,p2);
    t72 = AB3(p3,p6,p4,p5,p2);
    t74 = num(t71-t72);
    t75 = t74*t74;
    return(-TWO*t5*t8*t10*t12*t14*t16*t18*t22*t23*t26*t29*t33*t40+TWO/t44/t46*
t5*t14*t13*t52*t18*t17*t55*t56*t60*t61*t26*t25*t68*t75*t33*t40);
  }
}


std::complex<DOUBLE> C4234mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t18;
  FourMomentum<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  {
    t1 = B(p2,p6);
    t2 = p3+p2;
    t3 = AB(p1,t2,p4);
    t5 = AB(p6,p5,p1);
    t7 = B(p5,p6);
    t8 = AB(p5,p1,p2);
    t10 = AB(p6,t2,p4);
    t13 = ONE/(-t1*t3*t5+t7*t8*t10);
    t14 = B(p2,p4);
    t15 = t14*t14;
    t16 = t15*t15;
    t18 = IA(p1,p5);
    t20 = IA(p1,p6);
    t21 = t20*t20;
    t22 = IA(p5,p6);
    t24 = IB(p1,p5);
    t25 = IB(p1,p6);
    t26 = t25*t25;
    t30 = IB(p2,p3);
    t31 = IB(p3,p4);
    t33 = IB(p5,p6);
    t34 = t33*t33;
    t37 = IS(p4+p3+p2);
    t38 = AB3(p1,p1,p5,p6,p6);
    t39 = AB3(p1,p6,p5,p1,p6);
    t41 = num(-t38+t39);
    t42 = t41*t41;
    t44 = AB3(p5,p1,p5,p6,p6);
    t45 = AB3(p5,p6,p5,p1,p6);
    t47 = num(-t44+t45);
    t48 = t47*I;
    t52 = tr5(p2,p1,p5,p6);
    t54 = tr5(p4,p1,p5,p6);
    t57 = t18*t18;
    t61 = IA(p2,p3);
    t63 = IA(p3,p4);
    t64 = t22*t22;
    t68 = t24*t24;
    t74 = AB3(p3,p1,p5,p6,p2);
    t75 = AB3(p3,p6,p5,p1,p2);
    t77 = num(-t74+t75);
    t78 = t77*t77;
    t79 = AB3(p3,p1,p5,p6,p4);
    t80 = AB3(p3,p6,p5,p1,p4);
    t82 = num(-t79+t80);
    t83 = t82*t82;
    return(-TWO*t13*t16*t18*t21*t22*t24*t26*t30*t31*t34*t37*t42*t48+TWO/t52/t54
*t13*t57*t21*t20*t61*t63*t64*t68*t26*t25*t34*t33*t42*t78*t83*t48);
  }
}


std::complex<DOUBLE> C4345mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  {
    t1 = p4+p3;
    t2 = AB(p5,t1,p1);
    t3 = AA(p1,p6,p2,p3);
    t5 = AA4(p3,p1,p6,p2,t1,p5);
    t7 = ONE/(t2*t3+t5);
    t8 = A(p3,p5);
    t9 = t8*t8;
    t10 = t9*t9;
    t12 = IA(p1,p2);
    t13 = t12*t12;
    t15 = IA(p1,p6);
    t16 = t15*t15;
    t17 = IA(p2,p6);
    t19 = IA(p3,p4);
    t20 = IA(p4,p5);
    t24 = IB(p1,p2);
    t25 = t24*t24;
    t26 = IB(p1,p6);
    t28 = IB(p2,p6);
    t31 = IS(p5+p4+p3);
    t32 = AB3(p1,p1,p6,p2,p2);
    t33 = AB3(p1,p2,p6,p1,p2);
    t35 = num(t32-t33);
    t36 = t35*t35;
    t38 = AB3(p1,p1,p6,p2,p6);
    t39 = AB3(p1,p2,p6,p1,p6);
    t41 = num(t38-t39);
    t46 = tr5(p3,p1,p6,p2);
    t48 = tr5(p5,p1,p6,p2);
    t55 = t17*t17;
    t58 = t26*t26;
    t62 = t28*t28;
    t63 = IB(p3,p4);
    t65 = IB(p4,p5);
    t68 = AB3(p3,p1,p6,p2,p4);
    t69 = AB3(p3,p2,p6,p1,p4);
    t71 = num(t68-t69);
    t72 = t71*t71;
    t74 = AB3(p5,p1,p6,p2,p4);
    t75 = AB3(p5,p2,p6,p1,p4);
    t77 = num(t74-t75);
    t78 = t77*t77;
    return(TWO*t7*t10*t13*t16*t17*t19*t20*t25*t26*t28*t31*t36*t41*I-TWO/t46/t48
*t7*t13*t12*t16*t15*t55*t25*t24*t58*t62*t63*t65*t36*t41*t72*t78*I);
  }
}


std::complex<DOUBLE> C4456mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  {
    t1 = B(p2,p4);
    t2 = AB(p2,p1,p3);
    t4 = p5+p4;
    t5 = AB(p3,t4,p6);
    t7 = B(p3,p4);
    t8 = AB(p2,t4,p6);
    t10 = AB(p3,p1,p2);
    t13 = ONE/(t1*t2*t5-t7*t8*t10);
    t14 = B(p4,p6);
    t15 = t14*t14;
    t16 = t15*t15;
    t18 = IA(p1,p2);
    t20 = IA(p1,p3);
    t21 = IA(p2,p3);
    t22 = t21*t21;
    t24 = IB(p1,p2);
    t25 = t24*t24;
    t26 = IB(p1,p3);
    t30 = IB(p2,p3);
    t31 = t30*t30;
    t32 = IB(p4,p5);
    t34 = IB(p5,p6);
    t37 = IS(p6+p5+p4);
    t38 = AB3(p1,p2,p1,p3,p2);
    t39 = AB3(p1,p3,p1,p2,p2);
    t41 = num(t38-t39);
    t43 = AB3(p3,p2,p1,p3,p2);
    t44 = AB3(p3,p3,p1,p2,p2);
    t46 = num(t43-t44);
    t47 = t46*t46;
    t52 = tr5(p2,p1,p3,p4);
    t54 = tr5(p2,p1,p3,p6);
    t57 = t18*t18;
    t60 = t20*t20;
    t63 = IA(p4,p5);
    t64 = IA(p5,p6);
    t69 = t26*t26;
    t74 = AB3(p5,p2,p1,p3,p4);
    t75 = AB3(p5,p3,p1,p2,p4);
    t77 = num(t74-t75);
    t78 = t77*t77;
    t80 = AB3(p5,p2,p1,p3,p6);
    t81 = AB3(p5,p3,p1,p2,p6);
    t83 = num(t80-t81);
    t84 = t83*t83;
    return(-TWO*t13*t16*t18*t20*t22*t25*t26*t31*t32*t34*t37*t41*t47*I+TWO/t52/
t54*t13*t57*t60*t22*t21*t63*t64*t25*t24*t69*t31*t30*t41*t47*t78*t84*I);
  }
}


std::complex<DOUBLE> C4561mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t9;
  {
    t1 = p6+p5;
    t2 = AB(p1,t1,p3);
    t3 = AA(p3,p2,p4,p5);
    t5 = AB(p1,t1,p4);
    t6 = AA(p4,p2,p3,p5);
    t9 = ONE/(t2*t3-t5*t6);
    t10 = A(p1,p5);
    t11 = t10*t10;
    t12 = t11*t11;
    t14 = IA(p1,p6);
    t16 = IA(p2,p3);
    t17 = t16*t16;
    t18 = IA(p2,p4);
    t20 = IA(p3,p4);
    t21 = t20*t20;
    t22 = IA(p5,p6);
    t26 = IB(p2,p3);
    t27 = IB(p2,p4);
    t29 = IB(p3,p4);
    t30 = t29*t29;
    t33 = IS(p6+p5+p1);
    t34 = AB3(p3,p3,p2,p4,p2);
    t35 = AB3(p3,p4,p2,p3,p2);
    t37 = num(t34-t35);
    t39 = AB3(p3,p3,p2,p4,p4);
    t40 = AB3(p3,p4,p2,p3,p4);
    t42 = num(t39-t40);
    t43 = t42*t42;
    t48 = tr5(p3,p2,p4,p5);
    t50 = tr5(p3,p1,p4,p2);
    t56 = t18*t18;
    t59 = IB(p1,p6);
    t60 = t26*t26;
    t64 = t27*t27;
    t67 = IB(p5,p6);
    t68 = AB3(p1,p3,p2,p4,p6);
    t69 = AB3(p1,p4,p2,p3,p6);
    t71 = num(t68-t69);
    t72 = t71*t71;
    t76 = AB3(p5,p3,p2,p4,p6);
    t77 = AB3(p5,p4,p2,p3,p6);
    t79 = num(t76-t77);
    t80 = t79*t79;
    return(TWO*t9*t12*t14*t17*t18*t21*t22*t26*t27*t30*t33*t37*t43*I+TWO/t48/t50
*t9*t17*t16*t56*t21*t20*t59*t60*t64*t30*t29*t67*t72*t37*t43*t80*I);
  }
}


std::complex<DOUBLE> C4612mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  {
    t1 = B(p4,p6);
    t2 = AB(p4,p3,p5);
    t4 = p6+p1;
    t5 = AB(p5,t4,p2);
    t7 = B(p5,p6);
    t8 = AB(p4,t4,p2);
    t10 = AB(p5,p3,p4);
    t13 = ONE/(t1*t2*t5-t7*t8*t10);
    t14 = B(p2,p6);
    t15 = t14*t14;
    t16 = t15*t15;
    t18 = IA(p3,p4);
    t20 = IA(p3,p5);
    t21 = IA(p4,p5);
    t22 = t21*t21;
    t24 = IB(p1,p2);
    t25 = IB(p1,p6);
    t29 = IB(p3,p4);
    t30 = t29*t29;
    t31 = IB(p3,p5);
    t33 = IB(p4,p5);
    t34 = t33*t33;
    t37 = IS(p6+p2+p1);
    t38 = AB3(p3,p4,p3,p5,p4);
    t39 = AB3(p3,p5,p3,p4,p4);
    t41 = num(t38-t39);
    t43 = AB3(p5,p4,p3,p5,p4);
    t44 = AB3(p5,p5,p3,p4,p4);
    t46 = num(t43-t44);
    t47 = t46*t46;
    t48 = t47*I;
    t52 = tr5(p4,p3,p5,p6);
    t54 = tr5(p4,p2,p5,p3);
    t57 = IA(p1,p2);
    t60 = IA(p1,p6);
    t61 = t18*t18;
    t63 = t20*t20;
    t69 = t31*t31;
    t72 = AB3(p1,p4,p3,p5,p2);
    t73 = AB3(p1,p5,p3,p4,p2);
    t75 = num(t72-t73);
    t76 = t75*t75;
    t79 = AB3(p1,p4,p3,p5,p6);
    t80 = AB3(p1,p5,p3,p4,p6);
    t82 = num(t79-t80);
    t83 = t82*t82;
    return(TWO*t13*t16*t18*t20*t22*t24*t25*t30*t31*t34*t37*t41*t48+TWO/t52/t54*
t13*t57*t60*t61*t63*t22*t21*t30*t29*t69*t34*t33*t76*t83*t41*t48);
  }
}


std::complex<DOUBLE> C412x34mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t99;
  {
    t1 = tr5(p5,p2,p6,p1);
    t3 = tr5(p5,p4,p6,p3);
    t5 = ONE/t1/t3;
    t6 = A(p1,p5);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p3,p5);
    t10 = t9*t9;
    t11 = B(p2,p6);
    t12 = t11*t11;
    t14 = B(p4,p6);
    t15 = t14*t14;
    t16 = IA(p1,p2);
    t17 = t15*t16;
    t20 = IA(p3,p4);
    t21 = IA(p5,p6);
    t22 = t20*t21;
    t23 = IB(p1,p2);
    t24 = t22*t23;
    t25 = IB(p3,p4);
    t26 = IB(p5,p6);
    t27 = t25*t26;
    t28 = p2+p1;
    t29 = AB(p6,t28,p5);
    t30 = t29*t29;
    t38 = A(p3,p6);
    t40 = B(p4,p5);
    t41 = t40*t14;
    t46 = t16*t20*t21*t23;
    t47 = AB(p5,t28,p6);
    t51 = t46*t27*t47*t29*I;
    t54 = t38*t38;
    t56 = t40*t40;
    t57 = t56*t16;
    t60 = t47*t47;
    t63 = t24*t27*t60*I;
    t66 = A(p1,p6);
    t68 = t5*t6*t66;
    t69 = B(p2,p5);
    t83 = t23*t25;
    t95 = IAB(p6,t28,p5);
    t99 = t46*t27*t60*t47*t95*I;
    t102 = t66*t66;
    t103 = t5*t102;
    t104 = t69*t69;
    t121 = t60*t60;
    t123 = t95*t95;
    return(-TWO*t8*t10*t12*t17*t24*t27*t30*I+FOUR*t5*t7*t9*t38*t12*t41*t51-TWO*
t8*t54*t12*t57*t63+FOUR*t68*t10*t69*t11*t15*t51-EIGHT*t68*t9*t38*t69*t11*t40*t14*
t16*t22*t83*t26*t60*I+FOUR*t68*t54*t69*t11*t56*t99-TWO*t103*t10*t104*t17*t63+FOUR
*t5*t102*t9*t38*t104*t41*t99-TWO*t103*t54*t104*t57*t22*t83*t26*t121*t123*I);
  }
}


std::complex<DOUBLE> C423x45mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t99;
  {
    t1 = tr5(p1,p5,p6,p4);
    t3 = tr5(p1,p3,p6,p2);
    t5 = ONE/t1/t3;
    t6 = A(p1,p3);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p1,p5);
    t10 = t9*t9;
    t11 = B(p2,p6);
    t12 = t11*t11;
    t14 = B(p4,p6);
    t15 = t14*t14;
    t16 = IA(p1,p6);
    t17 = t15*t16;
    t20 = IA(p2,p3);
    t21 = IA(p4,p5);
    t22 = t20*t21;
    t23 = IB(p1,p6);
    t24 = t22*t23;
    t25 = IB(p2,p3);
    t26 = IB(p4,p5);
    t27 = t25*t26;
    t28 = p3+p2;
    t29 = AB(p6,t28,p1);
    t30 = t29*t29;
    t38 = A(p5,p6);
    t39 = B(p1,p4);
    t46 = t16*t20*t21*t23;
    t47 = AB(p1,t28,p6);
    t51 = t46*t27*t47*t29*I;
    t54 = t38*t38;
    t55 = t39*t39;
    t60 = t47*t47;
    t63 = t24*t27*t60*I;
    t68 = A(p3,p6);
    t69 = B(p1,p2);
    t85 = t23*t25;
    t99 = IAB(p6,t28,p1);
    t103 = t46*t27*t60*t47*t99*I;
    t107 = t68*t68;
    t108 = t69*t69;
    t128 = t60*t60;
    t130 = t99*t99;
    return(-TWO*t8*t10*t12*t17*t24*t27*t30*I+FOUR*t5*t7*t9*t38*t39*t12*t14*t51
-TWO*t8*t54*t55*t12*t16*t63+FOUR*t5*t6*t10*t68*t69*t11*t15*t51-EIGHT*t5*t6*t9*t68*
t38*t69*t39*t11*t14*t16*t22*t85*t26*t60*I+FOUR*t5*t6*t68*t54*t69*t55*t11*t103
-TWO*t5*t10*t107*t108*t17*t63+FOUR*t5*t9*t107*t38*t108*t39*t14*t103-TWO*t5*t107*
t54*t108*t55*t16*t22*t85*t26*t128*t130*I);
  }
}


std::complex<DOUBLE> C434x56mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t99;
  {
    t1 = tr5(p1,p4,p2,p3);
    t3 = tr5(p1,p6,p2,p5);
    t5 = ONE/t1/t3;
    t6 = A(p1,p3);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p1,p5);
    t10 = t9*t9;
    t11 = B(p2,p4);
    t12 = t11*t11;
    t14 = B(p2,p6);
    t15 = t14*t14;
    t16 = IA(p1,p2);
    t17 = t15*t16;
    t20 = IA(p3,p4);
    t21 = IA(p5,p6);
    t22 = t20*t21;
    t23 = IB(p1,p2);
    t24 = t22*t23;
    t25 = IB(p3,p4);
    t26 = IB(p5,p6);
    t27 = t25*t26;
    t28 = p4+p3;
    t29 = AB(p2,t28,p1);
    t30 = t29*t29;
    t38 = A(p2,p5);
    t39 = B(p1,p6);
    t46 = t16*t20*t21*t23;
    t47 = AB(p1,t28,p2);
    t51 = t46*t27*t47*t29*I;
    t54 = t38*t38;
    t55 = t39*t39;
    t60 = t47*t47;
    t63 = t24*t27*t60*I;
    t68 = A(p2,p3);
    t69 = B(p1,p4);
    t85 = t23*t25;
    t99 = IAB(p2,t28,p1);
    t103 = t46*t27*t60*t47*t99*I;
    t107 = t68*t68;
    t108 = t69*t69;
    t128 = t60*t60;
    t130 = t99*t99;
    return(-TWO*t8*t10*t12*t17*t24*t27*t30*I+FOUR*t5*t7*t9*t38*t39*t12*t14*t51
-TWO*t8*t54*t55*t12*t16*t63+FOUR*t5*t6*t10*t68*t69*t11*t15*t51-EIGHT*t5*t6*t9*t68*
t38*t69*t39*t11*t14*t16*t22*t85*t26*t60*I+FOUR*t5*t6*t68*t54*t69*t55*t11*t103
-TWO*t5*t10*t107*t108*t17*t63+FOUR*t5*t9*t107*t38*t108*t39*t14*t103-TWO*t5*t107*
t54*t108*t55*t16*t22*t85*t26*t128*t130*I);
  }
}


std::complex<DOUBLE> C445x61mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  {
    t1 = tr5(p2,p5,p3,p4);
    t3 = tr5(p2,p1,p3,p6);
    t5 = ONE/t1/t3;
    t6 = A(p1,p2);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p2,p5);
    t10 = t9*t9;
    t11 = B(p3,p4);
    t12 = t11*t11;
    t14 = B(p3,p6);
    t15 = t14*t14;
    t16 = IA(p1,p6);
    t17 = t15*t16;
    t20 = IA(p2,p3);
    t21 = IA(p4,p5);
    t22 = t20*t21;
    t23 = IB(p1,p6);
    t24 = IB(p2,p3);
    t25 = t23*t24;
    t27 = IB(p4,p5);
    t28 = p5+p4;
    t29 = AB(p3,t28,p2);
    t30 = t29*t29;
    t31 = t30*t30;
    t33 = IAB(p2,t28,p3);
    t34 = t33*t33;
    t42 = A(p3,p5);
    t43 = B(p2,p4);
    t44 = t42*t43;
    t50 = t16*t20*t21*t23;
    t51 = t24*t27;
    t56 = t50*t51*t30*t29*t33*I;
    t59 = t42*t42;
    t60 = t43*t43;
    t61 = t59*t60;
    t64 = t22*t23;
    t67 = t64*t51*t30*I;
    t70 = A(p1,p3);
    t72 = t5*t6*t70;
    t73 = B(p2,p6);
    t96 = AB(p2,t28,p3);
    t100 = t50*t51*t96*t29*I;
    t103 = t70*t70;
    t104 = t5*t103;
    t105 = t73*t73;
    t122 = t96*t96;
    return(-TWO*t8*t10*t12*t17*t22*t25*t27*t31*t34*I+FOUR*t5*t7*t9*t44*t11*t15*
t56-TWO*t8*t61*t17*t67+FOUR*t72*t10*t73*t12*t14*t56-EIGHT*t72*t9*t42*t43*t73*t11*
t14*t16*t22*t25*t27*t30*I+FOUR*t72*t61*t73*t14*t100-TWO*t104*t10*t105*t12*t16*
t67+FOUR*t5*t103*t9*t44*t105*t11*t100-TWO*t104*t61*t105*t16*t64*t51*t122*I);
  }
}


std::complex<DOUBLE> C456x12mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t99;
  {
    t1 = tr5(p3,p6,p4,p5);
    t3 = tr5(p3,p2,p4,p1);
    t5 = ONE/t1/t3;
    t6 = A(p1,p3);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p3,p5);
    t10 = t9*t9;
    t11 = B(p2,p4);
    t12 = t11*t11;
    t14 = B(p4,p6);
    t15 = t14*t14;
    t16 = IA(p1,p2);
    t17 = t15*t16;
    t20 = IA(p3,p4);
    t21 = IA(p5,p6);
    t22 = t20*t21;
    t23 = IB(p1,p2);
    t24 = t22*t23;
    t25 = IB(p3,p4);
    t26 = IB(p5,p6);
    t27 = t25*t26;
    t28 = p2+p1;
    t29 = AB(p4,t28,p3);
    t30 = t29*t29;
    t38 = A(p4,p5);
    t40 = B(p3,p6);
    t41 = t40*t14;
    t46 = t16*t20*t21*t23;
    t47 = AB(p3,t28,p4);
    t51 = t46*t27*t47*t29*I;
    t54 = t38*t38;
    t56 = t40*t40;
    t57 = t56*t16;
    t60 = t47*t47;
    t63 = t24*t27*t60*I;
    t66 = A(p1,p4);
    t68 = t5*t6*t66;
    t69 = B(p2,p3);
    t83 = t23*t25;
    t95 = IAB(p4,t28,p3);
    t99 = t46*t27*t60*t47*t95*I;
    t102 = t66*t66;
    t103 = t5*t102;
    t104 = t69*t69;
    t121 = t60*t60;
    t123 = t95*t95;
    return(-TWO*t8*t10*t12*t17*t24*t27*t30*I+FOUR*t5*t7*t9*t38*t12*t41*t51-TWO*
t8*t54*t12*t57*t63+FOUR*t68*t10*t69*t11*t15*t51-EIGHT*t68*t9*t38*t69*t11*t40*t14*
t16*t22*t83*t26*t60*I+FOUR*t68*t54*t69*t11*t56*t99-TWO*t103*t10*t104*t17*t63+FOUR
*t5*t102*t9*t38*t104*t41*t99-TWO*t103*t54*t104*t57*t22*t83*t26*t121*t123*I);
  }
}


std::complex<DOUBLE> C461x23mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t27;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t98;
  {
    t1 = tr5(p4,p1,p5,p6);
    t3 = tr5(p4,p3,p5,p2);
    t5 = ONE/t1/t3;
    t6 = A(p1,p4);
    t7 = t6*t6;
    t8 = t5*t7;
    t9 = A(p3,p4);
    t10 = t9*t9;
    t11 = B(p2,p5);
    t12 = t11*t11;
    t13 = t10*t12;
    t14 = B(p5,p6);
    t15 = t14*t14;
    t16 = IA(p1,p6);
    t17 = t15*t16;
    t20 = IA(p2,p3);
    t21 = IA(p4,p5);
    t22 = t20*t21;
    t23 = IB(p1,p6);
    t24 = IB(p2,p3);
    t25 = t23*t24;
    t27 = IB(p4,p5);
    t28 = p6+p1;
    t29 = AB(p5,t28,p4);
    t30 = t29*t29;
    t31 = t30*t30;
    t33 = IAB(p4,t28,p5);
    t34 = t33*t33;
    t42 = A(p3,p5);
    t43 = B(p2,p4);
    t44 = t42*t43;
    t50 = t16*t20*t21*t23;
    t51 = t24*t27;
    t56 = t50*t51*t30*t29*t33*I;
    t59 = t42*t42;
    t60 = t43*t43;
    t61 = t59*t60;
    t64 = t22*t23;
    t67 = t64*t51*t30*I;
    t70 = A(p1,p5);
    t72 = t5*t6*t70;
    t73 = B(p4,p6);
    t74 = t73*t14;
    t94 = AB(p4,t28,p5);
    t98 = t50*t51*t94*t29*I;
    t101 = t70*t70;
    t102 = t5*t101;
    t103 = t73*t73;
    t104 = t103*t16;
    t118 = t94*t94;
    return(-TWO*t8*t13*t17*t22*t25*t27*t31*t34*I+FOUR*t5*t7*t9*t44*t11*t15*t56
-TWO*t8*t61*t17*t67+FOUR*t72*t13*t74*t56-EIGHT*t72*t9*t42*t43*t73*t11*t14*t16*t22*
t25*t27*t30*I+FOUR*t72*t61*t74*t98-TWO*t102*t13*t104*t67+FOUR*t5*t101*t9*t44*t11*
t103*t98-TWO*t102*t61*t104*t64*t51*t118*I);
  }
}


std::complex<DOUBLE> C412x45mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t99;
  {
    t1 = p2+p1;
    t2 = tr5(p3,t1,p6,p1);
    t4 = tr5(p3,t1,p6,p4);
    t6 = ONE/t2/t4;
    t7 = A(p1,p3);
    t8 = t7*t7;
    t9 = t6*t8;
    t10 = A(p3,p5);
    t11 = t10*t10;
    t12 = B(p2,p6);
    t13 = t12*t12;
    t15 = B(p4,p6);
    t16 = t15*t15;
    t17 = IA(p1,p2);
    t18 = t16*t17;
    t21 = IA(p3,p6);
    t22 = IA(p4,p5);
    t23 = t21*t22;
    t24 = IB(p1,p2);
    t25 = t23*t24;
    t26 = IB(p3,p6);
    t27 = IB(p4,p5);
    t28 = t26*t27;
    t29 = AB(p6,t1,p3);
    t30 = t29*t29;
    t38 = A(p5,p6);
    t40 = B(p3,p4);
    t41 = t40*t15;
    t46 = t17*t21*t22*t24;
    t47 = AB(p3,t1,p6);
    t51 = t46*t28*t47*t29*I;
    t54 = t38*t38;
    t56 = t40*t40;
    t57 = t56*t17;
    t60 = t47*t47;
    t63 = t25*t28*t60*I;
    t66 = A(p1,p6);
    t68 = t6*t7*t66;
    t69 = B(p2,p3);
    t83 = t24*t26;
    t95 = IAB(p6,t1,p3);
    t99 = t46*t28*t60*t47*t95*I;
    t102 = t66*t66;
    t103 = t6*t102;
    t104 = t69*t69;
    t121 = t60*t60;
    t123 = t95*t95;
    return(TWO*t9*t11*t13*t18*t25*t28*t30*I-FOUR*t6*t8*t10*t38*t13*t41*t51+TWO*
t9*t54*t13*t57*t63-FOUR*t68*t11*t69*t12*t16*t51+EIGHT*t68*t10*t38*t69*t12*t40*t15*
t17*t23*t83*t27*t60*I-FOUR*t68*t54*t69*t12*t56*t99+TWO*t103*t11*t104*t18*t63-FOUR
*t6*t102*t10*t38*t104*t41*t99+TWO*t103*t54*t104*t57*t23*t83*t27*t121*t123*I);
  }
}


std::complex<DOUBLE> C423x56mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t1 = p3+p2;
    t2 = tr5(p1,t1,p4,p2);
    t4 = tr5(p1,t1,p4,p5);
    t6 = ONE/t2/t4;
    t7 = A(p1,p3);
    t8 = t7*t7;
    t9 = t6*t8;
    t10 = A(p1,p4);
    t11 = B(p2,p4);
    t12 = t11*t11;
    t14 = B(p4,p6);
    t15 = t14*t14;
    t18 = IA(p2,p3);
    t19 = IA(p5,p6);
    t20 = t18*t19;
    t21 = IB(p1,p4);
    t22 = t20*t21;
    t23 = IB(p2,p3);
    t24 = IB(p5,p6);
    t25 = t23*t24;
    t26 = AB(p5,t1,p1);
    t27 = t26*t26;
    t33 = A(p4,p5);
    t34 = t33*t33;
    t35 = B(p1,p6);
    t36 = t35*t35;
    t38 = IA(p1,p4);
    t42 = AB(p1,t1,p4);
    t43 = t42*t42;
    t46 = t22*t25*t43*I;
    t52 = t14*t38;
    t55 = t21*t23;
    t56 = t20*t55;
    t57 = AB(p1,t1,p1);
    t58 = t24*t57;
    t65 = t15*t38;
    t68 = t57*t57;
    t79 = t26*I;
    t81 = t56*t24*t42*t79;
    t88 = t19*t21;
    t94 = A(p1,p5);
    t96 = t6*t7*t94;
    t97 = A(p3,p4);
    t98 = t97*t33;
    t99 = B(p1,p2);
    t116 = t38*t18*t88;
    t136 = IAB(p4,t1,p1);
    t140 = t116*t25*t43*t42*t136*I;
    t143 = t94*t94;
    t145 = t97*t97;
    t146 = t99*t99;
    t165 = t43*t43;
    t167 = t136*t136;
    t173 = TWO*t9*t10*t12*t15*t22*t25*t27*I+TWO*t9*t34*t36*t12*t38*t46-FOUR*t6*
t8*t34*t35*t12*t52*t56*t58*t42*I+TWO*t9*t34*t12*t65*t22*t25*t68*I-FOUR*t9*t33*
t35*t12*t14*t81+FOUR*t9*t33*t12*t15*t18*t88*t23*t58*t79+EIGHT*t96*t98*t99*t35*t11*
t52*t20*t55*t24*t43*I-FOUR*t96*t98*t99*t11*t15*t116*t25*t57*t42*I-FOUR*t96*t97*
t99*t11*t15*t81-FOUR*t6*t7*t97*t34*t99*t36*t11*t140+TWO*t6*t143*t145*t146*t65*
t46-FOUR*t6*t94*t145*t33*t146*t35*t14*t140+TWO*t6*t145*t34*t146*t36*t38*t56*t24*
t165*t167*I;
    return(t173);
  }
}


std::complex<DOUBLE> C434x61mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  {
    t1 = p4+p3;
    t2 = tr5(p2,t1,p5,p3);
    t4 = tr5(p2,t1,p5,p6);
    t6 = ONE/t2/t4;
    t7 = A(p1,p2);
    t8 = t7*t7;
    t9 = t6*t8;
    t10 = A(p2,p3);
    t11 = t10*t10;
    t12 = B(p4,p5);
    t13 = t12*t12;
    t15 = B(p5,p6);
    t16 = t15*t15;
    t17 = IA(p1,p6);
    t18 = t16*t17;
    t21 = IA(p2,p5);
    t22 = IA(p3,p4);
    t23 = t21*t22;
    t24 = IB(p1,p6);
    t25 = IB(p2,p5);
    t26 = t24*t25;
    t28 = IB(p3,p4);
    t29 = AB(p5,t1,p2);
    t30 = t29*t29;
    t31 = t30*t30;
    t33 = IAB(p2,t1,p5);
    t34 = t33*t33;
    t42 = A(p3,p5);
    t43 = B(p2,p4);
    t44 = t42*t43;
    t50 = t17*t21*t22*t24;
    t51 = t25*t28;
    t56 = t50*t51*t30*t29*t33*I;
    t59 = t42*t42;
    t60 = t43*t43;
    t61 = t59*t60;
    t64 = t23*t24;
    t67 = t64*t51*t30*I;
    t70 = A(p1,p5);
    t72 = t6*t7*t70;
    t73 = B(p2,p6);
    t96 = AB(p2,t1,p5);
    t100 = t50*t51*t96*t29*I;
    t103 = t70*t70;
    t104 = t6*t103;
    t105 = t73*t73;
    t122 = t96*t96;
    return(TWO*t9*t11*t13*t18*t23*t26*t28*t31*t34*I-FOUR*t6*t8*t10*t44*t12*t16*
t56+TWO*t9*t61*t18*t67-FOUR*t72*t11*t73*t13*t15*t56+EIGHT*t72*t10*t42*t43*t73*t12*
t15*t17*t23*t26*t28*t30*I-FOUR*t72*t61*t73*t15*t100+TWO*t104*t11*t105*t13*t17*
t67-FOUR*t6*t103*t10*t44*t105*t12*t100+TWO*t104*t61*t105*t17*t64*t51*t122*I);
  }
}


std::complex<DOUBLE> C31234mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t216;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t265;
  FourMomentum<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t319;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t336;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t340;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t356;
  std::complex<DOUBLE> t358;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t389;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t399;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t410;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t418;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t444;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t462;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t467;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t475;
  std::complex<DOUBLE> t478;
  std::complex<DOUBLE> t479;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t486;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t495;
  std::complex<DOUBLE> t496;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t506;
  std::complex<DOUBLE> t507;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t512;
  std::complex<DOUBLE> t525;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t553;
  std::complex<DOUBLE> t560;
  std::complex<DOUBLE> t579;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t97;
  {
    t1 = A(p1,p2);
    t2 = A(p1,p5);
    t4 = A(p3,p6);
    t5 = t4*t4;
    t6 = t5*t5;
    t7 = t1*t2*t6;
    t8 = B(p1,p6);
    t9 = B(p2,p5);
    t10 = t9*t9;
    t11 = t10*t10;
    t12 = t8*t11;
    t13 = IA(p1,p6);
    t16 = IA(p3,p4);
    t17 = IA(p4,p6);
    t18 = t16*t17;
    t19 = IB(p1,p5);
    t20 = t19*t19;
    t21 = t18*t20;
    t22 = p2+p1;
    t23 = IAB(p3,t22,p5);
    t24 = IAB(p6,t22,p5);
    t25 = t24*t24;
    t26 = t23*t25;
    t27 = p4+p3;
    t28 = IAB(p6,t27,p2);
    t29 = t28*I;
    t30 = t26*t29;
    t33 = t11*t13;
    t34 = t33*t16;
    t36 = t17*t19;
    t37 = AB(p3,t22,p6);
    t38 = t36*t37;
    t39 = t23*t23;
    t41 = t39*t25*t29;
    t44 = t1*t6;
    t45 = A(p4,p5);
    t47 = B(p4,p6);
    t48 = t11*t47;
    t49 = t48*t16;
    t51 = t17*t17;
    t53 = IB(p4,p5);
    t54 = t51*t19*t53;
    t58 = t47*t16;
    t59 = t58*t17;
    t61 = t19*t53;
    t62 = AB(p5,t27,p2);
    t63 = t61*t62;
    t64 = t28*t28;
    t65 = t64*I;
    t66 = t26*t65;
    t69 = A(p1,p3);
    t70 = t69*t69;
    t71 = t70*t70;
    t72 = t71*t5;
    t73 = B(p5,p6);
    t75 = IA(p1,p2);
    t78 = IA(p2,p3);
    t80 = t13*t78*t16;
    t85 = t2*t6;
    t86 = t85*t45;
    t87 = t13*t16;
    t88 = t12*t87;
    t90 = IB(p1,p2);
    t91 = t51*t90;
    t93 = t23*t24;
    t94 = t93*t29;
    t97 = t16*t51;
    t100 = t90*t19;
    t102 = t39*t24;
    t103 = t102*t29;
    t104 = t100*t37*t103;
    t106 = t85*t8;
    t107 = t33*t18;
    t108 = t106*t107;
    t109 = t90*t20;
    t111 = t93*t65;
    t114 = AB(p6,t22,p6);
    t118 = t85*t11;
    t119 = t17*t90;
    t121 = t118*t87*t119;
    t122 = t19*t37;
    t124 = t102*t65;
    t130 = t6*t45;
    t132 = t48*t97;
    t135 = t109*t53*t94;
    t137 = t130*t11;
    t139 = t137*t58*t91;
    t141 = t61*t37*t103;
    t144 = t61*t114*t30;
    t148 = AB(p5,t22,p6);
    t150 = t19*t148*t23;
    t152 = t25*t28*I;
    t153 = t150*t152;
    t156 = t6*t8*t11;
    t157 = t58*t119;
    t163 = t18*t90;
    t169 = t6*t11;
    t170 = t169*t47;
    t172 = t170*t18*t100;
    t181 = t7*t12*t13*t21*t30+t7*t34*t38*t41+t44*t45*t49*t54*t30+t44*t11*t59*
t63*t66+t72*t45*t73*t75*t80*t51*t39*I+t86*t88*t91*t20*t94+t86*t33*t97*t104+t108
*t109*t62*t111-t108*t109*t114*t30+t121*t122*t62*t124-t121*t122*t114*t41-t130*t8
*t132*t135-t139*t141-t139*t144+t137*t97*t90*t153-t156*t157*t20*t53*t62*t111-
t156*t163*t20*t148*t23*t152-t172*t53*t37*t62*t124-t172*t53*t62*t114*t66;
    t183 = t119*t19;
    t184 = t169*t16*t183;
    t195 = A(p5,p6);
    t197 = B(p2,p4);
    t198 = t197*t197;
    t199 = t198*t198;
    t203 = IB(p2,p3);
    t205 = IB(p3,p4);
    t208 = t53*t64*I;
    t211 = t10*t9;
    t212 = B(p2,p6);
    t213 = t211*t212;
    t216 = t18*t19;
    t220 = A(p3,p5);
    t222 = t5*t4;
    t223 = t1*t220*t222;
    t225 = t36*t53;
    t229 = t213*t16;
    t231 = t36*t23;
    t240 = t25*t24*t28*I;
    t258 = t39*t23;
    t264 = B(p4,p5);
    t265 = t264*t264;
    t270 = IB(p5,p6);
    t271 = t37*t37;
    t274 = IAB(p1,p3+p2,p4);
    t277 = IS(p3+p2+p1);
    t284 = t47*t47;
    t306 = t69*t2*t222;
    t319 = t10*t212;
    t326 = t69*t220*t5;
    t328 = t211*t47*t16;
    t336 = t24*t28*I;
    t340 = t69*t222;
    t348 = t63*t111;
    t356 = -t184*t37*t148*t39*t152+t184*t148*t62*t23*t25*t64*I+t195*t8*t199*t10
*t90*t20*t203*t205*t208-TWO*t7*t213*t13*t216*t30-TWO*t223*t49*t225*t30-TWO*t223
*t229*t231*t152+TWO*t44*t11*t16*t17*t150*t240-TWO*t71*t220*t4*t73*t75*t80*t17*
t39*I-TWO*t72*t73*t75*t13*t78*t16*t17*t37*t258*I+TWO*t71*t265*t47*t75*t78*t270*
t271*t274*t258*t277*I-FOUR*t71*t264*t284*t75*t78*t270*t37*t274*t39*t277*I+TWO*
t71*t284*t47*t75*t78*t270*t274*t23*t277*I+TWO*t306*t8*t211*t13*t21*t94+TWO*t306
*t211*t13*t16*t38*t103-FOUR*t306*t319*t13*t216*t94-FOUR*t326*t328*t225*t94-TWO*
t326*t319*t16*t231*t336+TWO*t340*t45*t328*t54*t94+TWO*t340*t211*t59*t348+TWO*
t340*t211*t16*t17*t153;
    t358 = t2*t2;
    t361 = IA(p5,p6);
    t365 = IAB(p1,p6+p5,p4);
    t368 = IS(p4+p3+p2);
    t374 = A(p1,p6);
    t377 = t199*t361*t203;
    t386 = t374*t374;
    t389 = t62*t62;
    t392 = t64*t28;
    t399 = t2*t220*t222;
    t402 = t119*t20*t94;
    t408 = t213*t87;
    t410 = t183*t94;
    t415 = t91*t19*t94;
    t418 = t8*t8;
    t444 = t85*t211*t212*t13*t18;
    t448 = t100*t62*t111;
    t452 = t100*t114*t30;
    t456 = t212*t212;
    t462 = t220*t220;
    t463 = t462*t5;
    t467 = t100*t53*t94;
    t472 = t212*t16*t17;
    t475 = t100*t23*t336;
    t478 = t220*t222;
    t479 = t478*t45;
    t486 = TWO*t358*t2*t199*t361*t203*t205*t365*t28*t368*I-FOUR*t358*t374*t377*
t205*t62*t365*t64*t368*I+TWO*t2*t386*t377*t205*t389*t365*t392*t368*I-TWO*t399*
t88*t402-TWO*t399*t107*t104+FOUR*t399*t408*t410-TWO*t86*t408*t415-TWO*t85*t418*
t34*t119*t20*t19*t94-TWO*t108*t109*t37*t103+SIX*t106*t408*t402-TWO*t118*t87*t17
*t100*t271*t258*t24*t29+SIX*t444*t104-TWO*t444*t448+TWO*t444*t452-SIX*t85*t10*
t456*t13*t16*t410+TWO*t463*t11*t59*t467+FOUR*t463*t211*t472*t475-FOUR*t479*t132*
t467-TWO*t479*t229*t415;
    t487 = t478*t8;
    t495 = t478*t11;
    t496 = t495*t157;
    t506 = t478*t211;
    t507 = t212*t47;
    t512 = t506*t472;
    t525 = t45*t45;
    t553 = t6*t211*t212;
    t560 = t195*t199;
    t579 = TWO*t487*t48*t18*t135+TWO*t487*t229*t402+TWO*t496*t141-FOUR*t496*t348
+TWO*t496*t144-TWO*t495*t163*t153-EIGHT*t506*t507*t18*t467+TWO*t512*t104-TWO*t512
*t448+TWO*t512*t452-SIX*t478*t10*t456*t16*t17*t475+TWO*t6*t525*t11*t58*t51*t17*
t467+TWO*t139*t348+FOUR*t130*t211*t507*t97*t467+TWO*t170*t163*t61*t389*t93*t392*
I-TWO*t184*t148*t114*t23*t240+FOUR*t553*t157*t348+FOUR*t553*t163*t153-TWO*t560*
t10*t90*t19*t203*t205*t53*t62*t392*I-TWO*t560*t9*t212*t90*t19*t203*t205*t208;
    return(t181+t356+t486+t579);
  }
}


std::complex<DOUBLE> C32345mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t245;
  FourMomentum<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t268;
  FourMomentum<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t301;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t310;
  std::complex<DOUBLE> t314;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t333;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t343;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t355;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t362;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t369;
  std::complex<DOUBLE> t375;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t403;
  std::complex<DOUBLE> t405;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t421;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t433;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t440;
  std::complex<DOUBLE> t449;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t491;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t497;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t501;
  std::complex<DOUBLE> t503;
  std::complex<DOUBLE> t505;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t510;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t547;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t557;
  std::complex<DOUBLE> t566;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  {
    t1 = A(p1,p2);
    t2 = A(p1,p5);
    t4 = A(p3,p6);
    t5 = t4*t4;
    t6 = t5*t5;
    t7 = t1*t2*t6;
    t8 = B(p1,p4);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = B(p5,p6);
    t12 = t10*t11;
    t13 = IA(p2,p3);
    t14 = IA(p2,p6);
    t15 = t14*t14;
    t16 = t13*t15;
    t19 = IA(p5,p6);
    t20 = IB(p1,p5);
    t21 = t20*t20;
    t23 = IB(p4,p5);
    t24 = t19*t21*t23;
    t25 = p5+p4;
    t26 = IAB(p3,t25,p1);
    t27 = p3+p2;
    t28 = IAB(p6,t27,p1);
    t29 = t26*t28;
    t30 = IAB(p6,t27,p4);
    t31 = t30*I;
    t32 = t29*t31;
    t33 = t24*t32;
    t35 = t10*t13;
    t39 = t20*t23;
    t40 = AB(p3,t25,p6);
    t41 = t39*t40;
    t42 = t26*t26;
    t43 = t42*t28;
    t44 = t43*t31;
    t45 = t41*t44;
    t47 = A(p3,p5);
    t48 = t47*t47;
    t49 = t48*t48;
    t51 = B(p1,p6);
    t55 = IA(p3,p4);
    t57 = IA(p4,p5);
    t60 = t19*t42*I;
    t63 = t1*t6;
    t64 = t63*t10;
    t65 = B(p2,p3);
    t66 = B(p2,p6);
    t67 = t65*t66;
    t70 = IB(p1,p2);
    t71 = t70*t20;
    t72 = t71*t23;
    t73 = t28*t28;
    t74 = t26*t73;
    t75 = t74*t31;
    t76 = t72*t75;
    t78 = t66*t11;
    t83 = t70*t21*t23*t32;
    t85 = t66*t13;
    t88 = t64*t85*t15*t70;
    t89 = AB(p1,t27,p1);
    t91 = t39*t89*t75;
    t96 = AB(p1,t27,p6);
    t98 = t23*t96*t26;
    t100 = t73*t30*I;
    t101 = t98*t100;
    t103 = t2*t6;
    t104 = t103*t10;
    t110 = t65*t14;
    t114 = t42*t73*t31;
    t117 = t11*t13;
    t118 = t14*t19;
    t120 = t104*t117*t118;
    t121 = t21*t23;
    t125 = AB(p1,t27,p4);
    t127 = t30*t30;
    t128 = t127*I;
    t129 = t29*t128;
    t130 = t121*t125*t129;
    t132 = t13*t14;
    t133 = t19*t20;
    t135 = t104*t132*t133;
    t136 = t23*t89;
    t143 = t23*t125*t40*t43*t128;
    t145 = A(p1,p6);
    t146 = t145*t9;
    t147 = B(p2,p4);
    t148 = t147*t147;
    t149 = t148*t148;
    t153 = IB(p2,p3);
    t155 = IB(p3,p4);
    t158 = t23*t127*I;
    t161 = t6*t10;
    t164 = t66*t14*t70;
    t166 = t39*t125;
    t167 = t74*t128;
    t170 = t161*t66;
    t171 = t14*t70;
    t176 = t170*t132*t71;
    t181 = -t7*t12*t16*t33+t7*t35*t15*t19*t45+t1*t49*t5*t51*t13*t15*t55*t57*t60
-t64*t67*t15*t76+t64*t78*t16*t83+t88*t91-t88*t45-t64*t16*t20*t101-t104*t65*t11*
t14*t24*t75+t104*t110*t19*t41*t114+t120*t121*t89*t75+t120*t130-t135*t136*t40*
t114-t135*t143+t146*t149*t11*t70*t21*t153*t155*t158+t161*t65*t164*t166*t167-
t170*t117*t171*t130-t176*t136*t125*t167+t176*t143;
    t189 = t161*t13*t14*t20*t23;
    t200 = t1*t1;
    t206 = t72*t32;
    t209 = A(p1,p3);
    t211 = t5*t4;
    t212 = t1*t209*t211;
    t218 = t9*t8;
    t219 = B(p4,p6);
    t220 = t218*t219;
    t230 = t133*t23;
    t231 = t230*t32;
    t234 = t166*t129;
    t238 = t66*t219;
    t245 = t147*t66;
    t250 = t209*t209;
    t251 = t250*t5;
    t258 = t219*t13;
    t259 = t258*t14;
    t261 = t39*t26;
    t264 = t261*t28*t30*I;
    t268 = t209*t2*t211;
    t290 = t209*t211;
    t291 = t290*t10;
    t301 = t291*t85*t171;
    t306 = t161*t11*t132*t21*t101+t189*t125*t96*t26*t73*t127*I-t189*t96*t40*t42
*t100-TWO*t200*t6*t10*t85*t15*t14*t206+SIX*t212*t10*t66*t16*t206+TWO*t212*t220*
t13*t15*t20*t23*t32+TWO*t7*t220*t16*t231+TWO*t88*t234-TWO*t63*t218*t238*t16*
t206+TWO*t1*t211*t218*t245*t15*t206-SIX*t251*t10*t85*t14*t206-SIX*t251*t218*
t259*t264+FOUR*t268*t12*t132*t33-FOUR*t268*t35*t118*t45-EIGHT*t268*t220*t132*t231
-TWO*t209*t49*t4*t51*t13*t14*t55*t57*t60+TWO*t291*t67*t14*t76-TWO*t291*t78*t132
*t83-TWO*t301*t91-SIX*t301*t234;
    t310 = t132*t20;
    t314 = t290*t218;
    t316 = t65*t219*t14;
    t325 = t219*t11;
    t333 = t314*t259;
    t341 = t219*t219;
    t343 = t341*t13*t14;
    t347 = t209*t5;
    t355 = t147*t219*t14;
    t359 = t2*t2;
    t362 = IA(p1,p6);
    t366 = IAB(p5,p4+p3,p2);
    t369 = IS(p4+p3+p2);
    t375 = A(p5,p6);
    t378 = t149*t362*t153;
    t387 = t11*t11;
    t403 = t40*t40;
    t405 = t42*t26;
    t411 = t103*t218;
    t421 = t411*t258*t118;
    t424 = TWO*t301*t45+FOUR*t291*t310*t101+TWO*t314*t316*t261*t100+FOUR*t314*
t238*t132*t206-TWO*t314*t325*t13*t14*t21*t23*t32-TWO*t333*t91-TWO*t333*t234+TWO
*t333*t45+FOUR*t290*t9*t343*t264-FOUR*t347*t218*t245*t14*t206-TWO*t347*t9*t355*
t264+TWO*t359*t2*t149*t362*t153*t155*t366*t30*t369*I+FOUR*t359*t375*t378*t155*
t125*t366*t127*t369*I+TWO*t104*t387*t13*t14*t19*t21*t20*t23*t32-TWO*t120*t121*
t40*t44+TWO*t104*t132*t19*t39*t403*t405*t28*t31+TWO*t411*t316*t230*t75-FOUR*t411
*t325*t132*t33-TWO*t421*t91;
    t433 = t2*t211;
    t434 = t433*t218;
    t440 = t147*t14;
    t449 = t375*t375;
    t452 = t125*t125;
    t455 = t127*t30;
    t484 = t55*t57;
    t491 = B(p1,p2);
    t492 = t491*t491;
    t497 = IB(p1,p6);
    t501 = IAB(p5,p6+p1,p2);
    t503 = IS(p5+p4+p3);
    t505 = t501*t503*I;
    t510 = t66*t66;
    t530 = t73*t28*t30*I;
    t547 = t6*t218;
    t557 = t211*t218;
    t566 = -TWO*t421*t234+FOUR*t421*t45+TWO*t103*t9*t343*t231+TWO*t434*t147*t11*
t14*t33-TWO*t434*t440*t19*t45-FOUR*t433*t9*t355*t231+TWO*t2*t449*t378*t155*t452*
t366*t455*t369*I+TWO*t146*t149*t70*t20*t153*t155*t23*t125*t455*I-TWO*t145*t8*
t149*t219*t70*t20*t153*t155*t158+TWO*t49*t5*t51*t13*t14*t484*t19*t40*t405*I-TWO
*t49*t492*t66*t55*t57*t497*t403*t405*t505-FOUR*t49*t491*t510*t55*t57*t497*t40*
t42*t505-TWO*t49*t510*t66*t484*t497*t26*t505-TWO*t161*t110*t20*t98*t530-TWO*
t170*t132*t70*t39*t452*t29*t455*I+TWO*t189*t89*t96*t26*t530+TWO*t547*t66*t258*
t171*t234-TWO*t547*t219*t310*t101-TWO*t557*t147*t164*t234+TWO*t557*t440*t20*
t101;
    return(t181+t306+t424+t566);
  }
}


std::complex<DOUBLE> C33456mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  FourMomentum<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t219;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t285;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t296;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t312;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t319;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t336;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t355;
  std::complex<DOUBLE> t358;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t370;
  std::complex<DOUBLE> t375;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t388;
  std::complex<DOUBLE> t394;
  std::complex<DOUBLE> t403;
  std::complex<DOUBLE> t405;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t419;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t421;
  std::complex<DOUBLE> t422;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t430;
  std::complex<DOUBLE> t433;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t473;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t513;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t536;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t572;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  {
    t1 = A(p1,p5);
    t2 = A(p2,p5);
    t3 = t2*t2;
    t5 = A(p3,p5);
    t6 = t1*t3*t5;
    t7 = B(p1,p4);
    t8 = t7*t7;
    t9 = t8*t7;
    t10 = B(p2,p6);
    t11 = t10*t9;
    t12 = IA(p2,p6);
    t15 = IA(p5,p6);
    t16 = IB(p1,p3);
    t17 = t15*t16;
    t18 = IB(p1,p6);
    t19 = t17*t18;
    t20 = p4+p3;
    t21 = IAB(p2,t20,p1);
    t22 = p6+p5;
    t23 = IAB(p2,t22,p4);
    t24 = t21*t23;
    t25 = IAB(p5,t20,p1);
    t26 = t25*I;
    t27 = t24*t26;
    t28 = t19*t27;
    t31 = A(p1,p3);
    t32 = t3*t3;
    t33 = t31*t32;
    t34 = A(p3,p4);
    t35 = t33*t34;
    t36 = B(p2,p4);
    t37 = t9*t36;
    t38 = IA(p2,p3);
    t41 = t12*t15;
    t42 = t41*t16;
    t43 = t21*t21;
    t44 = t43*t23;
    t45 = t44*t26;
    t49 = t8*t8;
    t50 = t33*t49;
    t51 = B(p2,p3);
    t52 = t51*t51;
    t56 = t16*t16;
    t59 = IB(p3,p4);
    t66 = t50*t51*t38*t41;
    t67 = t56*t59;
    t68 = AB(p5,t20,p2);
    t70 = t25*t25;
    t71 = t70*I;
    t72 = t24*t71;
    t76 = t5*t5;
    t77 = t76*t76;
    t78 = t10*t10;
    t81 = IA(p3,p4);
    t82 = IA(p4,p5);
    t83 = t81*t82;
    t85 = IB(p1,p2);
    t87 = IAB(p3,p5+p4,p6);
    t90 = IS(p5+p4+p3);
    t96 = t1*t1;
    t97 = t96*t3;
    t100 = t36*t12*t15;
    t102 = t16*t59;
    t105 = t23*t25*I;
    t106 = t102*t21*t105;
    t109 = A(p1,p6);
    t111 = t3*t2;
    t112 = t1*t109*t111;
    t113 = t49*t10;
    t114 = t12*t12;
    t115 = t114*t15;
    t118 = t16*t18;
    t120 = t118*t59*t27;
    t125 = t17*t59;
    t126 = t125*t27;
    t129 = t33*t9;
    t130 = t51*t36;
    t131 = t38*t12;
    t134 = t15*t56;
    t136 = t134*t59*t27;
    t141 = t129*t36*t38*t41;
    t142 = AB(p1,t22,p4);
    t144 = t23*t23;
    t145 = t21*t144;
    t146 = t145*t26;
    t147 = t102*t142*t146;
    t150 = AB(p2,t20,p2);
    t152 = t102*t150*t45;
    t156 = t102*t68*t72;
    t160 = t36*t36;
    t167 = t31*t111*t5;
    t171 = t41*t56;
    t176 = t31*t109*t32;
    t177 = t49*t51;
    t178 = t38*t114;
    t182 = t49*t38;
    t192 = t17*t68;
    t193 = t44*t71;
    t199 = -FOUR*t6*t11*t12*t28-TWO*t35*t37*t38*t42*t45-TWO*t50*t52*t38*t12*t15*
t56*t16*t59*t27-TWO*t66*t67*t68*t72-TWO*t77*t78*t10*t83*t85*t87*t25*t90*I+FOUR*
t97*t9*t100*t106-FOUR*t112*t113*t115*t120-TWO*t112*t37*t114*t126+SIX*t129*t130*
t131*t136-TWO*t141*t147+TWO*t141*t152+SIX*t141*t156-SIX*t33*t8*t160*t38*t12*
t126+TWO*t167*t9*t51*t38*t171*t27+t176*t177*t178*t136+t176*t182*t115*t156+t35*
t177*t38*t171*t45+t35*t182*t12*t192*t193+t66*t67*t142*t146;
    t200 = t1*t111;
    t201 = t200*t34;
    t204 = t19*t45;
    t213 = t200*t49;
    t214 = t51*t10;
    t219 = t56*t18*t59*t27;
    t222 = t10*t12;
    t223 = t222*t17;
    t224 = t213*t223;
    t225 = t18*t59;
    t227 = t225*t142*t146;
    t231 = t225*t150*t45;
    t235 = t225*t68*t72;
    t239 = AB(p1,t20,p2);
    t242 = t59*t239*t43*t105;
    t245 = t200*t9;
    t250 = t36*t10;
    t255 = t245*t100;
    t258 = t32*t49;
    t260 = t258*t12*t125;
    t267 = A(p1,p2);
    t268 = t267*t8;
    t269 = B(p4,p6);
    t270 = t269*t269;
    t271 = t270*t270;
    t275 = IB(p4,p5);
    t277 = IB(p5,p6);
    t279 = t144*t23;
    t285 = t109*t32;
    t290 = t285*t49;
    t296 = t290*t10*t114*t17;
    t306 = t225*t275*t277*t144*I;
    t309 = t31*t31;
    t312 = IA(p1,p2);
    t317 = IAB(p3,p2+p1,p6);
    t319 = IS(p6+p5+p4);
    t321 = t317*t319*I;
    t325 = A(p2,p3);
    t328 = t271*t312*t275;
    t336 = t31*t1*t111;
    t341 = -TWO*t201*t113*t12*t204-TWO*t201*t37*t12*t17*t43*t105+TWO*t213*t214*
t41*t219-FOUR*t224*t227+TWO*t224*t231+TWO*t224*t235-TWO*t213*t42*t242+TWO*t245*
t130*t12*t136-EIGHT*t245*t250*t41*t120-TWO*t255*t147-t260*t239*t68*t43*t23*t70*I
-TWO*t268*t271*t16*t18*t59*t275*t277*t142*t279*I+t285*t34*t113*t114*t204-t290*
t214*t115*t219-t296*t231-t296*t235-TWO*t267*t7*t36*t271*t16*t306-TWO*t309*t31*
t271*t312*t275*t277*t23*t321+FOUR*t309*t325*t328*t277*t142*t144*t321-TWO*t336*
t177*t131*t136;
    t355 = t325*t325;
    t358 = t142*t142;
    t368 = B(p1,p2);
    t370 = t77*t368*t38;
    t375 = t15*t70*I;
    t378 = t32*t34;
    t380 = t222*t15;
    t382 = t118*t142;
    t384 = t43*t144*t26;
    t388 = t43*t21;
    t394 = t32*t9*t36;
    t403 = t68*t68;
    t405 = t70*t25;
    t413 = t258*t51;
    t419 = t258*t10;
    t421 = t419*t41*t118;
    t422 = t59*t142;
    t430 = t50*t131*t17;
    t433 = t422*t68*t145*t71;
    t441 = -TWO*t336*t182*t41*t156+FOUR*t336*t37*t131*t126-TWO*t176*t37*t178*
t126-TWO*t31*t355*t328*t277*t358*t279*t321+t290*t115*t16*t242+t109*t3*t370*t114
*t81*t82*t375+t378*t49*t380*t382*t384-TWO*t260*t239*t150*t388*t105+FOUR*t394*
t223*t227+FOUR*t394*t42*t242-TWO*t50*t131*t15*t102*t403*t24*t405*I+TWO*t255*t152
-t413*t222*t134*t227-t413*t171*t242-t421*t422*t150*t384-t66*t67*t150*t45+t430*
t433-t430*t59*t150*t68*t193+TWO*t255*t156;
    t448 = t111*t5;
    t457 = t16*t239;
    t472 = B(p1,p6);
    t473 = t472*t472;
    t513 = t8*t36;
    t536 = t109*t109;
    t572 = -SIX*t200*t8*t160*t12*t15*t106+TWO*t448*t9*t380*t382*t146+TWO*t448*
t9*t12*t15*t457*t43*t105-TWO*t3*t77*t368*t38*t12*t83*t15*t68*t405*I-TWO*t77*
t473*t10*t81*t82*t85*t403*t87*t405*t90*I+FOUR*t77*t472*t78*t81*t82*t85*t68*t87*
t70*t90*I+t268*t51*t271*t56*t306-t421*t433+t260*t239*t142*t43*t144*t25*I+TWO*
t167*t9*t38*t12*t192*t72-FOUR*t167*t513*t38*t42*t27+TWO*t97*t49*t380*t120-TWO*t6
*t513*t12*t17*t21*t105-TWO*t1*t2*t370*t12*t81*t82*t375+TWO*t536*t32*t49*t10*
t114*t12*t15*t120+TWO*t296*t227+FOUR*t285*t9*t250*t115*t120+TWO*t109*t111*t5*t11
*t114*t28+TWO*t378*t49*t12*t15*t457*t388*t105+TWO*t419*t42*t225*t358*t21*t279*
t26;
    return(t199+t341+t441+t572);
  }
}


std::complex<DOUBLE> C34561mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t218;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t284;
  FourMomentum<DOUBLE> t29;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t300;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t311;
  std::complex<DOUBLE> t314;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t355;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t367;
  std::complex<DOUBLE> t370;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t393;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t444;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t462;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t464;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t478;
  std::complex<DOUBLE> t482;
  std::complex<DOUBLE> t485;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t519;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t532;
  std::complex<DOUBLE> t533;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t558;
  std::complex<DOUBLE> t573;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  {
    t1 = A(p1,p3);
    t2 = A(p2,p5);
    t3 = t2*t2;
    t4 = t3*t3;
    t5 = t1*t4;
    t6 = B(p1,p2);
    t7 = t5*t6;
    t8 = B(p3,p6);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = IA(p1,p2);
    t12 = t10*t11;
    t13 = IA(p2,p4);
    t14 = IA(p4,p5);
    t15 = t13*t14;
    t16 = t12*t15;
    t17 = t7*t16;
    t18 = IB(p1,p3);
    t19 = t18*t18;
    t20 = IB(p1,p6);
    t21 = t19*t20;
    t22 = p5+p4;
    t23 = AB(p3,t22,p6);
    t25 = IAB(p2,t22,p3);
    t26 = IAB(p2,t22,p6);
    t27 = t26*t26;
    t28 = t25*t27;
    t29 = p6+p1;
    t30 = IAB(p5,t29,p3);
    t31 = t30*I;
    t32 = t28*t31;
    t35 = t5*t10;
    t36 = B(p4,p5);
    t40 = t18*t20;
    t41 = AB(p5,t29,p2);
    t42 = t40*t41;
    t43 = t25*t25;
    t44 = t43*t26;
    t45 = t30*t30;
    t46 = t45*I;
    t47 = t44*t46;
    t50 = t11*t13;
    t51 = t14*t18;
    t53 = t35*t50*t51;
    t54 = AB(p3,t22,p3);
    t59 = B(p2,p4);
    t60 = t4*t59;
    t61 = t60*t10;
    t63 = t36*t13*t18;
    t65 = IB(p3,p4);
    t66 = t20*t65;
    t67 = t66*t23;
    t69 = t43*t27*t31;
    t73 = t61*t15*t40;
    t80 = t28*t46;
    t83 = t4*t10;
    t85 = t51*t20;
    t86 = t83*t13*t85;
    t87 = AB(p3,t22,p2);
    t100 = A(p1,p2);
    t101 = t100*t100;
    t103 = B(p4,p6);
    t104 = t103*t103;
    t105 = t104*t104;
    t106 = IA(p2,p3);
    t108 = IB(p4,p5);
    t109 = t105*t106*t108;
    t111 = IB(p5,p6);
    t112 = t23*t23;
    t115 = IAB(p1,p6+p5,p4);
    t117 = t27*t26;
    t119 = IS(p6+p5+p4);
    t125 = A(p3,p4);
    t126 = t5*t125;
    t127 = t6*t10;
    t128 = t13*t13;
    t129 = t11*t128;
    t132 = t14*t19;
    t134 = t25*t26;
    t135 = t134*t31;
    t136 = t132*t20*t135;
    t138 = t128*t14;
    t141 = t134*t46;
    t142 = t42*t141;
    t144 = t36*t10;
    t148 = t13*t19*t20;
    t149 = t44*t31;
    t155 = t3*t2;
    t156 = A(p3,p5);
    t157 = t155*t156;
    t158 = t157*t59;
    t159 = B(p2,p6);
    t160 = t9*t8;
    t161 = t159*t160;
    t164 = t40*t65;
    t165 = t164*t135;
    t168 = t10*t13;
    t170 = t158*t168*t51;
    t172 = t66*t54*t149;
    t175 = t67*t32;
    t179 = t66*t41*t141;
    t182 = A(p1,p5);
    t183 = t182*t182;
    t184 = t183*t183;
    t185 = t184*t3;
    t186 = B(p2,p3);
    t188 = IA(p1,p6);
    t191 = IA(p5,p6);
    t193 = t45*t30;
    t206 = t191*t45*I;
    t210 = -t17*t21*t23*t32+t35*t36*t11*t13*t42*t47-t53*t20*t54*t41*t47+t61*t63
*t67*t69-t73*t65*t54*t23*t69+t73*t65*t23*t41*t80-t86*t87*t23*t43*t27*t30*I+t86*
t87*t41*t43*t26*t45*I+TWO*t101*t1*t109*t111*t112*t115*t117*t119*I-t126*t127*
t129*t136-t126*t12*t138*t142+t7*t144*t11*t148*t149-t17*t21*t54*t149+FOUR*t158*
t161*t15*t165+TWO*t170*t172+SIX*t170*t175-TWO*t170*t179-TWO*t185*t186*t11*t188*
t15*t191*t41*t193*I-TWO*t184*t2*t156*t186*t11*t188*t13*t14*t206;
    t211 = t59*t59;
    t216 = IB(p2,p3);
    t218 = IAB(p1,p3+p2,p4);
    t221 = IS(p6+p5+p1);
    t228 = B(p3,p4);
    t240 = t228*t228;
    t244 = t41*t41;
    t252 = A(p2,p3);
    t260 = t66*t108*t111*t27*I;
    t274 = t125*t125;
    t283 = t4*t125;
    t284 = t283*t59;
    t291 = t284*t10*t128*t51;
    t295 = t160*t13;
    t300 = t15*t18;
    t311 = t20*t87;
    t314 = t26*t30*I;
    t315 = t311*t43*t314;
    t318 = t6*t6;
    t328 = t161*t50;
    t332 = t5*t159;
    t335 = t332*t160*t11*t15;
    t337 = t40*t54*t149;
    t341 = t40*t23*t32;
    t354 = t1*t155;
    t355 = t354*t156;
    t361 = t85*t135;
    t367 = TWO*t184*t211*t59*t188*t191*t216*t218*t30*t221*I-FOUR*t184*t211*t228*
t188*t191*t216*t41*t218*t45*t221*I+TWO*t184*t59*t240*t188*t191*t216*t244*t218*
t193*t221*I-TWO*t252*t159*t8*t105*t18*t260-TWO*t252*t9*t105*t18*t20*t65*t108*
t111*t23*t117*I-TWO*t4*t274*t59*t10*t128*t13*t14*t165-TWO*t284*t161*t138*t165
-TWO*t291*t175-TWO*t60*t159*t295*t51*t175-TWO*t61*t300*t66*t112*t25*t117*t31
-TWO*t4*t159*t160*t300*t315+TWO*t5*t318*t12*t13*t14*t19*t18*t20*t135-FOUR*t7*
t328*t136+TWO*t335*t337+TWO*t335*t341-FOUR*t335*t142+TWO*t35*t50*t14*t40*t244*
t134*t193*I+FOUR*t355*t127*t50*t136-EIGHT*t355*t328*t361+FOUR*t355*t16*t142;
    t370 = t160*t103;
    t377 = t9*t103;
    t381 = t13*t18*t20;
    t391 = t157*t159;
    t392 = t295*t14;
    t393 = t391*t392;
    t407 = t103*t13*t18;
    t411 = t1*t1;
    t435 = t159*t159;
    t441 = t40*t25*t314;
    t444 = t160*t36;
    t452 = t43*t25;
    t462 = t155*t125;
    t463 = t462*t156;
    t464 = t59*t10;
    t465 = t464*t138;
    t478 = t157*t6;
    t482 = t21*t65*t135;
    t485 = TWO*t354*t6*t370*t11*t148*t135-FOUR*t354*t159*t377*t11*t381*t135+TWO*
t354*t160*t103*t11*t13*t142+TWO*t393*t337+TWO*t393*t341-TWO*t393*t142+FOUR*t157*
t10*t300*t315+TWO*t155*t59*t160*t407*t175-FOUR*t100*t411*t109*t111*t23*t115*t27*
t119*I+TWO*t411*t1*t105*t106*t108*t111*t115*t26*t119*I+TWO*t126*t161*t129*t361+
FOUR*t157*t435*t9*t13*t14*t441-TWO*t391*t444*t13*t40*t43*t314+TWO*t83*t63*t311*
t452*t314-TWO*t86*t87*t54*t452*t314+SIX*t463*t465*t165+TWO*t463*t161*t128*t361+
TWO*t462*t59*t370*t128*t165-TWO*t478*t464*t15*t482;
    t511 = t164*t149;
    t519 = t4*t6;
    t532 = t156*t156;
    t533 = t3*t532;
    t558 = t3*t156;
    t573 = -TWO*t478*t161*t13*t136-t53*t20*t23*t41*t80+t185*t125*t186*t11*t188*
t128*t14*t206+t252*t6*t9*t105*t19*t260+t283*t6*t465*t482+t284*t144*t128*t511-
t291*t172+t291*t179-t283*t10*t138*t18*t315+t519*t59*t168*t132*t175+t519*t10*t15
*t19*t315+TWO*t155*t160*t407*t315-SIX*t533*t59*t168*t14*t165+TWO*t17*t21*t41*
t141+TWO*t5*t435*t9*t11*t13*t361-TWO*t332*t444*t11*t381*t149-SIX*t533*t159*t392
*t441-FOUR*t558*t59*t370*t13*t165-TWO*t558*t159*t377*t13*t441-TWO*t158*t144*t13*
t511;
    return(t210+t367+t485+t573);
  }
}


std::complex<DOUBLE> C35612mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t208;
  FourMomentum<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t239;
  FourMomentum<DOUBLE> t24;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t285;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t297;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t304;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t329;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t338;
  std::complex<DOUBLE> t339;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t344;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t376;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t396;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t404;
  std::complex<DOUBLE> t410;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t412;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t414;
  std::complex<DOUBLE> t416;
  std::complex<DOUBLE> t419;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t420;
  std::complex<DOUBLE> t423;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t427;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t469;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t514;
  std::complex<DOUBLE> t519;
  std::complex<DOUBLE> t521;
  std::complex<DOUBLE> t523;
  std::complex<DOUBLE> t527;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t537;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t555;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  {
    t1 = A(p1,p3);
    t2 = A(p1,p4);
    t3 = t2*t2;
    t4 = t3*t2;
    t5 = t1*t4;
    t6 = B(p2,p4);
    t7 = t5*t6;
    t8 = B(p3,p6);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = IA(p1,p2);
    t12 = t10*t11;
    t13 = IA(p2,p4);
    t14 = IB(p2,p3);
    t15 = t13*t14;
    t17 = t7*t12*t15;
    t18 = IB(p3,p5);
    t19 = IB(p5,p6);
    t20 = t18*t19;
    t21 = p2+p1;
    t22 = AB(p3,t21,p6);
    t24 = p6+p5;
    t25 = IAB(p1,t24,p3);
    t26 = IAB(p4,t21,p3);
    t27 = t25*t26;
    t28 = IAB(p4,t21,p6);
    t29 = t28*t28;
    t30 = t29*I;
    t31 = t27*t30;
    t32 = t20*t22*t31;
    t35 = A(p3,p5);
    t36 = t5*t35;
    t37 = B(p4,p5);
    t38 = t10*t37;
    t39 = t11*t13;
    t40 = t38*t39;
    t42 = IA(p4,p5);
    t43 = t18*t18;
    t46 = t28*I;
    t47 = t27*t46;
    t48 = t42*t43*t19*t47;
    t51 = t13*t42;
    t54 = AB(p1,t24,p4);
    t56 = t25*t25;
    t57 = t56*t26;
    t58 = t57*t46;
    t59 = t20*t54*t58;
    t62 = t9*t8;
    t63 = B(p4,p6);
    t64 = t62*t63;
    t65 = t64*t39;
    t67 = t42*t18;
    t69 = t67*t19*t47;
    t72 = A(p5,p6);
    t73 = t5*t72;
    t74 = t6*t10;
    t75 = t74*t11;
    t77 = t15*t18;
    t78 = t26*t26;
    t79 = t25*t78;
    t80 = t79*t46;
    t84 = t64*t11;
    t86 = t13*t18;
    t87 = t86*t25;
    t89 = t78*t28*I;
    t96 = t14*t43*t19*t47;
    t101 = t3*t3;
    t102 = t101*t72;
    t104 = t12*t13;
    t106 = t14*t18;
    t107 = t106*t22;
    t108 = t79*t30;
    t111 = t101*t6;
    t112 = t111*t10;
    t113 = t37*t11;
    t116 = t43*t19;
    t118 = t116*t22*t31;
    t121 = t112*t39*t106;
    t122 = t19*t54;
    t125 = t122*t22*t57*t30;
    t128 = AB(p4,t24,p4);
    t132 = t101*t10;
    t136 = AB(p3,t21,p4);
    t139 = t19*t136*t25*t89;
    t143 = t132*t11*t86*t19;
    t154 = A(p1,p5);
    t155 = t154*t154;
    t156 = t155*t155;
    t157 = t3*t156;
    t158 = A(p2,p3);
    t159 = B(p3,p4);
    t163 = IA(p1,p6);
    t164 = t13*t13;
    t167 = IA(p5,p6);
    t169 = t167*t56*I;
    t173 = t20*t128*t80;
    t177 = t106*t19;
    t178 = t177*t47;
    t182 = t39*t18;
    t186 = -FOUR*t17*t32-TWO*t36*t40*t48-TWO*t36*t12*t51*t59+FOUR*t36*t65*t69+TWO
*t73*t75*t77*t80+TWO*t73*t84*t87*t89+TWO*t7*t40*t96+TWO*t17*t59-t102*t6*t104*
t107*t108-t112*t113*t15*t118-t121*t125+t121*t19*t22*t128*t108-t132*t37*t39*t43*
t139-t143*t54*t136*t56*t89+t143*t136*t22*t25*t78*t29*I+t157*t158*t159*t11*t163*
t164*t42*t169-TWO*t17*t173-EIGHT*t7*t65*t178-TWO*t5*t10*t182*t139;
    t187 = t5*t62;
    t188 = t37*t63;
    t196 = t63*t11;
    t197 = t196*t13;
    t198 = t187*t197;
    t206 = t63*t63;
    t208 = t206*t11*t13;
    t212 = t26*t28*I;
    t213 = t20*t25*t212;
    t217 = t1*t3*t154;
    t219 = t6*t62*t11;
    t225 = t9*t63*t11;
    t235 = t4*t154;
    t238 = t164*t14;
    t239 = t238*t18;
    t243 = t235*t35;
    t247 = t51*t43;
    t252 = t62*t11*t13;
    t254 = t67*t54;
    t259 = t51*t18;
    t270 = t18*t136*t25;
    t278 = t56*t25;
    t284 = B(p2,p3);
    t285 = t284*t284;
    t290 = IB(p3,p4);
    t291 = t54*t54;
    t295 = IAB(p5,p6+p1,p2);
    t297 = IS(p6+p5+p1);
    t299 = t295*t297*I;
    t304 = t6*t6;
    t321 = t101*t35;
    t322 = t321*t10;
    t324 = t322*t113*t51;
    t329 = t322*t39*t67;
    t331 = TWO*t187*t188*t11*t13*t43*t19*t47+TWO*t198*t59-TWO*t198*t32-TWO*t198
*t173-SIX*t5*t9*t208*t213-FOUR*t217*t219*t77*t47-TWO*t217*t225*t87*t212+FOUR*t101
*t62*t63*t182*t139+TWO*t235*t158*t219*t239*t47+TWO*t243*t62*t37*t11*t247*t47+
TWO*t243*t252*t254*t58-FOUR*t243*t225*t259*t47+TWO*t235*t6*t252*t107*t31+TWO*
t235*t252*t270*t89-TWO*t157*t159*t11*t163*t51*t167*t54*t278*I-TWO*t156*t285*t6*
t163*t167*t290*t291*t278*t299+FOUR*t156*t284*t304*t163*t167*t290*t54*t56*t299
-TWO*t156*t304*t6*t163*t167*t290*t25*t299+t324*t116*t128*t80+t329*t125;
    t335 = t56*t78*t46;
    t338 = t101*t158;
    t339 = t338*t6;
    t341 = t339*t12*t238;
    t344 = t11*t164;
    t345 = t64*t344;
    t349 = t321*t72;
    t354 = t37*t37;
    t376 = t321*t62;
    t382 = t376*t196*t51;
    t396 = t78*t26*t28*I;
    t402 = t22*t22;
    t404 = t29*t28;
    t410 = A(p3,p4);
    t411 = B(p2,p6);
    t412 = t411*t411;
    t413 = t412*t412;
    t414 = t410*t413;
    t416 = IB(p1,p2);
    t419 = IB(p1,p6);
    t420 = t419*t14;
    t423 = t19*t29*I;
    t426 = t1*t1;
    t427 = t426*t3;
    t436 = t5*t158;
    t447 = t329*t122*t128*t335+TWO*t341*t32+FOUR*t339*t345*t178+TWO*t349*t84*
t259*t80-TWO*t322*t354*t11*t13*t42*t43*t18*t19*t47-TWO*t324*t116*t54*t58-TWO*
t322*t39*t42*t20*t291*t278*t26*t46+SIX*t376*t188*t39*t48+SIX*t382*t59-TWO*t382*
t32-TWO*t382*t173-SIX*t321*t9*t208*t69-TWO*t102*t104*t270*t396+TWO*t112*t39*t14
*t20*t402*t27*t404*I+t414*t9*t37*t416*t420*t43*t423+TWO*t427*t6*t104*t178+FOUR*
t427*t62*t197*t213-FOUR*t436*t74*t344*t178-TWO*t436*t84*t164*t18*t19*t47;
    t457 = t158*t158;
    t465 = t338*t35;
    t469 = t38*t344;
    t511 = t35*t35;
    t514 = IA(p3,p4);
    t519 = IAB(p5,p4+p3,p2);
    t521 = IS(p6+p2+p1);
    t523 = t519*t521*I;
    t527 = A(p4,p5);
    t530 = t413*t514*t416;
    t537 = t527*t527;
    t555 = -TWO*t1*t2*t156*t159*t11*t163*t13*t42*t169+TWO*t101*t457*t6*t12*t164
*t13*t178-TWO*t465*t345*t69+t465*t469*t48+t465*t12*t164*t42*t59-t338*t72*t75*
t239*t80-t339*t469*t96-t341*t59+t341*t173+t338*t10*t344*t18*t139-t349*t38*t11*
t247*t80-t349*t104*t254*t335+t324*t118-TWO*t414*t9*t416*t419*t177*t22*t404*I
-TWO*t414*t8*t63*t416*t420*t18*t423+TWO*t511*t35*t413*t514*t416*t419*t28*t523
-FOUR*t511*t527*t530*t419*t22*t29*t523+TWO*t35*t537*t530*t419*t402*t404*t523+FOUR
*t111*t62*t196*t15*t32+TWO*t143*t136*t128*t25*t396;
    return(t186+t331+t447+t555);
  }
}


std::complex<DOUBLE> C36123mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t239;
  FourMomentum<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t249;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t255;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t288;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t300;
  std::complex<DOUBLE> t301;
  std::complex<DOUBLE> t303;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t327;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t334;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t340;
  std::complex<DOUBLE> t343;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t369;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t394;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t418;
  std::complex<DOUBLE> t419;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t423;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t428;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t473;
  std::complex<DOUBLE> t477;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t542;
  std::complex<DOUBLE> t552;
  std::complex<DOUBLE> t553;
  std::complex<DOUBLE> t567;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t1 = A(p1,p4);
    t2 = t1*t1;
    t3 = t2*t2;
    t4 = A(p3,p5);
    t5 = t3*t4;
    t6 = B(p2,p5);
    t7 = t6*t6;
    t8 = t7*t7;
    t9 = t5*t8;
    t10 = B(p3,p4);
    t11 = t10*t10;
    t12 = IA(p1,p6);
    t14 = IA(p3,p4);
    t17 = IA(p4,p6);
    t18 = IB(p2,p3);
    t19 = t17*t18;
    t20 = IB(p3,p5);
    t21 = t20*t20;
    t24 = p3+p2;
    t25 = IAB(p1,t24,p5);
    t26 = p6+p1;
    t27 = IAB(p4,t26,p2);
    t28 = t25*t27;
    t29 = IAB(p4,t26,p5);
    t30 = t29*I;
    t31 = t28*t30;
    t36 = t14*t17;
    t38 = t9*t10*t12*t36;
    t39 = t18*t21;
    t40 = AB(p1,t24,p4);
    t42 = t25*t25;
    t43 = t42*t27;
    t44 = t43*t30;
    t48 = t12*t14;
    t51 = t18*t20;
    t52 = t40*t40;
    t54 = t42*t25;
    t60 = A(p5,p6);
    t61 = t60*t60;
    t64 = B(p4,p6);
    t65 = t64*t12;
    t66 = t17*t17;
    t70 = IB(p5,p6);
    t71 = t51*t70;
    t72 = t71*t31;
    t75 = t3*t60;
    t76 = B(p2,p4);
    t78 = t7*t6;
    t79 = t78*t64;
    t80 = t12*t66;
    t85 = t75*t8;
    t86 = t66*t18;
    t88 = t85*t65*t86;
    t89 = t70*t20;
    t90 = AB(p5,t26,p2);
    t91 = t89*t90;
    t92 = t27*t27;
    t93 = t25*t92;
    t94 = t93*t30;
    t95 = t91*t94;
    t98 = B(p1,p6);
    t99 = t3*t98;
    t103 = AB(p5,t26,p4);
    t105 = t20*t103*t25;
    t106 = t29*t29;
    t109 = t27*t106*t29*I;
    t114 = t3*t76*t78;
    t115 = t65*t19;
    t119 = t12*t17;
    t120 = t119*t18;
    t123 = t27*t106*I;
    t124 = t105*t123;
    t127 = t3*t8;
    t128 = t127*t64;
    t130 = t90*t90;
    t132 = t92*t27;
    t139 = t19*t20;
    t140 = t127*t12*t139;
    t141 = AB(p5,t26,p5);
    t147 = t2*t1;
    t148 = A(p1,p5);
    t149 = t147*t148;
    t150 = t149*t4;
    t151 = t76*t78;
    t152 = t151*t48;
    t154 = t139*t31;
    t157 = t8*t10;
    t158 = t157*t48;
    t160 = t19*t21;
    t161 = t160*t31;
    t164 = t8*t12;
    t167 = t51*t40;
    t168 = t167*t44;
    t171 = t149*t60;
    t175 = t86*t20*t31;
    t178 = t8*t64;
    t183 = t149*t98;
    t186 = t51*t25;
    t190 = A(p1,p3);
    t191 = t190*t190;
    t192 = t191*t191;
    t193 = t192*t2;
    t194 = B(p4,p5);
    t196 = IA(p1,p2);
    t199 = IA(p2,p3);
    t201 = t12*t199*t14;
    t206 = t64*t64;
    t208 = B(p5,p6);
    t212 = IB(p4,p5);
    t216 = IAB(p3,p5+p4,p6);
    t218 = IS(p3+p2+p1);
    t220 = t216*t218*I;
    t224 = TWO*t9*t11*t12*t14*t19*t21*t20*t31+TWO*t38*t39*t40*t44+TWO*t9*t48*
t17*t51*t52*t54*t27*t30-TWO*t3*t61*t8*t65*t66*t17*t72-TWO*t75*t76*t79*t80*t72
-TWO*t88*t95+TWO*t99*t8*t17*t18*t105*t109-TWO*t114*t115*t95-TWO*t114*t120*t124
-TWO*t128*t120*t89*t130*t25*t132*t30-TWO*t140*t103*t141*t25*t109-EIGHT*t150*t152*
t154+FOUR*t150*t158*t161+FOUR*t150*t164*t36*t168+TWO*t171*t151*t12*t175+SIX*t171*
t178*t80*t72-TWO*t183*t151*t17*t186*t123+t193*t60*t194*t196*t201*t66*t42*I+FOUR*
t192*t206*t208*t196*t199*t212*t40*t42*t220;
    t226 = t208*t208;
    t235 = t5*t60;
    t239 = t5*t98;
    t242 = t106*I;
    t243 = t28*t242;
    t247 = t76*t76;
    t249 = t7*t12;
    t254 = t5*t76;
    t255 = t78*t10;
    t260 = t78*t12;
    t262 = t254*t260*t36;
    t266 = t51*t90*t94;
    t270 = t51*t141*t243;
    t273 = t147*t4;
    t274 = t273*t78;
    t275 = B(p2,p6);
    t283 = t275*t64;
    t288 = t147*t78;
    t291 = t64*t17*t18;
    t300 = t148*t148;
    t301 = t2*t300;
    t303 = t260*t17;
    t307 = t186*t27*t29*I;
    t315 = t2*t148;
    t317 = t7*t275;
    t327 = A(p3,p4);
    t328 = t327*t327;
    t330 = t275*t275;
    t331 = t330*t330;
    t332 = IA(p4,p5);
    t334 = IB(p1,p2);
    t335 = t331*t332*t334;
    t337 = IB(p1,p6);
    t340 = IAB(p3,p2+p1,p6);
    t343 = IS(p6+p2+p1);
    t349 = t4*t4;
    t369 = A(p4,p5);
    t374 = t337*t18;
    t377 = t70*t92*I;
    t381 = -TWO*t192*t64*t226*t196*t199*t212*t52*t54*t220+TWO*t235*t152*t175
-TWO*t239*t151*t14*t139*t243+TWO*t5*t247*t249*t14*t154-FOUR*t254*t255*t48*t161
-FOUR*t262*t168+TWO*t262*t266+TWO*t262*t270+TWO*t274*t275*t14*t17*t168+TWO*t147*
t60*t78*t283*t66*t72+TWO*t288*t275*t291*t95+TWO*t288*t275*t17*t18*t124-SIX*t301
*t76*t303*t307-SIX*t301*t8*t65*t17*t72-TWO*t315*t76*t317*t17*t307-FOUR*t315*t78*
t283*t17*t72-TWO*t328*t4*t335*t337*t130*t340*t132*t343*I+FOUR*t327*t349*t335*
t337*t90*t340*t92*t343*I-TWO*t349*t4*t331*t332*t334*t337*t340*t27*t343*I-TWO*
t369*t76*t6*t331*t334*t374*t20*t377;
    t383 = t369*t7;
    t394 = t71*t243;
    t402 = t149*t76;
    t411 = t402*t303;
    t418 = t149*t8;
    t419 = t10*t64;
    t423 = t39*t70*t31;
    t426 = t418*t115;
    t428 = t89*t40*t44;
    t434 = t89*t141*t243;
    t457 = t93*t242;
    t460 = -TWO*t383*t331*t334*t337*t71*t90*t132*I-TWO*t183*t178*t17*t394+FOUR*
t149*t247*t249*t17*t307-TWO*t402*t255*t12*t161+FOUR*t402*t79*t119*t72-TWO*t411*
t168+TWO*t411*t266+TWO*t411*t270-TWO*t418*t419*t119*t423-TWO*t426*t428+SIX*t426
*t95+TWO*t426*t434+FOUR*t418*t120*t124-FOUR*t273*t76*t317*t14*t154+TWO*t274*t275*
t10*t14*t161+t88*t428-t88*t434-t85*t80*t18*t124+t99*t8*t291*t91*t457;
    t461 = t127*t10;
    t473 = t128*t119*t51;
    t477 = t42*t92*t30;
    t542 = t43*t242;
    t552 = t9*t48*t19;
    t553 = t20*t40;
    t567 = t461*t115*t21*t70*t90*t94+t461*t120*t21*t103*t25*t123+t473*t70*t40*
t90*t477-t473*t70*t90*t141*t457+t140*t40*t103*t42*t123-t140*t90*t103*t25*t92*
t106*I+t383*t331*t10*t334*t374*t21*t377-TWO*t193*t194*t196*t12*t199*t14*t17*t40
*t54*I-TWO*t192*t1*t148*t194*t196*t201*t17*t42*I-TWO*t192*t206*t64*t196*t199*
t212*t25*t220-t235*t158*t86*t21*t31-t235*t164*t14*t66*t168+t239*t157*t14*t160*
t243+t239*t8*t14*t17*t167*t542-t38*t39*t90*t94-t38*t39*t141*t243-t552*t553*t90*
t477-t552*t553*t141*t542+t75*t98*t178*t66*t394+t85*t419*t80*t423;
    return(t224+t381+t460+t567);
  }
}


std::complex<DOUBLE> C312x345mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t292;
  std::complex<DOUBLE> t293;
  std::complex<DOUBLE> t294;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t301;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t308;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t318;
  FourMomentum<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t323;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t344;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t356;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t360;
  std::complex<DOUBLE> t364;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t398;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t417;
  std::complex<DOUBLE> t421;
  std::complex<DOUBLE> t431;
  std::complex<DOUBLE> t432;
  std::complex<DOUBLE> t433;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t438;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t440;
  std::complex<DOUBLE> t443;
  std::complex<DOUBLE> t445;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t519;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t524;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t98;
  {
    t1 = A(p1,p2);
    t2 = B(p1,p2);
    t4 = p2+p1;
    t5 = IAB(p6,t4,p6);
    t8 = A(p1,p6);
    t9 = t8*t8;
    t10 = A(p5,p6);
    t11 = t10*t9;
    t12 = B(p2,p6);
    t13 = t11*t12;
    t14 = B(p3,p6);
    t15 = B(p4,p6);
    t16 = t14*t15;
    t17 = IA(p2,p6);
    t18 = IB(p3,p4);
    t19 = t17*t18;
    t22 = IB(p4,p5);
    t23 = AB(p6,t4,p4);
    t24 = t23*t23;
    t25 = t23*t24;
    t26 = t22*t25;
    t27 = IAB(p6,t4,p3);
    t28 = t27*t27;
    t30 = IAB(p6,t4,p5);
    t31 = t30*t5;
    t32 = p4+p3;
    t33 = IAA(p5,t32,t4,p6);
    t34 = t33*I;
    t35 = t31*t34;
    t39 = t15*t15;
    t40 = t39*t17;
    t41 = t40*t18;
    t44 = t22*t24*t27;
    t45 = t44*t35;
    t48 = B(p5,p6);
    t49 = t15*t48;
    t50 = t49*t19;
    t52 = t27*t26;
    t53 = t30*t30;
    t54 = t53*t5;
    t55 = t54*t34;
    t56 = t52*t55;
    t59 = t15*t17;
    t60 = t18*t22;
    t63 = AB(p5,t32,p6);
    t64 = t63*t25;
    t66 = t33*t33;
    t67 = t66*I;
    t68 = t31*t67;
    t72 = t48*t48;
    t73 = t72*t17;
    t74 = t73*t18;
    t76 = t24*t24;
    t77 = t22*t76;
    t78 = t77*t27;
    t79 = t53*t30;
    t82 = t78*t79*t5*t34;
    t86 = t12*t12;
    t87 = t8*t10*t86;
    t88 = IB(p1,p2);
    t89 = t88*t18;
    t92 = t52*t35;
    t95 = t48*t17;
    t98 = t78*t55;
    t102 = t39*t39;
    t104 = IB(p1,p6);
    t106 = IB(p3,p6);
    t107 = t104*t18*t106;
    t109 = IB(p5,p6);
    t110 = t22*t109;
    t111 = AB(p5,t4,p6);
    t113 = IAB(p5,t32,p6);
    t114 = t5*t5;
    t120 = B(p2,p4);
    t125 = t88*t104*t18;
    t128 = t106*t22*t109;
    t131 = t111*t113*t5*I;
    t135 = t1*t9;
    t136 = t10*t2;
    t137 = t135*t136;
    t138 = t12*t14;
    t142 = t60*t76*t28;
    t143 = t53*t114;
    t144 = t143*t34;
    t147 = t12*t48;
    t150 = t22*t63;
    t151 = t76*t27;
    t152 = t150*t151;
    t156 = t135*t10;
    t157 = B(p1,p6);
    t158 = t157*t12;
    t159 = t17*t17;
    t163 = t30*t114;
    t164 = t163*t34;
    t170 = t163*t67;
    t174 = t1*t8*t10;
    t178 = t77*t28;
    t181 = t86*t17;
    t184 = t63*t76;
    t185 = t184*t27;
    t188 = A(p1,p3);
    t189 = t188*t188;
    t191 = t189*t8*t10;
    t192 = IA(p1,p2);
    t194 = IA(p3,p6);
    t195 = t194*t194;
    t196 = t17*t195;
    t201 = t28*t30*t34;
    t204 = TWO*t13*t16*t19*t26*t28*t35-SIX*t13*t41*t45+SIX*t13*t50*t56+TWO*t13*
t59*t60*t64*t27*t68-TWO*t13*t74*t82+FOUR*t87*t59*t89*t92-TWO*t87*t95*t89*t98+TWO
*t8*t86*t102*t107*t110*t111*t113*t114*I-TWO*t120*t86*t39*t15*t125*t128*t131+
t137*t138*t95*t142*t144+t137*t147*t19*t152*t143*t67+t156*t158*t14*t159*t142*
t164+t156*t158*t159*t18*t152*t170+t174*t86*t14*t19*t178*t164+t174*t181*t60*t185
*t170-t191*t14*t192*t196*t60*t76*t201;
    t205 = t86*t12;
    t209 = AA(p5,t32,t4,p5);
    t211 = t113*t113;
    t219 = t60*t25*t28;
    t223 = t2*t12;
    t227 = t60*t24*t27;
    t231 = t12*t15;
    t234 = t25*t27;
    t235 = t60*t234;
    t241 = t150*t234;
    t247 = t60*t151;
    t253 = t157*t157;
    t259 = t247*t164;
    t262 = t15*t159;
    t268 = t48*t159;
    t275 = t159*t88;
    t292 = A(p1,p5);
    t293 = t292*t9;
    t294 = t293*t2;
    t301 = t293*t157;
    t307 = t192*t17;
    t308 = t307*t18;
    t315 = t205*t102*t88*t107*t110*t209*t211*t5*I-TWO*t137*t138*t59*t219*t164+
SIX*t156*t223*t40*t227*t164-SIX*t137*t231*t95*t235*t144-TWO*t137*t231*t19*t241*
t170+TWO*t156*t223*t73*t247*t79*t114*t34+TWO*t156*t253*t12*t159*t17*t88*t259
-SIX*t156*t158*t262*t235*t164+TWO*t156*t158*t268*t247*t144+FOUR*t174*t157*t86*
t275*t259-THREE*FOUR*t174*t86*t15*t19*t52*t164+FOUR*t174*t86*t48*t19*t78*t144-FOUR*t294
*t50*t56+TWO*t294*t74*t82-TWO*t301*t262*t18*t92-TWO*t293*t39*t308*t44*t30*t33*I
;
    t318 = t48*t192;
    t321 = t60*t25;
    t322 = t27*t53;
    t323 = t322*t34;
    t335 = t292*t8*t12;
    t344 = t11*t157;
    t347 = t235*t35;
    t356 = t15*t192;
    t359 = t27*t30;
    t360 = t359*t34;
    t364 = t188*t9;
    t365 = t364*t136;
    t366 = t17*t194;
    t372 = t364*t10;
    t384 = t194*t18;
    t391 = t159*t194;
    t398 = t384*t26;
    t417 = t359*t67;
    t421 = FOUR*t293*t15*t318*t17*t321*t323-TWO*t293*t72*t308*t78*t79*t33*I-FOUR*
t335*t59*t18*t92+TWO*t335*t95*t18*t98+TWO*t344*t231*t275*t347+TWO*t1*t10*t120*
t181*t89*t92+TWO*t191*t356*t196*t321*t360-TWO*t365*t16*t366*t219*t35+FOUR*t372*
t2*t39*t366*t227*t35-TWO*t365*t49*t366*t235*t55-TWO*t365*t59*t384*t241*t68-TWO*
t372*t157*t15*t391*t347+TWO*t372*t16*t307*t398*t201-FOUR*t372*t39*t192*t366*t60*
t24*t360+TWO*t372*t49*t307*t398*t323+TWO*t372*t356*t366*t60*t64*t417;
    t431 = A(p3,p5);
    t432 = t431*t431;
    t433 = t432*t432;
    t435 = IA(p3,p4);
    t438 = IA(p4,p5);
    t440 = AB(p6,t4,p6);
    t443 = IS(p5+p4+p3);
    t445 = t33*t443*I;
    t448 = t14*t48;
    t455 = t67*t54;
    t463 = t157*t159;
    t489 = t194*t438*t445;
    t515 = B(p2,p3);
    t519 = t106*t106;
    t524 = TWO*t294*t41*t45-t191*t307*t195*t18*t150*t76*t417+t364*t433*t307*
t435*t195*t438*t440*t445+t365*t448*t366*t142*t55+t365*t95*t384*t152*t455+t372*
t157*t14*t391*t142*t35+t372*t463*t384*t152*t68-t372*t448*t307*t384*t77*t28*t53*
t34-t372*t318*t366*t60*t184*t322*t67+t301*t268*t18*t98-t9*t8*t433*t463*t435*
t489-t9*t433*t12*t17*t435*t489-t344*t147*t275*t247*t55-t13*t448*t19*t178*t55-
t13*t95*t60*t185*t455-t10*t205*t102*t125*t128*t113*t5*I+t515*t86*t102*t125*t519
*t22*t109*t131;
    return(t204+t315+t421+t524);
  }
}


std::complex<DOUBLE> C323x456mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K2f;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t249;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t281;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t297;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t300;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t310;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t346;
  std::complex<DOUBLE> t351;
  std::complex<DOUBLE> t352;
  FourMomentum<DOUBLE> t36;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t388;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t393;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t403;
  std::complex<DOUBLE> t412;
  std::complex<DOUBLE> t416;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t442;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t480;
  std::complex<DOUBLE> t483;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t491;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t503;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t522;
  std::complex<DOUBLE> t527;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t536;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t544;
  std::complex<DOUBLE> t545;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t550;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t570;
  std::complex<DOUBLE> t581;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t611;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t99;
  {
    t1 = A(p2,p3);
    t2 = B(p2,p3);
    t4 = p3+p2;
    t5 = IAB(p1,t4,p1);
    K2f = p3+p2+t1*t2*t5*p1;
    t8 = A(p1,p2);
    t9 = A(p1,p4);
    t10 = t9*t8;
    t11 = B(p1,p2);
    t12 = t11*t11;
    t13 = t11*t12;
    t15 = B(p1,p6);
    t16 = IA(p4,p5);
    t17 = t15*t16;
    t18 = IA(p5,p6);
    t19 = IB(p1,p3);
    t20 = t19*t19;
    t21 = t18*t20;
    t22 = t17*t21;
    t24 = IB(p1,p4);
    t25 = AB(p5,t4,p1);
    t26 = t25*t25;
    t27 = t25*t26;
    t29 = AB(p5,t4,p4);
    t30 = t29*t5;
    t31 = t24*t27*t30;
    t32 = IAB(p4,t4,p1);
    t33 = t32*t32;
    t34 = IAB(p6,t4,p1);
    t36 = p5+p4;
    t37 = IBB(p1,t4,t36,p6);
    t38 = t37*I;
    t39 = t33*t34*t38;
    t40 = t31*t39;
    t42 = A(p1,p5);
    t43 = t8*t42;
    t44 = A(p1,p6);
    t46 = t13*t15;
    t47 = t16*t18;
    t48 = t46*t47;
    t52 = t34*t34;
    t53 = t32*t52;
    t54 = t53*t38;
    t57 = t1*t13;
    t59 = t15*t2;
    t60 = t59*t47;
    t62 = AB(p1,t36,p6);
    t64 = t5*t5;
    t65 = t27*t64;
    t67 = t32*t34;
    t68 = t37*t37;
    t69 = t68*I;
    t70 = t67*t69;
    t73 = t8*t44;
    t74 = A(p3,p5);
    t75 = t73*t74;
    t76 = t12*t15;
    t77 = t2*t16;
    t78 = t76*t77;
    t80 = t27*t5;
    t81 = t21*t80;
    t84 = IA(p2,p3);
    t85 = t84*t16;
    t86 = t76*t85;
    t92 = t47*t20;
    t94 = AB(p5,t4,p6);
    t95 = t27*t94;
    t96 = t95*t5;
    t97 = t96*t54;
    t99 = t8*t13;
    t105 = t62*t27*t30;
    t106 = t105*t70;
    t108 = B(p4,p6);
    t109 = t108*t108;
    t110 = t109*t109;
    t114 = IB(p4,p5);
    t115 = IB(p5,p6);
    t116 = t114*t115;
    t118 = IS(p6+p5+p4);
    t120 = t37*t118*I;
    t121 = t116*t120;
    t123 = A(p1,p3);
    t124 = t123*t123;
    t125 = t124*t123;
    t126 = t42*t42;
    t127 = t126*t126;
    t128 = t125*t127;
    t129 = B(K2f,p6);
    t131 = IA(p1,p2);
    t132 = IA(p1,p4);
    t134 = IA(p1,p6);
    t135 = t131*t132*t134;
    t137 = t85*t18;
    t138 = AB(K2f,t36,p6);
    t140 = IAB(p1,t36,p6);
    t141 = t140*t140;
    t146 = t128*t15;
    t149 = t5*t140*I;
    t153 = t132*t132;
    t157 = AB(p4,t4,p6);
    t164 = AB(p1,t4,p6);
    t167 = t64*t140*I;
    t170 = t123*t12;
    t175 = t9*t126;
    t176 = t1*t1;
    t178 = t46*t77;
    t180 = t18*t19;
    t181 = t26*t64;
    t182 = t180*t181;
    t185 = t9*t42;
    t189 = t180*t65;
    t191 = t33*t52*t38;
    t197 = t180*t80;
    t200 = t9*t44;
    t202 = t17*t180;
    t208 = IB(p2,p3);
    t209 = t24*t208;
    t210 = t27*t29;
    t216 = t24*t24;
    t217 = t19*t216;
    t218 = t47*t217;
    t220 = t208*t26;
    t221 = t29*t29;
    t225 = t10*t13*t22*t40+t43*t44*t48*t20*t27*t5*t54-t43*t57*t60*t20*t62*t65*
t70+t75*t78*t81*t54-t75*t86*t21*t27*t54+t73*t13*t92*t97+t99*t15*t47*t20*t24*
t106-t99*t110*t20*t24*t121+t128*t129*t135*t137*t138*t5*t141*I+t146*t135*t137*
t149-t128*t131*t153*t134*t84*t47*t157*t149+t124*t127*t11*t135*t47*t164*t167-
t170*t110*t19*t24*t121-t175*t176*t178*t182*t39+t185*t44*t176*t178*t189*t191-
t185*t44*t57*t17*t197*t191-t200*t57*t202*t31*t191+t200*t13*t202*t209*t210*t191+
t9*t13*t15*t218*t220*t221*t39;
    t229 = t19*t62;
    t233 = t42*t44;
    t237 = t229*t65;
    t238 = t53*t69;
    t241 = t233*t1;
    t246 = t44*t1;
    t249 = t47*t19*t24;
    t256 = t208*t62;
    t264 = t62*t26;
    t268 = B(p2,p4);
    t273 = AB(p1,t4,p1);
    t277 = t8*t8;
    t281 = t19*t20;
    t284 = t67*t38;
    t290 = t59*t85;
    t297 = t8*t123;
    t300 = t21*t65;
    t310 = t8*t126;
    t318 = t17*t18;
    t330 = t43*t74;
    t332 = t26*t5;
    t342 = t43*t13;
    t345 = t24*t26*t30;
    t346 = t345*t284;
    t351 = t26*t94*t5;
    t352 = t351*t284;
    t363 = -t126*t176*t13*t60*t229*t181*t70+t233*t176*t13*t60*t237*t238-t241*
t48*t229*t80*t238-t246*t46*t249*t105*t238+t44*t13*t15*t249*t256*t210*t238+t46*
t16*t180*t216*t208*t264*t221*t70+t12*t268*t110*t217*t208*t116*t273*t120+TWO*
t277*t42*t13*t60*t281*t27*t64*t284-TWO*t277*t74*t12*t290*t18*t281*t80*t284+FOUR*
t297*t42*t78*t300*t284-FOUR*t297*t74*t11*t290*t81*t284+SIX*t310*t1*t178*t21*t181
*t284-TWO*t310*t13*t318*t20*t26*t5*t284-TWO*t43*t246*t178*t300*t54-TWO*t330*t78
*t21*t332*t284+TWO*t330*t86*t21*t26*t284-TWO*t342*t22*t346-TWO*t342*t92*t352
-TWO*t146*t2*t131*t132*t134*t16*t18*t167;
    t365 = t126*t42;
    t378 = t19*t27;
    t388 = t378*t5;
    t392 = t123*t9;
    t393 = t42*t1;
    t403 = t123*t126;
    t412 = t19*t26*t5;
    t416 = t123*t42;
    t434 = t416*t74;
    t435 = t11*t15;
    t436 = t435*t77;
    t442 = t435*t85;
    t448 = t416*t12;
    t452 = t47*t19;
    t456 = t123*t44;
    t457 = t456*t74;
    t480 = TWO*t125*t365*t131*t132*t134*t84*t47*t94*t149+TWO*t124*t42*t11*t60*
t378*t64*t284-TWO*t124*t74*t15*t2*t84*t47*t388*t284-TWO*t392*t393*t78*t189*t39+
TWO*t392*t12*t202*t40+THREE*FOUR*t403*t1*t78*t182*t284-FOUR*t403*t12*t318*t412*t284
-FOUR*t416*t246*t78*t189*t54+TWO*t416*t44*t76*t47*t388*t54-TWO*t416*t1*t12*t60*
t237*t70-FOUR*t434*t436*t180*t332*t284+FOUR*t434*t442*t180*t26*t284-FOUR*t448*t202
*t346-FOUR*t448*t452*t352+TWO*t457*t436*t197*t54-TWO*t457*t442*t180*t27*t54+TWO*
t456*t12*t452*t97+TWO*t170*t15*t249*t106+TWO*t175*t1*t48*t412*t39;
    t483 = t26*t29;
    t484 = t209*t483;
    t491 = t19*t25;
    t503 = t126*t44;
    t515 = t126*t1*t13;
    t522 = t25*t94;
    t527 = t126*t13;
    t530 = t208*t25;
    t536 = t180*t208;
    t544 = t44*t44;
    t545 = t42*t544;
    t548 = t52*t34;
    t550 = t32*t548*t38;
    t570 = t233*t13;
    t581 = t42*t13*t15;
    t611 = -TWO*t185*t13*t202*t484*t39+FOUR*t365*t176*t13*t60*t491*t64*t284-SIX*
t365*t1*t13*t318*t491*t5*t284-SIX*t503*t176*t178*t182*t54+SIX*t503*t1*t48*t412*
t54+TWO*t515*t202*t264*t5*t70-TWO*t515*t452*t522*t5*t284+FOUR*t527*t15*t249*t530
*t29*t284+TWO*t527*t16*t536*t522*t32*t34*t37*I+TWO*t545*t176*t178*t189*t550-TWO
*t545*t1*t48*t388*t550+TWO*t233*t57*t202*t345*t54+FOUR*t241*t13*t16*t180*t351*
t54-TWO*t570*t202*t484*t54-FOUR*t570*t452*t220*t94*t54-TWO*t581*t218*t530*t221*
t284-TWO*t581*t249*t256*t483*t70-TWO*t544*t1*t13*t452*t96*t550+TWO*t544*t13*t16
*t536*t95*t32*t548*t37*I-t10*t393*t178*t300*t39;
    return(t225+t363+t480+t611);
  }
}


std::complex<DOUBLE> C334x561mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t281;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t294;
  std::complex<DOUBLE> t298;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t308;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t327;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t334;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t360;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t364;
  FourMomentum<DOUBLE> t37;
  std::complex<DOUBLE> t371;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t375;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t398;
  std::complex<DOUBLE> t399;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t423;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t432;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t443;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t464;
  std::complex<DOUBLE> t468;
  std::complex<DOUBLE> t471;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t490;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t498;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t532;
  std::complex<DOUBLE> t535;
  std::complex<DOUBLE> t540;
  std::complex<DOUBLE> t545;
  std::complex<DOUBLE> t547;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t562;
  std::complex<DOUBLE> t565;
  std::complex<DOUBLE> t569;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t594;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t99;
  {
    t1 = A(p3,p4);
    t2 = B(p3,p4);
    t4 = p4+p3;
    t5 = IAB(p2,t4,p2);
    t8 = A(p1,p2);
    t9 = A(p2,p3);
    t10 = t9*t9;
    t11 = t10*t9;
    t12 = t8*t11;
    t13 = B(p1,p2);
    t14 = t1*t13;
    t15 = t12*t14;
    t16 = B(p2,p3);
    t17 = B(p2,p6);
    t18 = t16*t17;
    t19 = IA(p2,p4);
    t20 = t19*t19;
    t21 = t2*t20;
    t24 = IB(p1,p6);
    t25 = IB(p5,p6);
    t26 = t24*t25;
    t27 = AB(p2,t4,p6);
    t28 = t27*t27;
    t29 = t28*t27;
    t30 = IAB(p2,t4,p1);
    t31 = t30*t30;
    t32 = t29*t31;
    t33 = t26*t32;
    t34 = t5*t5;
    t35 = IAB(p2,t4,p5);
    t36 = t34*t35;
    t37 = p6+p5;
    t38 = IAA(p1,t37,t4,p2);
    t39 = t38*I;
    t40 = t36*t39;
    t41 = t33*t40;
    t43 = B(p2,p5);
    t44 = t43*t17;
    t45 = t2*t2;
    t46 = t19*t45;
    t49 = t35*t35;
    t51 = t34*t49*t39;
    t55 = t19*t24;
    t58 = AB(p1,t37,p2);
    t59 = t25*t58;
    t60 = t59*t32;
    t61 = t38*t38;
    t62 = t61*I;
    t63 = t36*t62;
    t66 = t1*t16;
    t67 = t12*t66;
    t70 = t29*t30;
    t71 = t26*t70;
    t72 = t71*t51;
    t74 = t17*t2;
    t75 = t20*t24;
    t79 = t59*t70*t63;
    t81 = t12*t1;
    t82 = t17*t17;
    t86 = t28*t30;
    t87 = t26*t86;
    t96 = t12*t13;
    t99 = t25*t29;
    t101 = t5*t35;
    t102 = t101*t39;
    t103 = t99*t31*t102;
    t105 = t2*t19;
    t106 = t44*t105;
    t109 = t5*t49*t39;
    t114 = IA(p2,p5);
    t115 = t114*t24;
    t116 = t105*t115;
    t118 = AB(p5,t4,p6);
    t119 = t118*t31;
    t123 = t43*t19;
    t124 = IA(p3,p4);
    t125 = t114*t124;
    t128 = t29*t118;
    t134 = t74*t55;
    t136 = t101*t62;
    t141 = t19*t114;
    t142 = t141*t26;
    t144 = t58*t29;
    t148 = t124*t24;
    t149 = t141*t148;
    t152 = t31*t35;
    t156 = t12*t16;
    t160 = t118*t30;
    t162 = t99*t160*t109;
    t168 = t144*t160*t136;
    t170 = t12*t43;
    t171 = t114*t114;
    t173 = t19*t171*t148;
    t175 = t25*t28;
    t176 = t118*t118;
    t179 = t30*t49*t39;
    t186 = t58*t28;
    t188 = t30*t35;
    t189 = t188*t62;
    t192 = -t15*t18*t21*t41+t15*t44*t46*t33*t51-t15*t17*t45*t55*t60*t63+t67*t44
*t21*t72-t67*t74*t75*t79+t81*t43*t82*t46*t87*t51-t81*t82*t45*t55*t59*t86*t63+
t96*t18*t75*t103-t96*t106*t33*t109-t12*t13*t43*t116*t99*t119*t109+t96*t123*t125
*t26*t128*t31*t49*t39+t96*t134*t60*t136+t12*t13*t2*t142*t144*t119*t136-t96*t149
*t59*t128*t152*t62-t156*t43*t20*t115*t162+t156*t20*t114*t26*t168-t170*t173*t175
*t176*t179+t12*t19*t171*t124*t26*t186*t176*t189;
    t193 = t8*t10;
    t194 = t193*t1;
    t196 = B(p4,p6);
    t197 = t196*t20;
    t200 = t33*t102;
    t202 = t193*t13;
    t206 = IB(p3,p4);
    t207 = t206*t25;
    t209 = t152*t39;
    t210 = t207*t29*t209;
    t212 = t8*t1;
    t213 = B(p2,p4);
    t214 = t213*t213;
    t215 = t214*t213;
    t217 = t82*t82;
    t218 = IB(p1,p2);
    t222 = IB(p2,p3);
    t223 = IB(p2,p5);
    t226 = IAB(p1,t37,p2);
    t228 = t226*t34*I;
    t234 = t218*t24*t222;
    t237 = t223*t206*t25;
    t242 = A(p1,p5);
    t243 = t242*t242;
    t244 = t243*t243;
    t246 = IA(p1,p6);
    t250 = IA(p5,p6);
    t253 = IS(p6+p5+p1);
    t255 = t38*t253*I;
    t256 = t114*t250*t255;
    t258 = t244*t10;
    t259 = A(p3,p5);
    t265 = AB(p2,t4,p2);
    t273 = t11*t13;
    t275 = t75*t25;
    t277 = AB(p1,t4,p2);
    t278 = t28*t28;
    t279 = t277*t278;
    t281 = t279*t31*t40;
    t290 = t214*t217;
    t291 = B(p4,p5);
    t294 = t223*t223;
    t298 = t5*I;
    t299 = t277*t226*t298;
    t307 = AB(p1,t37,p4);
    t308 = t226*t226;
    t321 = t16*t16;
    t324 = t20*t19*t24;
    t327 = t99*t30;
    t328 = t327*t40;
    t334 = t87*t40;
    t337 = t82*t17;
    t341 = t25*t27;
    t342 = t341*t30;
    t347 = t82*t2*t55;
    t360 = t17*t19;
    t361 = t360*t125;
    t363 = t28*t118;
    t364 = t26*t363;
    t371 = t175*t30;
    t372 = t371*t102;
    t375 = t194*t13*t16*t197*t200-t202*t16*t196*t75*t210+t212*t215*t217*t218*
t24*t222*t223*t25*t228-t8*t215*t217*t234*t237*t226*t5*I-t244*t11*t16*t246*t20*
t256+t258*t259*t246*t19*t171*t124*t250*t265*t255-t258*t213*t246*t19*t256-t273*
t16*t275*t281+t9*t214*t217*t234*t223*t25*t277*t228+t290*t291*t234*t294*t206*t25
*t299+t290*t218*t24*t222*t223*t207*t277*t307*t308*t298-TWO*t81*t13*t82*t46*t26*
t28*t31*t40-TWO*t81*t321*t17*t324*t328-FOUR*t81*t16*t82*t21*t334-TWO*t81*t337*
t45*t55*t342*t40+TWO*t96*t347*t175*t31*t102+TWO*t12*t13*t17*t116*t175*t119*t102
-TWO*t96*t361*t364*t209+TWO*t156*t82*t20*t24*t372;
    t377 = t17*t20;
    t381 = t175*t160*t102;
    t398 = t12*t82;
    t399 = t105*t26;
    t407 = t188*t39;
    t411 = t12*t17;
    t423 = t213*t17;
    t432 = t207*t70*t102;
    t443 = t87*t102;
    t447 = t193*t1*t213;
    t464 = t17*t196;
    t468 = t207*t28*t407;
    t471 = t193*t213;
    t481 = TWO*t156*t377*t115*t381-TWO*t170*t347*t371*t109+TWO*t170*t361*t364*
t179+FOUR*t12*t337*t105*t24*t342*t102+TWO*t398*t399*t186*t30*t136-FOUR*t398*t149*
t341*t118*t407+TWO*t411*t173*t341*t176*t407-TWO*t411*t149*t59*t363*t189-TWO*
t193*t14*t423*t105*t41+TWO*t194*t321*t196*t324*t432-FOUR*t194*t16*t213*t377*t71*
t40+TWO*t194*t18*t197*t443+TWO*t447*t106*t72-EIGHT*t194*t213*t82*t105*t334-TWO*
t447*t134*t79+TWO*t202*t423*t55*t103-TWO*t193*t16*t464*t75*t468-TWO*t471*t123*
t115*t162+FOUR*t471*t82*t19*t24*t372;
    t489 = t9*t8;
    t490 = t489*t1;
    t492 = t196*t19;
    t498 = t213*t196;
    t529 = t13*t13;
    t530 = t11*t529;
    t532 = t55*t25;
    t535 = t279*t31*t30;
    t540 = t148*t25;
    t545 = t273*t17;
    t547 = t277*t29;
    t548 = t547*t31;
    t562 = t547*t30*t40;
    t565 = t11*t82;
    t569 = t277*t28*t30;
    t594 = FOUR*t471*t360*t115*t381+TWO*t471*t142*t168+TWO*t490*t13*t213*t492*
t200+FOUR*t489*t66*t498*t75*t432-TWO*t490*t214*t17*t55*t328+FOUR*t490*t423*t492*
t443-TWO*t489*t13*t498*t55*t210-FOUR*t489*t213*t464*t55*t468+TWO*t212*t214*t492*
t24*t206*t327*t102-TWO*t530*t2*t532*t535*t40+TWO*t530*t19*t540*t535*t102-FOUR*
t545*t399*t548*t40+FOUR*t545*t19*t124*t26*t548*t102-TWO*t11*t16*t17*t275*t562
-TWO*t565*t2*t532*t569*t40+TWO*t565*t19*t540*t569*t102-TWO*t10*t13*t213*t532*
t281-FOUR*t10*t213*t17*t532*t562-TWO*t214*t337*t196*t234*t237*t299;
    return(t192+t375+t481+t594);
  }
}


std::complex<DOUBLE> C345x612mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K2f;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t339;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t352;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t376;
  std::complex<DOUBLE> t379;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t393;
  std::complex<DOUBLE> t396;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t401;
  std::complex<DOUBLE> t404;
  std::complex<DOUBLE> t412;
  std::complex<DOUBLE> t439;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t459;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t462;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t485;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t493;
  std::complex<DOUBLE> t498;
  std::complex<DOUBLE> t499;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t519;
  FourMomentum<DOUBLE> t52;
  std::complex<DOUBLE> t528;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t564;
  std::complex<DOUBLE> t575;
  std::complex<DOUBLE> t588;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t612;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t99;
  {
    t1 = A(p4,p5);
    t2 = B(p4,p5);
    t4 = p5+p4;
    t5 = IAB(p3,t4,p3);
    K2f = p5+p4+t1*t2*t5*p3;
    t8 = A(p1,p3);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = A(p3,p5);
    t12 = t11*t11;
    t13 = t11*t12;
    t15 = B(p2,p3);
    t16 = t10*t13*t15;
    t17 = IA(p1,p2);
    t18 = IA(p1,p6);
    t19 = t17*t18;
    t20 = IA(p2,p3);
    t23 = IA(p3,p4);
    t24 = IA(p3,p6);
    t26 = IA(p4,p5);
    t28 = IAB(p3,t4,p2);
    t30 = t28*t5*I;
    t33 = t10*t12;
    t34 = A(p5,p6);
    t38 = t20*t23;
    t39 = t24*t24;
    t45 = B(K2f,p4);
    t49 = t24*t26;
    t50 = IB(p4,p5);
    t52 = p6+p1;
    t53 = AB(K2f,t52,p2);
    t59 = A(p2,p3);
    t60 = t8*t59;
    t61 = A(p3,p4);
    t64 = B(p3,p4);
    t65 = t64*t64;
    t66 = t65*t64;
    t67 = t15*t66;
    t68 = t17*t2;
    t69 = t67*t68;
    t71 = IB(p3,p5);
    t72 = t71*t71;
    t73 = t18*t72;
    t74 = AB(p1,t4,p3);
    t75 = t74*t74;
    t76 = t75*t74;
    t77 = IAB(p2,t4,p3);
    t78 = t77*t77;
    t79 = t76*t78;
    t81 = t5*t5;
    t82 = IAB(p6,t4,p3);
    t83 = t81*t82;
    t84 = IBB(p3,t4,t52,p2);
    t85 = t84*I;
    t86 = t83*t85;
    t89 = t59*t61;
    t90 = t89*t11;
    t91 = t15*t65;
    t92 = t91*t68;
    t94 = t75*t75;
    t95 = t94*t78;
    t99 = t91*t19;
    t101 = t26*t72;
    t103 = t5*t82;
    t104 = t103*t85;
    t108 = t19*t72;
    t110 = AB(p1,t4,p2);
    t111 = t110*t76;
    t113 = t111*t78*t104;
    t115 = t59*t11;
    t116 = A(p3,p6);
    t117 = t116*t1;
    t120 = t18*t71;
    t122 = t82*t82;
    t124 = t81*t122*t85;
    t129 = t71*t94;
    t130 = t129*t78;
    t132 = t5*t122*t85;
    t135 = t1*t15;
    t138 = t65*t2*t19;
    t140 = AB(p3,t4,p2);
    t141 = t140*t78;
    t143 = t84*t84;
    t144 = t143*I;
    t145 = t83*t144;
    t149 = t65*t17;
    t150 = t149*t120;
    t152 = t94*t140;
    t154 = t103*t144;
    t157 = t59*t116;
    t159 = B(p4,p6);
    t161 = t65*t159*t19;
    t163 = IB(p3,p6);
    t164 = t71*t163;
    t170 = t50*t94;
    t178 = t159*t17;
    t179 = t178*t120;
    t181 = t163*t94;
    t188 = t163*t50;
    t190 = t78*t82;
    t194 = t61*t11;
    t197 = t94*t77;
    t204 = t140*t77;
    t208 = t16*t19*t20*t23*t24*t26*t30-t33*t34*t17*t18*t38*t39*t26*t5*I-t33*t45
*t19*t38*t49*t50*t53*t28*t5*I-t60*t61*t1*t69*t73*t79*t86+t90*t92*t73*t95*t86-
t90*t99*t101*t95*t104+t89*t66*t108*t113-t115*t117*t92*t120*t95*t124+t115*t116*
t99*t130*t132+t115*t135*t138*t129*t141*t145-t115*t15*t150*t152*t78*t154-t157*
t135*t161*t164*t95*t132+t157*t15*t161*t164*t170*t78*t122*t85+t59*t1*t91*t179*
t181*t141*t154-t59*t15*t65*t179*t188*t152*t190*t144-t194*t116*t92*t73*t197*t124
+t194*t15*t138*t72*t94*t204*t145;
    t212 = t72*t163;
    t223 = B(p2,p6);
    t224 = t223*t223;
    t225 = t224*t224;
    t227 = IB(p1,p2);
    t229 = IB(p1,p6);
    t233 = IS(p6+p2+p1);
    t235 = t84*t233*I;
    t241 = t64*t2*t19;
    t243 = t129*t77;
    t261 = t159*t159;
    t263 = t261*t17*t120;
    t265 = t163*t163;
    t266 = t265*t50;
    t269 = t77*t122*t85;
    t278 = t77*t82;
    t279 = t278*t144;
    t287 = AB(p3,t4,p3);
    t299 = t9*t8;
    t309 = t1*t1;
    t316 = t71*t74*t77;
    t322 = t66*t17;
    t328 = t9*t59;
    t337 = t67*t19;
    t339 = t71*t75;
    t347 = t75*t77;
    t352 = t9*t11;
    t366 = -t61*t116*t15*t161*t212*t197*t132+t61*t15*t65*t178*t73*t181*t204*
t154+t61*t225*t66*t227*t229*t212*t235-t12*t116*t15*t241*t243*t124+t12*t15*t64*
t68*t120*t152*t77*t145+t11*t225*t65*t227*t229*t164*t235+t116*t15*t64*t263*t266*
t94*t269-t15*t64*t261*t19*t71*t265*t170*t140*t279-t225*t65*t159*t227*t229*t71*
t266*t287*t235-THREE*t16*t68*t18*t38*t24*t28*t81*I+TWO*t299*t13*t17*t18*t20*t23*
t49*t110*t30-TWO*t299*t309*t15*t66*t2*t19*t316*t86+TWO*t299*t1*t15*t322*t18*
t316*t104+FOUR*t328*t309*t69*t120*t75*t78*t86-FOUR*t328*t1*t337*t339*t78*t104+TWO
*t9*t61*t1*t69*t73*t347*t86+EIGHT*t352*t1*t92*t120*t347*t86-FOUR*t352*t15*t149*t18
*t339*t77*t104;
    t368 = t1*t9;
    t376 = t19*t71;
    t379 = t110*t74*t77;
    t387 = t278*t85;
    t393 = t120*t50;
    t396 = t82*t84*I;
    t400 = t59*t59;
    t401 = t8*t400;
    t404 = t78*t77;
    t412 = t71*t76;
    t439 = t110*t75;
    t447 = t164*t50*t76;
    t448 = t190*t85;
    t459 = t8*t61;
    t460 = t459*t11;
    t462 = t76*t77;
    t481 = t439*t77*t104;
    t485 = t8*t12*t15;
    t487 = t412*t77;
    t493 = t64*t17*t18*t26;
    t498 = FOUR*t368*t15*t161*t164*t347*t104+TWO*t368*t66*t376*t379*t104-FOUR*t9*
t15*t65*t179*t188*t75*t387-TWO*t9*t66*t17*t393*t379*t396-TWO*t401*t309*t69*t120
*t76*t404*t86+TWO*t401*t1*t337*t412*t404*t104-FOUR*t60*t11*t1*t92*t120*t79*t86+
TWO*t60*t11*t99*t412*t78*t104-TWO*t60*t135*t161*t164*t79*t104-FOUR*t60*t1*t322*
t120*t439*t78*t104+TWO*t60*t15*t161*t447*t448+FOUR*t60*t66*t376*t50*t110*t75*
t448-FOUR*t460*t92*t73*t462*t86+TWO*t460*t99*t101*t462*t104-TWO*t459*t15*t161*
t212*t462*t104-TWO*t459*t66*t108*t481-SIX*t485*t241*t487*t86+FOUR*t485*t493*t487
*t104;
    t499 = t8*t11;
    t519 = t76*t140;
    t528 = t8*t116;
    t548 = t8*t15;
    t564 = t111*t404;
    t575 = t59*t12*t15;
    t588 = t61*t61;
    t612 = TWO*t499*t117*t92*t120*t462*t124-TWO*t499*t116*t99*t487*t132-TWO*
t499*t135*t138*t412*t204*t145+TWO*t499*t15*t150*t519*t77*t154-FOUR*t499*t65*t376
*t481+TWO*t528*t135*t161*t164*t462*t132-TWO*t528*t15*t161*t447*t269-TWO*t8*t1*
t91*t179*t163*t76*t204*t154+TWO*t548*t65*t179*t188*t519*t279+TWO*t548*t64*t263*
t266*t76*t387+TWO*t400*t1*t66*t376*t564*t104-TWO*t400*t66*t17*t393*t564*t396+
TWO*t575*t241*t130*t86-TWO*t575*t493*t130*t104+TWO*t115*t65*t376*t113+TWO*t588*
t11*t15*t138*t26*t72*t71*t197*t86+FOUR*t61*t12*t15*t241*t101*t197*t86+TWO*t13*
t15*t2*t19*t26*t243*t86;
    return(t208+t366+t498+t612);
  }
}


std::complex<DOUBLE> C356x123mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t246;
  FourMomentum<DOUBLE> t25;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t286;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t298;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t313;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t351;
  std::complex<DOUBLE> t358;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t39;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t423;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t431;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t440;
  std::complex<DOUBLE> t449;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t454;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t478;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t505;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t520;
  std::complex<DOUBLE> t523;
  std::complex<DOUBLE> t527;
  std::complex<DOUBLE> t528;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t545;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t551;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t577;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t1 = A(p5,p6);
    t2 = B(p5,p6);
    t4 = p6+p5;
    t5 = IAB(p4,t4,p4);
    t8 = A(p1,p3);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = A(p4,p5);
    t12 = t11*t11;
    t13 = t11*t12;
    t14 = t10*t13;
    t15 = B(p4,p5);
    t16 = IA(p1,p2);
    t18 = IA(p1,p4);
    t21 = IA(p2,p3);
    t22 = IA(p4,p6);
    t23 = t22*t22;
    t25 = p2+p1;
    t26 = IAA(p3,t25,t4,p4);
    t28 = IS(p3+p2+p1);
    t30 = t26*t28*I;
    t33 = t18*t18;
    t37 = IA(p5,p6);
    t38 = t22*t37;
    t39 = AB(p1,t4,p4);
    t43 = A(p3,p4);
    t44 = t43*t13;
    t45 = t44*t1;
    t46 = B(p1,p4);
    t47 = B(p2,p4);
    t48 = t47*t47;
    t50 = t2*t2;
    t51 = t50*t22;
    t54 = IB(p1,p2);
    t55 = IB(p2,p3);
    t56 = t54*t55;
    t57 = AB(p4,t4,p2);
    t58 = t57*t57;
    t59 = IAB(p4,t25,p3);
    t60 = t58*t59;
    t61 = t56*t60;
    t62 = IAB(p4,t4,p1);
    t63 = t62*t62;
    t64 = t5*t5;
    t66 = t26*I;
    t67 = t63*t64*t66;
    t70 = t1*t46;
    t71 = t44*t70;
    t72 = B(p3,p4);
    t73 = t47*t72;
    t76 = t58*t57;
    t77 = t59*t59;
    t78 = t76*t77;
    t79 = t56*t78;
    t82 = t47*t15;
    t83 = t2*t23;
    t86 = t76*t59;
    t87 = t56*t86;
    t88 = t87*t67;
    t91 = t22*t54;
    t94 = AB(p3,t25,p4);
    t95 = t55*t94;
    t97 = t62*t64;
    t98 = t26*t26;
    t99 = t98*I;
    t100 = t97*t99;
    t103 = t1*t47;
    t104 = t44*t103;
    t105 = t72*t15;
    t108 = t97*t66;
    t109 = t79*t108;
    t114 = t95*t78;
    t118 = t23*t54;
    t122 = t95*t86*t100;
    t124 = t44*t46;
    t125 = t2*t22;
    t129 = t63*t5*t66;
    t135 = t2*t18*t91;
    t137 = AB(p1,t4,p2);
    t138 = t55*t137;
    t143 = t72*t18*t38;
    t152 = t15*t18*t118;
    t155 = t138*t86*t129;
    t158 = t37*t54;
    t159 = t33*t22*t158;
    t161 = t137*t137;
    t162 = t55*t161;
    t165 = t59*t63*t66;
    t168 = t44*t47;
    t169 = t72*t2;
    t170 = t169*t91;
    t172 = t62*t5;
    t173 = t172*t99;
    t177 = t18*t22;
    t178 = t177*t56;
    t180 = t137*t94;
    t185 = t177*t158;
    t189 = t77*t62;
    t193 = -t14*t15*t16*t18*t21*t23*t30+t14*t16*t33*t21*t38*t39*t30-t45*t46*t48
*t51*t61*t67-t71*t73*t51*t79*t67+t71*t82*t83*t88-t45*t48*t50*t91*t95*t60*t100-
t104*t105*t83*t109-t104*t72*t50*t91*t114*t100+t104*t15*t2*t118*t122+t124*t73*
t125*t79*t129+t44*t46*t72*t135*t138*t78*t129-t124*t143*t56*t137*t76*t77*t63*t66
-t124*t152*t155+t124*t159*t162*t58*t165+t168*t170*t114*t173+t44*t169*t178*t180*
t78*t173-t44*t72*t185*t138*t94*t76*t189*t99;
    t199 = t180*t86*t173;
    t206 = t59*t62;
    t207 = t206*t99;
    t210 = t43*t12;
    t211 = t210*t1;
    t212 = B(p4,p6);
    t213 = t212*t23;
    t216 = t58*t58;
    t217 = t216*t77;
    t223 = t15*t212*t118;
    t225 = IB(p5,p6);
    t226 = t55*t225;
    t228 = t172*t66;
    t231 = t48*t48;
    t233 = t212*t212;
    t234 = t233*t212;
    t236 = IB(p1,p4);
    t240 = IB(p3,p4);
    t241 = IB(p4,p5);
    t244 = IAB(p3,t25,p4);
    t246 = t244*t5*I;
    t251 = t118*t55;
    t253 = AB(p3,t4,p2);
    t254 = t253*t76;
    t256 = t254*t77*t228;
    t260 = t236*t236;
    t264 = t241*t225;
    t265 = AB(p3,t4,p1);
    t272 = t236*t55*t240;
    t274 = AB(p3,t25,p6);
    t286 = t48*t47;
    t291 = t55*t57*t59;
    t298 = t58*t77;
    t306 = t61*t108;
    t309 = t72*t72;
    t313 = t77*t59;
    t322 = t55*t58;
    t331 = t56*t137*t58;
    t341 = t44*t48;
    t347 = -t44*t15*t18*t23*t56*t199+t44*t33*t38*t56*t161*t94*t58*t207+t211*
t105*t213*t56*t217*t108-t210*t72*t223*t226*t217*t228-t43*t231*t234*t54*t236*t55
*t240*t241*t225*t246+t13*t72*t15*t251*t256+t231*t234*t54*t260*t55*t240*t264*
t265*t246-t231*t233*t54*t272*t264*t274*t246-TWO*t10*t12*t212*t16*t18*t21*t22*
t30+FOUR*t45*t286*t50*t91*t291*t108+SIX*t45*t48*t72*t51*t56*t298*t108-FOUR*t45*
t48*t15*t83*t306+TWO*t45*t47*t309*t51*t56*t76*t313*t108+TWO*t124*t48*t2*t91*
t322*t59*t129-TWO*t124*t47*t18*t38*t331*t165-SIX*t44*t286*t125*t54*t291*t228
-SIX*t341*t170*t322*t77*t228;
    t349 = t125*t56;
    t351 = t94*t58;
    t358 = t206*t66;
    t365 = t55*t76;
    t377 = t189*t66;
    t383 = t138*t60*t228;
    t407 = t210*t103;
    t423 = t15*t15;
    t431 = t226*t216*t59*t108;
    t436 = t212*t18*t91;
    t440 = t210*t47;
    t449 = TWO*t341*t349*t351*t59*t173+FOUR*t341*t185*t138*t57*t358-TWO*t168*
t309*t2*t91*t365*t313*t228-TWO*t44*t73*t135*t138*t298*t228+TWO*t168*t143*t331*
t377+TWO*t168*t152*t383-TWO*t168*t159*t162*t57*t358-TWO*t168*t185*t138*t351*
t207+TWO*t210*t70*t47*t212*t125*t88-EIGHT*t211*t48*t212*t125*t306-TWO*t407*t72*
t212*t125*t109+TWO*t211*t82*t213*t87*t108+TWO*t407*t212*t2*t91*t122-TWO*t211*
t423*t212*t23*t22*t54*t431-TWO*t210*t46*t436*t155-TWO*t440*t223*t226*t86*t228+
FOUR*t440*t436*t383;
    t454 = t43*t11;
    t455 = t454*t1;
    t478 = t233*t22*t56;
    t487 = t225*t216;
    t492 = t43*t1;
    t505 = t91*t55;
    t511 = t13*t48;
    t515 = t253*t57*t59;
    t520 = t158*t55;
    t523 = t62*t26*I;
    t527 = t13*t47;
    t528 = t527*t72;
    t530 = t253*t58;
    t545 = t530*t59*t228;
    t548 = t13*t309;
    t551 = t254*t313;
    t577 = -TWO*t210*t212*t178*t199+FOUR*t455*t47*t233*t91*t365*t59*t108+TWO*
t455*t72*t233*t91*t55*t216*t77*t108-FOUR*t455*t15*t233*t118*t431-FOUR*t454*t47*
t478*t225*t76*t59*t228-TWO*t454*t72*t478*t487*t77*t228+TWO*t492*t231*t234*t54*
t236*t55*t240*t241*t244*t64*I-TWO*t492*t234*t505*t487*t59*t108-TWO*t511*t2*t505
*t515*t228+TWO*t511*t22*t520*t515*t523-FOUR*t528*t349*t530*t77*t228+FOUR*t528*t38
*t54*t55*t253*t58*t377+TWO*t527*t15*t251*t545-TWO*t548*t2*t505*t551*t228+TWO*
t548*t22*t520*t551*t523+FOUR*t12*t47*t212*t505*t545+TWO*t12*t72*t212*t505*t256
-TWO*t286*t234*t54*t272*t264*t253*t246;
    return(t193+t347+t449+t577);
  }
}


std::complex<DOUBLE> C361x234mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K2f;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t285;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t298;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t305;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t339;
  FourMomentum<DOUBLE> t34;
  std::complex<DOUBLE> t340;
  std::complex<DOUBLE> t343;
  std::complex<DOUBLE> t346;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t389;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t393;
  std::complex<DOUBLE> t395;
  std::complex<DOUBLE> t399;
  FourMomentum<DOUBLE> t4;
  std::complex<DOUBLE> t403;
  std::complex<DOUBLE> t404;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t451;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t497;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t546;
  std::complex<DOUBLE> t547;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t552;
  std::complex<DOUBLE> t555;
  std::complex<DOUBLE> t556;
  std::complex<DOUBLE> t567;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t98;
  {
    t1 = A(p1,p6);
    t2 = B(p1,p6);
    t4 = p6+p1;
    t5 = IAB(p5,t4,p5);
    K2f = p6+p1+t1*t2*t5*p5;
    t8 = A(p1,p5);
    t9 = A(p3,p5);
    t10 = t9*t8;
    t11 = A(p5,p6);
    t12 = t10*t11;
    t13 = B(p4,p5);
    t14 = B(p5,p6);
    t15 = t14*t14;
    t16 = t13*t15;
    t17 = IA(p1,p6);
    t18 = IA(p2,p3);
    t19 = t17*t18;
    t20 = t16*t19;
    t22 = IA(p3,p4);
    t23 = IB(p1,p5);
    t24 = t23*t23;
    t25 = t22*t24;
    t26 = AB(p3,t4,p5);
    t27 = t26*t26;
    t28 = t27*t26;
    t29 = IAB(p2,t4,p5);
    t30 = t28*t29;
    t31 = t25*t30;
    t32 = IAB(p4,t4,p5);
    t33 = t32*t5;
    t34 = p3+p2;
    t35 = IBB(p5,t4,t34,p4);
    t36 = t35*I;
    t37 = t33*t36;
    t42 = t15*t18;
    t43 = t22*t23;
    t44 = t42*t43;
    t46 = AB(p5,t34,p4);
    t47 = t28*t46;
    t49 = t35*t35;
    t50 = t49*I;
    t51 = t33*t50;
    t55 = B(p4,p6);
    t57 = t14*t18;
    t58 = t57*t22;
    t60 = t23*t28;
    t61 = t60*t29;
    t62 = t61*t37;
    t65 = B(p2,p6);
    t70 = IB(p2,p5);
    t71 = t27*t27;
    t72 = t70*t71;
    t73 = t46*t29;
    t74 = t72*t73;
    t77 = B(p2,p4);
    t78 = t77*t77;
    t79 = t78*t78;
    t82 = IB(p1,p6);
    t83 = t23*t82;
    t84 = IB(p2,p3);
    t87 = t70*t70;
    t88 = IB(p3,p4);
    t90 = AB(p5,t4,p5);
    t93 = IS(p4+p3+p2);
    t95 = t35*t93*I;
    t98 = t8*t1;
    t101 = t18*t22;
    t102 = t16*t101;
    t105 = t5*t5;
    t106 = t32*t105;
    t107 = t106*t50;
    t111 = A(p4,p5);
    t112 = t111*t111;
    t115 = t2*t13*t42;
    t117 = t71*t29;
    t119 = t32*t32;
    t120 = t119*t32;
    t126 = A(p2,p5);
    t127 = t8*t126;
    t130 = t29*t29;
    t135 = t9*t9;
    t138 = t42*t22;
    t142 = t23*t27*t29*t37;
    t147 = t119*t5;
    t148 = t147*t36;
    t149 = t61*t148;
    t152 = t8*t111;
    t153 = t152*t11;
    t155 = t25*t117;
    t156 = t119*t105;
    t157 = t156*t36;
    t163 = t23*t71;
    t164 = t163*t29;
    t165 = t164*t148;
    t168 = t1*t9;
    t175 = t65*t13;
    t178 = t70*t28;
    t188 = t164*t120*t5*t36;
    t191 = t126*t9;
    t194 = t83*t178;
    t196 = t130*t32*t36;
    t200 = -TWO*t12*t20*t31*t37+TWO*t10*t13*t44*t47*t29*t51+FOUR*t10*t55*t58*t62
+t11*t65*t13*t42*t25*t74*t51-t79*t65*t15*t83*t84*t87*t88*t90*t95-TWO*t98*t9*t2*
t102*t60*t73*t107-TWO*t98*t112*t115*t43*t117*t120*t105*t36-TWO*t127*t9*t102*t60
*t130*t37+SIX*t8*t135*t13*t138*t142-SIX*t10*t111*t102*t149-TWO*t153*t115*t155*
t157-TWO*t152*t55*t58*t165+FOUR*t168*t111*t55*t15*t101*t149-TWO*t168*t175*t44*
t178*t73*t51-TWO*t1*t112*t55*t138*t188-TWO*t191*t65*t102*t194*t196;
    t204 = t82*t70;
    t206 = t29*t32;
    t207 = t206*t36;
    t213 = t101*t23;
    t222 = t9*t111;
    t225 = t29*t119;
    t226 = t225*t36;
    t236 = t9*t11;
    t239 = t24*t70;
    t251 = t65*t65;
    t254 = t57*t43;
    t256 = t82*t87;
    t261 = t9*t65;
    t265 = t206*t50;
    t270 = t106*t36;
    t274 = t8*t8;
    t275 = t274*t111;
    t277 = t13*t14;
    t278 = t277*t101;
    t285 = t14*t17*t101;
    t298 = t28*t130;
    t305 = t27*t29;
    t321 = t1*t135;
    t326 = FOUR*t135*t65*t13*t44*t204*t27*t207+TWO*t135*t55*t15*t213*t82*t27*t29
*t32*t35*I-TWO*t222*t65*t102*t194*t226-FOUR*t222*t55*t138*t83*t28*t226+TWO*t236*
t65*t102*t239*t30*t37+TWO*t236*t55*t138*t24*t28*t29*t37-TWO*t9*t251*t13*t254*
t256*t28*t207+TWO*t261*t13*t44*t204*t47*t265+SIX*t12*t115*t31*t270-FOUR*t275*t2*
t278*t164*t157+TWO*t275*t13*t285*t165-FOUR*t274*t11*t2*t277*t19*t155*t270+TWO*
t98*t191*t115*t43*t298*t270-SIX*t98*t135*t115*t43*t305*t270+SIX*t98*t222*t115*
t43*t30*t157+TWO*t8*t112*t13*t138*t188-TWO*t321*t55*t138*t142;
    t328 = t111*t65;
    t331 = t23*t70;
    t339 = t71*t46;
    t340 = t339*t29;
    t343 = t126*t111;
    t346 = t71*t130;
    t359 = t163*t130;
    t372 = t147*t50;
    t378 = t24*t71;
    t382 = t274*t8;
    t383 = t135*t135;
    t384 = t382*t383;
    t386 = IA(p2,p5);
    t389 = IA(p4,p5);
    t391 = IA(p5,p6);
    t392 = t22*t389*t391;
    t393 = IAB(p5,t34,p4);
    t395 = t393*t5*I;
    t399 = t386*t386;
    t403 = t389*t391;
    t404 = AB(p2,t4,p4);
    t426 = t393*t105*I;
    t434 = t18*t386*t22;
    t436 = AB(p3,t4,p4);
    t441 = t11*t11;
    t451 = TWO*t168*t328*t102*t331*t30*t148+t274*t2*t13*t254*t340*t107-t98*t343
*t115*t43*t346*t157+t98*t111*t2*t102*t163*t73*t156*t50+t127*t111*t102*t359*t148
-t127*t11*t115*t25*t346*t270+t153*t20*t155*t148-t152*t13*t44*t340*t372+t8*t11*
t2*t102*t378*t73*t107-t384*t13*t19*t386*t392*t395+t384*t17*t18*t399*t22*t403*
t404*t395-t274*t126*t2*t278*t359*t270-t251*t13*t14*t101*t83*t87*t71*t46*t265+
TWO*t384*t2*t13*t18*t386*t392*t426-TWO*t382*t135*t9*t17*t434*t403*t436*t395-TWO
*t8*t441*t2*t20*t22*t24*t23*t117*t270;
    t452 = t1*t126;
    t472 = t274*t9;
    t497 = t70*t88*t95;
    t546 = t274*t383;
    t547 = B(K2f,p6);
    t552 = AB(K2f,t34,p4);
    t555 = AB(p5,t4,p4);
    t556 = t393*t393;
    t567 = TWO*t452*t261*t102*t331*t298*t37-FOUR*t321*t65*t102*t331*t305*t37-TWO
*t382*t2*t13*t19*t22*t164*t270+TEN*t472*t2*t278*t61*t270-FOUR*t472*t13*t285*t62
+TWO*t112*t55*t15*t213*t82*t71*t29*t120*t35*I+t8*t79*t15*t23*t84*t497-t452*t328
*t102*t331*t346*t148+t1*t111*t175*t44*t74*t372+t343*t65*t102*t83*t72*t130*t119*
t36-t126*t11*t65*t102*t239*t346*t37+t126*t251*t13*t254*t256*t71*t196-t111*t11*
t55*t138*t378*t29*t148-t328*t13*t44*t204*t339*t225*t50+t11*t79*t15*t14*t24*t84*
t497-t546*t547*t19*t386*t22*t403*t82*t552*t555*t556*t5*I+t546*t14*t434*t403*
t555*t426;
    return(t200+t326+t451+t567);
  }
}


std::complex<DOUBLE> C3123x45mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t229;
  FourMomentum<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t249;
  FourMomentum<DOUBLE> t25;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t304;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t334;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t340;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t350;
  std::complex<DOUBLE> t353;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t367;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t406;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t419;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t469;
  std::complex<DOUBLE> t486;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t545;
  std::complex<DOUBLE> t568;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  {
    t1 = p5+p4;
    t2 = IAB(p6,t1,p6);
    t3 = p3+p2+p1;
    t4 = S(t3);
    t7 = A(p1,p3);
    t8 = A(p5,p6);
    t9 = t8*t8;
    t10 = t9*t8;
    t11 = t7*t10;
    t12 = IA(p1,p6);
    t13 = t11*t12;
    t14 = IA(p3,p6);
    t15 = IA(p4,p5);
    t16 = t14*t15;
    t17 = IA(p4,p6);
    t18 = t17*t17;
    t19 = IB(p1,p2);
    t20 = t18*t19;
    t23 = p2+p1;
    t24 = AB(p3,t23,p6);
    t25 = p3+p1;
    t26 = AB(p4,t25,p2);
    t27 = t24*t26;
    t28 = AB(p6,t25,p2);
    t29 = t28*t28;
    t31 = IAB(p6,t23,p3);
    t33 = IAB(p6,p3+p2,p1);
    t34 = t31*t33;
    t35 = IS(t3);
    t36 = t35*t35;
    t37 = t36*I;
    t38 = t34*t37;
    t41 = t7*t9;
    t42 = B(p2,p6);
    t43 = t41*t42;
    t44 = B(p4,p6);
    t46 = t15*t17;
    t49 = IB(p2,p3);
    t50 = IB(p4,p5);
    t51 = t49*t50;
    t53 = t2*I;
    t54 = t34*t53;
    t61 = t50*t26;
    t62 = t61*t29;
    t63 = t35*I;
    t64 = t34*t63;
    t67 = B(p1,p6);
    t68 = t10*t67;
    t69 = t68*t42;
    t70 = B(p3,p6);
    t72 = t17*t19;
    t75 = t28*t29;
    t76 = t49*t75;
    t77 = t31*t31;
    t79 = t33*t33;
    t80 = t2*t2;
    t82 = t4*I;
    t83 = t79*t80*t82;
    t86 = t10*t42;
    t87 = t86*t70;
    t88 = t16*t72;
    t90 = t49*t24;
    t91 = t90*t75;
    t92 = t77*t33;
    t93 = t80*I;
    t97 = t15*t18;
    t98 = t97*t19;
    t100 = t49*t26;
    t101 = t100*t29;
    t102 = t92*t53;
    t106 = t19*t49;
    t107 = t97*t106;
    t109 = t29*t31;
    t111 = t33*t2;
    t112 = t111*t63;
    t115 = t9*t67;
    t116 = t115*t42;
    t117 = t44*t15;
    t120 = t75*t31;
    t127 = t31*t79;
    t128 = t127*t53;
    t131 = t9*t42;
    t132 = t131*t44;
    t134 = t24*t75;
    t136 = t34*t93;
    t139 = t9*t44;
    t141 = t20*t51;
    t146 = t11*t42;
    t148 = t70*t14*t15;
    t150 = t17*t49;
    t151 = t150*t29;
    t155 = t12*t14;
    t156 = t155*t46;
    t158 = t19*t24;
    t163 = t12*t15;
    t164 = t163*t18;
    t166 = t19*t26;
    t171 = t14*t14;
    t172 = t171*t15;
    t173 = t172*t150;
    t176 = t24*t29*t31;
    t180 = t16*t18;
    t186 = -t13*t16*t20*t27*t29*t38-t43*t44*t14*t46*t51*t29*t54-t41*t44*t16*t18
*t49*t62*t64+t69*t70*t15*t72*t76*t77*t83+t87*t88*t91*t92*t93-t87*t98*t101*t102-
t86*t14*t107*t27*t109*t112+t116*t117*t72*t51*t120*t83+t115*t44*t107*t61*t75*
t128+t132*t88*t51*t134*t136+t139*t16*t141*t27*t120*t112-THREE*t146*t148*t151*t102
-SIX*t146*t156*t158*t109*t112+TWO*t146*t164*t166*t28*t64-THREE*t146*t173*t176*
t112+TWO*t146*t180*t100*t28*t64;
    t189 = t24*t24;
    t191 = t189*t75*t31;
    t197 = t163*t17;
    t199 = AB(p5,t25,p2);
    t200 = t19*t199;
    t202 = t127*t63;
    t206 = t42*t42;
    t207 = t10*t206;
    t213 = t111*I;
    t217 = t206*t206;
    t218 = t44*t44;
    t219 = t218*t44;
    t222 = IB(p1,p6);
    t223 = t222*t49;
    t224 = IB(p3,p6);
    t225 = t224*t224;
    t228 = IB(p5,p6);
    t229 = t50*t228;
    t230 = AB(p3,t23,p3);
    t232 = IAB(p3,t23,p6);
    t234 = t232*t2*I;
    t241 = AB(p3,t23,p4);
    t245 = t7*t7;
    t246 = t245*t245;
    t247 = t246*t9;
    t249 = IA(p1,p2);
    t251 = IA(p2,p3);
    t252 = t249*t12*t251;
    t253 = t247*t44*t252;
    t254 = t16*t17;
    t255 = AB(p6,t1,p6);
    t261 = t245*t10;
    t263 = t155*t15;
    t268 = t33*t35*I;
    t272 = t67*t67;
    t277 = t79*t33;
    t283 = t11*t67;
    t304 = t11*t206;
    t322 = t46*t106;
    t324 = t24*t199;
    t334 = t26*t26;
    t340 = TWO*t13*t172*t72*t191*t111*t37+TWO*t41*t67*t197*t200*t29*t202+FOUR*
t207*t15*t20*t49*t26*t28*t31*t213+t217*t219*t19*t223*t225*t229*t230*t234-t217*
t218*t19*t223*t224*t229*t241*t234+TWO*t253*t254*t50*t255*t37+TWO*t261*t42*t263*
t17*t28*t31*t268+TWO*t11*t272*t197*t19*t75*t31*t277*t2*I-SIX*t283*t42*t12*t15*
t72*t29*t128-THREE*t283*t42*t14*t15*t151*t128+TWO*t283*t156*t158*t120*t79*t2*t63+
SIX*t304*t197*t19*t28*t31*t213+EIGHT*t304*t254*t49*t28*t31*t213-THREE*t132*t107*t62
*t54+TWO*t131*t14*t322*t324*t109*t112-TWO*t139*t15*t18*t17*t19*t49*t50*t334*t29
*t64;
    t345 = t50*t199;
    t350 = t206*t42;
    t353 = t72*t49;
    t363 = t46*t19;
    t365 = t49*t29;
    t367 = t33*t80;
    t368 = t367*t82;
    t384 = t49*t199;
    t391 = t384*t29;
    t392 = t92*t63;
    t406 = t345*t29;
    t415 = t224*t50*t228;
    t419 = A(p3,p6);
    t441 = t172*t17;
    t456 = AB(p4,p6+p5,p4);
    t469 = -TWO*t8*t67*t44*t322*t345*t75*t128+THREE*FOUR*t10*t350*t15*t353*t28*t31*
t33*t80*t4*I-THREE*t207*t70*t363*t365*t77*t368-EIGHT*t207*t14*t363*t90*t29*t136-FOUR
*t43*t197*t200*t28*t64-FOUR*t43*t254*t384*t28*t64+TWO*t41*t70*t254*t391*t392+TWO
*t41*t12*t88*t324*t29*t38+TWO*t7*t8*t44*t16*t150*t406*t64+TWO*t7*t350*t219*t222
*t49*t415*t234+TWO*t419*t217*t219*t19*t222*t49*t415*t232*t80*t82+TWO*t10*t272*
t42*t363*t76*t31*t277*t80*t82-t246*t10*t252*t441*t24*t36*I-t253*t254*t50*t35*I+
t247*t249*t12*t251*t14*t97*t50*t456*t255*t37-t261*t67*t263*t17*t29*t31*t79*t35*
I;
    t486 = t11*t70;
    t511 = t8*t44;
    t529 = t199*t199;
    t545 = t9*t206;
    t568 = -t261*t12*t441*t176*t33*t36*I+t283*t148*t150*t75*t77*t79*t53-t283*
t164*t166*t29*t202+t486*t173*t134*t77*t112-t486*t180*t101*t392-EIGHT*t68*t206*
t363*t365*t31*t83+TWO*t69*t88*t91*t127*t93+SIX*t8*t42*t44*t322*t406*t54-TWO*
t511*t16*t72*t51*t324*t120*t112+FOUR*t511*t15*t141*t26*t199*t29*t64-TWO*t117*t17
*t106*t50*t529*t29*t31*t268+TWO*t86*t171*t322*t191*t367*t63+TWO*t116*t363*t391*
t128-THREE*t545*t44*t322*t50*t29*t31*t368-EIGHT*t545*t15*t353*t199*t28*t31*t213+TWO
*t131*t70*t363*t391*t102-t69*t98*t101*t128;
    return(t186+t340+t469+t568);
  }
}


std::complex<DOUBLE> C3234x56mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t258;
  FourMomentum<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t289;
  FourMomentum<DOUBLE> t29;
  std::complex<DOUBLE> t296;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t319;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t327;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t353;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t362;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t367;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t390;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t397;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t406;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t420;
  std::complex<DOUBLE> t429;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t454;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t477;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t495;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t516;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t557;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t571;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t1 = p6+p5;
    t2 = IAB(p1,t1,p1);
    t3 = p4+p3+p2;
    t4 = S(t3);
    t7 = A(p1,p4);
    t8 = A(p1,p5);
    t9 = t7*t8;
    t10 = B(p1,p6);
    t11 = t10*t10;
    t12 = t9*t11;
    t13 = B(p2,p4);
    t14 = IA(p3,p4);
    t15 = t13*t14;
    t16 = IA(p5,p6);
    t17 = t15*t16;
    t19 = IB(p1,p5);
    t20 = IB(p5,p6);
    t21 = t19*t20;
    t22 = p4+p2;
    t23 = AB(p3,t22,p1);
    t24 = t23*t23;
    t25 = t24*t23;
    t27 = p4+p3;
    t28 = IAB(p2,t27,p1);
    t29 = p3+p2;
    t30 = IAB(p4,t29,p1);
    t31 = t30*t30;
    t32 = t28*t31;
    t33 = IBB(p1,t1,t29,p4);
    t34 = t33*I;
    t35 = t32*t34;
    t39 = B(p1,p4);
    t40 = t8*t39;
    t41 = t11*t13;
    t43 = IA(p2,p3);
    t44 = t43*t16;
    t45 = IB(p1,p2);
    t46 = t45*t19;
    t47 = t44*t46;
    t49 = AB(p1,t29,p4);
    t50 = t20*t49;
    t51 = t25*t2;
    t53 = t28*t30;
    t54 = t33*t33;
    t56 = t54*t4*I;
    t57 = t53*t56;
    t58 = t50*t51*t57;
    t61 = t43*t11;
    t62 = t40*t61;
    t63 = t14*t16;
    t64 = t63*t21;
    t66 = t49*t49;
    t67 = t24*t24;
    t68 = t66*t67;
    t69 = t2*t2;
    t70 = t69*t28;
    t72 = t54*t33;
    t74 = t4*t4;
    t75 = t74*I;
    t81 = t43*t14;
    t82 = t19*t19;
    t84 = t81*t16*t82;
    t86 = t20*t24;
    t87 = AB(p3,t22,p5);
    t88 = AB(p3,t22,p6);
    t91 = t34*t53;
    t98 = t49*t25;
    t99 = t88*t2;
    t105 = t63*t19;
    t107 = t88*t88;
    t112 = A(p1,p2);
    t113 = A(p1,p3);
    t114 = t112*t113;
    t115 = t39*t11;
    t116 = t114*t115;
    t118 = t20*t25;
    t120 = AB(p5,p6+p1,p5);
    t121 = t120*t2;
    t122 = t118*t121;
    t123 = t28*t28;
    t124 = t123*t30;
    t126 = t33*t4*I;
    t127 = t124*t126;
    t130 = t112*t7;
    t132 = t61*t63;
    t134 = t67*t69;
    t138 = t33*t74*I;
    t144 = AB(p5,t29,p4);
    t145 = t67*t144;
    t146 = t2*t123;
    t149 = t4*I;
    t150 = t31*t33*t149;
    t153 = t112*t8;
    t154 = t153*t115;
    t156 = t87*t2;
    t160 = t112*t39;
    t163 = t16*t45;
    t164 = t13*t43*t163;
    t166 = t82*t20;
    t167 = t25*t120;
    t169 = t124*t34;
    t172 = t113*t7;
    t175 = t25*t144;
    t176 = t2*t28;
    t177 = t175*t176;
    t181 = t113*t8;
    t182 = t181*t115;
    t186 = t53*t126;
    t190 = t8*t11;
    t191 = t13*t13;
    t192 = t191*t191;
    t194 = t163*t19;
    t196 = IB(p2,p3);
    t197 = IB(p3,p4);
    t200 = AB(p1,t1,p1);
    t202 = IS(t3);
    t203 = t202*I;
    t216 = t163*t21;
    t218 = AB(p5,t27,p2);
    t219 = t218*t2;
    t227 = TWO*t12*t17*t21*t25*t35+TWO*t40*t41*t47*t58+TWO*t62*t64*t68*t70*t30*
t72*t75-FOUR*t40*t10*t84*t86*t87*t88*t91-TWO*t40*t10*t43*t64*t98*t99*t57+TWO*t40
*t43*t105*t86*t107*t91+t116*t84*t122*t127+t130*t40*t132*t21*t134*t123*t31*t138-
t130*t61*t64*t145*t146*t150-t154*t84*t118*t156*t127+t160*t11*t164*t166*t167*
t169+TWO*t172*t61*t64*t177*t150+FOUR*t182*t164*t21*t24*t2*t186-TWO*t190*t192*
t194*t196*t197*t20*t200*t33*t203-TWO*t8*t10*t13*t105*t86*t88*t91-TWO*t115*t81*
t216*t68*t219*t53*t72*t4*I;
    t228 = t113*t39;
    t229 = t228*t61;
    t230 = t63*t166;
    t236 = t16*t19;
    t237 = t81*t236;
    t239 = t50*t134;
    t241 = t54*t74*I;
    t245 = t15*t236;
    t250 = t7*t11;
    t252 = t63*t82;
    t258 = t236*t50;
    t284 = t45*t82;
    t289 = t53*t54*I;
    t296 = t197*t20;
    t316 = t113*t113;
    t317 = t316*t316;
    t318 = t8*t8;
    t319 = t317*t318;
    t321 = IA(p1,p2);
    t322 = IA(p1,p4);
    t324 = IA(p1,p6);
    t326 = t319*t10*t321*t322*t324;
    t327 = t81*t16;
    t332 = t40*t11;
    t342 = t87*t87;
    t348 = t21*t51;
    t349 = t348*t127;
    t353 = t41*t44;
    t359 = -t229*t230*t98*t121*t57-t9*t115*t237*t239*t32*t241-t12*t245*t118*
t176*t150+t250*t13*t252*t118*t120*t35+t250*t81*t258*t145*t176*t31*t54*t149+t62*
t230*t98*t156*t57+t190*t192*t16*t45*t19*t196*t197*t20*t33*I+t190*t13*t252*t86*
t87*t91-t115*t13*t44*t284*t50*t167*t289+t11*t10*t192*t46*t196*t296*t49*t54*I-
t11*t192*t16*t284*t196*t296*t200*t120*t33*t203+t112*t11*t13*t47*t118*t144*t169-
t326*t327*t20*t2*I-TWO*t332*t164*t21*t98*t289+TWO*t332*t81*t16*t82*t19*t86*t342
*t91-TWO*t154*t164*t349+TWO*t153*t39*t353*t46*t118*t169;
    t362 = t322*t322;
    t366 = t63*t20;
    t367 = AB(p4,t29,p6);
    t374 = t32*t126;
    t382 = t112*t112;
    t386 = t20*t67;
    t387 = t123*t28;
    t390 = t30*t33;
    t391 = t390*t75;
    t397 = t63*t46;
    t406 = t114*t40*t132;
    t407 = t25*t69;
    t408 = t21*t407;
    t420 = t39*t10;
    t429 = t63*t46*t20;
    t454 = t316*t8*t115*t237;
    t460 = t390*t149;
    t461 = t86*t176*t460;
    t465 = t316*t39*t61;
    t472 = t24*t120;
    t477 = t319*t321*t362*t324*t43*t366*t367*t2*I+t172*t115*t84*t122*t374-t41*
t43*t216*t98*t144*t289+TWO*t382*t8*t115*t237*t386*t69*t387*t391-TWO*t382*t39*
t61*t397*t386*t219*t387*t30*t126-NINE*t406*t408*t124*t138+TWO*t406*t349-TWO*t154
*t237*t239*t124*t241+TWO*t153*t420*t237*t118*t99*t127+TWO*t160*t61*t429*t49*t67
*t219*t124*t56+TWO*t326*t327*t20*t69*t149-TWO*t316*t113*t318*t321*t322*t324*t43
*t366*t99*I+THREE*FIVE*t454*t86*t70*t391-EIGHT*t454*t461-TWO*t465*t397*t86*t219*t186
-FOUR*t465*t230*t472*t176*t460;
    t487 = t172*t40*t132;
    t495 = t181*t11;
    t515 = t113*t11;
    t516 = t515*t13;
    t557 = t182*t237;
    t571 = -SIX*t316*t11*t43*t64*t24*t144*t176*t460-THREE*t487*t408*t32*t138+TWO*
t487*t348*t374+THREE*t495*t245*t461-FOUR*t495*t17*t21*t24*t91-TWO*t228*t11*t164*
t166*t472*t91-FOUR*t229*t429*t98*t219*t57-TWO*t516*t47*t86*t144*t91-TWO*t516*
t252*t86*t120*t91-THREE*t515*t81*t258*t177*t30*t54*t149+FOUR*t116*t81*t194*t118*
t219*t127+THREE*t114*t61*t64*t175*t146*t460-FOUR*t181*t39*t353*t46*t86*t91+THREE*
t182*t84*t86*t156*t186+NINE*t557*t50*t407*t53*t241-TWO*t557*t58-SIX*t181*t420*
t237*t86*t99*t186;
    return(t227+t359+t477+t571);
  }
}


std::complex<DOUBLE> C3345x61mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t228;
  FourMomentum<DOUBLE> t23;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t276;
  std::complex<DOUBLE> t277;
  FourMomentum<DOUBLE> t28;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t295;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t301;
  FourMomentum<DOUBLE> t31;
  std::complex<DOUBLE> t312;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t320;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t328;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t339;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t352;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t371;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t393;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t425;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t431;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t443;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t470;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t486;
  std::complex<DOUBLE> t490;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t500;
  std::complex<DOUBLE> t507;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t510;
  std::complex<DOUBLE> t514;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t531;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t99;
  {
    t1 = p6+p1;
    t2 = IAB(p2,t1,p2);
    t3 = p5+p4+p3;
    t4 = S(t3);
    t7 = A(p1,p2);
    t8 = t7*t7;
    t9 = B(p2,p3);
    t10 = t9*t9;
    t11 = t8*t10;
    t12 = IA(p1,p6);
    t14 = IA(p2,p3);
    t15 = IA(p2,p6);
    t17 = IB(p1,p6);
    t18 = IB(p3,p4);
    t19 = t17*t18;
    t20 = t14*t15*t19;
    t22 = IB(p4,p5);
    t23 = p5+p3;
    t24 = AB(p2,t23,p4);
    t25 = t24*t24;
    t26 = t25*t25;
    t27 = t22*t26;
    t28 = p5+p4;
    t29 = AB(p3,t28,p6);
    t31 = p4+p3;
    t32 = IAB(p2,t31,p5);
    t33 = IAB(p2,t28,p3);
    t34 = t33*t33;
    t35 = t34*t33;
    t37 = t2*I;
    t42 = t9*t8;
    t43 = B(p2,p4);
    t44 = t42*t43;
    t45 = B(p2,p6);
    t46 = t45*t12;
    t47 = t15*t17;
    t48 = t46*t47;
    t50 = t18*t22;
    t51 = t25*t24;
    t52 = t51*t32;
    t54 = t2*t2;
    t55 = t34*t54;
    t56 = t4*I;
    t57 = t55*t56;
    t61 = t12*t14;
    t66 = t32*t34;
    t67 = t66*t37;
    t71 = IA(p2,p5);
    t72 = t12*t71;
    t73 = t47*t72;
    t75 = AB(p5,t31,p6);
    t76 = t51*t75;
    t77 = t50*t76;
    t81 = t42*t45;
    t83 = AB(p5,t31,p2);
    t84 = t26*t83;
    t85 = t50*t84;
    t86 = t54*I;
    t92 = t71*t15;
    t93 = t92*t19;
    t95 = t29*t83;
    t97 = IS(t3);
    t99 = t2*t97*I;
    t104 = t43*t43;
    t105 = t8*t104;
    t108 = t12*t15*t19;
    t110 = t22*t25;
    t112 = t54*t33;
    t113 = t112*t56;
    t117 = t105*t12;
    t120 = t32*t33;
    t121 = t120*t37;
    t130 = t15*t15;
    t131 = t130*t17;
    t135 = AB(p6,p2+p1,p6);
    t136 = t110*t135;
    t140 = A(p3,p5);
    t141 = t8*t140;
    t142 = t141*t9;
    t143 = t61*t92;
    t146 = t97*I;
    t147 = t66*t146;
    t150 = B(p2,p5);
    t151 = t7*t150;
    t154 = t32*t32;
    t155 = t75*t154;
    t157 = t33*t2;
    t158 = t157*t146;
    t162 = t45*t45;
    t163 = t7*t162;
    t165 = t47*t50;
    t168 = t112*t146;
    t175 = AB(p6,t23,p4);
    t182 = t14*t71;
    t185 = t26*t29;
    t188 = t97*t97;
    t189 = t188*I;
    t190 = t157*t189;
    t194 = TWO*t11*t12*t20*t27*t29*t32*t35*t37+FIVE*t44*t48*t50*t52*t57-FOUR*t44*
t61*t47*t50*t51*t29*t67-THREE*t44*t73*t77*t67-TWO*t81*t73*t85*t66*t86+TWO*t42*t61
*t93*t27*t95*t66*t99-THREE*t105*t45*t108*t110*t32*t113+TWO*t117*t20*t110*t29*t121
+SIX*t117*t93*t110*t75*t121+FOUR*t117*t131*t18*t136*t121-t142*t143*t19*t76*t147+
TWO*t151*t45*t93*t27*t155*t158-FOUR*t163*t71*t165*t84*t32*t168+FOUR*t163*t130*t19
*t22*t51*t175*t32*t158+TWO*t7*t45*t182*t165*t185*t83*t32*t190;
    t195 = A(p2,p5);
    t196 = t104*t104;
    t198 = t162*t45;
    t200 = IB(p1,p2);
    t202 = IB(p2,p3);
    t205 = IB(p2,p5);
    t208 = IAB(p5,t31,p2);
    t222 = t2*t208*I;
    t228 = AB(p1,t23,p4);
    t235 = t8*t43;
    t236 = t235*t150;
    t238 = t154*t33;
    t239 = t238*t37;
    t243 = t235*t45;
    t245 = t51*t83;
    t253 = t22*t51;
    t259 = t71*t71;
    t260 = t12*t259;
    t263 = t75*t32;
    t269 = t140*t140;
    t270 = t269*t269;
    t274 = IA(p3,p4);
    t275 = t15*t274;
    t276 = IA(p4,p5);
    t277 = t275*t276;
    t282 = t8*t270;
    t295 = AB(p2,t1,p2);
    t301 = t12*t130*t19;
    t312 = t253*t135;
    t316 = t110*t175;
    t320 = t131*t50;
    t326 = -TWO*t195*t196*t198*t200*t17*t202*t205*t18*t22*t54*t208*t56-TWO*t140
*t104*t43*t198*t200*t17*t202*t205*t22*t222+TWO*t162*t15*t17*t50*t228*t52*t33*
t99-TWO*t236*t73*t77*t239+FIVE*t243*t73*t50*t245*t120*t86-FOUR*t235*t61*t93*t253*
t95*t120*t99-THREE*t235*t260*t165*t245*t263*t158+t8*t7*t270*t61*t259*t277*t83*
t188*I+t282*t45*t61*t71*t277*t17*t97*I+t282*t12*t182*t130*t274*t276*t17*t295*
t135*t189+t81*t301*t253*t175*t67+t236*t48*t50*t51*t154*t113-t236*t301*t312*t239
+t243*t301*t316*t121-t235*t72*t320*t245*t135*t32*t158;
    t328 = t8*t150;
    t339 = t8*t45;
    t348 = t17*t202;
    t349 = t205*t205;
    t352 = AB(p5,t31,p5);
    t371 = t141*t43;
    t372 = t46*t92;
    t374 = t17*t22;
    t382 = t120*t146;
    t386 = t61*t131;
    t393 = t72*t131;
    t400 = t27*t32;
    t407 = t18*t51;
    t411 = t141*t150;
    t424 = -t328*t45*t73*t85*t238*t86+t328*t260*t165*t84*t155*t158+t339*t72*
t320*t245*t175*t32*t158-t196*t198*t200*t348*t349*t50*t352*t222+t196*t162*t200*
t348*t205*t50*t75*t222+TWO*t282*t45*t14*t71*t275*t276*t188*I+THREE*t371*t372*t374
*t25*t121+TWO*t371*t143*t19*t25*t75*t382+TWO*t371*t386*t18*t25*t135*t382+TWO*
t371*t393*t136*t382-TWO*t11*t45*t108*t400*t35*t54*t56-t142*t386*t407*t135*t147-
t411*t372*t374*t51*t239-t411*t393*t312*t238*t146-t141*t45*t393*t316*t382;
    t425 = t141*t12;
    t426 = t14*t259;
    t431 = t120*t189;
    t443 = t42*t150;
    t456 = t339*t12;
    t460 = t83*t83;
    t461 = t460*t32;
    t470 = t175*t175;
    t486 = t253*t32;
    t490 = t7*t9;
    t492 = t47*t18;
    t500 = t29*t32;
    t507 = t7*t43;
    t510 = t112*I;
    t514 = t507*t45;
    t531 = -t425*t426*t47*t407*t83*t75*t431-t425*t182*t131*t407*t83*t135*t431-
t44*t301*t312*t67-t443*t48*t50*t26*t154*t57+t443*t73*t50*t26*t75*t154*t34*t37
-TWO*t456*t259*t15*t19*t27*t461*t168-TWO*t456*t130*t15*t17*t18*t110*t470*t382+
TWO*t8*t12*t426*t165*t185*t461*t190+TWO*t7*t140*t162*t92*t17*t486*t158-FOUR*t490
*t162*t492*t400*t55*I+TWO*t490*t45*t20*t27*t500*t34*t2*t146+THREE*FOUR*t507*t162*t492
*t486*t510-FOUR*t514*t20*t253*t500*t158-FOUR*t514*t93*t253*t263*t158-TWO*t151*
t162*t492*t27*t154*t510;
    return(t194+t326+t424+t531);
  }
}


std::complex<DOUBLE> C3456x12mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t245;
  FourMomentum<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t280;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t288;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t291;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t297;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t304;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t323;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t327;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t371;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t405;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t412;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t417;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t439;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t446;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t451;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t490;
  std::complex<DOUBLE> t491;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t497;
  std::complex<DOUBLE> t500;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t510;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t552;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t573;
  std::complex<DOUBLE> t575;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t611;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t1 = p2+p1;
    t2 = IAB(p3,t1,p3);
    t3 = p6+p5+p4;
    t4 = S(t3);
    t7 = B(p2,p3);
    t8 = t7*t7;
    t9 = t7*t8;
    t10 = B(p3,p6);
    t11 = t9*t10;
    t12 = B(p4,p6);
    t14 = IA(p4,p5);
    t15 = IB(p1,p2);
    t16 = t14*t15;
    t17 = IB(p1,p3);
    t18 = IB(p3,p4);
    t19 = t17*t18;
    t20 = t16*t19;
    t22 = p5+p4;
    t23 = AB(p3,t22,p6);
    t24 = t23*t23;
    t25 = p6+p4;
    t26 = AB(p5,t25,p3);
    t27 = t26*t26;
    t28 = t27*t26;
    t29 = t24*t28;
    t30 = IAB(p4,t1,p3);
    t31 = t2*t30;
    t33 = IAB(p6,t22,p3);
    t34 = IBB(p3,t1,t22,p6);
    t35 = t34*t34;
    t37 = t33*t35*t34;
    t38 = t4*I;
    t43 = A(p1,p3);
    t44 = t43*t43;
    t45 = t44*t43;
    t46 = A(p3,p5);
    t47 = t46*t46;
    t48 = t47*t47;
    t49 = t45*t48;
    t51 = IA(p1,p2);
    t52 = IA(p2,p3);
    t54 = IA(p3,p4);
    t55 = t51*t52*t54;
    t57 = IA(p3,p6);
    t59 = IA(p5,p6);
    t60 = t57*t14*t59;
    t61 = t2*t2;
    t62 = IAB(p3,t22,p6);
    t69 = t57*t57;
    t72 = t14*t59;
    t73 = AB(p6,t22,p6);
    t76 = t2*t62*I;
    t81 = t44*t48*t7*t55;
    t90 = A(p2,p3);
    t91 = t43*t90;
    t92 = A(p3,p4);
    t93 = t92*t8;
    t95 = t10*t51;
    t96 = t95*t72;
    t98 = t17*t17;
    t99 = t27*t27;
    t101 = t30*t30;
    t102 = t61*t101;
    t104 = t34*t33;
    t105 = t104*t38;
    t108 = t8*t10;
    t110 = t51*t14;
    t115 = t61*t30;
    t117 = t33*t35;
    t118 = t117*t38;
    t119 = t23*t99*t115*t118;
    t121 = t91*t8;
    t122 = t12*t51;
    t123 = t122*t59;
    t125 = t98*t28;
    t127 = t30*t33;
    t128 = t34*I;
    t129 = t127*t128;
    t132 = t43*t92;
    t133 = t46*t8;
    t135 = t132*t133*t96;
    t136 = t15*t17;
    t137 = t28*t61;
    t138 = t136*t137;
    t139 = t101*t33;
    t140 = t4*t4;
    t142 = t34*t140*I;
    t147 = t122*t16;
    t149 = t2*t28;
    t152 = t34*t4*I;
    t153 = t139*t152;
    t156 = t43*t46;
    t157 = A(p3,p6);
    t158 = t157*t8;
    t160 = t156*t158*t96;
    t161 = t33*t33;
    t162 = t30*t161;
    t166 = t156*t108;
    t167 = t59*t15;
    t168 = t110*t167;
    t169 = t166*t168;
    t170 = t17*t23;
    t178 = t43*t157*t8;
    t179 = t122*t167;
    t181 = t17*t28;
    t183 = t161*t34;
    t184 = t183*t38;
    t187 = t8*t43;
    t188 = t10*t12;
    t196 = t127*t35*t4*I;
    t199 = t12*t12;
    t200 = t199*t199;
    t204 = IB(p4,p5);
    t206 = IB(p5,p6);
    t211 = t46*t92;
    t212 = t157*t9;
    t215 = t10*t14*t167;
    t218 = t140*I;
    t219 = t183*t218;
    t225 = AB(p1,p3+p2,p1);
    t226 = t98*t225;
    t227 = t226*t149;
    t230 = -TWO*t11*t12*t20*t29*t31*t37*t38-t49*t10*t55*t60*t61*t62*t38-t49*t51
*t52*t54*t69*t72*t73*t76-t81*t60*t15*t2*I+t81*t60*t15*t61*t38+t91*t93*t96*t98*
t99*t102*t105+t91*t108*t110*t59*t98*t119+t121*t123*t125*t2*t129-t135*t138*t139*
t142-t132*t108*t147*t19*t149*t153+t160*t138*t162*t142-t169*t170*t137*t127*t35*
t140*I+t178*t179*t181*t31*t184-t187*t188*t110*t136*t18*t23*t149*t196+t187*t200*
t51*t15*t19*t204*t206*t34*I+t211*t212*t215*t181*t102*t219+t211*t108*t168*t227*
t153;
    t234 = t12*t59*t136;
    t239 = t92*t9;
    t241 = t199*t15*t17;
    t243 = t18*t27;
    t245 = t104*I;
    t250 = t98*t18;
    t253 = t139*t128;
    t256 = t46*t157;
    t258 = t72*t136;
    t260 = t23*t28;
    t262 = t161*t35;
    t268 = t162*t152;
    t272 = t15*t98;
    t273 = t72*t272;
    t280 = t167*t170;
    t282 = t149*t30;
    t287 = t51*t59;
    t288 = t287*t15;
    t291 = t162*t128;
    t295 = t136*t18;
    t297 = t204*t206;
    t304 = t23*t27;
    t315 = t127*t35*I;
    t323 = AB(p3,t1,p3);
    t325 = IS(t3);
    t327 = t323*t34*t325*I;
    t330 = t46*t47;
    t354 = t44*t7;
    t371 = AB(p5,t25,p2);
    t377 = t92*t157*t9*t234*t149*t101*t184-t239*t241*t243*t101*t245+t93*t10*
t147*t250*t225*t28*t253+t256*t11*t258*t260*t115*t262*t218-t256*t108*t168*t227*
t268+t133*t95*t273*t225*t23*t149*t196+t212*t12*t280*t282*t262*t38-t158*t12*t288
*t226*t28*t291+t9*t200*t295*t297*t23*t35*I-t9*t199*t295*t304*t30*t117*I+t108*
t12*t110*t272*t18*t225*t260*t315-t8*t200*t51*t272*t18*t297*t225*t327-TWO*t45*
t330*t122*t52*t54*t57*t59*t76+TWO*t44*t92*t7*t96*t17*t99*t102*t105+EIGHT*t44*t46*
t7*t96*t181*t115*t105+TWO*t354*t10*t110*t59*t17*t119+TWO*t354*t12*t287*t17*t282
*t245+TWO*t44*t10*t51*t72*t17*t371*t28*t2*t129;
    t387 = AB(p5,t25,p1);
    t400 = t136*t149;
    t405 = t188*t110;
    t407 = t18*t28;
    t412 = t47*t9;
    t413 = t412*t10;
    t415 = t12*t14*t136;
    t417 = t18*t26;
    t424 = t117*t218;
    t439 = t225*t27;
    t446 = t27*t2;
    t447 = t446*t30;
    t451 = t46*t9;
    t489 = t43*t47*t108*t168;
    t490 = t17*t27;
    t491 = t490*t115;
    t492 = t104*t218;
    t497 = t490*t31*t105;
    t500 = FIVE*t91*t133*t96*t125*t115*t105-TWO*t121*t96*t98*t17*t387*t149*t129+
FOUR*t91*t7*t96*t98*t371*t149*t129+TWO*t135*t400*t153+TWO*t132*t8*t405*t136*t407
*t253-FOUR*t413*t415*t417*t31*t105-SEVEN*t413*t258*t304*t115*t424-SIX*t412*t12*
t167*t17*t26*t2*t30*t105+FOUR*t47*t8*t95*t273*t439*t31*t105+TWO*t256*t9*t234*
t447*t184-FIVE*t451*t188*t20*t304*t31*t118-TWO*t451*t10*t258*t29*t115*t37*t218
-TWO*t451*t241*t417*t30*t245-THREE*t451*t12*t280*t447*t118+TWO*t133*t10*t147*t250
*t439*t129+TWO*t133*t12*t288*t226*t27*t129-FOUR*t489*t491*t492+EIGHT*t489*t497;
    t510 = t156*t8;
    t548 = t92*t92;
    t552 = t101*t30;
    t573 = t211*t11;
    t575 = t2*t101;
    t611 = -TWO*t160*t400*t268-TWO*t166*t147*t19*t446*t127*t152+FOUR*t510*t405*
t136*t243*t129+TWO*t169*t170*t149*t196-TWO*t510*t179*t497+FOUR*t510*t123*t136*
t27*t129-TWO*t178*t123*t136*t28*t291+TWO*t187*t10*t147*t19*t260*t315-TWO*t187*
t200*t51*t15*t17*t18*t204*t206*t327-TWO*t548*t46*t9*t215*t181*t61*t552*t492-TWO
*t548*t9*t10*t415*t407*t2*t552*t105-SEVEN*t92*t47*t9*t215*t490*t102*t492-FIVE*t573
*t415*t243*t575*t105-TWO*t573*t258*t260*t102*t424-THREE*t211*t9*t234*t446*t101*
t105-TWO*t239*t188*t20*t260*t575*t118-EIGHT*t330*t9*t10*t258*t26*t61*t30*t492+TWO
*t47*t157*t9*t215*t491*t219;
    return(t230+t377+t500+t611);
  }
}


std::complex<DOUBLE> C3561x23mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t207;
  FourMomentum<DOUBLE> t21;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t238;
  FourMomentum<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t249;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t281;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t289;
  std::complex<DOUBLE> t293;
  std::complex<DOUBLE> t294;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t304;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t370;
  std::complex<DOUBLE> t373;
  std::complex<DOUBLE> t390;
  std::complex<DOUBLE> t394;
  std::complex<DOUBLE> t395;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t406;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t419;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t432;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t462;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t469;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t473;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t498;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t517;
  std::complex<DOUBLE> t519;
  std::complex<DOUBLE> t522;
  std::complex<DOUBLE> t531;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t567;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  {
    t1 = p3+p2;
    t2 = IAB(p4,t1,p4);
    t3 = p6+p5+p1;
    t4 = S(t3);
    t7 = A(p3,p4);
    t8 = B(p2,p4);
    t9 = t7*t8;
    t10 = B(p4,p6);
    t12 = IA(p1,p4);
    t13 = IA(p2,p4);
    t15 = IB(p1,p6);
    t16 = IB(p2,p3);
    t17 = t15*t16;
    t18 = t12*t13*t17;
    t20 = IB(p5,p6);
    t21 = p6+p5;
    t22 = AB(p1,t21,p2);
    t23 = t20*t22;
    t24 = p5+p1;
    t25 = AB(p4,t24,p6);
    t26 = t25*t25;
    t27 = t26*t25;
    t28 = t27*t2;
    t31 = IAB(p4,p6+p1,p5);
    t32 = IAB(p4,t21,p1);
    t33 = t31*t32;
    t34 = IS(t3);
    t35 = t34*I;
    t36 = t33*t35;
    t37 = t23*t28*t36;
    t40 = t7*t7;
    t41 = B(p1,p4);
    t42 = t40*t41;
    t43 = B(p4,p5);
    t45 = IA(p2,p3);
    t46 = t45*t12;
    t47 = t13*t15;
    t48 = t46*t47;
    t50 = t16*t20;
    t51 = t26*t26;
    t54 = t31*t31;
    t55 = t2*t54;
    t56 = t32*t32;
    t57 = t56*I;
    t61 = t12*t12;
    t62 = t61*t45;
    t64 = t47*t50;
    t66 = AB(p1,t21,p4);
    t67 = t22*t66;
    t68 = t51*t2;
    t70 = t31*t56;
    t71 = t70*t35;
    t74 = t40*t8;
    t75 = B(p3,p4);
    t76 = t75*t74;
    t77 = t13*t13;
    t81 = t50*t51;
    t82 = t2*t2;
    t84 = t32*I;
    t85 = t82*t54*t84;
    t89 = t12*t77*t17;
    t91 = t20*t66;
    t92 = t51*t82;
    t97 = t77*t15;
    t98 = t97*t50;
    t100 = AB(p2,t24,p6);
    t106 = t40*t43*t10;
    t107 = t45*t77;
    t111 = AB(p2,p4+p3,p2);
    t112 = t20*t111;
    t113 = t112*t27;
    t114 = t55*t84;
    t117 = t40*t10;
    t124 = t8*t8;
    t126 = t10*t10;
    t127 = t126*t126;
    t129 = IB(p1,p4);
    t130 = t129*t129;
    t132 = IB(p3,p4);
    t133 = t17*t132;
    t135 = IB(p4,p5);
    t136 = t135*t20;
    t137 = AB(p1,t21,p1);
    t139 = IAB(p1,t21,p4);
    t141 = t139*t2*I;
    t150 = A(p1,p5);
    t151 = t150*t150;
    t152 = t151*t151;
    t153 = t152*t40;
    t155 = IA(p1,p6);
    t158 = t153*t8*t12*t155*t45;
    t159 = IA(p4,p5);
    t160 = t13*t159;
    t161 = IA(p5,p6);
    t162 = t160*t161;
    t163 = AB(p4,t1,p4);
    t165 = t34*t34;
    t166 = t165*I;
    t171 = t40*t7;
    t172 = t150*t171;
    t173 = t43*t43;
    t175 = t45*t13;
    t176 = t175*t159;
    t178 = t20*t27;
    t181 = t54*t31*t32;
    t186 = t172*t43;
    t187 = t10*t45;
    t190 = t159*t20;
    t195 = t46*t160;
    t197 = t91*t28;
    t198 = t54*t32;
    t199 = t198*t35;
    t207 = t33*I;
    t218 = FOUR*t9*t10*t18*t37+t42*t43*t48*t50*t22*t51*t55*t57-t42*t62*t64*t67*
t68*t71+t76*t43*t77*t15*t81*t85-t76*t89*t91*t92*t36+t74*t46*t98*t66*t100*t28*
t36+t106*t107*t17*t113*t114-t117*t46*t98*t66*t111*t28*t36+t124*t8*t127*t130*
t133*t136*t137*t141-t124*t127*t129*t133*t136*t22*t141+TWO*t158*t162*t16*t163*
t166+TWO*t172*t173*t176*t178*t2*t181*I-FOUR*t186*t187*t13*t190*t26*t114-TWO*t186
*t195*t197*t199+TWO*t172*t126*t176*t20*t25*t2*t207+FOUR*t172*t10*t195*t91*t26*t2
*t36;
    t222 = t66*t66;
    t223 = t222*t27;
    t225 = t33*t166;
    t229 = t150*t40;
    t230 = t229*t8;
    t231 = t43*t45;
    t233 = t230*t231*t160;
    t234 = t50*t27;
    t238 = t234*t114;
    t242 = t230*t187*t160;
    t243 = t50*t26;
    t247 = t2*t31;
    t248 = t247*t84;
    t249 = t243*t248;
    t253 = t66*t27;
    t254 = t50*t253;
    t260 = t159*t16;
    t261 = t175*t260;
    t263 = t197*t36;
    t266 = t229*t10;
    t273 = t107*t260;
    t275 = t112*t26;
    t281 = t175*t15;
    t283 = t178*t82;
    t284 = t4*I;
    t289 = t171*t43;
    t293 = t20*t26*t82;
    t294 = t198*t284;
    t304 = t126*t10;
    t318 = AB(p3,t24,p6);
    t354 = TWO*t172*t61*t175*t190*t223*t2*t225+TWO*t233*t234*t199-TWO*t233*t238
-FOUR*t242*t243*t36+FOUR*t242*t249-TWO*t230*t195*t254*t225+TWO*t229*t8*t12*t261*
t263-TWO*t266*t195*t50*t22*t26*t36-TWO*t266*t273*t275*t36+TWO*t171*t173*t10*
t281*t283*t181*t284-FOUR*t289*t126*t281*t293*t294-TWO*t289*t10*t48*t91*t27*t85+
TWO*t171*t304*t45*t47*t20*t25*t82*t31*t32*t4*I-TWO*t9*t46*t64*t66*t318*t28*t36+
TWO*t124*t304*t45*t129*t15*t16*t132*t135*t20*t318*t2*I-TWO*t124*t13*t15*t50*
t318*t28*t31*t32*t34*I-t152*t171*t61*t155*t45*t162*t66*t165*I;
    t370 = t229*t43;
    t373 = t50*t22*t27;
    t390 = t42*t8;
    t394 = t82*t31;
    t395 = t394*t57;
    t406 = t66*t51;
    t413 = t23*t68*t71;
    t419 = t394*t84;
    t432 = t187*t47;
    t435 = t50*t27*t82;
    t462 = t74*t43*t432;
    t465 = -t158*t162*t16*t34*I+t153*t12*t155*t45*t77*t159*t161*t16*t111*t163*
t166+t370*t195*t373*t199+t370*t273*t113*t199-t229*t61*t261*t23*t253*t225-t229*
t12*t273*t91*t111*t27*t225-t390*t75*t77*t15*t81*t395-t390*t231*t47*t50*t92*t54*
t56*t284+t390*t48*t50*t406*t395+t42*t75*t89*t413+FOUR*t171*t126*t12*t281*t91*t26
*t419+TWO*t171*t10*t61*t175*t15*t20*t223*t82*t36+TWO*t390*t432*t435*t70*t284
-TWO*t42*t10*t48*t373*t247*t57-SIX*t76*t10*t77*t15*t234*t419+TWO*t76*t77*t13*
t15*t50*t100*t27*t2*t36+TWO*t462*t238;
    t469 = t74*t126;
    t473 = t175*t17;
    t498 = t40*t126;
    t517 = t7*t41;
    t519 = t47*t16;
    t522 = t20*t51*t82;
    t531 = t7*t124;
    t567 = -FOUR*t462*t435*t294-FOUR*t469*t281*t249+EIGHT*t469*t473*t293*t33*t284
-TWO*t74*t10*t12*t473*t263+THREE*t74*t10*t48*t254*t419+TWO*t40*t75*t10*t89*t37+
THREE*t106*t48*t373*t114-SIX*t498*t12*t473*t23*t26*t248-TWO*t498*t45*t97*t16*t275
*t248-THREE*t117*t62*t64*t67*t28*t36-TWO*t517*t124*t519*t522*t70*I+TWO*t517*t8*
t18*t413+TWO*t531*t43*t519*t522*t198*I+TWO*t531*t127*t45*t129*t15*t16*t132*t135
*t20*t82*t284-TEN*t531*t10*t519*t283*t207-TWO*t531*t12*t64*t406*t82*t36-FOUR*t9
*t75*t98*t318*t27*t2*t36;
    return(t218+t354+t465+t567);
  }
}


std::complex<DOUBLE> C3612x34mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t258;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t297;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t311;
  std::complex<DOUBLE> t312;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t323;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t333;
  std::complex<DOUBLE> t338;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t356;
  std::complex<DOUBLE> t364;
  std::complex<DOUBLE> t376;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t385;
  FourMomentum<DOUBLE> t39;
  std::complex<DOUBLE> t390;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t404;
  std::complex<DOUBLE> t405;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t423;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t479;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t507;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t512;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t535;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t550;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t58;
  FourMomentum<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  FourMomentum<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  {
    t1 = p4+p3;
    t2 = IAB(p5,t1,p5);
    t3 = p6+p2+p1;
    t4 = S(t3);
    t7 = A(p1,p5);
    t8 = t7*t7;
    t9 = t8*t8;
    t10 = A(p3,p5);
    t11 = t10*t10;
    t12 = t10*t11;
    t13 = t9*t12;
    t14 = IA(p1,p2);
    t16 = IA(p1,p6);
    t17 = IA(p2,p5);
    t18 = t17*t17;
    t20 = IA(p3,p4);
    t23 = IA(p4,p5);
    t24 = IA(p5,p6);
    t25 = t24*t23;
    t26 = p6+p1;
    t27 = AB(p2,t26,p2);
    t29 = IAB(p5,t26,p2);
    t31 = t2*t29*I;
    t39 = p5+p4;
    t40 = AB(p3,t39,p2);
    t44 = A(p2,p5);
    t45 = t44*t7;
    t46 = A(p5,p6);
    t47 = B(p2,p5);
    t48 = t46*t47;
    t50 = B(p4,p5);
    t51 = t50*t50;
    t52 = t50*t51;
    t53 = t52*t14;
    t54 = IB(p3,p4);
    t55 = t16*t54;
    t56 = t53*t55;
    t58 = IB(p3,p5);
    t59 = p6+p2;
    t60 = AB(p1,t59,p5);
    t61 = t60*t60;
    t62 = t61*t60;
    t63 = t58*t62;
    t64 = IAB(p2,t26,p5);
    t65 = t64*t64;
    t66 = t2*t2;
    t67 = t65*t66;
    t69 = p2+p1;
    t70 = IAB(p6,t69,p5);
    t71 = t70*t70;
    t72 = IBB(p5,t1,t26,p2);
    t73 = t71*t72;
    t74 = t4*t4;
    t75 = t74*I;
    t76 = t73*t75;
    t79 = t47*t52;
    t80 = t45*t79;
    t81 = t14*t16;
    t82 = t58*t58;
    t83 = t54*t82;
    t84 = t81*t83;
    t86 = AB(p1,t59,p3);
    t87 = t86*t61;
    t88 = t65*t2;
    t90 = t70*t72;
    t91 = t4*I;
    t92 = t90*t91;
    t95 = t54*t58;
    t96 = t81*t95;
    t98 = AB(p5,t26,p2);
    t101 = t72*t72;
    t102 = t70*t101;
    t103 = t102*t75;
    t106 = t7*t10;
    t108 = t51*t14;
    t109 = t16*t20;
    t112 = t62*t64;
    t114 = t66*t71;
    t120 = t47*t51;
    t122 = t20*t54;
    t123 = t81*t122;
    t124 = t106*t120*t123;
    t125 = t98*t64;
    t127 = t66*t70;
    t133 = B(p2,p6);
    t137 = t58*t61;
    t138 = t64*t2;
    t144 = t53*t95;
    t148 = t73*t91;
    t151 = t44*t133;
    t152 = t151*t52;
    t153 = t14*t54;
    t154 = t153*t82;
    t157 = t90*I;
    t164 = t102*t91;
    t167 = t10*t46;
    t170 = t82*t86;
    t172 = t2*t71;
    t174 = t72*t4*I;
    t175 = t172*t174;
    t178 = t10*t47;
    t182 = t86*t62;
    t184 = t2*t70;
    t186 = t101*t4*I;
    t187 = t184*t186;
    t190 = t133*t133;
    t191 = t190*t190;
    t192 = t10*t191;
    t194 = IB(p1,p2);
    t197 = IB(p1,p6);
    t200 = IB(p5,p6);
    t205 = t13*t14*t16*t18*t20*t25*t27*t31+t9*t11*t14*t16*t17*t20*t25*t40*t31+
t45*t48*t56*t63*t67*t76+t80*t84*t87*t88*t92-t80*t96*t62*t98*t67*t103+t106*t48*
t108*t109*t95*t112*t114*t72*t74*I-t124*t63*t125*t127*t101*t74*I+t106*t133*t108*
t122*t137*t138*t92+t44*t46*t133*t144*t62*t65*t2*t148+t152*t154*t87*t65*t157-
t152*t153*t63*t98*t65*t2*t164+t167*t120*t123*t170*t112*t175-t178*t108*t109*t83*
t182*t125*t187-t192*t51*t20*t194*t197*t54*t58*t200*t72*I;
    t206 = t10*t133;
    t209 = t14*t20*t54;
    t213 = t72*I;
    t214 = t64*t70*t213;
    t217 = t48*t53;
    t218 = t82*t200;
    t219 = t55*t218;
    t221 = AB(p5,t69,p6);
    t222 = t221*t64;
    t224 = t172*t213;
    t229 = t58*t200;
    t230 = t153*t229;
    t232 = t62*t221;
    t236 = t79*t81;
    t240 = t232*t125;
    t242 = t184*t101*I;
    t243 = t240*t242;
    t247 = t194*t197*t54;
    t256 = AB(p3,t39,p3);
    t258 = AB(p5,t1,p5);
    t260 = IS(t3);
    t262 = t258*t72*t260*I;
    t267 = t200*t62;
    t278 = t20*t23*t24;
    t284 = t7*t8;
    t295 = t60*t64;
    t297 = t90*t75;
    t311 = t64*t66;
    t312 = t137*t311;
    t322 = t8*t47;
    t323 = t322*t52;
    t325 = t86*t60;
    t331 = t55*t229;
    t333 = t61*t221;
    t338 = t206*t51*t209*t170*t61*t214+t217*t219*t182*t222*t224+t46*t133*t52*
t230*t232*t64*t224-t236*t83*t200*t86*t243-t191*t52*t247*t229*t98*t101*I+t191*
t51*t20*t247*t218*t256*t262-t133*t52*t14*t95*t267*t221*t98*t64*t242+TWO*t13*t47
*t81*t17*t278*t66*t29*t91+TWO*t284*t12*t133*t14*t17*t278*t31-SIX*t284*t47*t52*
t96*t295*t66*t297+THREE*t8*t44*t47*t56*t137*t67*t297+THREE*t8*t10*t120*t123*t312*
t297-TWO*t8*t46*t47*t56*t312*t76-TWO*t323*t84*t325*t138*t92+SIX*t322*t53*t331*
t333*t311*t92;
    t348 = AB(p1,t59,p4);
    t349 = t348*t60;
    t356 = t153*t58;
    t364 = t348*t61;
    t376 = t61*t64;
    t378 = t184*t174;
    t382 = t47*t50;
    t385 = t58*t348;
    t390 = t7*t46;
    t404 = t7*t47;
    t405 = t404*t53;
    t408 = t184*t213;
    t423 = t7*t133;
    t424 = t423*t52;
    t435 = TWO*t323*t96*t61*t98*t311*t103+FOUR*t322*t51*t96*t349*t138*t92-SIX*t8
*t133*t52*t356*t295*t2*t92-TWO*t45*t120*t96*t364*t88*t92+THREE*t45*t133*t144*t61*
t65*t2*t92+THREE*t124*t170*t376*t378-SIX*t106*t382*t123*t385*t376*t378+SIX*t390*
t79*t96*t267*t222*t114*t174-TWO*t390*t133*t144*t376*t2*t148+TWO*t405*t219*t87*
t222*t408-SIX*t405*t331*t240*t127*t186-FOUR*t404*t108*t331*t364*t222*t408-TWO*
t424*t154*t325*t64*t157+TWO*t424*t230*t333*t64*t408;
    t463 = t86*t86;
    t479 = t348*t62;
    t487 = t348*t348;
    t507 = t46*t46;
    t511 = t61*t61;
    t512 = t511*t221;
    t535 = t98*t98;
    t550 = TWO*t424*t153*t137*t125*t2*t164+FOUR*t423*t51*t356*t349*t64*t157-TWO*
t151*t51*t356*t364*t65*t157-TWO*t167*t382*t123*t385*t112*t175+TWO*t178*t51*t123
*t82*t58*t463*t61*t214-FOUR*t178*t50*t123*t170*t364*t214+TWO*t178*t50*t14*t109*
t95*t479*t125*t187+TWO*t178*t14*t109*t54*t58*t487*t61*t214+TWO*t192*t51*t20*
t194*t197*t95*t200*t262-TWO*t206*t50*t209*t385*t61*t214+TWO*t507*t47*t53*t331*
t512*t311*t71*t70*t72*t91-TWO*t217*t331*t512*t125*t114*t186-TWO*t48*t108*t331*
t479*t222*t224+TWO*t236*t95*t200*t511*t221*t535*t311*t70*t101*t72*t91+TWO*t120*
t81*t95*t200*t348*t243;
    return(t205+t338+t435+t550);
  }
}


std::complex<DOUBLE> C312x34x56mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  FourMomentum<std::complex<DOUBLE> > K1f1;
  FourMomentum<std::complex<DOUBLE> > K2f;
  FourMomentum<std::complex<DOUBLE> > K2f1;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> gg1;
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t1000;
  std::complex<DOUBLE> t1008;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t1017;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t1020;
  std::complex<DOUBLE> t1022;
  std::complex<DOUBLE> t1023;
  std::complex<DOUBLE> t1026;
  std::complex<DOUBLE> t1027;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t1034;
  std::complex<DOUBLE> t1035;
  std::complex<DOUBLE> t1040;
  std::complex<DOUBLE> t1043;
  std::complex<DOUBLE> t1044;
  std::complex<DOUBLE> t1045;
  std::complex<DOUBLE> t1046;
  std::complex<DOUBLE> t1047;
  std::complex<DOUBLE> t1048;
  std::complex<DOUBLE> t1049;
  std::complex<DOUBLE> t1052;
  std::complex<DOUBLE> t1053;
  std::complex<DOUBLE> t1054;
  std::complex<DOUBLE> t1055;
  std::complex<DOUBLE> t1056;
  std::complex<DOUBLE> t1057;
  std::complex<DOUBLE> t1058;
  std::complex<DOUBLE> t1059;
  std::complex<DOUBLE> t1060;
  std::complex<DOUBLE> t1061;
  std::complex<DOUBLE> t1063;
  std::complex<DOUBLE> t1064;
  std::complex<DOUBLE> t1065;
  std::complex<DOUBLE> t1066;
  std::complex<DOUBLE> t1067;
  std::complex<DOUBLE> t1068;
  std::complex<DOUBLE> t1069;
  std::complex<DOUBLE> t1070;
  std::complex<DOUBLE> t1072;
  std::complex<DOUBLE> t1075;
  std::complex<DOUBLE> t1076;
  std::complex<DOUBLE> t1078;
  std::complex<DOUBLE> t1079;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t1080;
  std::complex<DOUBLE> t1081;
  std::complex<DOUBLE> t1082;
  std::complex<DOUBLE> t1083;
  std::complex<DOUBLE> t1084;
  std::complex<DOUBLE> t1085;
  std::complex<DOUBLE> t1086;
  std::complex<DOUBLE> t1087;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t1090;
  std::complex<DOUBLE> t1092;
  std::complex<DOUBLE> t1093;
  std::complex<DOUBLE> t1094;
  std::complex<DOUBLE> t1095;
  std::complex<DOUBLE> t1096;
  std::complex<DOUBLE> t1097;
  std::complex<DOUBLE> t1098;
  std::complex<DOUBLE> t1099;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t1100;
  std::complex<DOUBLE> t1105;
  std::complex<DOUBLE> t1106;
  std::complex<DOUBLE> t1107;
  std::complex<DOUBLE> t1108;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t1111;
  std::complex<DOUBLE> t1113;
  std::complex<DOUBLE> t1114;
  std::complex<DOUBLE> t1116;
  std::complex<DOUBLE> t1117;
  std::complex<DOUBLE> t1118;
  std::complex<DOUBLE> t1119;
  std::complex<DOUBLE> t1120;
  std::complex<DOUBLE> t1121;
  std::complex<DOUBLE> t1123;
  std::complex<DOUBLE> t1124;
  std::complex<DOUBLE> t1128;
  std::complex<DOUBLE> t1129;
  std::complex<DOUBLE> t1133;
  std::complex<DOUBLE> t1135;
  std::complex<DOUBLE> t1136;
  std::complex<DOUBLE> t1137;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t1140;
  std::complex<DOUBLE> t1141;
  std::complex<DOUBLE> t1144;
  std::complex<DOUBLE> t1145;
  std::complex<DOUBLE> t1148;
  std::complex<DOUBLE> t1149;
  std::complex<DOUBLE> t1150;
  std::complex<DOUBLE> t1151;
  std::complex<DOUBLE> t1155;
  std::complex<DOUBLE> t1156;
  std::complex<DOUBLE> t1158;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t1160;
  std::complex<DOUBLE> t1161;
  std::complex<DOUBLE> t1162;
  std::complex<DOUBLE> t1165;
  std::complex<DOUBLE> t1169;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t1171;
  std::complex<DOUBLE> t1172;
  std::complex<DOUBLE> t1174;
  std::complex<DOUBLE> t1175;
  std::complex<DOUBLE> t1176;
  std::complex<DOUBLE> t1177;
  std::complex<DOUBLE> t1178;
  std::complex<DOUBLE> t1180;
  std::complex<DOUBLE> t1181;
  std::complex<DOUBLE> t1182;
  std::complex<DOUBLE> t1183;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t1193;
  std::complex<DOUBLE> t1194;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t1204;
  std::complex<DOUBLE> t1205;
  std::complex<DOUBLE> t1206;
  std::complex<DOUBLE> t1207;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t1212;
  std::complex<DOUBLE> t1213;
  std::complex<DOUBLE> t1215;
  std::complex<DOUBLE> t1218;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t1223;
  std::complex<DOUBLE> t1224;
  std::complex<DOUBLE> t1227;
  std::complex<DOUBLE> t1228;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t1231;
  std::complex<DOUBLE> t1233;
  std::complex<DOUBLE> t1234;
  std::complex<DOUBLE> t1236;
  std::complex<DOUBLE> t1238;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t1240;
  std::complex<DOUBLE> t1243;
  std::complex<DOUBLE> t1244;
  std::complex<DOUBLE> t1246;
  std::complex<DOUBLE> t1250;
  std::complex<DOUBLE> t1251;
  std::complex<DOUBLE> t1253;
  std::complex<DOUBLE> t1254;
  std::complex<DOUBLE> t1255;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t1260;
  std::complex<DOUBLE> t1267;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t1271;
  std::complex<DOUBLE> t1281;
  std::complex<DOUBLE> t1282;
  std::complex<DOUBLE> t1295;
  std::complex<DOUBLE> t1296;
  std::complex<DOUBLE> t1298;
  std::complex<DOUBLE> t1299;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t1300;
  std::complex<DOUBLE> t1301;
  std::complex<DOUBLE> t1308;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t1311;
  std::complex<DOUBLE> t1312;
  std::complex<DOUBLE> t1317;
  std::complex<DOUBLE> t1319;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t1320;
  std::complex<DOUBLE> t1323;
  std::complex<DOUBLE> t1324;
  std::complex<DOUBLE> t1330;
  std::complex<DOUBLE> t1332;
  std::complex<DOUBLE> t1338;
  std::complex<DOUBLE> t1341;
  std::complex<DOUBLE> t1343;
  std::complex<DOUBLE> t1359;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t1360;
  std::complex<DOUBLE> t1362;
  std::complex<DOUBLE> t1365;
  std::complex<DOUBLE> t1367;
  std::complex<DOUBLE> t1368;
  std::complex<DOUBLE> t1370;
  std::complex<DOUBLE> t1371;
  std::complex<DOUBLE> t1374;
  std::complex<DOUBLE> t1375;
  std::complex<DOUBLE> t1377;
  std::complex<DOUBLE> t1378;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t1385;
  std::complex<DOUBLE> t1386;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t1391;
  std::complex<DOUBLE> t1392;
  std::complex<DOUBLE> t1394;
  std::complex<DOUBLE> t1399;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t1401;
  std::complex<DOUBLE> t1402;
  std::complex<DOUBLE> t1408;
  std::complex<DOUBLE> t1411;
  std::complex<DOUBLE> t1412;
  std::complex<DOUBLE> t1415;
  std::complex<DOUBLE> t1418;
  std::complex<DOUBLE> t1421;
  std::complex<DOUBLE> t1422;
  std::complex<DOUBLE> t1426;
  std::complex<DOUBLE> t1427;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t1430;
  std::complex<DOUBLE> t1431;
  std::complex<DOUBLE> t1437;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t1441;
  std::complex<DOUBLE> t1444;
  std::complex<DOUBLE> t1447;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t1450;
  std::complex<DOUBLE> t1452;
  std::complex<DOUBLE> t1463;
  std::complex<DOUBLE> t1464;
  std::complex<DOUBLE> t1468;
  std::complex<DOUBLE> t1476;
  std::complex<DOUBLE> t1478;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t1481;
  std::complex<DOUBLE> t1482;
  std::complex<DOUBLE> t1489;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t1497;
  std::complex<DOUBLE> t1498;
  std::complex<DOUBLE> t1502;
  std::complex<DOUBLE> t1509;
  std::complex<DOUBLE> t1510;
  std::complex<DOUBLE> t1517;
  std::complex<DOUBLE> t1518;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t1522;
  std::complex<DOUBLE> t1524;
  std::complex<DOUBLE> t1527;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t1533;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t1546;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t1575;
  std::complex<DOUBLE> t1576;
  std::complex<DOUBLE> t1587;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t1592;
  std::complex<DOUBLE> t1595;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t1603;
  std::complex<DOUBLE> t1615;
  std::complex<DOUBLE> t1617;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t1620;
  std::complex<DOUBLE> t1624;
  std::complex<DOUBLE> t1630;
  std::complex<DOUBLE> t1631;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t1644;
  std::complex<DOUBLE> t1646;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t1650;
  std::complex<DOUBLE> t1651;
  std::complex<DOUBLE> t1656;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t1663;
  std::complex<DOUBLE> t1666;
  std::complex<DOUBLE> t1667;
  std::complex<DOUBLE> t1669;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t1672;
  std::complex<DOUBLE> t1674;
  std::complex<DOUBLE> t1680;
  std::complex<DOUBLE> t1681;
  std::complex<DOUBLE> t1682;
  std::complex<DOUBLE> t1685;
  std::complex<DOUBLE> t1686;
  std::complex<DOUBLE> t1688;
  std::complex<DOUBLE> t1694;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t1702;
  std::complex<DOUBLE> t1705;
  std::complex<DOUBLE> t1707;
  std::complex<DOUBLE> t1717;
  std::complex<DOUBLE> t1721;
  std::complex<DOUBLE> t1724;
  std::complex<DOUBLE> t1726;
  std::complex<DOUBLE> t1729;
  std::complex<DOUBLE> t1732;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t1776;
  std::complex<DOUBLE> t1785;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t1791;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t1821;
  std::complex<DOUBLE> t1828;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t1832;
  std::complex<DOUBLE> t1834;
  std::complex<DOUBLE> t1845;
  std::complex<DOUBLE> t1849;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t1855;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t1870;
  std::complex<DOUBLE> t1874;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t1880;
  std::complex<DOUBLE> t1884;
  std::complex<DOUBLE> t1889;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t1906;
  std::complex<DOUBLE> t1918;
  std::complex<DOUBLE> t1919;
  std::complex<DOUBLE> t1920;
  std::complex<DOUBLE> t1925;
  std::complex<DOUBLE> t1933;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t1981;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t1993;
  FourMomentum<DOUBLE> t2;
  std::complex<DOUBLE> t2001;
  std::complex<DOUBLE> t2010;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t249;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t288;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t301;
  std::complex<DOUBLE> t302;
  std::complex<DOUBLE> t304;
  std::complex<DOUBLE> t305;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t314;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t323;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t329;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t336;
  std::complex<DOUBLE> t338;
  std::complex<DOUBLE> t344;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t371;
  std::complex<DOUBLE> t373;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t376;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t397;
  std::complex<DOUBLE> t398;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t405;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t414;
  std::complex<DOUBLE> t417;
  std::complex<DOUBLE> t418;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t421;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t427;
  std::complex<DOUBLE> t428;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t432;
  std::complex<DOUBLE> t433;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t437;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t443;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t450;
  std::complex<DOUBLE> t453;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t458;
  std::complex<DOUBLE> t469;
  std::complex<DOUBLE> t470;
  std::complex<DOUBLE> t474;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t482;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t488;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t491;
  std::complex<DOUBLE> t496;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t504;
  std::complex<DOUBLE> t505;
  std::complex<DOUBLE> t509;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t516;
  std::complex<DOUBLE> t517;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t524;
  std::complex<DOUBLE> t525;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t531;
  std::complex<DOUBLE> t534;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t540;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t553;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t582;
  std::complex<DOUBLE> t583;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t594;
  std::complex<DOUBLE> t599;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t602;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t610;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t622;
  std::complex<DOUBLE> t624;
  std::complex<DOUBLE> t627;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t631;
  std::complex<DOUBLE> t637;
  std::complex<DOUBLE> t638;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t651;
  std::complex<DOUBLE> t653;
  std::complex<DOUBLE> t657;
  std::complex<DOUBLE> t658;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t663;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t670;
  std::complex<DOUBLE> t673;
  std::complex<DOUBLE> t674;
  std::complex<DOUBLE> t676;
  std::complex<DOUBLE> t679;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t681;
  std::complex<DOUBLE> t687;
  std::complex<DOUBLE> t688;
  std::complex<DOUBLE> t689;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t692;
  std::complex<DOUBLE> t693;
  std::complex<DOUBLE> t695;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t701;
  std::complex<DOUBLE> t709;
  std::complex<DOUBLE> t712;
  std::complex<DOUBLE> t714;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t724;
  std::complex<DOUBLE> t728;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t731;
  std::complex<DOUBLE> t732;
  std::complex<DOUBLE> t734;
  std::complex<DOUBLE> t739;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t783;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t792;
  std::complex<DOUBLE> t798;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t824;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t835;
  std::complex<DOUBLE> t839;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t841;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t852;
  std::complex<DOUBLE> t856;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t862;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t877;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t881;
  std::complex<DOUBLE> t887;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t891;
  std::complex<DOUBLE> t896;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t913;
  std::complex<DOUBLE> t921;
  std::complex<DOUBLE> t926;
  std::complex<DOUBLE> t927;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t932;
  std::complex<DOUBLE> t940;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t988;
  std::complex<DOUBLE> t99;
  {
    t1 = p6+p5;
    t2 = p2+p1;
    t3 = dot(t1,t2);
    t4 = t3*t3;
    t5 = S(t2);
    t6 = S(t1);
    t7 = t5*t6;
    t9 = sqrt(t4-t7);
    gg1 = t3+t9;
    t10 = gg1*gg1;
    t12 = ONE/(t10-t7);
    t13 = t12*t10;
    t16 = A(p5,p6);
    t17 = t12*t16;
    t18 = B(p5,p6);
    t19 = t18*gg1;
    K1f1 = t13*p6+t13*p5+t17*t19*p2+t17*t19*p1;
    t26 = A(p1,p2);
    t27 = t12*t26;
    t28 = B(p1,p2);
    t29 = t28*gg1;
    K2f1 = t13*p2+t13*p1+t27*t29*p6+t27*t29*p5;
    t35 = t16*t28*t18;
    t36 = t27*t35;
    t39 = A(K1f1,p1);
    t40 = A(K1f1,p3);
    t41 = t39*t40;
    t42 = A(K2f1,p5);
    t43 = B(K1f1,p1);
    t44 = t42*t43;
    t45 = B(K1f1,p5);
    t48 = B(K2f1,p2);
    t49 = t48*t48;
    t50 = B(K2f1,p4);
    t51 = t50*t50;
    t52 = t49*t51;
    t53 = B(K2f1,p6);
    t54 = t53*t53;
    t55 = IA(p1,p2);
    t56 = t54*t55;
    t57 = IA(p3,p4);
    t58 = t56*t57;
    t59 = t52*t58;
    t61 = IA(p5,p6);
    t62 = IB(K2f1,p1);
    t63 = t62*t62;
    t64 = t61*t63;
    t65 = IB(K2f1,p3);
    t66 = IB(K2f1,p5);
    t67 = t66*t66;
    t68 = t65*t67;
    t69 = IB(p1,p2);
    t70 = t68*t69;
    t72 = IB(p3,p4);
    t73 = IB(p5,p6);
    t74 = t72*t73;
    t75 = IAB(K1f1,t2,K1f1);
    t78 = t74*t75*(t36-t27*t29)*I;
    t79 = t64*t70*t78;
    t81 = B(K1f1,p2);
    t82 = t81*t81;
    t83 = t42*t82;
    t84 = t41*t83;
    t85 = B(K1f1,p4);
    t86 = t85*t85;
    t87 = B(K1f1,p6);
    t88 = t87*t87;
    t89 = t86*t88;
    t90 = B(K2f1,p1);
    t93 = t89*t90*t55*t57;
    t95 = IB(K1f1,p1);
    t96 = t95*t95;
    t97 = t61*t96;
    t98 = IB(K1f1,p3);
    t99 = IB(K1f1,p5);
    t100 = t98*t99;
    t101 = t100*t69;
    t102 = t97*t101;
    t103 = t102*t78;
    t108 = t97*t100;
    t109 = t75*t75;
    t110 = t109*I;
    t111 = t74*t110;
    t114 = B(K2f1,p3);
    t116 = t114*t55*t57;
    t117 = t89*t116;
    t119 = t61*t95;
    t120 = t98*t98;
    t121 = t120*t99;
    t122 = t121*t69;
    t123 = t119*t122;
    t124 = t123*t78;
    t126 = B(K1f1,p3);
    t127 = t42*t126;
    t131 = t61*t62;
    t132 = t65*t65;
    t136 = t131*t132*t67*t69*t78;
    t138 = t41*t127;
    t139 = t45*t49;
    t140 = t51*t54;
    t143 = t57*t61;
    t144 = t62*t132;
    t145 = t143*t144;
    t148 = t73*t109*I;
    t149 = t67*t72*t148;
    t152 = A(K1f1,p5);
    t153 = t39*t152;
    t154 = A(K2f1,p3);
    t155 = t154*t43;
    t159 = t132*t66;
    t160 = t159*t69;
    t162 = t64*t160*t78;
    t164 = t154*t82;
    t165 = t153*t164;
    t166 = t165*t93;
    t167 = t69*t72;
    t170 = t167*t73*t75*I;
    t174 = B(K2f1,p5);
    t176 = t174*t55*t57;
    t177 = t89*t176;
    t179 = t99*t99;
    t180 = t98*t179;
    t181 = t180*t69;
    t182 = t119*t181;
    t183 = t182*t78;
    t185 = t154*t126;
    t186 = t153*t185;
    t187 = t140*t55;
    t188 = t139*t187;
    t198 = t39*t154;
    t199 = t198*t83;
    t209 = t83*t86;
    t210 = t198*t209;
    t211 = t88*t90;
    t212 = t211*t116;
    t217 = t74*t75*(-t36+t17*t19)*I;
    t218 = t97*t122*t217;
    t220 = t211*t176;
    t223 = t97*t181*t217;
    t228 = t55*t57;
    t229 = t96*t98;
    t230 = t228*t229;
    t233 = t72*t109*I;
    t234 = t179*t69*t233;
    t237 = t198*t127;
    t239 = t131*t160;
    t240 = t239*t217;
    t242 = t52*t56;
    t244 = t57*t62;
    t246 = t167*t110;
    t249 = t41*t44*t45*t59*t79-t84*t93*t103-t84*t89*t90*t57*t108*t111-t84*t117*
t124+t41*t127*t45*t59*t136+t138*t139*t140*t145*t149+t153*t155*t126*t59*t162-
t166*t108*t170-t166*t103-t165*t177*t183+t186*t188*t143*t144*t67*t170+t153*t185*
t45*t59*t136-t199*t89*t90*t114*t55*t143*t96*t120*t99*t170+t210*t212*t218+t210*
t220*t223-t199*t89*t90*t174*t230*t234-t237*t59*t240+t237*t242*t244*t159*t246;
    t250 = t42*t45;
    t252 = t198*t250*t59;
    t256 = t131*t70;
    t257 = t256*t217;
    t259 = t40*t152;
    t260 = A(K2f1,p1);
    t261 = t260*t43;
    t266 = t259*t261;
    t273 = t66*t72*t148;
    t277 = t63*t65;
    t287 = t259*t260*t82;
    t288 = t287*t117;
    t301 = t40*t260;
    t302 = t301*t44;
    t304 = t65*t66;
    t305 = t304*t69;
    t306 = t64*t305;
    t307 = t306*t217;
    t314 = t301*t209;
    t317 = t88*t114;
    t318 = t317*t176;
    t323 = t119*t120*t179*t69*t217;
    t325 = t301*t83;
    t326 = t114*t174;
    t329 = t95*t120;
    t330 = t228*t329;
    t336 = t152*t260;
    t338 = t336*t155*t59;
    t344 = t336*t164*t86;
    t347 = t252*t131*t68*t170-t252*t257+t259*t261*t126*t59*t162+t266*t126*t49*
t140*t143*t63*t132*t273+t266*t188*t143*t277*t67*t170+t259*t261*t45*t59*t79-t288
*t119*t121*t170-t288*t124-t287*t177*t183-t287*t89*t174*t57*t119*t180*t111-t302*
t59*t307+t302*t242*t57*t63*t304*t246+t314*t212*t218+t314*t318*t323-t325*t89*
t326*t330*t234-t301*t250*t59*t257+t338*t64*t304*t170-t338*t307+t344*t220*t223;
    t349 = t336*t164;
    t365 = t53*t55*t57;
    t366 = t52*t365;
    t368 = t306*t78;
    t371 = t85*t88;
    t373 = t50*t55*t57;
    t374 = t371*t373;
    t376 = t119*t101;
    t377 = t376*t78;
    t380 = t42*t81;
    t381 = t41*t380;
    t383 = t48*t55*t57;
    t384 = t89*t383;
    t391 = t119*t100;
    t392 = t391*t111;
    t397 = t48*t51;
    t398 = t397*t58;
    t400 = t256*t78;
    t405 = t397*t365;
    t407 = t131*t305;
    t408 = t407*t78;
    t414 = t239*t78;
    t417 = t87*t49;
    t418 = t51*t53;
    t421 = t145*t273;
    t424 = t42*t85;
    t427 = t49*t50;
    t428 = t427*t58;
    t432 = t41*t424;
    t433 = t50*t54;
    t436 = t62*t65;
    t437 = t143*t436;
    t443 = t427*t365;
    t447 = t50*t53;
    t450 = t437*t273;
    t453 = t45*t45;
    t456 = t41*t42*t453*t59;
    t458 = t65*t67*t66;
    t469 = t418*t55;
    t470 = t417*t469;
    t474 = t143*t436*t67*t170;
    t482 = -t349*t89*t326*t55*t143*t329*t179*t170+t344*t318*t323-t336*t185*t59*
t240-TWO*t41*t44*t87*t366*t368+TWO*t84*t374*t377+TWO*t381*t384*t377+TWO*t381*
t89*t48*t57*t392-TWO*t41*t380*t45*t398*t400+FOUR*t41*t380*t87*t405*t408-TWO*t41*
t127*t87*t366*t414-TWO*t138*t417*t418*t421-TWO*t41*t424*t45*t428*t400-TWO*t432*
t139*t433*t437*t149+FOUR*t41*t424*t87*t443*t408+FOUR*t432*t417*t447*t450+TWO*t456
*t131*t458*t170+TWO*t456*t131*t458*t69*t78-FOUR*t41*t250*t470*t474-FOUR*t41*t250*
t87*t366*t400;
    t484 = t41*t42*t88;
    t487 = t131*t304;
    t488 = t487*t170;
    t491 = t228*t61;
    t496 = t436*t66*t69*t78;
    t504 = t86*t87;
    t505 = t504*t365;
    t509 = t154*t81;
    t516 = t153*t509*t384;
    t517 = t391*t170;
    t524 = t48*t50;
    t525 = t524*t58;
    t529 = t126*t126;
    t531 = t153*t154*t529;
    t534 = t132*t65*t66;
    t540 = t54*t57;
    t553 = t85*t49*t433;
    t582 = t154*t85;
    t583 = t153*t582;
    t594 = t447*t55;
    t599 = t143*t436*t66*t170;
    t602 = TWO*t484*t52*t228*t488+TWO*t484*t52*t491*t496-TWO*t153*t155*t85*t428
*t368+TWO*t165*t505*t377-TWO*t153*t509*t126*t398*t414+TWO*t516*t517+TWO*t516*
t377+FOUR*t153*t509*t85*t525*t408+TWO*t531*t59*t131*t534*t69*t78+TWO*t531*t52*
t540*t131*t534*t111-FOUR*t153*t185*t85*t428*t414-FOUR*t186*t553*t421-TWO*t186*
t470*t143*t144*t66*t170-TWO*t153*t185*t87*t366*t414+TWO*t153*t154*t86*t49*t54*
t491*t496+TWO*t153*t154*t86*t49*t540*t487*t111-TWO*t583*t139*t433*t55*t474-TWO*
t153*t582*t45*t428*t400+FOUR*t583*t417*t594*t599;
    t610 = t90*t90;
    t622 = t87*t90*t365;
    t624 = t102*t217;
    t627 = t198*t424;
    t631 = t244*t304*t246;
    t637 = t99*t69;
    t638 = t637*t233;
    t651 = t83*t85;
    t653 = t211*t373;
    t657 = t380*t86;
    t658 = t198*t657;
    t663 = t198*t380;
    t670 = t143*t329*t99*t170;
    t673 = t88*t48;
    t674 = t673*t116;
    t676 = t123*t217;
    t679 = t673*t176;
    t681 = t182*t217;
    t687 = t95*t98;
    t688 = t228*t687;
    t689 = t688*t234;
    t692 = t87*t48;
    t693 = t692*t365;
    t695 = t376*t217;
    t701 = t688*t638;
    t709 = t143*t687*t99*t170;
    t712 = t380*t85;
    t714 = t673*t373;
    t724 = t687*t637*t217;
    t728 = t407*t217;
    t731 = FOUR*t153*t582*t87*t443*t408+TWO*t199*t89*t610*t55*t57*t61*t96*t95*
t101*t217-TWO*t210*t622*t624-TWO*t627*t427*t56*t631+TWO*t199*t504*t90*t53*t230*
t638+TWO*t199*t371*t90*t50*t55*t143*t229*t99*t170-TWO*t198*t651*t653*t624-FOUR*
t658*t211*t383*t624+TWO*t663*t89*t48*t114*t55*t670-TWO*t658*t674*t676-TWO*t658*
t679*t681+TWO*t663*t89*t48*t174*t689+FOUR*t658*t693*t695-FOUR*t663*t504*t48*t53*
t701-FOUR*t663*t371*t524*t55*t709+FOUR*t198*t712*t714*t695+TWO*t198*t42*t86*t88*
t49*t491*t724+TWO*t627*t428*t728;
    t732 = t42*t87;
    t734 = t198*t732*t366;
    t739 = t43*t43;
    t783 = t287*t374;
    t792 = t260*t81;
    t798 = t259*t792;
    t824 = -TWO*t734*t488+TWO*t734*t728+TWO*t259*t260*t739*t59*t61*t63*t62*t305
*t78-FOUR*t259*t261*t81*t398*t368-TWO*t259*t261*t85*t428*t368-TWO*t266*t553*t143
*t277*t273-TWO*t266*t470*t143*t277*t66*t170-TWO*t259*t261*t87*t366*t368+TWO*
t287*t505*t377+TWO*t287*t504*t53*t57*t392+TWO*t783*t517+TWO*t783*t377+TWO*t287*
t140*t491*t496-TWO*t259*t792*t126*t398*t414-TWO*t798*t126*t48*t140*t421+FOUR*
t259*t792*t85*t525*t408+FOUR*t798*t85*t48*t433*t450-TWO*t798*t45*t48*t187*t474
-TWO*t259*t792*t45*t398*t400;
    t835 = t114*t114;
    t839 = t325*t89*t835*t55*t57;
    t841 = t120*t98*t99;
    t852 = t87*t114*t365;
    t856 = t114*t53;
    t862 = t301*t651;
    t877 = t88*t50*t176;
    t881 = t50*t174;
    t887 = t87*t50*t365;
    t891 = t85*t87;
    t896 = t88*t51;
    t913 = t301*t380;
    t921 = FOUR*t798*t692*t469*t599+FOUR*t259*t792*t87*t405*t408-TWO*t839*t119*
t841*t170+TWO*t839*t119*t841*t69*t217-TWO*t314*t852*t676+TWO*t325*t504*t856*
t330*t638-TWO*t862*t653*t624+FOUR*t325*t371*t114*t50*t55*t670-FOUR*t862*t317*t373
*t676-TWO*t862*t877*t681+TWO*t325*t371*t881*t689+FOUR*t862*t887*t695-FOUR*t325*
t891*t447*t701-TWO*t325*t896*t228*t517+TWO*t325*t896*t491*t724-TWO*t301*t657*
t674*t676+FOUR*t301*t712*t714*t695+TWO*t913*t398*t728-TWO*t913*t397*t56*t631;
    t926 = t174*t174;
    t927 = t926*t55;
    t932 = t98*t179*t99;
    t940 = t57*t95;
    t988 = t336*t164*t85;
    t1000 = t336*t509*t86;
    t1008 = t336*t509*t398;
    t1017 = TWO*t301*t732*t366*t728+TWO*t349*t89*t927*t57*t119*t932*t69*t217
-TWO*t349*t89*t927*t940*t932*t246-TWO*t344*t622*t624+TWO*t349*t504*t856*t55*
t670-TWO*t344*t852*t676-FOUR*t344*t87*t174*t365*t681+FOUR*t349*t504*t174*t53*t689
+TWO*t349*t86*t54*t491*t724-TWO*t336*t154*t82*t86*t56*t940*t100*t246+TWO*t349*
t371*t881*t55*t143*t687*t179*t170-TWO*t988*t877*t681-FOUR*t349*t891*t594*t709+
FOUR*t988*t887*t695-TWO*t1000*t679*t681+FOUR*t1000*t693*t695-TWO*t1008*t488+TWO*
t1008*t728+TWO*t336*t582*t428*t728;
    gg = t3-t9;
    t1020 = gg*gg;
    t1022 = ONE/(t1020-t7);
    t1023 = t1022*t1020;
    t1026 = t1022*t16;
    t1027 = t18*gg;
    K1f = t1023*p6+t1023*p5+t1026*t1027*p2+t1026*t1027*p1;
    t1034 = t1022*t26;
    t1035 = t28*gg;
    K2f = t1023*p2+t1023*p1+t1034*t1035*p6+t1034*t1035*p5;
    t1040 = t1034*t35;
    t1043 = A(K1f,p1);
    t1044 = A(K1f,p3);
    t1045 = t1043*t1044;
    t1046 = A(K2f,p5);
    t1047 = B(K1f,p1);
    t1048 = t1046*t1047;
    t1049 = B(K1f,p5);
    t1052 = B(K2f,p2);
    t1053 = t1052*t1052;
    t1054 = B(K2f,p4);
    t1055 = t1054*t1054;
    t1056 = t1053*t1055;
    t1057 = B(K2f,p6);
    t1058 = t1057*t1057;
    t1059 = t1058*t55;
    t1060 = t1059*t57;
    t1061 = t1056*t1060;
    t1063 = IB(K2f,p1);
    t1064 = t1063*t1063;
    t1065 = t61*t1064;
    t1066 = IB(K2f,p3);
    t1067 = IB(K2f,p5);
    t1068 = t1067*t1067;
    t1069 = t1066*t1068;
    t1070 = t1069*t69;
    t1072 = IAB(K1f,t2,K1f);
    t1075 = t74*t1072*(t1040-t1034*t1035)*I;
    t1076 = t1065*t1070*t1075;
    t1078 = B(K1f,p2);
    t1079 = t1078*t1078;
    t1080 = t1046*t1079;
    t1081 = t1045*t1080;
    t1082 = B(K1f,p4);
    t1083 = t1082*t1082;
    t1084 = B(K1f,p6);
    t1085 = t1084*t1084;
    t1086 = t1083*t1085;
    t1087 = B(K2f,p1);
    t1090 = t1086*t1087*t55*t57;
    t1092 = IB(K1f,p1);
    t1093 = t1092*t1092;
    t1094 = t61*t1093;
    t1095 = IB(K1f,p3);
    t1096 = IB(K1f,p5);
    t1097 = t1095*t1096;
    t1098 = t1097*t69;
    t1099 = t1094*t1098;
    t1100 = t1099*t1075;
    t1105 = t1094*t1097;
    t1106 = t1072*t1072;
    t1107 = t1106*I;
    t1108 = t74*t1107;
    t1111 = B(K2f,p3);
    t1113 = t1111*t55*t57;
    t1114 = t1086*t1113;
    t1116 = t61*t1092;
    t1117 = t1095*t1095;
    t1118 = t1117*t1096;
    t1119 = t1118*t69;
    t1120 = t1116*t1119;
    t1121 = t1120*t1075;
    t1123 = B(K1f,p3);
    t1124 = t1046*t1123;
    t1128 = t61*t1063;
    t1129 = t1066*t1066;
    t1133 = t1128*t1129*t1068*t69*t1075;
    t1135 = t1045*t1124;
    t1136 = t1049*t1053;
    t1137 = t1055*t1058;
    t1140 = t1063*t1129;
    t1141 = t143*t1140;
    t1144 = t73*t1106*I;
    t1145 = t1068*t72*t1144;
    t1148 = A(K1f,p5);
    t1149 = t1043*t1148;
    t1150 = A(K2f,p3);
    t1151 = t1150*t1047;
    t1155 = t1129*t1067;
    t1156 = t1155*t69;
    t1158 = t1065*t1156*t1075;
    t1160 = t1150*t1079;
    t1161 = t1149*t1160;
    t1162 = t1161*t1090;
    t1165 = t167*t73*t1072*I;
    t1169 = B(K2f,p5);
    t1171 = t1169*t55*t57;
    t1172 = t1086*t1171;
    t1174 = t1096*t1096;
    t1175 = t1095*t1174;
    t1176 = t1175*t69;
    t1177 = t1116*t1176;
    t1178 = t1177*t1075;
    t1180 = t1150*t1123;
    t1181 = t1149*t1180;
    t1182 = t1137*t55;
    t1183 = t1136*t1182;
    t1193 = t1043*t1150;
    t1194 = t1193*t1080;
    t1204 = t1080*t1083;
    t1205 = t1193*t1204;
    t1206 = t1085*t1087;
    t1207 = t1206*t1113;
    t1212 = t74*t1072*(-t1040+t1026*t1027)*I;
    t1213 = t1094*t1119*t1212;
    t1215 = t1206*t1171;
    t1218 = t1094*t1176*t1212;
    t1223 = t1093*t1095;
    t1224 = t228*t1223;
    t1227 = t72*t1106*I;
    t1228 = t1174*t69*t1227;
    t1231 = t1193*t1124;
    t1233 = t1128*t1156;
    t1234 = t1233*t1212;
    t1236 = t1056*t1059;
    t1238 = t57*t1063;
    t1240 = t167*t1107;
    t1243 = t1045*t1048*t1049*t1061*t1076-t1081*t1090*t1100-t1081*t1086*t1087*
t57*t1105*t1108-t1081*t1114*t1121+t1045*t1124*t1049*t1061*t1133+t1135*t1136*
t1137*t1141*t1145+t1149*t1151*t1123*t1061*t1158-t1162*t1105*t1165-t1162*t1100-
t1161*t1172*t1178+t1181*t1183*t143*t1140*t1068*t1165+t1149*t1180*t1049*t1061*
t1133-t1194*t1086*t1087*t1111*t55*t143*t1093*t1117*t1096*t1165+t1205*t1207*
t1213+t1205*t1215*t1218-t1194*t1086*t1087*t1169*t1224*t1228-t1231*t1061*t1234+
t1231*t1236*t1238*t1155*t1240;
    t1244 = t1046*t1049;
    t1246 = t1193*t1244*t1061;
    t1250 = t1128*t1070;
    t1251 = t1250*t1212;
    t1253 = t1044*t1148;
    t1254 = A(K2f,p1);
    t1255 = t1254*t1047;
    t1260 = t1253*t1255;
    t1267 = t1067*t72*t1144;
    t1271 = t1064*t1066;
    t1281 = t1253*t1254*t1079;
    t1282 = t1281*t1114;
    t1295 = t1044*t1254;
    t1296 = t1295*t1048;
    t1298 = t1066*t1067;
    t1299 = t1298*t69;
    t1300 = t1065*t1299;
    t1301 = t1300*t1212;
    t1308 = t1295*t1204;
    t1311 = t1085*t1111;
    t1312 = t1311*t1171;
    t1317 = t1116*t1117*t1174*t69*t1212;
    t1319 = t1295*t1080;
    t1320 = t1111*t1169;
    t1323 = t1092*t1117;
    t1324 = t228*t1323;
    t1330 = t1148*t1254;
    t1332 = t1330*t1151*t1061;
    t1338 = t1330*t1160*t1083;
    t1341 = t1246*t1128*t1069*t1165-t1246*t1251+t1253*t1255*t1123*t1061*t1158+
t1260*t1123*t1053*t1137*t143*t1064*t1129*t1267+t1260*t1183*t143*t1271*t1068*
t1165+t1253*t1255*t1049*t1061*t1076-t1282*t1116*t1118*t1165-t1282*t1121-t1281*
t1172*t1178-t1281*t1086*t1169*t57*t1116*t1175*t1108-t1296*t1061*t1301+t1296*
t1236*t57*t1064*t1298*t1240+t1308*t1207*t1213+t1308*t1312*t1317-t1319*t1086*
t1320*t1324*t1228-t1295*t1244*t1061*t1251+t1332*t1065*t1298*t1165-t1332*t1301+
t1338*t1215*t1218;
    t1343 = t1330*t1160;
    t1359 = t1057*t55*t57;
    t1360 = t1056*t1359;
    t1362 = t1300*t1075;
    t1365 = t1082*t1085;
    t1367 = t1054*t55*t57;
    t1368 = t1365*t1367;
    t1370 = t1116*t1098;
    t1371 = t1370*t1075;
    t1374 = t1046*t1078;
    t1375 = t1045*t1374;
    t1377 = t1052*t55*t57;
    t1378 = t1086*t1377;
    t1385 = t1116*t1097;
    t1386 = t1385*t1108;
    t1391 = t1052*t1055;
    t1392 = t1391*t1060;
    t1394 = t1250*t1075;
    t1399 = t1391*t1359;
    t1401 = t1128*t1299;
    t1402 = t1401*t1075;
    t1408 = t1233*t1075;
    t1411 = t1084*t1053;
    t1412 = t1055*t1057;
    t1415 = t1141*t1267;
    t1418 = t1046*t1082;
    t1421 = t1053*t1054;
    t1422 = t1421*t1060;
    t1426 = t1045*t1418;
    t1427 = t1054*t1058;
    t1430 = t1063*t1066;
    t1431 = t143*t1430;
    t1437 = t1421*t1359;
    t1441 = t1054*t1057;
    t1444 = t1431*t1267;
    t1447 = t1049*t1049;
    t1450 = t1045*t1046*t1447*t1061;
    t1452 = t1066*t1068*t1067;
    t1463 = t1412*t55;
    t1464 = t1411*t1463;
    t1468 = t143*t1430*t1068*t1165;
    t1476 = -t1343*t1086*t1320*t55*t143*t1323*t1174*t1165+t1338*t1312*t1317-
t1330*t1180*t1061*t1234-TWO*t1045*t1048*t1084*t1360*t1362+TWO*t1081*t1368*t1371
+TWO*t1375*t1378*t1371+TWO*t1375*t1086*t1052*t57*t1386-TWO*t1045*t1374*t1049*
t1392*t1394+FOUR*t1045*t1374*t1084*t1399*t1402-TWO*t1045*t1124*t1084*t1360*t1408
-TWO*t1135*t1411*t1412*t1415-TWO*t1045*t1418*t1049*t1422*t1394-TWO*t1426*t1136*
t1427*t1431*t1145+FOUR*t1045*t1418*t1084*t1437*t1402+FOUR*t1426*t1411*t1441*t1444
+TWO*t1450*t1128*t1452*t1165+TWO*t1450*t1128*t1452*t69*t1075-FOUR*t1045*t1244*
t1464*t1468-FOUR*t1045*t1244*t1084*t1360*t1394;
    t1478 = t1045*t1046*t1085;
    t1481 = t1128*t1298;
    t1482 = t1481*t1165;
    t1489 = t1430*t1067*t69*t1075;
    t1497 = t1083*t1084;
    t1498 = t1497*t1359;
    t1502 = t1150*t1078;
    t1509 = t1149*t1502*t1378;
    t1510 = t1385*t1165;
    t1517 = t1052*t1054;
    t1518 = t1517*t1060;
    t1522 = t1123*t1123;
    t1524 = t1149*t1150*t1522;
    t1527 = t1129*t1066*t1067;
    t1533 = t1058*t57;
    t1546 = t1082*t1053*t1427;
    t1575 = t1150*t1082;
    t1576 = t1149*t1575;
    t1587 = t1441*t55;
    t1592 = t143*t1430*t1067*t1165;
    t1595 = TWO*t1478*t1056*t228*t1482+TWO*t1478*t1056*t491*t1489-TWO*t1149*
t1151*t1082*t1422*t1362+TWO*t1161*t1498*t1371-TWO*t1149*t1502*t1123*t1392*t1408
+TWO*t1509*t1510+TWO*t1509*t1371+FOUR*t1149*t1502*t1082*t1518*t1402+TWO*t1524*
t1061*t1128*t1527*t69*t1075+TWO*t1524*t1056*t1533*t1128*t1527*t1108-FOUR*t1149*
t1180*t1082*t1422*t1408-FOUR*t1181*t1546*t1415-TWO*t1181*t1464*t143*t1140*t1067*
t1165-TWO*t1149*t1180*t1084*t1360*t1408+TWO*t1149*t1150*t1083*t1053*t1058*t491*
t1489+TWO*t1149*t1150*t1083*t1053*t1533*t1481*t1108-TWO*t1576*t1136*t1427*t55*
t1468-TWO*t1149*t1575*t1049*t1422*t1394+FOUR*t1576*t1411*t1587*t1592;
    t1603 = t1087*t1087;
    t1615 = t1084*t1087*t1359;
    t1617 = t1099*t1212;
    t1620 = t1193*t1418;
    t1624 = t1238*t1298*t1240;
    t1630 = t1096*t69;
    t1631 = t1630*t1227;
    t1644 = t1080*t1082;
    t1646 = t1206*t1367;
    t1650 = t1374*t1083;
    t1651 = t1193*t1650;
    t1656 = t1193*t1374;
    t1663 = t143*t1323*t1096*t1165;
    t1666 = t1085*t1052;
    t1667 = t1666*t1113;
    t1669 = t1120*t1212;
    t1672 = t1666*t1171;
    t1674 = t1177*t1212;
    t1680 = t1092*t1095;
    t1681 = t228*t1680;
    t1682 = t1681*t1228;
    t1685 = t1084*t1052;
    t1686 = t1685*t1359;
    t1688 = t1370*t1212;
    t1694 = t1681*t1631;
    t1702 = t143*t1680*t1096*t1165;
    t1705 = t1374*t1082;
    t1707 = t1666*t1367;
    t1717 = t1680*t1630*t1212;
    t1721 = t1401*t1212;
    t1724 = t1046*t1084;
    t1726 = t1193*t1724*t1360;
    t1729 = FOUR*t1149*t1575*t1084*t1437*t1402+TWO*t1194*t1086*t1603*t55*t57*t61
*t1093*t1092*t1098*t1212-TWO*t1205*t1615*t1617-TWO*t1620*t1421*t1059*t1624+TWO*
t1194*t1497*t1087*t1057*t1224*t1631+TWO*t1194*t1365*t1087*t1054*t55*t143*t1223*
t1096*t1165-TWO*t1193*t1644*t1646*t1617-FOUR*t1651*t1206*t1377*t1617+TWO*t1656*
t1086*t1052*t1111*t55*t1663-TWO*t1651*t1667*t1669-TWO*t1651*t1672*t1674+TWO*
t1656*t1086*t1052*t1169*t1682+FOUR*t1651*t1686*t1688-FOUR*t1656*t1497*t1052*t1057
*t1694-FOUR*t1656*t1365*t1517*t55*t1702+FOUR*t1193*t1705*t1707*t1688+TWO*t1193*
t1046*t1083*t1085*t1053*t491*t1717+TWO*t1620*t1422*t1721-TWO*t1726*t1482;
    t1732 = t1047*t1047;
    t1776 = t1281*t1368;
    t1785 = t1254*t1078;
    t1791 = t1253*t1785;
    t1821 = TWO*t1726*t1721+TWO*t1253*t1254*t1732*t1061*t61*t1064*t1063*t1299*
t1075-FOUR*t1253*t1255*t1078*t1392*t1362-TWO*t1253*t1255*t1082*t1422*t1362-TWO*
t1260*t1546*t143*t1271*t1267-TWO*t1260*t1464*t143*t1271*t1067*t1165-TWO*t1253*
t1255*t1084*t1360*t1362+TWO*t1281*t1498*t1371+TWO*t1281*t1497*t1057*t57*t1386+
TWO*t1776*t1510+TWO*t1776*t1371+TWO*t1281*t1137*t491*t1489-TWO*t1253*t1785*
t1123*t1392*t1408-TWO*t1791*t1123*t1052*t1137*t1415+FOUR*t1253*t1785*t1082*t1518
*t1402+FOUR*t1791*t1082*t1052*t1427*t1444-TWO*t1791*t1049*t1052*t1182*t1468-TWO*
t1253*t1785*t1049*t1392*t1394+FOUR*t1791*t1685*t1463*t1592;
    t1828 = t1111*t1111;
    t1832 = t1319*t1086*t1828*t55*t57;
    t1834 = t1117*t1095*t1096;
    t1845 = t1084*t1111*t1359;
    t1849 = t1111*t1057;
    t1855 = t1295*t1644;
    t1870 = t1085*t1054*t1171;
    t1874 = t1054*t1169;
    t1880 = t1084*t1054*t1359;
    t1884 = t1082*t1084;
    t1889 = t1085*t1055;
    t1906 = t1295*t1374;
    t1918 = FOUR*t1253*t1785*t1084*t1399*t1402-TWO*t1832*t1116*t1834*t1165+TWO*
t1832*t1116*t1834*t69*t1212-TWO*t1308*t1845*t1669+TWO*t1319*t1497*t1849*t1324*
t1631-TWO*t1855*t1646*t1617+FOUR*t1319*t1365*t1111*t1054*t55*t1663-FOUR*t1855*
t1311*t1367*t1669-TWO*t1855*t1870*t1674+TWO*t1319*t1365*t1874*t1682+FOUR*t1855*
t1880*t1688-FOUR*t1319*t1884*t1441*t1694-TWO*t1319*t1889*t228*t1510+TWO*t1319*
t1889*t491*t1717-TWO*t1295*t1650*t1667*t1669+FOUR*t1295*t1705*t1707*t1688+TWO*
t1906*t1392*t1721-TWO*t1906*t1391*t1059*t1624+TWO*t1295*t1724*t1360*t1721;
    t1919 = t1169*t1169;
    t1920 = t1919*t55;
    t1925 = t1095*t1174*t1096;
    t1933 = t57*t1092;
    t1981 = t1330*t1160*t1082;
    t1993 = t1330*t1502*t1083;
    t2001 = t1330*t1502*t1392;
    t2010 = t249+TWO*t1343*t1086*t1920*t57*t1116*t1925*t69*t1212+t482+TWO*t1330
*t1575*t1422*t1721-FOUR*t1343*t1884*t1587*t1702-TWO*t2001*t1482+TWO*t2001*t1721
-TWO*t1338*t1615*t1617+FOUR*t1993*t1686*t1688-TWO*t1993*t1672*t1674+FOUR*t1981*
t1880*t1688-TWO*t1338*t1845*t1669+TWO*t1343*t1083*t1058*t491*t1717+FOUR*t1343*
t1497*t1169*t1057*t1682-FOUR*t1338*t1084*t1169*t1359*t1674+TWO*t1343*t1497*t1849
*t55*t1663-TWO*t1343*t1086*t1920*t1933*t1925*t1240+t731+t824+t921+t1017+t347+
TWO*t1343*t1365*t1874*t55*t143*t1680*t1174*t1165-TWO*t1330*t1150*t1079*t1083*
t1059*t1933*t1097*t1240-TWO*t1981*t1870*t1674+t602;
    return(t1243/TWO+t1341/TWO+t1476/TWO+t1595/TWO+t1729/TWO+t1821/TWO+t1918/
TWO+t2010/TWO);
  }
}


std::complex<DOUBLE> C323x45x61mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  FourMomentum<std::complex<DOUBLE> > K1f1;
  FourMomentum<std::complex<DOUBLE> > K2f;
  FourMomentum<std::complex<DOUBLE> > K2f1;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> gg1;
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t1000;
  std::complex<DOUBLE> t1001;
  std::complex<DOUBLE> t1002;
  std::complex<DOUBLE> t1003;
  std::complex<DOUBLE> t1004;
  std::complex<DOUBLE> t1005;
  std::complex<DOUBLE> t1006;
  std::complex<DOUBLE> t1007;
  std::complex<DOUBLE> t1008;
  std::complex<DOUBLE> t1009;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t1010;
  std::complex<DOUBLE> t1012;
  std::complex<DOUBLE> t1013;
  std::complex<DOUBLE> t1014;
  std::complex<DOUBLE> t1015;
  std::complex<DOUBLE> t1019;
  std::complex<DOUBLE> t1023;
  std::complex<DOUBLE> t1024;
  std::complex<DOUBLE> t1025;
  std::complex<DOUBLE> t1026;
  std::complex<DOUBLE> t1027;
  std::complex<DOUBLE> t1028;
  std::complex<DOUBLE> t1030;
  std::complex<DOUBLE> t1031;
  std::complex<DOUBLE> t1034;
  std::complex<DOUBLE> t1035;
  std::complex<DOUBLE> t1038;
  std::complex<DOUBLE> t1040;
  std::complex<DOUBLE> t1041;
  std::complex<DOUBLE> t1042;
  std::complex<DOUBLE> t1043;
  std::complex<DOUBLE> t1044;
  std::complex<DOUBLE> t1045;
  std::complex<DOUBLE> t1046;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t1051;
  std::complex<DOUBLE> t1052;
  std::complex<DOUBLE> t1053;
  std::complex<DOUBLE> t1054;
  std::complex<DOUBLE> t1057;
  std::complex<DOUBLE> t1058;
  std::complex<DOUBLE> t1059;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t1064;
  std::complex<DOUBLE> t1065;
  std::complex<DOUBLE> t1066;
  std::complex<DOUBLE> t1067;
  std::complex<DOUBLE> t1068;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t1070;
  std::complex<DOUBLE> t1073;
  std::complex<DOUBLE> t1076;
  std::complex<DOUBLE> t1078;
  std::complex<DOUBLE> t1079;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t1085;
  std::complex<DOUBLE> t1086;
  std::complex<DOUBLE> t1087;
  std::complex<DOUBLE> t1088;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t1090;
  std::complex<DOUBLE> t1093;
  std::complex<DOUBLE> t1096;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t1100;
  std::complex<DOUBLE> t1101;
  std::complex<DOUBLE> t1103;
  std::complex<DOUBLE> t1105;
  std::complex<DOUBLE> t1106;
  std::complex<DOUBLE> t1107;
  std::complex<DOUBLE> t1110;
  std::complex<DOUBLE> t1113;
  std::complex<DOUBLE> t1114;
  std::complex<DOUBLE> t1116;
  std::complex<DOUBLE> t1118;
  std::complex<DOUBLE> t1119;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t1122;
  std::complex<DOUBLE> t1125;
  std::complex<DOUBLE> t1128;
  std::complex<DOUBLE> t1129;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t1130;
  std::complex<DOUBLE> t1133;
  std::complex<DOUBLE> t1135;
  std::complex<DOUBLE> t1138;
  std::complex<DOUBLE> t1139;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t1141;
  std::complex<DOUBLE> t1145;
  std::complex<DOUBLE> t1146;
  std::complex<DOUBLE> t1147;
  std::complex<DOUBLE> t1148;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t1150;
  std::complex<DOUBLE> t1151;
  std::complex<DOUBLE> t1156;
  std::complex<DOUBLE> t1160;
  std::complex<DOUBLE> t1161;
  std::complex<DOUBLE> t1162;
  std::complex<DOUBLE> t1168;
  std::complex<DOUBLE> t1170;
  std::complex<DOUBLE> t1171;
  std::complex<DOUBLE> t1174;
  std::complex<DOUBLE> t1175;
  std::complex<DOUBLE> t1176;
  std::complex<DOUBLE> t1177;
  std::complex<DOUBLE> t1178;
  std::complex<DOUBLE> t1179;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t1180;
  std::complex<DOUBLE> t1181;
  std::complex<DOUBLE> t1183;
  std::complex<DOUBLE> t1184;
  std::complex<DOUBLE> t1185;
  std::complex<DOUBLE> t1187;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t1191;
  std::complex<DOUBLE> t1195;
  std::complex<DOUBLE> t1198;
  std::complex<DOUBLE> t1199;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t1200;
  std::complex<DOUBLE> t1202;
  std::complex<DOUBLE> t1203;
  std::complex<DOUBLE> t1206;
  std::complex<DOUBLE> t1207;
  std::complex<DOUBLE> t1208;
  std::complex<DOUBLE> t1210;
  std::complex<DOUBLE> t1211;
  std::complex<DOUBLE> t1215;
  std::complex<DOUBLE> t1216;
  std::complex<DOUBLE> t1218;
  std::complex<DOUBLE> t1219;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t1220;
  std::complex<DOUBLE> t1221;
  std::complex<DOUBLE> t1222;
  std::complex<DOUBLE> t1226;
  std::complex<DOUBLE> t1228;
  std::complex<DOUBLE> t1230;
  std::complex<DOUBLE> t1232;
  std::complex<DOUBLE> t1235;
  std::complex<DOUBLE> t1237;
  std::complex<DOUBLE> t1238;
  std::complex<DOUBLE> t1239;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t1243;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t1250;
  std::complex<DOUBLE> t1252;
  std::complex<DOUBLE> t1255;
  std::complex<DOUBLE> t1258;
  std::complex<DOUBLE> t1259;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t1260;
  std::complex<DOUBLE> t1261;
  std::complex<DOUBLE> t1265;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t1272;
  std::complex<DOUBLE> t1274;
  std::complex<DOUBLE> t1276;
  std::complex<DOUBLE> t1279;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t1281;
  std::complex<DOUBLE> t1286;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t1291;
  std::complex<DOUBLE> t1292;
  std::complex<DOUBLE> t1295;
  std::complex<DOUBLE> t1298;
  std::complex<DOUBLE> t1299;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t1305;
  std::complex<DOUBLE> t1309;
  std::complex<DOUBLE> t1314;
  std::complex<DOUBLE> t1324;
  std::complex<DOUBLE> t1330;
  std::complex<DOUBLE> t1335;
  std::complex<DOUBLE> t1337;
  std::complex<DOUBLE> t1339;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t1351;
  std::complex<DOUBLE> t1353;
  std::complex<DOUBLE> t1357;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t1361;
  std::complex<DOUBLE> t1362;
  std::complex<DOUBLE> t1363;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t1371;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t1384;
  std::complex<DOUBLE> t1385;
  std::complex<DOUBLE> t1389;
  std::complex<DOUBLE> t1393;
  std::complex<DOUBLE> t1395;
  std::complex<DOUBLE> t1399;
  std::complex<DOUBLE> t1402;
  std::complex<DOUBLE> t1404;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t1415;
  std::complex<DOUBLE> t1417;
  std::complex<DOUBLE> t1418;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t1428;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t1432;
  std::complex<DOUBLE> t1434;
  std::complex<DOUBLE> t1440;
  std::complex<DOUBLE> t1442;
  std::complex<DOUBLE> t1448;
  std::complex<DOUBLE> t1449;
  std::complex<DOUBLE> t1453;
  std::complex<DOUBLE> t1459;
  std::complex<DOUBLE> t1460;
  std::complex<DOUBLE> t1466;
  std::complex<DOUBLE> t1472;
  std::complex<DOUBLE> t1475;
  std::complex<DOUBLE> t1477;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t1486;
  std::complex<DOUBLE> t1487;
  std::complex<DOUBLE> t1488;
  std::complex<DOUBLE> t1489;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t1490;
  std::complex<DOUBLE> t1492;
  std::complex<DOUBLE> t1496;
  std::complex<DOUBLE> t1498;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t1502;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t1511;
  std::complex<DOUBLE> t1512;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t1527;
  std::complex<DOUBLE> t1532;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t1541;
  std::complex<DOUBLE> t1550;
  std::complex<DOUBLE> t1560;
  std::complex<DOUBLE> t1564;
  std::complex<DOUBLE> t1566;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t1576;
  std::complex<DOUBLE> t1590;
  std::complex<DOUBLE> t1591;
  std::complex<DOUBLE> t1597;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t1608;
  std::complex<DOUBLE> t1615;
  std::complex<DOUBLE> t162;
  std::complex<DOUBLE> t1625;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t1637;
  std::complex<DOUBLE> t1638;
  std::complex<DOUBLE> t1643;
  std::complex<DOUBLE> t1647;
  std::complex<DOUBLE> t1662;
  std::complex<DOUBLE> t1663;
  std::complex<DOUBLE> t1664;
  std::complex<DOUBLE> t1665;
  std::complex<DOUBLE> t1681;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t1693;
  std::complex<DOUBLE> t1698;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t1703;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t1710;
  std::complex<DOUBLE> t1718;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t1725;
  std::complex<DOUBLE> t1734;
  std::complex<DOUBLE> t1735;
  std::complex<DOUBLE> t1736;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t1742;
  std::complex<DOUBLE> t1746;
  std::complex<DOUBLE> t1747;
  std::complex<DOUBLE> t1748;
  std::complex<DOUBLE> t1757;
  std::complex<DOUBLE> t1761;
  std::complex<DOUBLE> t1765;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t1773;
  std::complex<DOUBLE> t1778;
  std::complex<DOUBLE> t1783;
  std::complex<DOUBLE> t1784;
  std::complex<DOUBLE> t1790;
  std::complex<DOUBLE> t1797;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t1807;
  std::complex<DOUBLE> t1825;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t1847;
  std::complex<DOUBLE> t1848;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t1852;
  std::complex<DOUBLE> t1856;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  FourMomentum<DOUBLE> t2;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t285;
  std::complex<DOUBLE> t286;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t289;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t290;
  std::complex<DOUBLE> t293;
  std::complex<DOUBLE> t294;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t297;
  std::complex<DOUBLE> t298;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t302;
  std::complex<DOUBLE> t303;
  std::complex<DOUBLE> t305;
  std::complex<DOUBLE> t306;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t308;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t313;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t319;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t339;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t346;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t352;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t373;
  std::complex<DOUBLE> t378;
  std::complex<DOUBLE> t379;
  std::complex<DOUBLE> t382;
  std::complex<DOUBLE> t385;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t399;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t401;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t417;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t422;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t438;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t440;
  std::complex<DOUBLE> t444;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t449;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t450;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t462;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t464;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t466;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t474;
  std::complex<DOUBLE> t476;
  std::complex<DOUBLE> t479;
  std::complex<DOUBLE> t480;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t482;
  std::complex<DOUBLE> t487;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t492;
  std::complex<DOUBLE> t496;
  std::complex<DOUBLE> t498;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t502;
  std::complex<DOUBLE> t509;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t511;
  std::complex<DOUBLE> t513;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t523;
  std::complex<DOUBLE> t524;
  std::complex<DOUBLE> t525;
  std::complex<DOUBLE> t528;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t532;
  std::complex<DOUBLE> t539;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t561;
  std::complex<DOUBLE> t562;
  std::complex<DOUBLE> t566;
  std::complex<DOUBLE> t570;
  std::complex<DOUBLE> t572;
  std::complex<DOUBLE> t576;
  std::complex<DOUBLE> t579;
  std::complex<DOUBLE> t581;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t592;
  std::complex<DOUBLE> t594;
  std::complex<DOUBLE> t595;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t608;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t610;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t623;
  std::complex<DOUBLE> t626;
  std::complex<DOUBLE> t627;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t643;
  std::complex<DOUBLE> t646;
  std::complex<DOUBLE> t647;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t669;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t670;
  std::complex<DOUBLE> t671;
  std::complex<DOUBLE> t675;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t690;
  std::complex<DOUBLE> t697;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t701;
  std::complex<DOUBLE> t704;
  std::complex<DOUBLE> t706;
  std::complex<DOUBLE> t718;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t720;
  std::complex<DOUBLE> t724;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t733;
  std::complex<DOUBLE> t734;
  std::complex<DOUBLE> t749;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t760;
  std::complex<DOUBLE> t769;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t779;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t782;
  std::complex<DOUBLE> t785;
  std::complex<DOUBLE> t787;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t797;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t811;
  std::complex<DOUBLE> t812;
  std::complex<DOUBLE> t818;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t835;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t845;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t857;
  std::complex<DOUBLE> t858;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t861;
  std::complex<DOUBLE> t867;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t882;
  std::complex<DOUBLE> t883;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t899;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t914;
  std::complex<DOUBLE> t919;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t926;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t943;
  std::complex<DOUBLE> t946;
  std::complex<DOUBLE> t948;
  std::complex<DOUBLE> t949;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t952;
  std::complex<DOUBLE> t953;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t960;
  std::complex<DOUBLE> t961;
  std::complex<DOUBLE> t966;
  std::complex<DOUBLE> t969;
  std::complex<DOUBLE> t970;
  std::complex<DOUBLE> t971;
  std::complex<DOUBLE> t972;
  std::complex<DOUBLE> t973;
  std::complex<DOUBLE> t974;
  std::complex<DOUBLE> t975;
  std::complex<DOUBLE> t976;
  std::complex<DOUBLE> t979;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t980;
  std::complex<DOUBLE> t981;
  std::complex<DOUBLE> t982;
  std::complex<DOUBLE> t983;
  std::complex<DOUBLE> t984;
  std::complex<DOUBLE> t985;
  std::complex<DOUBLE> t986;
  std::complex<DOUBLE> t989;
  std::complex<DOUBLE> t992;
  std::complex<DOUBLE> t993;
  std::complex<DOUBLE> t996;
  std::complex<DOUBLE> t997;
  std::complex<DOUBLE> t998;
  std::complex<DOUBLE> t999;
  {
    t1 = p6+p1;
    t2 = p3+p2;
    t3 = dot(t1,t2);
    t4 = t3*t3;
    t5 = S(t2);
    t6 = S(t1);
    t7 = t5*t6;
    t9 = sqrt(t4-t7);
    gg1 = t3+t9;
    t10 = gg1*gg1;
    t12 = ONE/(t10-t7);
    t13 = t12*t10;
    t16 = A(p1,p6);
    t17 = t12*t16;
    t18 = B(p1,p6);
    t19 = t18*gg1;
    K1f1 = t13*p6+t13*p1+t17*t19*p3+t17*t19*p2;
    t26 = A(p2,p3);
    t27 = t12*t26;
    t28 = B(p2,p3);
    t29 = t28*gg1;
    K2f1 = t13*p3+t13*p2+t27*t29*p6+t27*t29*p1;
    t35 = t26*t18*t28;
    t36 = t17*t35;
    t39 = A(K1f1,p3);
    t40 = t39*t39;
    t41 = A(K1f1,p5);
    t42 = t41*t41;
    t43 = t40*t42;
    t44 = A(K2f1,p1);
    t45 = t44*t44;
    t46 = B(K1f1,p6);
    t49 = B(K2f1,p2);
    t50 = B(K2f1,p4);
    t51 = t49*t50;
    t52 = IA(K1f1,p2);
    t53 = IA(K1f1,p4);
    t54 = t52*t53;
    t55 = IA(K1f1,p6);
    t56 = t54*t55;
    t59 = IA(p1,p6);
    t60 = IA(p2,p3);
    t61 = t59*t60;
    t62 = IA(p4,p5);
    t63 = IB(p1,p6);
    t64 = t62*t63;
    t65 = t61*t64;
    t66 = IB(p2,p3);
    t67 = IB(p4,p5);
    t68 = t66*t67;
    t69 = IAB(K1f1,t2,K1f1);
    t72 = t68*t69*(-t36+t17*t19)*I;
    t73 = t65*t72;
    t76 = A(K1f1,p1);
    t77 = A(K1f1,p4);
    t78 = t76*t77;
    t79 = A(K2f1,p3);
    t80 = t79*t79;
    t81 = t44*t80;
    t82 = t78*t81;
    t83 = A(K2f1,p5);
    t84 = t83*t83;
    t85 = B(K1f1,p2);
    t86 = t84*t85;
    t87 = t46*t50;
    t88 = IA(K2f1,p2);
    t89 = t87*t88;
    t90 = t86*t89;
    t92 = IA(K2f1,p4);
    t93 = t92*t92;
    t94 = IA(K2f1,p6);
    t95 = t93*t94;
    t96 = t61*t62;
    t98 = t63*t66;
    t101 = t98*t67*t69*I;
    t105 = t81*t84;
    t106 = t78*t105;
    t107 = t85*t46;
    t108 = t50*t88;
    t109 = t108*t93;
    t110 = t107*t109;
    t112 = t94*t59;
    t113 = t60*t62;
    t114 = t113*t63;
    t115 = t112*t114;
    t118 = t68*t69*(t36-t27*t29)*I;
    t119 = t115*t118;
    t122 = t76*t41;
    t124 = t122*t81*t83;
    t125 = B(K1f1,p4);
    t126 = t85*t125;
    t127 = B(K2f1,p6);
    t128 = t127*t88;
    t129 = t128*t92;
    t130 = t126*t129;
    t135 = t80*t84;
    t136 = t77*t45*t135;
    t137 = t85*t50;
    t138 = t137*t128;
    t141 = t69*t69;
    t142 = t141*I;
    t143 = t68*t142;
    t148 = t39*t45*t79*t84;
    t149 = t46*t49;
    t150 = t108*t92;
    t151 = t149*t150;
    t152 = t148*t151;
    t154 = t112*t113*t101;
    t157 = t115*t72;
    t160 = t77*t77;
    t162 = t160*t45*t135;
    t163 = t93*t92;
    t169 = t76*t40;
    t170 = t42*t44;
    t171 = t169*t170;
    t172 = A(K2f1,p4);
    t174 = t51*t52;
    t177 = t53*t53;
    t180 = t177*t55*t96*t101;
    t184 = t169*t170*t172;
    t185 = t50*t52;
    t187 = t149*t185*t177;
    t189 = t55*t59;
    t190 = t189*t114;
    t191 = t190*t72;
    t194 = A(K2f1,p6);
    t197 = t185*t53;
    t198 = t149*t197;
    t200 = t55*t55;
    t202 = t200*t59*t114;
    t203 = t202*t72;
    t206 = t194*t46;
    t209 = t200*t60;
    t211 = t62*t66;
    t213 = t67*t141*I;
    t214 = t211*t213;
    t215 = t54*t209*t214;
    t218 = t107*t197;
    t220 = t190*t118;
    t223 = t125*t46;
    t224 = t49*t52;
    t226 = t223*t224*t53;
    t230 = t76*t76;
    t231 = t230*t39;
    t232 = t42*t79;
    t233 = t231*t232;
    t235 = t50*t127;
    t236 = t194*t85*t235;
    t241 = t231*t232*t194;
    t245 = t127*t52;
    t246 = t245*t53;
    t247 = t126*t246;
    t253 = t53*t55;
    t254 = t59*t62;
    t256 = t63*t67;
    t257 = t256*t142;
    t258 = t253*t254*t257;
    t261 = TWO*t43*t45*t46*t51*t56*t73-TWO*t82*t90*t95*t96*t101-TWO*t106*t110*
t119+FOUR*t124*t130*t119+t136*t138*t95*t113*t143-TWO*t152*t154+TWO*t152*t157+TWO
*t162*t107*t108*t163*t119+TWO*t171*t172*t46*t174*t180-TWO*t184*t187*t191-FOUR*
t169*t170*t194*t198*t203+FOUR*t171*t206*t51*t215+TWO*t171*t218*t220+TWO*t171*
t226*t220+TWO*t233*t236*t215-TWO*t241*t198*t203+TWO*t233*t247*t220+TWO*t233*
t126*t245*t258;
    t262 = A(K1f1,p6);
    t263 = t41*t262;
    t264 = t45*t80;
    t265 = t263*t264;
    t266 = t83*t85;
    t267 = t125*t127;
    t268 = t266*t267;
    t270 = t88*t92;
    t271 = t94*t94;
    t272 = t271*t59;
    t274 = t64*t213;
    t278 = t266*t89;
    t282 = t92*t271*t96*t101;
    t285 = t264*t83;
    t286 = t263*t285;
    t287 = t107*t150;
    t289 = t272*t114;
    t290 = t289*t118;
    t293 = t230*t40;
    t294 = t41*t83;
    t295 = t293*t294;
    t297 = t49*t127;
    t298 = t194*t125*t297;
    t302 = A(K2f1,p2);
    t303 = t42*t302;
    t305 = t293*t303*t172;
    t306 = t125*t49;
    t307 = t52*t52;
    t308 = t127*t307;
    t309 = t308*t177;
    t313 = t293*t303;
    t315 = t50*t307*t53;
    t317 = t313*t107*t315;
    t319 = t189*t113*t101;
    t322 = t42*t172;
    t324 = t293*t322*t194;
    t325 = t245*t177;
    t326 = t306*t325;
    t330 = t149*t185;
    t337 = t84*t125;
    t339 = t270*t94;
    t342 = t65*t118;
    t345 = t39*t77;
    t346 = t45*t79;
    t347 = t346*t84;
    t348 = t345*t347;
    t352 = t49*t88;
    t359 = t337*t149;
    t361 = t88*t93;
    t363 = t361*t112*t274;
    t366 = t39*t41;
    t368 = t366*t346*t83;
    t373 = t223*t352*t92;
    t378 = t83*t125;
    t379 = t378*t149;
    t382 = t270*t112*t274;
    t385 = t39*t262;
    t386 = t385*t347;
    t392 = t337*t149*t88;
    t399 = -TWO*t265*t268*t270*t272*t274-TWO*t265*t278*t282-TWO*t286*t287*t290+
TWO*t295*t298*t215+t305*t306*t309*t191-t317*t319-t317*t220+t324*t326*t203-TWO*
t43*t45*t330*t253*t113*t143+TWO*t40*t45*t337*t149*t339*t342-TWO*t348*t110*t119
-TWO*t348*t223*t352*t93*t119-TWO*t345*t346*t359*t363+FOUR*t368*t287*t119+FOUR*
t368*t373*t119+FOUR*t366*t346*t379*t382-TWO*t386*t130*t290-TWO*t385*t346*t392*
t282-TWO*t386*t373*t290;
    t401 = t302*t302;
    t411 = t172*t172;
    t417 = t293*t42*t411*t306*t245*t177*t53;
    t422 = t194*t194;
    t424 = t293*t42*t422;
    t426 = t200*t55;
    t438 = t41*t302;
    t440 = t235*t307;
    t444 = t253*t96*t101;
    t448 = t293*t438*t83;
    t449 = t308*t53;
    t450 = t137*t449;
    t455 = t230*t80*t86;
    t460 = A(K1f1,p2);
    t461 = t76*t460;
    t462 = t461*t105;
    t463 = t88*t88;
    t464 = t127*t463;
    t465 = t464*t92;
    t466 = t126*t465;
    t472 = t337*t149*t463;
    t474 = t92*t94;
    t476 = t474*t96*t101;
    t479 = t460*t77;
    t480 = t264*t84;
    t481 = t479*t480;
    t482 = t50*t463;
    t487 = t49*t463;
    t492 = t460*t262;
    t496 = t492*t480;
    t498 = t223*t487*t92;
    t502 = t460*t45*t135;
    t509 = t474*t113*t143;
    t511 = t482*t92;
    t513 = t502*t149*t511;
    t515 = t306*t129;
    t523 = TWO*t293*t42*t401*t137*t127*t307*t52*t53*t191-TWO*t417*t319+TWO*t417
*t191+TWO*t424*t198*t426*t59*t114*t72-TWO*t424*t330*t53*t426*t113*t143+TWO*t293
*t438*t266*t440*t444-TWO*t448*t450*t191+TWO*t455*t267*t339*t342-TWO*t462*t466*
t119-TWO*t461*t81*t472*t476+t481*t107*t482*t93*t119+t481*t223*t487*t93*t119+
t492*t264*t472*t282+t496*t498*t290-t502*t306*t465*t157+t502*t306*t464*t509+t513
*t154+TWO*t148*t515*t157-TWO*t148*t306*t128*t509;
    t524 = t77*t262;
    t525 = t524*t480;
    t528 = t128*t93;
    t532 = t107*t108;
    t539 = t77*t41;
    t548 = t41*t172;
    t561 = t293*t294*t194;
    t562 = t306*t246;
    t566 = t223*t224;
    t570 = t41*t44;
    t572 = t169*t570*t83;
    t576 = t169*t570;
    t579 = t55*t60;
    t581 = t54*t579*t214;
    t592 = t76*t39;
    t594 = t592*t170*t79;
    t595 = t137*t246;
    t608 = t44*t79;
    t610 = t592*t608*t84;
    t623 = t525*t110*t290-t136*t137*t528*t157+TWO*t162*t532*t163*t94*t254*t257
-FOUR*t539*t285*t110*t119-TWO*t448*t306*t449*t191+FOUR*t293*t548*t378*t297*t52*
t180-FOUR*t293*t548*t83*t326*t191-TWO*t561*t562*t203+TWO*t171*t566*t258+FOUR*t572
*t562*t191-FOUR*t576*t378*t297*t581-FOUR*t576*t83*t46*t174*t444+FOUR*t572*t198*
t191+FOUR*t594*t595*t191-FOUR*t592*t170*t79*t85*t235*t581+FOUR*t594*t198*t191+FOUR*
t610*t130*t119+FOUR*t592*t608*t392*t476-FOUR*t539*t264*t266*t87*t363;
    t626 = t42*t45;
    t627 = t80*t85;
    t643 = t126*t528;
    t646 = t524*t264;
    t647 = t86*t267;
    t669 = t262*t45*t135;
    t670 = t137*t129;
    t671 = t669*t670;
    t675 = t289*t72;
    t690 = t41*t45*t80*t83;
    t697 = TWO*t626*t627*t87*t339*t342+TWO*t626*t80*t532*t474*t254*t257-TWO*
t286*t130*t290-t513*t157+t525*t643*t290+t646*t647*t361*t272*t274+t479*t264*t359
*t463*t93*t112*t274+t496*t466*t290+t646*t90*t93*t271*t96*t101-t136*t149*t109*
t157+t671*t272*t113*t101-t671*t675-t669*t515*t675+FOUR*t610*t373*t119-TWO*t106*
t643*t119-TWO*t82*t647*t363+TWO*t690*t670*t157-TWO*t690*t138*t509;
    t701 = t262*t262;
    t704 = t701*t45*t135*t130;
    t706 = t271*t94*t59;
    t718 = t293*t322;
    t720 = t52*t177;
    t724 = t206*t174;
    t733 = t169*t170*t302;
    t734 = t149*t315;
    t749 = t53*t200;
    t760 = t295*t226;
    t769 = t307*t53;
    t779 = t718*t223*t224*t177;
    t782 = TWO*t690*t151*t157+TWO*t704*t706*t113*t101+TWO*t704*t706*t114*t118
-TWO*t462*t498*t119-t718*t298*t720*t209*t214-t718*t724*t177*t200*t96*t101+t324*
t187*t203-TWO*t733*t734*t191-TWO*t184*t326*t191+TWO*t171*t172*t125*t297*t720*
t579*t214+TWO*t295*t724*t749*t96*t101-TWO*t561*t198*t203+TWO*t295*t247*t220+TWO
*t760*t319-TWO*t733*t450*t191+TWO*t171*t302*t85*t235*t769*t579*t214-t718*t126*
t325*t220-t779*t319-t779*t220;
    t785 = t293*t42*t194;
    t787 = t202*t118;
    t797 = t293*t337;
    t811 = t172*t85;
    t812 = t235*t52;
    t818 = t231*t232*t172;
    t835 = t81*t122;
    t845 = t76*t262;
    t857 = t76*t44*t135;
    t858 = t857*t670;
    t861 = -t785*t218*t787-t785*t226*t787-t785*t566*t749*t254*t257+TWO*t760*
t220-TWO*t797*t297*t54*t319+TWO*t797*t297*t56*t73-FOUR*t231*t303*t79*t450*t191+
TWO*t233*t811*t812*t180-TWO*t818*t137*t325*t191-TWO*t818*t326*t191-TWO*t241*
t595*t203-t313*t811*t440*t180+t305*t137*t309*t191+FOUR*t835*t268*t382+FOUR*t835*
t278*t476+FOUR*t124*t287*t119-FOUR*t845*t81*t86*t267*t88*t282-FOUR*t845*t105*t130*
t290-TWO*t858*t154;
    t867 = t460*t460;
    t882 = t460*t41;
    t883 = t882*t285;
    t899 = t293*t303*t194;
    t914 = t233*t218;
    t919 = t41*t79;
    t926 = t231*t919*t83;
    t943 = TWO*t858*t157+TWO*t857*t515*t157+TWO*t867*t45*t135*t223*t49*t463*t88
*t92*t119-FOUR*t460*t39*t347*t498*t119-TWO*t883*t107*t511*t119-TWO*t883*t498*
t119-TWO*t882*t264*t379*t463*t92*t112*t274+t899*t450*t203-t313*t236*t769*t209*
t214+t899*t734*t203-t313*t126*t449*t220-t313*t126*t308*t258+TWO*t914*t319+TWO*
t914*t220-FOUR*t231*t919*t266*t812*t444+FOUR*t926*t595*t191+FOUR*t926*t562*t191+
TWO*t230*t42*t627*t235*t56*t73+TWO*t455*t267*t270*t154;
    gg = t3-t9;
    t946 = gg*gg;
    t948 = ONE/(t946-t7);
    t949 = t948*t946;
    t952 = t948*t16;
    t953 = t18*gg;
    K1f = t949*p6+t949*p1+t952*t953*p3+t952*t953*p2;
    t960 = t948*t26;
    t961 = t28*gg;
    K2f = t949*p3+t949*p2+t960*t961*p6+t960*t961*p1;
    t966 = t952*t35;
    t969 = A(K1f,p3);
    t970 = t969*t969;
    t971 = A(K1f,p5);
    t972 = t971*t971;
    t973 = t970*t972;
    t974 = A(K2f,p1);
    t975 = t974*t974;
    t976 = B(K1f,p6);
    t979 = B(K2f,p2);
    t980 = B(K2f,p4);
    t981 = t979*t980;
    t982 = IA(K1f,p2);
    t983 = IA(K1f,p4);
    t984 = t982*t983;
    t985 = IA(K1f,p6);
    t986 = t984*t985;
    t989 = IAB(K1f,t2,K1f);
    t992 = t68*t989*(-t966+t952*t953)*I;
    t993 = t65*t992;
    t996 = A(K1f,p1);
    t997 = A(K1f,p4);
    t998 = t996*t997;
    t999 = A(K2f,p3);
    t1000 = t999*t999;
    t1001 = t974*t1000;
    t1002 = t998*t1001;
    t1003 = A(K2f,p5);
    t1004 = t1003*t1003;
    t1005 = B(K1f,p2);
    t1006 = t1004*t1005;
    t1007 = t976*t980;
    t1008 = IA(K2f,p2);
    t1009 = t1007*t1008;
    t1010 = t1006*t1009;
    t1012 = IA(K2f,p4);
    t1013 = t1012*t1012;
    t1014 = IA(K2f,p6);
    t1015 = t1013*t1014;
    t1019 = t98*t67*t989*I;
    t1023 = t1001*t1004;
    t1024 = t998*t1023;
    t1025 = t1005*t976;
    t1026 = t980*t1008;
    t1027 = t1026*t1013;
    t1028 = t1025*t1027;
    t1030 = t1014*t59;
    t1031 = t1030*t114;
    t1034 = t68*t989*(t966-t960*t961)*I;
    t1035 = t1031*t1034;
    t1038 = t996*t971;
    t1040 = t1038*t1001*t1003;
    t1041 = B(K1f,p4);
    t1042 = t1005*t1041;
    t1043 = B(K2f,p6);
    t1044 = t1043*t1008;
    t1045 = t1044*t1012;
    t1046 = t1042*t1045;
    t1051 = t1000*t1004;
    t1052 = t997*t975*t1051;
    t1053 = t1005*t980;
    t1054 = t1053*t1044;
    t1057 = t989*t989;
    t1058 = t1057*I;
    t1059 = t68*t1058;
    t1064 = t969*t975*t999*t1004;
    t1065 = t976*t979;
    t1066 = t1026*t1012;
    t1067 = t1065*t1066;
    t1068 = t1064*t1067;
    t1070 = t1030*t113*t1019;
    t1073 = t1031*t992;
    t1076 = t997*t997;
    t1078 = t1076*t975*t1051;
    t1079 = t1013*t1012;
    t1085 = t996*t970;
    t1086 = t972*t974;
    t1087 = t1085*t1086;
    t1088 = A(K2f,p4);
    t1090 = t981*t982;
    t1093 = t983*t983;
    t1096 = t1093*t985*t96*t1019;
    t1100 = t1085*t1086*t1088;
    t1101 = t980*t982;
    t1103 = t1065*t1101*t1093;
    t1105 = t985*t59;
    t1106 = t1105*t114;
    t1107 = t1106*t992;
    t1110 = A(K2f,p6);
    t1113 = t1101*t983;
    t1114 = t1065*t1113;
    t1116 = t985*t985;
    t1118 = t1116*t59*t114;
    t1119 = t1118*t992;
    t1122 = t1110*t976;
    t1125 = t1116*t60;
    t1128 = t67*t1057*I;
    t1129 = t211*t1128;
    t1130 = t984*t1125*t1129;
    t1133 = t1025*t1113;
    t1135 = t1106*t1034;
    t1138 = t1041*t976;
    t1139 = t979*t982;
    t1141 = t1138*t1139*t983;
    t1145 = t996*t996;
    t1146 = t1145*t969;
    t1147 = t972*t999;
    t1148 = t1146*t1147;
    t1150 = t980*t1043;
    t1151 = t1110*t1005*t1150;
    t1156 = t1146*t1147*t1110;
    t1160 = t1043*t982;
    t1161 = t1160*t983;
    t1162 = t1042*t1161;
    t1168 = t983*t985;
    t1170 = t256*t1058;
    t1171 = t1168*t254*t1170;
    t1174 = TWO*t973*t975*t976*t981*t986*t993-TWO*t1002*t1010*t1015*t96*t1019
-TWO*t1024*t1028*t1035+FOUR*t1040*t1046*t1035+t1052*t1054*t1015*t113*t1059-TWO*
t1068*t1070+TWO*t1068*t1073+TWO*t1078*t1025*t1026*t1079*t1035+TWO*t1087*t1088*
t976*t1090*t1096-TWO*t1100*t1103*t1107-FOUR*t1085*t1086*t1110*t1114*t1119+FOUR*
t1087*t1122*t981*t1130+TWO*t1087*t1133*t1135+TWO*t1087*t1141*t1135+TWO*t1148*
t1151*t1130-TWO*t1156*t1114*t1119+TWO*t1148*t1162*t1135+TWO*t1148*t1042*t1160*
t1171;
    t1175 = A(K1f,p6);
    t1176 = t971*t1175;
    t1177 = t975*t1000;
    t1178 = t1176*t1177;
    t1179 = t1003*t1005;
    t1180 = t1041*t1043;
    t1181 = t1179*t1180;
    t1183 = t1008*t1012;
    t1184 = t1014*t1014;
    t1185 = t1184*t59;
    t1187 = t64*t1128;
    t1191 = t1179*t1009;
    t1195 = t1012*t1184*t96*t1019;
    t1198 = t1177*t1003;
    t1199 = t1176*t1198;
    t1200 = t1025*t1066;
    t1202 = t1185*t114;
    t1203 = t1202*t1034;
    t1206 = t1145*t970;
    t1207 = t971*t1003;
    t1208 = t1206*t1207;
    t1210 = t979*t1043;
    t1211 = t1110*t1041*t1210;
    t1215 = A(K2f,p2);
    t1216 = t972*t1215;
    t1218 = t1206*t1216*t1088;
    t1219 = t1041*t979;
    t1220 = t982*t982;
    t1221 = t1043*t1220;
    t1222 = t1221*t1093;
    t1226 = t1206*t1216;
    t1228 = t980*t1220*t983;
    t1230 = t1226*t1025*t1228;
    t1232 = t1105*t113*t1019;
    t1235 = t972*t1088;
    t1237 = t1206*t1235*t1110;
    t1238 = t1160*t1093;
    t1239 = t1219*t1238;
    t1243 = t1065*t1101;
    t1250 = t1004*t1041;
    t1252 = t1183*t1014;
    t1255 = t65*t1034;
    t1258 = t969*t997;
    t1259 = t975*t999;
    t1260 = t1259*t1004;
    t1261 = t1258*t1260;
    t1265 = t979*t1008;
    t1272 = t1250*t1065;
    t1274 = t1008*t1013;
    t1276 = t1274*t1030*t1187;
    t1279 = t969*t971;
    t1281 = t1279*t1259*t1003;
    t1286 = t1138*t1265*t1012;
    t1291 = t1003*t1041;
    t1292 = t1291*t1065;
    t1295 = t1183*t1030*t1187;
    t1298 = t969*t1175;
    t1299 = t1298*t1260;
    t1305 = t1250*t1065*t1008;
    t1309 = t1218*t1219*t1222*t1107-TWO*t1258*t1259*t1272*t1276+FOUR*t1279*t1259
*t1292*t1295-TWO*t1298*t1259*t1305*t1195-TWO*t1178*t1181*t1183*t1185*t1187-TWO*
t1261*t1138*t1265*t1013*t1035-TWO*t973*t975*t1243*t1168*t113*t1059+TWO*t970*
t975*t1250*t1065*t1252*t1255-t1230*t1232-t1230*t1135-TWO*t1199*t1200*t1203+FOUR*
t1281*t1200*t1035-TWO*t1261*t1028*t1035-TWO*t1178*t1191*t1195+TWO*t1208*t1211*
t1130+t1237*t1239*t1119+FOUR*t1281*t1286*t1035-TWO*t1299*t1046*t1203+t623+t697+
t782+t861+t943+t261+t399+t523;
    t1314 = t1215*t1215;
    t1324 = t1088*t1088;
    t1330 = t1206*t972*t1324*t1219*t1160*t1093*t983;
    t1335 = t1110*t1110;
    t1337 = t1206*t972*t1335;
    t1339 = t1116*t985;
    t1351 = t971*t1215;
    t1353 = t1150*t1220;
    t1357 = t1168*t96*t1019;
    t1361 = t1206*t1351*t1003;
    t1362 = t1221*t983;
    t1363 = t1053*t1362;
    t1371 = t971*t1088;
    t1384 = t1206*t1207*t1110;
    t1385 = t1219*t1161;
    t1389 = t1138*t1139;
    t1393 = t971*t974;
    t1395 = t1085*t1393*t1003;
    t1399 = t1085*t1393;
    t1402 = t985*t60;
    t1404 = t984*t1402*t1129;
    t1415 = t996*t969;
    t1417 = t1415*t1086*t999;
    t1418 = t1053*t1161;
    t1428 = -TWO*t1299*t1286*t1203+TWO*t1206*t972*t1314*t1053*t1043*t1220*t982*
t983*t1107-TWO*t1330*t1232+TWO*t1330*t1107+TWO*t1337*t1114*t1339*t59*t114*t992
-TWO*t1337*t1243*t983*t1339*t113*t1059+TWO*t1206*t1351*t1179*t1353*t1357-TWO*
t1361*t1363*t1107-TWO*t1361*t1219*t1362*t1107+FOUR*t1206*t1371*t1291*t1210*t982*
t1096-FOUR*t1206*t1371*t1003*t1239*t1107-TWO*t1384*t1385*t1119+TWO*t1087*t1389*
t1171+FOUR*t1395*t1385*t1107-FOUR*t1399*t1291*t1210*t1404-FOUR*t1399*t1003*t976*
t1090*t1357+FOUR*t1395*t1114*t1107+FOUR*t1417*t1418*t1107-FOUR*t1415*t1086*t999*
t1005*t1150*t1404;
    t1432 = t974*t999;
    t1434 = t1415*t1432*t1004;
    t1440 = t1012*t1014;
    t1442 = t1440*t96*t1019;
    t1448 = t1044*t1013;
    t1449 = t1042*t1448;
    t1453 = t1006*t1180;
    t1459 = t971*t975*t1000*t1003;
    t1460 = t1053*t1045;
    t1466 = t1440*t113*t1059;
    t1472 = t1175*t1175;
    t1475 = t1472*t975*t1051*t1046;
    t1477 = t1184*t1014*t59;
    t1486 = A(K1f,p2);
    t1487 = t996*t1486;
    t1488 = t1487*t1023;
    t1489 = t1008*t1008;
    t1490 = t979*t1489;
    t1492 = t1138*t1490*t1012;
    t1496 = t1206*t1235;
    t1498 = t982*t1093;
    t1502 = t1122*t1090;
    t1511 = t1085*t1086*t1215;
    t1512 = t1065*t1228;
    t1527 = t983*t1116;
    t1532 = FOUR*t1417*t1114*t1107+FOUR*t1434*t1046*t1035+FOUR*t1415*t1432*t1305*
t1442+FOUR*t1434*t1286*t1035-TWO*t1024*t1449*t1035-TWO*t1002*t1453*t1276+TWO*
t1459*t1460*t1073-TWO*t1459*t1054*t1466+TWO*t1459*t1067*t1073+TWO*t1475*t1477*
t113*t1019+TWO*t1475*t1477*t114*t1034-TWO*t1488*t1492*t1035-t1496*t1211*t1498*
t1125*t1129-t1496*t1502*t1093*t1116*t96*t1019+t1237*t1103*t1119-TWO*t1511*t1512
*t1107-TWO*t1100*t1239*t1107+TWO*t1087*t1088*t1041*t1210*t1498*t1402*t1129+TWO*
t1208*t1502*t1527*t96*t1019;
    t1541 = t1208*t1141;
    t1550 = t1220*t983;
    t1560 = t1496*t1138*t1139*t1093;
    t1564 = t1206*t972*t1110;
    t1566 = t1118*t1034;
    t1576 = t1206*t1250;
    t1590 = t1088*t1005;
    t1591 = t1150*t982;
    t1597 = t1146*t1147*t1088;
    t1608 = -TWO*t1384*t1114*t1119+TWO*t1208*t1162*t1135+TWO*t1541*t1232-TWO*
t1511*t1363*t1107+TWO*t1087*t1215*t1005*t1150*t1550*t1402*t1129-t1496*t1042*
t1238*t1135-t1560*t1232-t1560*t1135-t1564*t1133*t1566-t1564*t1141*t1566-t1564*
t1389*t1527*t254*t1170+TWO*t1541*t1135-TWO*t1576*t1210*t984*t1232+TWO*t1576*
t1210*t986*t993-FOUR*t1146*t1216*t999*t1363*t1107+TWO*t1148*t1590*t1591*t1096
-TWO*t1597*t1053*t1238*t1107-TWO*t1597*t1239*t1107-TWO*t1156*t1418*t1119;
    t1615 = t1038*t1001;
    t1625 = t996*t1175;
    t1637 = t996*t974*t1051;
    t1638 = t1637*t1460;
    t1643 = t1219*t1045;
    t1647 = t1486*t1486;
    t1662 = t1486*t971;
    t1663 = t1662*t1198;
    t1664 = t980*t1489;
    t1665 = t1664*t1012;
    t1681 = t1206*t1216*t1110;
    t1693 = -t1226*t1590*t1353*t1096+t1218*t1053*t1222*t1107+FOUR*t1615*t1181*
t1295+FOUR*t1615*t1191*t1442+FOUR*t1040*t1200*t1035-FOUR*t1625*t1001*t1006*t1180*
t1008*t1195-FOUR*t1625*t1023*t1046*t1203-TWO*t1638*t1070+TWO*t1638*t1073+TWO*
t1637*t1643*t1073+TWO*t1647*t975*t1051*t1138*t979*t1489*t1008*t1012*t1035-FOUR*
t1486*t969*t1260*t1492*t1035-TWO*t1663*t1025*t1665*t1035-TWO*t1663*t1492*t1035
-TWO*t1662*t1177*t1292*t1489*t1012*t1030*t1187+t1681*t1363*t1119-t1226*t1151*
t1550*t1125*t1129+t1681*t1512*t1119-t1226*t1042*t1362*t1135;
    t1698 = t1148*t1133;
    t1703 = t971*t999;
    t1710 = t1146*t1703*t1003;
    t1718 = t1000*t1005;
    t1725 = t1145*t1000*t1006;
    t1734 = t1043*t1489;
    t1735 = t1734*t1012;
    t1736 = t1042*t1735;
    t1742 = t1250*t1065*t1489;
    t1746 = t1486*t997;
    t1747 = t1177*t1004;
    t1748 = t1746*t1747;
    t1757 = t1486*t1175;
    t1761 = t1757*t1747;
    t1765 = t1486*t975*t1051;
    t1773 = t1765*t1065*t1665;
    t1778 = -t1226*t1042*t1221*t1171+TWO*t1698*t1232+TWO*t1698*t1135-FOUR*t1146*
t1703*t1179*t1591*t1357+FOUR*t1710*t1418*t1107+FOUR*t1710*t1385*t1107+TWO*t1145*
t972*t1718*t1150*t986*t993+TWO*t1725*t1180*t1183*t1070+TWO*t1725*t1180*t1252*
t1255-TWO*t1488*t1736*t1035-TWO*t1487*t1001*t1742*t1442+t1748*t1025*t1664*t1013
*t1035+t1748*t1138*t1490*t1013*t1035+t1757*t1177*t1742*t1195+t1761*t1492*t1203-
t1765*t1219*t1735*t1073+t1765*t1219*t1734*t1466+t1773*t1070+TWO*t1064*t1643*
t1073;
    t1783 = t997*t1175;
    t1784 = t1783*t1747;
    t1790 = t1025*t1026;
    t1797 = t997*t971;
    t1807 = t972*t975;
    t1825 = t1783*t1177;
    t1847 = t1175*t975*t1051;
    t1848 = t1847*t1460;
    t1852 = t1202*t992;
    t1856 = -TWO*t1064*t1219*t1044*t1466+t1784*t1028*t1203-t1052*t1053*t1448*
t1073+TWO*t1078*t1790*t1079*t1014*t254*t1170-FOUR*t1797*t1198*t1028*t1035-FOUR*
t1797*t1177*t1179*t1007*t1276+TWO*t1807*t1718*t1007*t1252*t1255+TWO*t1807*t1000
*t1790*t1440*t254*t1170-TWO*t1199*t1046*t1203-t1773*t1073+t1784*t1449*t1203+
t1825*t1453*t1274*t1185*t1187+t1746*t1177*t1272*t1489*t1013*t1030*t1187+t1761*
t1736*t1203+t1825*t1010*t1013*t1184*t96*t1019-t1052*t1065*t1027*t1073+t1848*
t1185*t113*t1019-t1848*t1852-t1847*t1643*t1852;
    return(t1174/TWO+t1309/TWO+t1428/TWO+t1532/TWO+t1608/TWO+t1693/TWO+t1778/
TWO+t1856/TWO);
  }
}

std::complex<DOUBLE> C212mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  {
    return(ZERO);
  }
}


std::complex<DOUBLE> C223mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t1002;
  std::complex<DOUBLE> t1006;
  std::complex<DOUBLE> t1013;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t1022;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t1071;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t249;
  FourMomentum<DOUBLE> t25;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t255;
  FourMomentum<DOUBLE> t256;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t276;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t281;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t286;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t303;
  std::complex<DOUBLE> t308;
  FourMomentum<DOUBLE> t31;
  std::complex<DOUBLE> t310;
  std::complex<DOUBLE> t311;
  std::complex<DOUBLE> t314;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t320;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t344;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t351;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t358;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t360;
  std::complex<DOUBLE> t362;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t369;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t389;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t397;
  std::complex<DOUBLE> t399;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t404;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t420;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t436;
  std::complex<DOUBLE> t437;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t443;
  std::complex<DOUBLE> t444;
  std::complex<DOUBLE> t445;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t449;
  std::complex<DOUBLE> t450;
  std::complex<DOUBLE> t453;
  std::complex<DOUBLE> t454;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t463;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t470;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t476;
  std::complex<DOUBLE> t480;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t493;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t500;
  std::complex<DOUBLE> t507;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t526;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t533;
  std::complex<DOUBLE> t541;
  std::complex<DOUBLE> t543;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t553;
  std::complex<DOUBLE> t559;
  std::complex<DOUBLE> t560;
  std::complex<DOUBLE> t561;
  std::complex<DOUBLE> t562;
  std::complex<DOUBLE> t563;
  std::complex<DOUBLE> t569;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t575;
  std::complex<DOUBLE> t578;
  std::complex<DOUBLE> t582;
  std::complex<DOUBLE> t586;
  std::complex<DOUBLE> t589;
  std::complex<DOUBLE> t592;
  std::complex<DOUBLE> t593;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t612;
  std::complex<DOUBLE> t619;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t630;
  std::complex<DOUBLE> t632;
  std::complex<DOUBLE> t638;
  std::complex<DOUBLE> t642;
  std::complex<DOUBLE> t648;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t650;
  std::complex<DOUBLE> t651;
  std::complex<DOUBLE> t654;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t661;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t672;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t682;
  std::complex<DOUBLE> t686;
  std::complex<DOUBLE> t688;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t694;
  std::complex<DOUBLE> t697;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t701;
  std::complex<DOUBLE> t708;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t713;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t729;
  std::complex<DOUBLE> t731;
  std::complex<DOUBLE> t734;
  std::complex<DOUBLE> t736;
  std::complex<DOUBLE> t737;
  std::complex<DOUBLE> t748;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t753;
  std::complex<DOUBLE> t754;
  std::complex<DOUBLE> t755;
  std::complex<DOUBLE> t756;
  std::complex<DOUBLE> t757;
  std::complex<DOUBLE> t758;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t765;
  std::complex<DOUBLE> t766;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t781;
  std::complex<DOUBLE> t786;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t790;
  std::complex<DOUBLE> t797;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t804;
  std::complex<DOUBLE> t808;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t832;
  std::complex<DOUBLE> t837;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t842;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t855;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t870;
  std::complex<DOUBLE> t871;
  std::complex<DOUBLE> t883;
  std::complex<DOUBLE> t893;
  std::complex<DOUBLE> t898;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t912;
  std::complex<DOUBLE> t927;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t933;
  std::complex<DOUBLE> t938;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t948;
  std::complex<DOUBLE> t949;
  std::complex<DOUBLE> t963;
  std::complex<DOUBLE> t964;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t982;
  {
    t1 = A(p2,p5);
    t2 = t1*t1;
    t3 = t2*t2;
    t4 = A(p3,p4);
    t5 = t3*t4;
    t6 = B(p1,p2);
    t7 = t5*t6;
    t8 = B(p3,p6);
    t9 = t8*t8;
    t10 = t9*t9;
    t11 = IA(p2,p3);
    t12 = t10*t11;
    t13 = IA(p2,p4);
    t14 = t13*t13;
    t15 = IA(p4,p5);
    t16 = t14*t15;
    t17 = t12*t16;
    t18 = t7*t17;
    t19 = IB(p1,p3);
    t20 = t19*t19;
    t21 = IB(p1,p6);
    t22 = t20*t21;
    t23 = IB(p2,p3);
    t24 = t22*t23;
    t25 = p5+p4;
    t26 = AB(p3,t25,p6);
    t27 = IAB(p2,t25,p3);
    t29 = IAB(p2,t25,p6);
    t30 = t29*t29;
    t31 = p6+p1;
    t32 = IAB(p5,t31,p3);
    t33 = t30*t32;
    t34 = t26*t27*t33;
    t35 = t24*t34;
    t38 = AB(p5,t31,p2);
    t39 = t38*t27;
    t40 = t32*t32;
    t41 = t29*t40;
    t42 = t39*t41;
    t43 = t24*t42;
    t46 = B(p2,p4);
    t47 = t46*t46;
    t50 = t19*t21;
    t51 = t50*t23;
    t52 = IB(p3,p4);
    t53 = t52*t52;
    t55 = t29*t32;
    t57 = t51*t53*t27*t55;
    t60 = t3*t6;
    t61 = t60*t10;
    t62 = t11*t13;
    t63 = t15*t20;
    t65 = t61*t62*t63;
    t66 = t21*t23;
    t67 = AB(p2,t25,p2);
    t68 = t66*t67;
    t69 = t27*t27;
    t70 = t38*t69;
    t71 = t70*t41;
    t72 = t68*t71;
    t75 = t62*t15;
    t76 = t61*t75;
    t77 = AB(p3,t25,p2);
    t79 = t77*t69*t55;
    t83 = AB(p3,t25,p3);
    t84 = t83*t83;
    t85 = t69*t27;
    t87 = t84*t85*t55;
    t91 = t66*t83;
    t93 = t26*t69*t33;
    t94 = t91*t93;
    t97 = t91*t71;
    t100 = t26*t26;
    t102 = t30*t29;
    t103 = t102*t32;
    t104 = t100*t27*t103;
    t108 = t66*t26;
    t109 = t30*t40;
    t110 = t39*t109;
    t111 = t108*t110;
    t114 = t38*t38;
    t115 = t114*t27;
    t116 = t40*t32;
    t117 = t29*t116;
    t118 = t115*t117;
    t124 = t15*t19;
    t125 = t62*t124;
    t131 = TWO/THREE*t18*t35-TWO/THREE*t18*t43-TWO*t5*t47*t17*t57+TWO/THREE*t65*t72
-FOUR/THREE*t76*t24*t79+TWO/THREE*t76*t24*t87+TWO/THREE*t65*t94-TWO/THREE*t65*t97+TWO/
THREE*t76*t24*t104-TWO/THREE*t65*t111+TWO/THREE*t76*t24*t118-TWO*t3*t47*t10*t125*t66*
t53*t34;
    t132 = t3*t46;
    t133 = B(p2,p6);
    t135 = t9*t8;
    t136 = t135*t11;
    t137 = t13*t15;
    t138 = t136*t137;
    t140 = t23*t52;
    t141 = t50*t140;
    t142 = t141*t34;
    t146 = t132*t10*t125;
    t157 = t66*t52;
    t166 = t133*t133;
    t171 = t27*t29;
    t172 = t171*t32;
    t173 = t51*t172;
    t176 = t5*t10;
    t177 = t11*t14;
    t179 = t176*t177*t124;
    t183 = t176*t177*t15;
    t184 = t51*t104;
    t189 = t51*t118;
    t192 = t6*t6;
    t197 = t20*t20;
    t203 = t3*t192;
    t205 = t136*t13;
    t207 = t20*t19;
    t211 = t23*t27*t55;
    t212 = t15*t207*t21*t211;
    t216 = t203*t10*t75;
    t218 = t207*t21*t23;
    t220 = t67*t69*t55;
    t224 = EIGHT*t132*t133*t138*t142-TWO*t146*t66*t52*t67*t93+TWO*t146*t66*t52*
t83*t93+FOUR*t146*t157*t104-TWO*t146*t66*t52*t26*t110-EIGHT/THREE*t3*t166*t133*t8*
t75*t173+TWO/THREE*t179*t97-TWO/THREE*t183*t184+TWO/THREE*t179*t111-TWO/THREE*t183*t189
+TWO/THREE*t3*t192*t6*t10*t75*t197*t21*t23*t172-EIGHT/THREE*t203*t133*t205*t212+TWO/
THREE*t216*t218*t220;
    t227 = t83*t69*t55;
    t237 = A(p1,p3);
    t238 = B(p4,p6);
    t239 = t238*t238;
    t240 = t239*t239;
    t242 = t11*t11;
    t243 = t23*t23;
    t246 = IB(p4,p5);
    t248 = t237*t240*t242*t243*t23*t246;
    t249 = IB(p5,p6);
    t251 = IAB(p1,p6+p5,p4);
    t252 = t249*t251;
    t254 = IS(p6+p5+p4);
    t255 = t102*t254;
    t256 = p3+p2;
    t257 = AB3(p1,t256,p2,t25,p6);
    t258 = t257*t257;
    t264 = t254*t257;
    t265 = AB3(p1,t256,p3,t25,p6);
    t270 = t265*t265;
    t275 = A(p1,p5);
    t276 = t275*t275;
    t277 = t276*t276;
    t278 = t277*t2;
    t279 = IA(p1,p2);
    t281 = IA(p1,p6);
    t284 = IA(p5,p6);
    t286 = t15*t284*t40;
    t299 = A(p3,p5);
    t303 = t62*t286;
    t308 = t281*t11;
    t310 = t284*t243;
    t311 = IAB(p1,t256,p4);
    t314 = IS(p6+p5+p1);
    t320 = B(p3,p4);
    t331 = t320*t320;
    t341 = -TWO/THREE*t216*t218*t227-TWO/THREE*t216*t218*t34+TWO/THREE*t216*t218*t42+
EIGHT/THREE*t248*t252*t255*t258-TWO*EIGHT/THREE*t248*t252*t102*t264*t265+EIGHT/THREE*t248*t252
*t255*t270-TWO*t278*t4*t279*t281*t177*t286+FOUR*t278*t279*t281*t11*t137*t284*t38
*t116+FOUR*t277*t1*t299*t279*t281*t303-TEN/THREE*t277*t47*t46*t308*t310*t311*t32*
t314+SIX*t277*t47*t320*t281*t11*t310*t38*t311*t40*t314-TWO*t277*t46*t331*t281*
t11*t310*t114*t311*t116*t314;
    t344 = t3*t166*t9*t75;
    t345 = t51*t220;
    t348 = t51*t227;
    t351 = t51*t34;
    t354 = t51*t42;
    t358 = t3*t133*t135;
    t359 = t358*t75;
    t360 = t67*t67;
    t362 = t360*t85*t55;
    t363 = t51*t362;
    t366 = t358*t125;
    t368 = t83*t85*t55;
    t369 = t68*t368;
    t372 = t68*t93;
    t377 = t51*t79;
    t380 = t51*t87;
    t389 = FOUR*t344*t345-FOUR*t344*t348-FOUR*t344*t351+FOUR*t344*t354-EIGHT/THREE*t359
*t363+TWO*EIGHT/THREE*t366*t369+EIGHT/THREE*t366*t372-EIGHT/THREE*t366*t72+TWO*EIGHT/THREE*t359*t377
-EIGHT/THREE*t359*t380-EIGHT/THREE*t366*t94+EIGHT/THREE*t366*t97-EIGHT/THREE*t359*t184;
    t397 = t3*t10*t11;
    t399 = t397*t137*t19;
    t402 = t69*t69;
    t404 = t402*t29*t32;
    t408 = t66*t360;
    t415 = t26*t85*t33;
    t420 = t38*t85*t41;
    t434 = t137*t50;
    t435 = t397*t434;
    t436 = t23*t67;
    t437 = t436*t83;
    t441 = t2*t1;
    t443 = t441*t4*t299;
    t444 = t10*t46;
    t445 = t444*t177;
    t447 = t124*t66;
    t449 = t52*t27*t55;
    t450 = t447*t449;
    t453 = t133*t135;
    t454 = t453*t177;
    t457 = t124*t21*t211;
    t460 = t443*t17;
    t463 = EIGHT/THREE*t366*t111-EIGHT/THREE*t359*t189+TWO/THREE*t399*t66*t360*t67*t404
-TWO*t399*t408*t83*t402*t55-TWO/THREE*t399*t408*t415+TWO/THREE*t399*t408*t420-EIGHT/
THREE*t399*t68*t77*t85*t55+TWO*t399*t68*t84*t402*t55+FOUR/THREE*t435*t437*t415-THREE*FOUR*
t443*t445*t450+20.0/THREE*t443*t454*t457-EIGHT/THREE*t460*t345;
    t470 = t441*t299;
    t472 = t12*t13;
    t476 = t5*t46;
    t480 = t476*t17;
    t481 = t141*t220;
    t484 = t141*t227;
    t489 = t141*t42;
    t493 = t9*t11;
    t500 = t5*t133*t136*t16;
    t507 = EIGHT/THREE*t460*t348+EIGHT/THREE*t460*t351-EIGHT/THREE*t460*t354+EIGHT/THREE*t470*
t192*t472*t212+EIGHT*t476*t454*t450-TWO*t480*t481+TWO*t480*t484+FOUR*t480*t142-TWO
*t480*t489-FOUR*t5*t166*t493*t14*t457+EIGHT/THREE*t500*t345-EIGHT/THREE*t500*t348-EIGHT/
THREE*t500*t351;
    t526 = t470*t133*t138;
    t529 = t470*t10;
    t530 = t529*t75;
    t533 = t529*t125;
    t541 = EIGHT/THREE*t500*t354-TWO/THREE*t183*t363+FOUR/THREE*t179*t369+TWO/THREE*t179*
t372-TWO/THREE*t179*t72+FOUR/THREE*t183*t377-TWO/THREE*t183*t380-TWO/THREE*t179*t94-20.0
/THREE*t526*t354+EIGHT/THREE*t530*t363-TWO*EIGHT/THREE*t533*t369-EIGHT/THREE*t533*t372-FOUR/THREE*
t435*t437*t420;
    t543 = t100*t69*t103;
    t548 = t70*t109;
    t553 = t279*t279;
    t559 = t237*t3;
    t560 = t559*t4;
    t561 = t6*t10;
    t562 = t279*t11;
    t563 = t561*t562;
    t569 = t453*t562;
    t575 = t10*t279;
    t578 = t447*t42;
    t582 = t575*t62;
    t586 = t559*t6;
    t589 = t137*t22*t211;
    t592 = t586*t582;
    t593 = t63*t66;
    t612 = TWO/THREE*t399*t68*t543-TWO/THREE*t435*t436*t26*t548-TWO*t237*t277*t2*
t553*t281*t303+TWO*t560*t563*t16*t22*t211-FOUR*t560*t569*t16*t50*t211+TWO*t560*
t575*t177*t578-FOUR*t559*t192*t582*t212+THREE*FOUR*t586*t569*t589-TWO*t592*t593*t220+
TWO*t592*t593*t227+TWO*t592*t593*t34-FOUR*t592*t593*t42-THREE*FOUR*t559*t166*t9*t279*
t62*t457;
    t619 = t559*t133*t135*t279*t62;
    t630 = t114*t38;
    t632 = t40*t40;
    t638 = t4*t4;
    t642 = t14*t14;
    t648 = t3*t638;
    t650 = t14*t13;
    t651 = t12*t650;
    t654 = t63*t21*t211;
    t661 = t51*t449;
    t672 = t648*t10*t11*t650*t15;
    t682 = t12*t14;
    t686 = FOUR*t619*t447*t220-FOUR*t619*t447*t227-TWO/THREE*t277*t331*t320*t308*
t284*t243*t630*t311*t632*t314-TWO/THREE*t3*t638*t4*t10*t11*t642*t15*t173+TWO/THREE*
t648*t6*t651*t654+FOUR*t648*t46*t12*t650*t15*t661-EIGHT/THREE*t648*t133*t136*t650*
t457+TWO/THREE*t672*t345-TWO/THREE*t672*t348-TWO/THREE*t672*t351+TWO/THREE*t672*t354
-TWO/THREE*t5*t192*t682*t212;
    t688 = t593*t449;
    t694 = t24*t220;
    t697 = t24*t227;
    t701 = t12*t137;
    t708 = t493*t13;
    t713 = t60*t133*t138;
    t729 = -TWO*t7*t445*t688+EIGHT/THREE*t7*t454*t654-TWO/THREE*t18*t694+TWO/THREE*t18*
t697-TWO*t60*t46*t701*t22*t140*t34+FOUR*t60*t166*t708*t654-EIGHT/THREE*t713*t694+EIGHT
/THREE*t713*t697+EIGHT/THREE*t713*t35-EIGHT/THREE*t713*t43+TWO/THREE*t76*t24*t362-FOUR/THREE*
t65*t369-TWO/THREE*t65*t372;
    t731 = t9*t240;
    t734 = t731*t279*t19*t21;
    t736 = t243*t52*t246;
    t737 = t249*t102;
    t748 = t249*t26;
    t753 = t240*t242;
    t754 = t243*t243;
    t755 = t754*t246;
    t756 = t753*t755;
    t757 = t30*t30;
    t758 = t757*t254;
    t765 = t753*t755*t249;
    t766 = t251*t757;
    t781 = t470*t6;
    t786 = t453*t62;
    t790 = t781*t701;
    t797 = -TWO*t734*t736*t737*t257+TWO*t734*t736*t737*t265+TWO*t731*t51*t52*
t246*t748*t102+TWO/THREE*t756*t252*t758*t258*t257-TWO*t765*t766*t254*t258*t265+
TWO*t765*t766*t264*t270-TWO/THREE*t756*t252*t758*t270*t265+FOUR*t781*t444*t62*t688
-20.0/THREE*t781*t786*t654+EIGHT/THREE*t790*t694-EIGHT/THREE*t790*t697-EIGHT/THREE*t790*t35;
    t804 = t470*t46;
    t808 = t804*t701;
    t832 = EIGHT/THREE*t790*t43+FOUR*t470*t47*t701*t57-TWO*EIGHT*t804*t786*t450+FOUR*t808*
t481-FOUR*t808*t484-THREE*FOUR*t808*t142+FOUR*t808*t489+FOUR*t470*t166*t708*t457-20.0/
THREE*t526*t345+20.0/THREE*t526*t348+20.0/THREE*t526*t351-FOUR*t619*t447*t34+THREE*FOUR*t619
*t578;
    t837 = t559*t10*t562*t137;
    t842 = t23*t83;
    t855 = t237*t441*t299;
    t870 = t246*t249;
    t871 = t870*t30;
    t883 = -TWO*t837*t50*t436*t71+TWO*t837*t50*t842*t71+TWO*t837*t50*t23*t26*
t110-FOUR*t837*t189-EIGHT*t855*t563*t589+TWO*EIGHT*t855*t569*t434*t211-EIGHT*t855*t582*
t578-TWO*t237*t9*t240*t279*t19*t157*t871+EIGHT/THREE*t533*t72-TWO*EIGHT/THREE*t530*t377+
EIGHT/THREE*t530*t380+EIGHT/THREE*t533*t94;
    t893 = t299*t299;
    t898 = t2*t893;
    t912 = t898*t10*t75;
    t927 = -EIGHT/THREE*t533*t97+EIGHT/THREE*t530*t184-EIGHT/THREE*t533*t111+EIGHT/THREE*t530*
t189-FOUR*t2*t4*t893*t682*t457+FOUR*t898*t6*t472*t654+THREE*FOUR*t898*t46*t701*t661-FOUR
*t898*t133*t205*t457+FOUR*t912*t345-FOUR*t912*t348-FOUR*t912*t351+FOUR*t912*t354+
EIGHT/THREE*t1*t893*t299*t10*t75*t173;
    t933 = t140*t871;
    t938 = t240*t19*t21;
    t948 = A(p1,p2);
    t949 = t237*t237;
    t963 = t948*t237*t240*t242*t243*t246;
    t964 = t748*t251;
    t982 = t949*t3;
    t1002 = t114*t69*t117;
    t1006 = t66*t77;
    t1013 = -TWO*t6*t9*t240*t20*t21*t933-TWO*t46*t9*t938*t23*t53*t871+FOUR*t133*
t8*t938*t933+TWO*EIGHT/THREE*t948*t949*t753*t23*t870*t26*t251*t30*t254+EIGHT/THREE*t963*
t964*t255*t257-EIGHT/THREE*t963*t964*t255*t265-TWO*EIGHT/THREE*t949*t237*t240*t242*t23*
t870*t251*t29*t254+TWO*t982*t6*t10*t553*t62*t654-FOUR*t982*t133*t135*t553*t62*
t457+TWO*t982*t10*t553*t11*t137*t354+TWO/THREE*t399*t68*t1002+EIGHT/THREE*t399*t1006*
t368+FOUR/THREE*t399*t1006*t93;
    t1022 = t66*t84;
    t1071 = -FOUR/THREE*t399*t1006*t71-TWO/THREE*t399*t66*t84*t83*t404-TWO/THREE*t399*
t1022*t415+TWO/THREE*t399*t1022*t420-TWO/THREE*t399*t91*t543+TWO/THREE*t435*t842*t26*
t548-TWO/THREE*t399*t91*t1002-TWO/THREE*t399*t66*t100*t26*t27*t757*t32+TWO/THREE*t399
*t66*t100*t39*t102*t40-TWO/THREE*t399*t108*t115*t30*t116+TWO/THREE*t399*t66*t630*
t171*t632+EIGHT/THREE*t441*t638*t299*t651*t457-EIGHT/THREE*t443*t561*t177*t654;
    return(t131+t224+t341+t389+t463+t507+t541+t612+t686+t729+t797+t832+t883+
t927+t1013+t1071);
  }
}

std::complex<DOUBLE> C234mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  {
    return(ZERO);
  }
}


std::complex<DOUBLE> C245mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t1000;
  std::complex<DOUBLE> t1001;
  std::complex<DOUBLE> t1007;
  std::complex<DOUBLE> t1012;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t1024;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t1041;
  std::complex<DOUBLE> t1042;
  std::complex<DOUBLE> t1043;
  std::complex<DOUBLE> t1052;
  std::complex<DOUBLE> t1059;
  std::complex<DOUBLE> t1060;
  std::complex<DOUBLE> t1061;
  std::complex<DOUBLE> t1062;
  std::complex<DOUBLE> t1069;
  std::complex<DOUBLE> t1070;
  std::complex<DOUBLE> t1085;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  FourMomentum<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t2;
  FourMomentum<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t284;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t288;
  std::complex<DOUBLE> t289;
  std::complex<DOUBLE> t292;
  std::complex<DOUBLE> t294;
  std::complex<DOUBLE> t296;
  std::complex<DOUBLE> t297;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t302;
  std::complex<DOUBLE> t303;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t313;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t320;
  std::complex<DOUBLE> t323;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t329;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t333;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t338;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t351;
  std::complex<DOUBLE> t355;
  std::complex<DOUBLE> t364;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t379;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t390;
  std::complex<DOUBLE> t393;
  std::complex<DOUBLE> t396;
  std::complex<DOUBLE> t399;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t406;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t408;
  std::complex<DOUBLE> t409;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t412;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t418;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t422;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t427;
  std::complex<DOUBLE> t428;
  std::complex<DOUBLE> t431;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t437;
  std::complex<DOUBLE> t439;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t440;
  std::complex<DOUBLE> t441;
  std::complex<DOUBLE> t442;
  std::complex<DOUBLE> t445;
  std::complex<DOUBLE> t446;
  std::complex<DOUBLE> t447;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t451;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t458;
  std::complex<DOUBLE> t459;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t464;
  std::complex<DOUBLE> t465;
  std::complex<DOUBLE> t466;
  std::complex<DOUBLE> t469;
  std::complex<DOUBLE> t470;
  std::complex<DOUBLE> t471;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t475;
  std::complex<DOUBLE> t476;
  std::complex<DOUBLE> t477;
  std::complex<DOUBLE> t478;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t493;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t503;
  std::complex<DOUBLE> t507;
  std::complex<DOUBLE> t508;
  std::complex<DOUBLE> t513;
  std::complex<DOUBLE> t519;
  std::complex<DOUBLE> t523;
  std::complex<DOUBLE> t526;
  std::complex<DOUBLE> t527;
  std::complex<DOUBLE> t546;
  std::complex<DOUBLE> t549;
  std::complex<DOUBLE> t553;
  std::complex<DOUBLE> t560;
  std::complex<DOUBLE> t562;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t575;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t581;
  std::complex<DOUBLE> t585;
  std::complex<DOUBLE> t586;
  std::complex<DOUBLE> t599;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t603;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t630;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t645;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t650;
  std::complex<DOUBLE> t652;
  std::complex<DOUBLE> t659;
  FourMomentum<DOUBLE> t66;
  std::complex<DOUBLE> t665;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t681;
  std::complex<DOUBLE> t682;
  std::complex<DOUBLE> t683;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t714;
  std::complex<DOUBLE> t715;
  std::complex<DOUBLE> t719;
  std::complex<DOUBLE> t720;
  std::complex<DOUBLE> t725;
  std::complex<DOUBLE> t730;
  std::complex<DOUBLE> t735;
  std::complex<DOUBLE> t753;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t782;
  std::complex<DOUBLE> t783;
  std::complex<DOUBLE> t794;
  std::complex<DOUBLE> t796;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t800;
  std::complex<DOUBLE> t815;
  std::complex<DOUBLE> t828;
  std::complex<DOUBLE> t837;
  std::complex<DOUBLE> t842;
  std::complex<DOUBLE> t849;
  std::complex<DOUBLE> t857;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t870;
  std::complex<DOUBLE> t877;
  std::complex<DOUBLE> t880;
  std::complex<DOUBLE> t883;
  std::complex<DOUBLE> t904;
  std::complex<DOUBLE> t905;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t910;
  std::complex<DOUBLE> t916;
  std::complex<DOUBLE> t923;
  std::complex<DOUBLE> t938;
  std::complex<DOUBLE> t940;
  std::complex<DOUBLE> t941;
  std::complex<DOUBLE> t942;
  std::complex<DOUBLE> t943;
  std::complex<DOUBLE> t944;
  std::complex<DOUBLE> t945;
  std::complex<DOUBLE> t948;
  std::complex<DOUBLE> t949;
  std::complex<DOUBLE> t952;
  std::complex<DOUBLE> t955;
  std::complex<DOUBLE> t962;
  std::complex<DOUBLE> t963;
  std::complex<DOUBLE> t964;
  std::complex<DOUBLE> t966;
  std::complex<DOUBLE> t967;
  std::complex<DOUBLE> t968;
  std::complex<DOUBLE> t973;
  std::complex<DOUBLE> t992;
  std::complex<DOUBLE> t996;
  std::complex<DOUBLE> t998;
  {
    t1 = A(p1,p3);
    t2 = t1*t1;
    t3 = t2*t2;
    t4 = A(p1,p4);
    t5 = t4*t4;
    t6 = t3*t5;
    t7 = A(p3,p5);
    t8 = IA(p1,p2);
    t10 = IA(p1,p6);
    t13 = IA(p2,p3);
    t14 = IA(p3,p4);
    t15 = t14*t14;
    t17 = IA(p4,p5);
    t18 = IA(p4,p6);
    t19 = t17*t18;
    t20 = p3+p2;
    t21 = IAB(p1,t20,p5);
    t22 = t21*t21;
    t23 = t19*t22;
    t27 = A(p5,p6);
    t31 = t14*t13;
    t32 = t18*t18;
    t33 = t17*t32;
    t41 = t14*t17;
    t42 = AB(p1,t20,p4);
    t44 = t22*t21;
    t50 = A(p1,p5);
    t57 = B(p4,p6);
    t58 = t57*t57;
    t61 = t8*t13;
    t63 = IB(p4,p5);
    t64 = t63*t63;
    t65 = t17*t64;
    t66 = p5+p4;
    t67 = IAB(p3,t66,p6);
    t70 = IS(p3+p2+p1);
    t76 = B(p5,p6);
    t87 = t76*t76;
    t91 = t42*t42;
    t102 = t91*t42;
    t104 = t22*t22;
    t110 = t5*t5;
    t111 = t7*t7;
    t112 = t110*t111;
    t113 = B(p2,p4);
    t115 = B(p2,p5);
    t116 = t115*t115;
    t117 = t116*t115;
    t118 = t117*t10;
    t119 = t15*t17;
    t122 = IB(p2,p3);
    t123 = t18*t122;
    t124 = IB(p3,p5);
    t127 = p6+p1;
    t128 = IAB(p4,t127,p2);
    t129 = IAB(p4,t127,p5);
    t130 = t128*t129;
    t131 = t63*t21*t130;
    t132 = t123*t124*t131;
    t135 = t116*t116;
    t136 = t112*t135;
    t137 = B(p3,p4);
    t138 = t137*t10;
    t141 = t124*t124;
    t143 = t123*t141*t131;
    t149 = t122*t124;
    t150 = t149*t63;
    t152 = t42*t22*t130;
    t153 = t150*t152;
    t156 = t110*t7;
    t157 = t156*t27;
    t158 = t113*t117;
    t159 = t10*t14;
    t160 = t158*t159;
    t166 = -TWO*t6*t7*t8*t10*t13*t15*t23-TWO*t6*t27*t8*t10*t31*t33*t22+FOUR*t6*
t8*t10*t13*t41*t18*t42*t44+FOUR*t3*t4*t50*t8*t10*t31*t23+TEN/THREE*t3*t58*t57*t61
*t65*t21*t67*t70-SIX*t3*t58*t76*t8*t13*t65*t42*t22*t67*t70+TWO*t3*t57*t87*t8*
t13*t65*t91*t44*t67*t70+TWO/THREE*t3*t87*t76*t61*t17*t64*t102*t104*t67*t70-FOUR*
t112*t113*t118*t119*t132+TWO*t136*t138*t119*t143+TWO*t136*t10*t15*t19*t153-FOUR*
t157*t160*t33*t149*t131;
    t167 = t135*t137;
    t168 = t167*t159;
    t170 = t122*t141;
    t175 = t135*t10;
    t176 = t175*t41;
    t178 = t32*t122;
    t179 = t124*t63;
    t180 = t178*t179;
    t184 = t113*t113;
    t186 = t116*t10;
    t191 = t156*t113;
    t192 = t117*t137;
    t196 = t19*t170*t131;
    t200 = t191*t118*t41;
    t201 = t123*t179;
    t202 = t201*t152;
    t205 = AB(p4,t127,p4);
    t207 = t129*t129;
    t208 = t128*t207;
    t209 = t205*t21*t208;
    t213 = AB(p5,t127,p2);
    t214 = t213*t21;
    t215 = t128*t128;
    t216 = t215*t129;
    t217 = t214*t216;
    t221 = AB(p5,t127,p5);
    t222 = t221*t21;
    t223 = t222*t208;
    t227 = t156*t135;
    t228 = t137*t137;
    t229 = t228*t10;
    t232 = t141*t124;
    t234 = t123*t232*t131;
    t238 = t227*t138*t41;
    t239 = t141*t63;
    t240 = t123*t239;
    t253 = TWO*t157*t168*t33*t170*t131+TWO*t157*t176*t180*t152-THREE*FOUR*t156*t184*
t186*t41*t132+THREE*FOUR*t191*t192*t159*t196+THREE*FOUR*t200*t202+FOUR*t200*t201*t209-FOUR*
t200*t201*t217-FOUR*t200*t201*t223-FOUR*t227*t229*t41*t234-FOUR*t238*t240*t152-TWO
*t238*t240*t209+TWO*t238*t240*t217+TWO*t238*t240*t223;
    t256 = t227*t159*t19;
    t258 = t91*t44*t130;
    t259 = t150*t258;
    t262 = t63*t42;
    t263 = t149*t262;
    t265 = t205*t22*t208;
    t269 = t213*t22;
    t270 = t269*t216;
    t274 = t221*t22;
    t275 = t274*t208;
    t279 = t27*t27;
    t283 = t10*t17;
    t284 = t32*t32;
    t287 = t21*t128;
    t288 = t287*t129;
    t289 = t150*t288;
    t292 = t110*t279;
    t294 = t118*t17;
    t296 = t32*t18;
    t297 = t296*t122;
    t299 = t297*t124*t131;
    t302 = t292*t135;
    t303 = t138*t17;
    t309 = t57*t10;
    t313 = IB(p5,p6);
    t315 = t313*t21*t130;
    t316 = t150*t315;
    t320 = t302*t283*t296;
    t323 = t150*t209;
    t326 = t150*t217;
    t329 = t150*t223;
    t332 = -FOUR*t256*t259-TWO*t256*t263*t265+TWO*t256*t263*t270+TWO*t256*t263*
t275-TWO/THREE*t110*t279*t27*t135*t283*t284*t289-EIGHT/THREE*t292*t113*t294*t299+TWO/
THREE*t302*t303*t297*t141*t131+FOUR*t302*t309*t17*t296*t316+TWO/THREE*t320*t153+TWO/
THREE*t320*t323-TWO/THREE*t320*t326-TWO/THREE*t320*t329;
    t333 = t110*t27;
    t335 = t186*t17;
    t338 = t178*t124*t131;
    t341 = t333*t113;
    t342 = t192*t283;
    t345 = t178*t141*t131;
    t349 = t117*t57*t283;
    t351 = t180*t315;
    t355 = t341*t118*t33;
    t364 = t333*t135;
    t365 = t229*t17;
    t372 = t137*t57*t283;
    t379 = t364*t138*t33;
    t380 = t170*t63;
    t381 = t380*t152;
    t384 = t380*t209;
    t387 = t380*t217;
    t390 = t380*t223;
    t393 = -FOUR*t333*t184*t335*t338+EIGHT/THREE*t341*t342*t345+EIGHT*t341*t349*t351+
EIGHT/THREE*t355*t153+EIGHT/THREE*t355*t323-EIGHT/THREE*t355*t326-EIGHT/THREE*t355*t329-TWO/THREE
*t364*t365*t178*t232*t131-TWO*t364*t372*t178*t239*t315-TWO/THREE*t379*t381-TWO/
THREE*t379*t384+TWO/THREE*t379*t387+TWO/THREE*t379*t390;
    t396 = t58*t10;
    t399 = t313*t313;
    t402 = t150*t399*t21*t130;
    t406 = t364*t309*t33;
    t407 = t63*t313;
    t408 = t149*t407;
    t409 = t408*t152;
    t412 = t408*t209;
    t415 = t408*t217;
    t418 = t408*t223;
    t422 = t364*t283*t32;
    t426 = t364*t283*t178;
    t427 = t179*t42;
    t428 = t427*t265;
    t431 = t427*t270;
    t434 = t427*t275;
    t437 = t205*t205;
    t439 = t207*t129;
    t440 = t128*t439;
    t441 = t437*t21*t440;
    t442 = t150*t441;
    t445 = t179*t205;
    t446 = t215*t207;
    t447 = t214*t446;
    t448 = t445*t447;
    t451 = t222*t440;
    t452 = t445*t451;
    t455 = -TWO*t364*t396*t33*t402-TWO*t406*t409-TWO*t406*t412+FOUR*t406*t415+
TWO*t406*t418-TWO/THREE*t422*t259-TWO/THREE*t426*t428+TWO/THREE*t426*t431+TWO/THREE*
t426*t434-TWO/THREE*t422*t442+TWO/THREE*t426*t448+FOUR/THREE*t426*t452;
    t456 = t213*t213;
    t457 = t456*t21;
    t458 = t215*t128;
    t459 = t458*t129;
    t460 = t457*t459;
    t461 = t150*t460;
    t464 = t179*t213;
    t465 = t222*t446;
    t466 = t464*t465;
    t469 = AB(p5,t127,p4);
    t470 = t469*t21;
    t471 = t470*t208;
    t472 = t150*t471;
    t475 = t221*t221;
    t476 = t475*t21;
    t477 = t476*t440;
    t478 = t150*t477;
    t484 = t283*t18;
    t489 = t110*t184*t116;
    t493 = t489*t484;
    t503 = t110*t113*t117;
    t507 = t138*t19;
    t508 = t503*t507;
    t513 = -TWO/THREE*t422*t461-TWO/THREE*t426*t466+FOUR/THREE*t422*t472-TWO/THREE*t422*
t478-EIGHT/THREE*t110*t184*t113*t115*t484*t289+FOUR*t489*t303*t143+FOUR*t493*t153+FOUR
*t493*t323-FOUR*t493*t326-FOUR*t493*t329-EIGHT/THREE*t503*t365*t234-EIGHT/THREE*t508*t381
-EIGHT/THREE*t508*t384;
    t519 = t309*t19;
    t523 = t503*t484;
    t526 = t283*t123;
    t527 = t503*t526;
    t546 = EIGHT/THREE*t508*t387+EIGHT/THREE*t508*t390+EIGHT*t503*t519*t415-EIGHT/THREE*t523*
t259-EIGHT/THREE*t527*t428+EIGHT/THREE*t527*t431+EIGHT/THREE*t527*t434-EIGHT/THREE*t523*t442+
EIGHT/THREE*t527*t448+TWO*EIGHT/THREE*t527*t452-EIGHT/THREE*t523*t461-EIGHT/THREE*t527*t466+TWO*EIGHT/
THREE*t523*t472;
    t549 = t110*t135;
    t553 = t141*t141;
    t560 = t549*t228*t484;
    t562 = t122*t232*t63;
    t575 = t549*t137;
    t581 = t575*t484;
    t585 = t575*t526;
    t586 = t239*t42;
    t599 = t239*t205;
    t603 = -EIGHT/THREE*t523*t478+TWO/THREE*t549*t228*t137*t484*t122*t553*t63*t288+
TWO/THREE*t560*t562*t152+TWO/THREE*t560*t562*t209-TWO/THREE*t560*t562*t217-TWO/THREE*
t560*t562*t223-TWO*t575*t519*t170*t407*t217+TWO/THREE*t581*t380*t258+TWO/THREE*t585
*t586*t265-TWO/THREE*t585*t586*t270-TWO/THREE*t585*t586*t275+TWO/THREE*t581*t380*t441
-TWO/THREE*t585*t599*t447;
    t630 = t549*t57*t526;
    t645 = t313*t213;
    t650 = t549*t10;
    t652 = t650*t19*t122;
    t659 = t179*t91;
    t665 = -FOUR/THREE*t585*t599*t451+TWO/THREE*t581*t380*t460+TWO/THREE*t585*t239*
t213*t465-FOUR/THREE*t581*t380*t471+TWO/THREE*t581*t380*t477-TWO*t549*t58*t526*t179*
t399*t217-TWO*t630*t179*t313*t42*t270-TWO*t630*t179*t313*t205*t447+FOUR*t630*
t179*t313*t460+TWO*t630*t179*t645*t465+TWO/THREE*t652*t179*t102*t104*t128*t129+
TWO/THREE*t652*t659*t205*t44*t208;
    t681 = t19*t149;
    t682 = t650*t681;
    t683 = t262*t205;
    t714 = t207*t207;
    t715 = t287*t714;
    t719 = t179*t437;
    t720 = t215*t439;
    t725 = t128*t714;
    t730 = t458*t207;
    t735 = -TWO/THREE*t652*t659*t213*t44*t216-TWO/THREE*t652*t659*t221*t44*t208+TWO
/THREE*t652*t427*t437*t22*t440-TWO/THREE*t682*t683*t269*t446-FOUR/THREE*t682*t683*t274
*t440+TWO/THREE*t652*t427*t456*t22*t459+TWO/THREE*t682*t262*t213*t274*t446-FOUR/THREE*
t652*t427*t469*t22*t208+TWO/THREE*t652*t427*t475*t22*t440+TWO/THREE*t652*t179*t437*
t205*t715-TWO/THREE*t652*t719*t214*t720-TWO*t652*t719*t222*t725+TWO/THREE*t652*t445
*t457*t730;
    t753 = t215*t215;
    t782 = t5*t4*t50;
    t783 = t782*t7;
    t794 = FOUR/THREE*t682*t63*t205*t213*t222*t720-EIGHT/THREE*t652*t445*t470*t440+TWO
*t652*t445*t476*t725-TWO/THREE*t652*t179*t456*t213*t21*t753*t129-TWO/THREE*t652*
t179*t456*t222*t730+FOUR/THREE*t652*t464*t470*t446-TWO/THREE*t652*t464*t476*t720+EIGHT
/THREE*t652*t179*t469*t451-TWO/THREE*t652*t179*t475*t221*t715+TWO*EIGHT*t783*t160*t681*
t131-EIGHT*t783*t168*t196-EIGHT*t783*t176*t202;
    t796 = t175*t17;
    t800 = t782*t27;
    t815 = t800*t175*t33;
    t828 = t782*t113;
    t837 = t828*t118*t19;
    t842 = EIGHT/THREE*t782*t279*t796*t299+20.0/THREE*t800*t158*t283*t338-EIGHT/THREE*
t800*t167*t283*t345-THREE*FOUR*t800*t135*t57*t283*t351-EIGHT/THREE*t815*t153-EIGHT/THREE*t815
*t323+EIGHT/THREE*t815*t326+EIGHT/THREE*t815*t329+FOUR*t782*t184*t335*t132-20.0/THREE*t828
*t342*t143-TWO*EIGHT*t828*t349*t201*t315-20.0/THREE*t837*t153-20.0/THREE*t837*t323;
    t849 = t782*t135;
    t857 = t849*t507;
    t870 = t849*t519;
    t877 = 20.0/THREE*t837*t326+20.0/THREE*t837*t329+EIGHT/THREE*t849*t365*t234+FOUR*
t849*t372*t240*t315+EIGHT/THREE*t857*t381+EIGHT/THREE*t857*t384-EIGHT/THREE*t857*t387-EIGHT/
THREE*t857*t390+FOUR*t849*t396*t19*t402+FOUR*t870*t409+FOUR*t870*t412-THREE*FOUR*t870*t415
;
    t880 = t849*t484;
    t883 = t849*t526;
    t904 = t50*t50;
    t905 = t5*t904;
    t910 = -FOUR*t870*t418+EIGHT/THREE*t880*t259+EIGHT/THREE*t883*t428-EIGHT/THREE*t883*t431
-EIGHT/THREE*t883*t434+EIGHT/THREE*t880*t442-EIGHT/THREE*t883*t448-TWO*EIGHT/THREE*t883*t452+EIGHT/
THREE*t880*t461+EIGHT/THREE*t883*t466-TWO*EIGHT/THREE*t880*t472+EIGHT/THREE*t880*t478-FOUR*t905*
t27*t796*t338;
    t916 = t905*t135;
    t923 = t916*t484;
    t938 = A(p3,p4);
    t940 = B(p2,p6);
    t941 = t940*t940;
    t942 = t941*t941;
    t943 = t17*t17;
    t944 = t942*t943;
    t945 = IB(p1,p2);
    t948 = IB(p1,p6);
    t949 = t948*t63;
    t952 = IAB(p3,p2+p1,p6);
    t955 = IS(p6+p2+p1);
    t962 = t943*t945;
    t963 = t962*t948;
    t964 = t938*t7*t942*t963;
    t966 = t64*t213*t952;
    t967 = t458*t955;
    t968 = AB3(p3,t66,p4,t127,p2);
    t973 = AB3(p3,t66,p5,t127,p2);
    t992 = t407*t215;
    t996 = -FOUR*t905*t113*t294*t132+FOUR*t916*t303*t143+THREE*FOUR*t916*t519*t316+FOUR*
t923*t153+FOUR*t923*t323-FOUR*t923*t326-FOUR*t923*t329+EIGHT/THREE*t4*t904*t50*t135*
t484*t289-TWO*EIGHT/THREE*t938*t111*t944*t945*t949*t213*t952*t215*t955-EIGHT/THREE*t964*
t966*t967*t968+EIGHT/THREE*t964*t966*t967*t973+TWO*EIGHT/THREE*t111*t7*t942*t962*t949*t952
*t128*t955-TWO*t7*t116*t942*t14*t945*t948*t122*t124*t992;
    t998 = t7*t942*t963;
    t1000 = t64*t63*t952;
    t1001 = t968*t968;
    t1007 = t955*t968;
    t1012 = t973*t973;
    t1024 = t116*t942;
    t1041 = t1024*t14*t945*t948;
    t1042 = t149*t64;
    t1043 = t313*t458;
    t1052 = t945*t948;
    t1059 = t944*t1052;
    t1060 = t64*t64;
    t1061 = t1060*t952;
    t1062 = t753*t955;
    t1069 = t944*t1052*t1060;
    t1070 = t952*t753;
    t1085 = -EIGHT/THREE*t998*t1000*t967*t1001+TWO*EIGHT/THREE*t998*t1000*t458*t1007*t973
-EIGHT/THREE*t998*t1000*t967*t1012+FOUR*t113*t115*t942*t945*t948*t149*t992-TWO*t1024
*t137*t945*t948*t170*t992-TWO*t1024*t57*t945*t948*t149*t63*t399*t215-TWO*t1041*
t1042*t1043*t968+TWO*t1041*t1042*t1043*t973+TWO*t1024*t1052*t122*t179*t645*t458
-TWO/THREE*t1059*t1061*t1062*t1001*t968+TWO*t1069*t1070*t955*t1001*t973-TWO*t1069
*t1070*t1007*t1012+TWO/THREE*t1059*t1061*t1062*t1012*t973;
    return(t166+t253+t332+t393+t455+t513+t546+t603+t665+t735+t794+t842+t877+
t910+t996+t1085);
  }
}

std::complex<DOUBLE> C256mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  {
    return(ZERO);
  }
}


std::complex<DOUBLE> C261mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t1030;
  std::complex<DOUBLE> t1033;
  std::complex<DOUBLE> t1041;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t1081;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t166;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t173;
  FourMomentum<DOUBLE> t175;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  FourMomentum<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t217;
  FourMomentum<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t278;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t295;
  std::complex<DOUBLE> t299;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t300;
  std::complex<DOUBLE> t303;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t311;
  std::complex<DOUBLE> t315;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t329;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t337;
  std::complex<DOUBLE> t338;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t348;
  std::complex<DOUBLE> t353;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t360;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t364;
  std::complex<DOUBLE> t365;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t369;
  std::complex<DOUBLE> t372;
  std::complex<DOUBLE> t375;
  std::complex<DOUBLE> t376;
  std::complex<DOUBLE> t377;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t391;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t395;
  std::complex<DOUBLE> t396;
  std::complex<DOUBLE> t397;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t403;
  std::complex<DOUBLE> t407;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t411;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t425;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t438;
  std::complex<DOUBLE> t439;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t442;
  std::complex<DOUBLE> t444;
  std::complex<DOUBLE> t445;
  std::complex<DOUBLE> t446;
  std::complex<DOUBLE> t448;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t450;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t453;
  std::complex<DOUBLE> t456;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t459;
  std::complex<DOUBLE> t460;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t480;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t482;
  std::complex<DOUBLE> t483;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t494;
  std::complex<DOUBLE> t499;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t514;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t525;
  std::complex<DOUBLE> t527;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t532;
  std::complex<DOUBLE> t536;
  std::complex<DOUBLE> t537;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t549;
  std::complex<DOUBLE> t550;
  std::complex<DOUBLE> t556;
  std::complex<DOUBLE> t560;
  std::complex<DOUBLE> t562;
  std::complex<DOUBLE> t564;
  std::complex<DOUBLE> t567;
  std::complex<DOUBLE> t568;
  std::complex<DOUBLE> t571;
  std::complex<DOUBLE> t573;
  std::complex<DOUBLE> t577;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t581;
  std::complex<DOUBLE> t582;
  std::complex<DOUBLE> t587;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t590;
  std::complex<DOUBLE> t594;
  std::complex<DOUBLE> t595;
  std::complex<DOUBLE> t596;
  std::complex<DOUBLE> t599;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t613;
  std::complex<DOUBLE> t619;
  std::complex<DOUBLE> t622;
  std::complex<DOUBLE> t625;
  std::complex<DOUBLE> t632;
  std::complex<DOUBLE> t637;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t647;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t651;
  std::complex<DOUBLE> t658;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t671;
  std::complex<DOUBLE> t673;
  std::complex<DOUBLE> t679;
  std::complex<DOUBLE> t699;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t706;
  std::complex<DOUBLE> t711;
  std::complex<DOUBLE> t717;
  std::complex<DOUBLE> t718;
  std::complex<DOUBLE> t720;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t739;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t750;
  std::complex<DOUBLE> t756;
  std::complex<DOUBLE> t757;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t774;
  std::complex<DOUBLE> t779;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t781;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t790;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t801;
  std::complex<DOUBLE> t804;
  std::complex<DOUBLE> t805;
  std::complex<DOUBLE> t806;
  std::complex<DOUBLE> t807;
  std::complex<DOUBLE> t814;
  std::complex<DOUBLE> t815;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t838;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t857;
  std::complex<DOUBLE> t859;
  std::complex<DOUBLE> t860;
  std::complex<DOUBLE> t873;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t886;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t916;
  std::complex<DOUBLE> t929;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t930;
  std::complex<DOUBLE> t934;
  std::complex<DOUBLE> t935;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t954;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t978;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t986;
  std::complex<DOUBLE> t99;
  std::complex<DOUBLE> t994;
  {
    t1 = A(p3,p6);
    t2 = t1*t1;
    t3 = t2*t2;
    t4 = B(p1,p4);
    t5 = t4*t4;
    t6 = t5*t5;
    t7 = t3*t6;
    t8 = IA(p1,p6);
    t9 = t7*t8;
    t10 = IA(p2,p3);
    t11 = IA(p2,p6);
    t12 = t10*t11;
    t13 = IB(p1,p5);
    t14 = IB(p1,p6);
    t15 = t14*t13;
    t17 = t9*t12*t15;
    t18 = IB(p4,p5);
    t19 = p3+p2;
    t20 = AB(p1,t19,p4);
    t21 = t18*t20;
    t22 = p5+p4;
    t23 = AB(p3,t22,p6);
    t25 = AB(p6,t19,p6);
    t26 = IAB(p3,t22,p1);
    t27 = t26*t26;
    t28 = t25*t27;
    t29 = IAB(p6,t19,p1);
    t30 = t29*t29;
    t31 = IAB(p6,t19,p4);
    t32 = t31*t31;
    t33 = t30*t32;
    t39 = t9*t12*t13;
    t40 = t14*t18;
    t41 = t40*t20;
    t42 = t25*t25;
    t43 = t42*t26;
    t44 = t30*t29;
    t45 = t44*t32;
    t50 = AB(p1,t19,p6);
    t51 = t40*t50;
    t52 = t23*t27;
    t53 = t30*t31;
    t54 = t52*t53;
    t58 = t25*t26;
    t59 = t44*t31;
    t60 = t58*t59;
    t64 = t23*t23;
    t65 = t64*t23;
    t67 = t27*t27;
    t73 = A(p1,p2);
    t74 = t73*t3;
    t75 = t74*t6;
    t76 = B(p2,p6);
    t77 = t76*t76;
    t78 = t77*t8;
    t79 = t11*t11;
    t80 = t10*t79;
    t83 = IB(p1,p2);
    t84 = t83*t83;
    t88 = t29*t31;
    t89 = t18*t26*t88;
    t90 = t84*t13*t14*t89;
    t93 = B(p5,p6);
    t95 = t8*t10;
    t96 = t76*t93*t95;
    t98 = t79*t83;
    t99 = t13*t13;
    t100 = t99*t14;
    t105 = t76*t8;
    t107 = t75*t105*t80;
    t108 = t83*t13;
    t109 = t108*t40;
    t110 = AB(p1,t19,p1);
    t112 = t110*t26*t53;
    t113 = t109*t112;
    t116 = A(p1,p5);
    t117 = t116*t3;
    t118 = t117*t6;
    t119 = t93*t93;
    t120 = t119*t8;
    t123 = IA(p5,p6);
    t124 = t99*t13;
    t130 = t93*t8;
    t131 = t130*t12;
    t132 = t118*t131;
    t134 = t123*t99*t40;
    t139 = t73*t116*t3;
    t140 = t5*t4;
    t141 = B(p4,p6);
    t143 = t140*t141*t95;
    t145 = t79*t123;
    t150 = A(p3,p5);
    t151 = t150*t150;
    t152 = t151*t151;
    t155 = t2*t8*t10;
    t157 = IA(p3,p4);
    t159 = IA(p4,p5);
    t161 = t159*t123*t27;
    t165 = TWO/THREE*t17*t21*t23*t28*t33+TWO/THREE*t39*t41*t43*t45+FOUR/THREE*t39*t51*
t54+EIGHT/THREE*t39*t51*t60-TWO/THREE*t39*t40*t65*t67*t29*t31-TWO*t75*t78*t80*t90-TWO
*t75*t96*t98*t100*t89-TWO*t107*t113-FOUR*t118*t120*t12*t123*t124*t14*t89-TWO*
t132*t134*t112-FOUR*t139*t143*t145*t15*t89-TWO*t73*t152*t155*t79*t157*t161;
    t166 = B(p1,p2);
    t167 = t166*t166;
    t170 = t8*t157;
    t173 = t14*t14;
    t175 = p6+p1;
    t176 = IAB(p5,t175,p2);
    t179 = IS(p5+p4+p3);
    t184 = A(p1,p3);
    t186 = t2*t1;
    t187 = t73*t184*t186;
    t188 = t6*t8;
    t189 = t188*t80;
    t190 = t187*t189;
    t191 = t15*t18;
    t192 = t191*t112;
    t195 = t20*t26;
    t196 = t29*t32;
    t197 = t195*t196;
    t198 = t191*t197;
    t202 = t184*t116*t186;
    t205 = t123*t13;
    t206 = t205*t40;
    t207 = t52*t88;
    t208 = t206*t207;
    t211 = t74*t140;
    t212 = t141*t8;
    t214 = t211*t212*t80;
    t217 = t191*t207;
    t220 = t58*t53;
    t221 = t191*t220;
    t225 = t141*t141;
    t226 = t225*t8;
    t227 = t226*t10;
    t229 = t79*t13;
    t231 = t229*t14*t89;
    t234 = t184*t184;
    t238 = t95*t11;
    t240 = t26*t29;
    t241 = t240*t31;
    t242 = t191*t241;
    t246 = t75*t130*t80;
    t247 = t100*t18;
    t248 = t247*t220;
    t251 = t7*t93;
    t252 = t251*t238;
    t253 = t110*t110;
    t255 = t253*t26*t59;
    t259 = t11*t99;
    t261 = t251*t95*t259;
    t262 = t40*t110;
    t263 = t195*t33;
    t264 = t262*t263;
    t267 = t73*t73;
    t268 = t267*t3;
    t269 = t268*t6;
    t270 = t79*t11;
    t272 = t269*t95*t270;
    t275 = -TWO/THREE*t152*t167*t166*t170*t159*t173*t65*t67*t176*t179-EIGHT/THREE*
t190*t192-EIGHT/THREE*t190*t198+EIGHT*t202*t188*t12*t208+EIGHT/THREE*t214*t198-EIGHT/THREE*
t214*t217-EIGHT/THREE*t214*t221-FOUR*t74*t5*t227*t231+EIGHT/THREE*t234*t184*t1*t6*t238*
t242+TWO/THREE*t246*t248+TWO/THREE*t252*t247*t255+TWO/THREE*t261*t264-TWO/THREE*t272*
t221;
    t278 = t212*t10;
    t282 = t270*t13*t14*t89;
    t287 = t188*t10;
    t295 = t98*t15*t89;
    t299 = t75*t95*t79;
    t300 = t191*t255;
    t303 = t43*t59;
    t309 = t30*t30;
    t311 = t26*t309*t31;
    t315 = t40*t253;
    t324 = t309*t31;
    t329 = t20*t20;
    t330 = t329*t26;
    t331 = t32*t31;
    t332 = t30*t331;
    t337 = t18*t110;
    t338 = t337*t20;
    t347 = -EIGHT/THREE*t268*t140*t278*t282-FOUR*t73*t234*t2*t287*t231-THREE*FOUR*t187*t6*
t76*t95*t295-TWO/THREE*t299*t300+TWO/THREE*t252*t247*t303+TWO/THREE*t39*t40*t253*t110
*t311+TWO/THREE*t39*t315*t195*t45-TWO/THREE*t39*t315*t52*t59-TWO*t39*t315*t58*t324+
TWO/THREE*t39*t262*t330*t332-TWO/THREE*t17*t338*t52*t33-FOUR/THREE*t17*t338*t58*t45;
    t348 = t50*t26;
    t353 = t27*t26;
    t354 = t64*t353;
    t359 = t184*t186;
    t360 = t359*t6;
    t361 = t360*t238;
    t364 = t11*t13;
    t365 = t95*t364;
    t366 = t360*t365;
    t369 = t262*t54;
    t372 = t262*t60;
    t375 = t29*t331;
    t376 = t330*t375;
    t377 = t191*t376;
    t380 = t52*t196;
    t381 = t41*t380;
    t391 = t360*t131;
    t392 = t247*t207;
    t395 = t105*t12;
    t396 = t360*t395;
    t397 = t109*t220;
    t400 = -EIGHT/THREE*t39*t262*t348*t59+TWO/THREE*t39*t262*t354*t53+EIGHT/THREE*t361*
t300+EIGHT/THREE*t366*t264-EIGHT/THREE*t366*t369-TWO*EIGHT/THREE*t366*t372+EIGHT/THREE*t361*t377
-EIGHT/THREE*t366*t381+EIGHT/THREE*t190*t217+EIGHT/THREE*t190*t221+20.0/THREE*t187*t143*t231
-EIGHT/THREE*t391*t392-FOUR*t396*t397;
    t403 = t120*t10;
    t407 = t11*t124*t14*t89;
    t411 = t7*t119*t238;
    t413 = t124*t14*t18;
    t425 = t79*t79;
    t438 = t108*t14;
    t439 = t438*t89;
    t442 = A(p5,p6);
    t444 = B(p2,p4);
    t445 = t444*t444;
    t446 = t445*t445;
    t448 = t8*t8;
    t450 = IB(p2,p3);
    t452 = t116*t442*t446*t448*t173*t450;
    t453 = IB(p3,p4);
    t456 = IAB(p5,p4+p3,p2);
    t457 = t453*t20*t456;
    t459 = IS(p4+p3+p2);
    t460 = t331*t459;
    t461 = AB3(p5,t175,p6,t19,p4);
    t472 = t453*t18*t32;
    t480 = t116*t446*t448*t173*t14*t450;
    t481 = t453*t456;
    t482 = AB3(p5,t175,p1,t19,p4);
    t483 = t482*t482;
    t489 = t459*t482;
    t494 = t461*t461;
    t499 = EIGHT/THREE*t360*t403*t407-TWO/THREE*t411*t413*t207-TWO/THREE*t411*t413*t220
-EIGHT/THREE*t391*t248-TWO/THREE*t267*t73*t3*t6*t95*t425*t242+EIGHT/THREE*t267*t184*t186*
t287*t282+FOUR*t269*t105*t10*t270*t439-EIGHT/THREE*t452*t457*t460*t461-TWO*t116*t5*
t446*t123*t83*t15*t450*t472+EIGHT/THREE*t480*t481*t460*t483-TWO*EIGHT/THREE*t480*t481*t331
*t489*t461+EIGHT/THREE*t480*t481*t460*t494;
    t514 = t40*t23;
    t525 = t3*t140;
    t527 = t212*t12;
    t529 = t109*t197;
    t532 = t525*t141;
    t536 = t532*t131;
    t537 = t247*t112;
    t548 = t234*t2;
    t549 = t548*t6;
    t550 = t549*t238;
    t556 = t364*t14*t89;
    t560 = t6*t93*t95;
    t562 = t11*t123;
    t564 = t562*t100*t89;
    t567 = t58*t33;
    t568 = t41*t567;
    t571 = -FOUR*t152*t2*t238*t157*t159*t123*t23*t353-TWO/THREE*t39*t40*t64*t25*
t353*t53-TWO/THREE*t39*t514*t42*t27*t59-TWO/THREE*t39*t40*t42*t25*t311-EIGHT*t525*t76
*t527*t529-EIGHT/THREE*t532*t403*t407-EIGHT/THREE*t536*t537-FOUR/THREE*t39*t41*t348*t33+
TWO/THREE*t39*t41*t354*t196-FOUR*t550*t221-FOUR*t548*t140*t278*t556-EIGHT*t202*t560*
t564-TWO/THREE*t261*t568;
    t573 = t348*t53;
    t577 = t354*t88;
    t581 = t359*t140;
    t582 = t581*t527;
    t587 = t191*t573;
    t590 = t191*t577;
    t594 = t75*t95*t229;
    t595 = t28*t53;
    t596 = t514*t595;
    t599 = t191*t303;
    t613 = t79*t99*t14*t89;
    t619 = t130*t10;
    t622 = t259*t14*t89;
    t625 = -FOUR/THREE*t252*t247*t573+TWO/THREE*t252*t247*t577-20.0/THREE*t582*t192
-20.0/THREE*t582*t198+FOUR/THREE*t299*t587-TWO/THREE*t299*t590-TWO/THREE*t594*t596-TWO/
THREE*t299*t599+TWO*t139*t560*t145*t100*t89-TWO*t139*t189*t208-EIGHT/THREE*t187*t560*
t613+THREE*FOUR*t549*t395*t439+FOUR*t549*t619*t622;
    t632 = t532*t365;
    t637 = t532*t238;
    t647 = t3*t5*t225;
    t651 = t647*t238;
    t658 = FOUR*t550*t192+FOUR*t550*t198-FOUR*t550*t217+EIGHT/THREE*t632*t381+EIGHT/THREE*
t632*t568+TWO*EIGHT/THREE*t637*t587-EIGHT/THREE*t637*t590-EIGHT/THREE*t632*t596-EIGHT/THREE*t637*
t599+FOUR*t647*t619*t622+FOUR*t651*t192+FOUR*t651*t198-EIGHT/THREE*t366*t568;
    t671 = t76*t141*t95;
    t673 = t11*t83;
    t679 = t141*t93*t95;
    t699 = -TWO*EIGHT/THREE*t361*t587+EIGHT/THREE*t361*t590+EIGHT/THREE*t366*t596+EIGHT/THREE*t361
*t599-TWO*EIGHT*t581*t671*t673*t15*t89-20.0/THREE*t581*t679*t622+TWO/THREE*t269*t619*
t270*t99*t14*t89+TWO/THREE*t272*t192+TWO/THREE*t272*t198-TWO/THREE*t272*t217+TWO/THREE*
t261*t596+20.0/THREE*t582*t217;
    t706 = t116*t116;
    t711 = t450*t453;
    t717 = t706*t3;
    t718 = t717*t6;
    t720 = t123*t123;
    t739 = t446*t448;
    t750 = t11*t157;
    t756 = t117*t140;
    t757 = t756*t527;
    t774 = t32*t32;
    t779 = 20.0/THREE*t582*t221+FOUR*t359*t5*t227*t556-TWO*EIGHT/THREE*t706*t116*t446*
t448*t14*t711*t456*t31*t459+TWO*t718*t131*t720*t99*t14*t89-TWO*t718*t95*t11*
t720*t217-FOUR*t717*t140*t527*t720*t13*t14*t89-TWO*EIGHT/THREE*t706*t442*t739*t14*t711*
t20*t456*t32*t459-TWO*t116*t152*t155*t750*t159*t720*t27+FOUR*t757*t206*t197-THREE*FOUR
*t757*t208+FOUR/THREE*t17*t337*t23*t28*t59+TWO*t39*t262*t43*t324+TWO/THREE*t39*t40*
t329*t20*t240*t774;
    t781 = t40*t329;
    t790 = t5*t446;
    t801 = t14*t450*t472;
    t804 = t173*t173;
    t805 = t804*t450;
    t806 = t739*t805;
    t807 = t774*t459;
    t814 = t739*t805*t453;
    t815 = t456*t774;
    t838 = -TWO/THREE*t39*t781*t52*t375-TWO/THREE*t39*t781*t58*t332-TWO*t790*t438*
t711*t21*t331+FOUR*t4*t446*t141*t83*t13*t801-TWO/THREE*t806*t481*t807*t483*t482+
TWO*t814*t815*t459*t483*t461-TWO*t814*t815*t489*t494+TWO/THREE*t806*t481*t807*
t494*t461+TWO*EIGHT/THREE*t632*t372-EIGHT/THREE*t637*t377-FOUR*t651*t217-FOUR*t651*t221;
    t857 = t790*t123*t83*t13;
    t859 = t173*t450*t453;
    t860 = t18*t331;
    t873 = t247*t197;
    t886 = -EIGHT/THREE*t3*t4*t225*t141*t238*t242-TWO*t790*t76*t84*t13*t801-TWO*
t790*t93*t83*t99*t801+TWO*t857*t859*t860*t482-TWO*t857*t859*t860*t461-TWO/THREE*
t594*t264+TWO/THREE*t594*t369-EIGHT/THREE*t536*t873+EIGHT/THREE*t536*t392+EIGHT/THREE*t536*
t248-EIGHT/THREE*t637*t300-EIGHT/THREE*t632*t264+EIGHT/THREE*t632*t369;
    t916 = FOUR/THREE*t594*t372-TWO/THREE*t299*t377+TWO/THREE*t594*t381+TWO/THREE*t594*
t568-TWO/THREE*t261*t369-FOUR/THREE*t261*t372+TWO/THREE*t252*t247*t376-TWO/THREE*t261*
t381+EIGHT*t211*t671*t295+EIGHT/THREE*t211*t679*t613+EIGHT/THREE*t214*t192+EIGHT/THREE*t391*
t537;
    t929 = t118*t95*t562;
    t930 = t15*t337;
    t934 = t15*t21;
    t935 = t934*t380;
    t954 = t159*t173;
    t978 = EIGHT/THREE*t391*t873-TWO*t132*t134*t197+FOUR*t132*t134*t207+TWO*t132*
t134*t220+TWO*t929*t930*t54+TWO*t929*t935-FOUR*t929*t590-TWO*t929*t15*t18*t23*
t595+THREE*FOUR*t756*t679*t564+FOUR*t757*t206*t112+TWO*t152*t167*t105*t157*t954*t64*
t353*t176*t179+SIX*t152*t166*t78*t157*t954*t23*t27*t176*t179+TEN/THREE*t152*t77*
t76*t170*t954*t26*t176*t179;
    t986 = t7*t76;
    t994 = t986*t95*t673;
    t1030 = t109*t207;
    t1033 = TWO*t7*t77*t95*t11*t84*t198+TWO*t986*t131*t83*t99*t40*t197+TWO*t994
*t930*t263+FOUR*t994*t377-TWO*t994*t935-TWO*t994*t934*t567+TWO*EIGHT*t202*t143*t562*
t15*t89+FOUR*t184*t152*t1*t8*t10*t750*t161+FOUR*t360*t78*t12*t90+FOUR*t360*t96*
t673*t100*t89+FOUR*t396*t113+THREE*FOUR*t396*t529-FOUR*t396*t1030;
    t1041 = t99*t99;
    t1081 = -TWO/THREE*t246*t873+TWO/THREE*t246*t392+TWO/THREE*t7*t119*t93*t238*t1041
*t14*t18*t241+TWO/THREE*t411*t413*t112+TWO/THREE*t411*t413*t197-FOUR*t757*t206*t220
-THREE*FOUR*t117*t5*t226*t12*t205*t14*t89+EIGHT/THREE*t452*t457*t460*t482-FOUR*t107*t529+
TWO*t107*t1030+TWO*t107*t397-TWO/THREE*t75*t403*t79*t124*t14*t89-TWO/THREE*t246*
t537;
    return(t165+t275+t347+t400+t499+t571+t625+t658+t699+t779+t838+t886+t916+
t978+t1033+t1081);
  }
}


std::complex<DOUBLE> C2123mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t280;
  std::complex<DOUBLE> t285;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t298;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t302;
  std::complex<DOUBLE> t307;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t311;
  std::complex<DOUBLE> t313;
  std::complex<DOUBLE> t322;
  std::complex<DOUBLE> t325;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t347;
  std::complex<DOUBLE> t348;
  FourMomentum<DOUBLE> t35;
  std::complex<DOUBLE> t356;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t361;
  std::complex<DOUBLE> t364;
  std::complex<DOUBLE> t367;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t381;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t384;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t410;
  std::complex<DOUBLE> t417;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t431;
  std::complex<DOUBLE> t457;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t472;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  {
    t2 = IAB(p1,p3+p2,p1);
    t3 = p3+p2+p1;
    t4 = S(t3);
    K1f = p3+p2+p1-t2*t4*p1;
    t7 = A(K1f,p3);
    t8 = A(K1f,p4);
    t10 = A(p1,p5);
    t11 = t10*t10;
    t12 = t11*t11;
    t13 = B(K1f,p2);
    t14 = t13*t13;
    t15 = t14*t14;
    t16 = t12*t15;
    t18 = B(K1f,p6);
    t19 = IA(p1,p4);
    t20 = t19*t19;
    t22 = IA(p1,p6);
    t23 = IA(p4,p5);
    t24 = t22*t23;
    t25 = t18*t20*t24;
    t27 = IA(p5,p6);
    t28 = IB(K1f,p1);
    t29 = t27*t28;
    t30 = IB(K1f,p3);
    t31 = IB(p1,p2);
    t33 = t29*t30*t31;
    t34 = IB(p2,p3);
    t35 = p5+p4;
    t36 = IAB(p1,t35,p6);
    t39 = IAB(p3,p2+p1,K1f);
    t40 = IS(t3);
    t41 = t39*t40;
    t43 = t33*t34*t36*t41;
    t46 = A(K1f,p5);
    t48 = t10*t11;
    t49 = t48*t15;
    t51 = t18*t19;
    t52 = t51*t24;
    t56 = A(K1f,p6);
    t59 = t22*t22;
    t60 = t59*t23;
    t61 = t51*t60;
    t65 = t7*t12;
    t66 = A(p2,p3);
    t68 = t15*t18;
    t69 = t19*t22;
    t70 = t68*t69;
    t72 = t23*t27;
    t73 = t28*t30;
    t74 = t72*t73;
    t75 = t34*t2;
    t76 = t39*t39;
    t77 = t36*t76;
    t79 = t74*t75*t77;
    t82 = A(p1,p3);
    t83 = t82*t12;
    t84 = t14*t13;
    t85 = t83*t84;
    t88 = t29*t30*t34;
    t89 = AB(K1f,t35,p6);
    t90 = t89*t2;
    t91 = t36*t36;
    t92 = t91*t39;
    t93 = t90*t92;
    t97 = B(p1,p6);
    t98 = t97*t19;
    t99 = t98*t24;
    t101 = t2*t2;
    t102 = t101*t36;
    t104 = t102*t39*t4;
    t105 = t88*t104;
    t110 = B(p1,p2);
    t117 = B(p4,p6);
    t118 = t117*t117;
    t119 = t118*t118;
    t121 = IB(K1f,p4);
    t124 = t31*t34;
    t125 = IB(p4,p5);
    t127 = IB(p5,p6);
    t130 = IS(p6+p5+p4);
    t136 = t46*t82;
    t137 = t136*t49;
    t139 = t34*t89;
    t142 = t33*t139*t92*t40;
    t146 = t36*t39;
    t147 = t75*t146;
    t148 = t33*t147;
    t152 = t84*t18;
    t155 = t74*t147;
    t158 = t56*t56;
    t167 = t82*t82;
    t168 = t167*t167;
    t171 = IA(K1f,p1);
    t172 = t18*t171;
    t173 = IA(p1,p2);
    t175 = t172*t173*t19;
    t177 = IA(p2,p3);
    t181 = t27*t36*t41;
    t185 = TWO/THREE*t7*t8*t16*t25*t43-EIGHT/THREE*t7*t46*t49*t52*t43+TWO/THREE*t7*t56*
t16*t61*t43+FOUR/THREE*t65*t66*t70*t79+FOUR/THREE*t85*t52*t88*t93+FOUR/THREE*t85*t99*
t105+EIGHT*t83*t14*t18*t110*t19*t24*t105+TWO*t82*t15*t119*t73*t121*t124*t125*t127
*t36*t39*t130+EIGHT/THREE*t137*t52*t142-FOUR/THREE*t137*t99*t148-TWO*EIGHT/THREE*t136*t48*
t152*t69*t155-TWO/THREE*t158*t82*t16*t51*t59*t22*t23*t43-TWO*t56*t168*t12*t175*
t59*t177*t23*t181;
    t186 = t56*t82;
    t187 = t12*t66;
    t189 = t19*t59;
    t190 = t68*t189;
    t194 = t186*t16;
    t195 = B(p1,p3);
    t196 = t18*t195;
    t199 = t30*t30;
    t201 = t72*t28*t199;
    t204 = t124*t2*t36*t39;
    t205 = t201*t204;
    t215 = t168*t12;
    t217 = t171*t173;
    t221 = t22*t177*t23;
    t222 = t27*t2;
    t232 = t221*t181;
    t235 = t8*t82;
    t237 = t20*t22;
    t238 = t68*t237;
    t242 = t235*t16;
    t247 = B(p1,p4);
    t248 = t18*t247;
    t254 = t29*t30*t121*t204;
    t270 = t46*t46;
    t277 = -FOUR/THREE*t186*t187*t190*t79+TWO/THREE*t194*t196*t189*t205-TWO/THREE*t194
*t61*t142+FOUR/THREE*t194*t98*t60*t148-TWO*t215*t97*t217*t19*t221*t222*t146-TWO*t8
*t168*t12*t172*t173*t20*t232-FOUR/THREE*t235*t187*t238*t79+TWO/THREE*t242*t196*t237*
t205+TWO*t242*t248*t237*t23*t254-TWO/THREE*t242*t25*t142-TWO/THREE*t242*t97*t20*t24
*t148+FOUR/THREE*t235*t12*t152*t237*t155-FOUR*t270*t82*t11*t15*t52*t43;
    t280 = t82*t48;
    t285 = t74*t124*t146*t40;
    t298 = t196*t69;
    t302 = t69*t23;
    t307 = t66*t66;
    t311 = t18*t110*t69;
    t313 = t34*t101;
    t322 = t83*t66*t15;
    t325 = t313*t77*t4;
    t347 = t83*t68;
    t348 = t195*t195;
    t356 = t313*t146*t4;
    t361 = t195*t19*t24;
    t364 = t29*t199*t31;
    t367 = t139*t2*t91*t39;
    t380 = EIGHT/THREE*t46*t56*t280*t190*t285+EIGHT*t46*t168*t48*t175*t232+TWO*EIGHT/THREE*
t136*t48*t66*t70*t79-EIGHT/THREE*t137*t298*t205-FOUR*t137*t248*t302*t254+TEN/THREE*
t83*t307*t15*t311*t74*t313*t36*t76*t39*t4+FOUR/THREE*t322*t298*t201*t325-FOUR/THREE*
t322*t52*t88*t90*t91*t76-FOUR/THREE*t322*t99*t88*t102*t76*t4-SEVEN*FOUR/THREE*t83*t66*t84*
t311*t74*t325-TWO/THREE*t347*t348*t19*t24*t29*t199*t30*t31*t356+TWO/THREE*t347*t361
*t364*t367+TWO*t347*t247*t19*t24*t27*t73*t121*t31*t367;
    t381 = t69*t72;
    t383 = t73*t124;
    t384 = t89*t89;
    t410 = t65*t68;
    t417 = t89*t91*t41;
    t431 = t8*t8;
    t457 = t215*t18;
    t472 = -TWO/THREE*t347*t381*t383*t384*t91*t36*t41+TWO/THREE*t83*t15*t195*t99*
t364*t356-TWO/THREE*t83*t15*t97*t381*t383*t93-FOUR/THREE*t83*t152*t361*t29*t199*t34*
t104+FOUR/THREE*t410*t361*t364*t147+TWO/THREE*t410*t381*t383*t417+TWO/THREE*t65*t15*
t99*t148-TWO*EIGHT/THREE*t65*t84*t52*t29*t30*t147-TWO/THREE*t431*t82*t16*t18*t20*t19*t24
*t43+EIGHT/THREE*t8*t46*t280*t238*t285-TWO/THREE*t8*t56*t83*t68*t20*t59*t285+FOUR/THREE*
t186*t12*t152*t189*t155+TWO*t457*t110*t171*t173*t302*t222*t77-TWO*t457*t217*t69
*t177*t23*t27*t417;
    return(t185+t277+t380+t472);
  }
}


std::complex<DOUBLE> C2234mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t1001;
  std::complex<DOUBLE> t1007;
  std::complex<DOUBLE> t1039;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t1057;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t1097;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t1103;
  std::complex<DOUBLE> t1104;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t1114;
  std::complex<DOUBLE> t1118;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t1133;
  std::complex<DOUBLE> t1134;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t1173;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t269;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t276;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t283;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t295;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t301;
  std::complex<DOUBLE> t308;
  std::complex<DOUBLE> t309;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t313;
  std::complex<DOUBLE> t316;
  std::complex<DOUBLE> t317;
  std::complex<DOUBLE> t318;
  std::complex<DOUBLE> t319;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t324;
  std::complex<DOUBLE> t329;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t330;
  std::complex<DOUBLE> t331;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t333;
  std::complex<DOUBLE> t335;
  std::complex<DOUBLE> t336;
  std::complex<DOUBLE> t339;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t340;
  std::complex<DOUBLE> t342;
  std::complex<DOUBLE> t345;
  std::complex<DOUBLE> t346;
  std::complex<DOUBLE> t349;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t352;
  std::complex<DOUBLE> t354;
  std::complex<DOUBLE> t358;
  std::complex<DOUBLE> t359;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t363;
  std::complex<DOUBLE> t366;
  std::complex<DOUBLE> t367;
  std::complex<DOUBLE> t368;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t370;
  std::complex<DOUBLE> t374;
  std::complex<DOUBLE> t375;
  std::complex<DOUBLE> t379;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t380;
  std::complex<DOUBLE> t383;
  std::complex<DOUBLE> t386;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t389;
  FourMomentum<DOUBLE> t39;
  std::complex<DOUBLE> t392;
  std::complex<DOUBLE> t395;
  std::complex<DOUBLE> t398;
  std::complex<DOUBLE> t399;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t401;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t409;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t413;
  std::complex<DOUBLE> t415;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t427;
  std::complex<DOUBLE> t429;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t434;
  std::complex<DOUBLE> t435;
  std::complex<DOUBLE> t437;
  std::complex<DOUBLE> t438;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t442;
  std::complex<DOUBLE> t445;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t453;
  std::complex<DOUBLE> t454;
  std::complex<DOUBLE> t455;
  std::complex<DOUBLE> t458;
  std::complex<DOUBLE> t461;
  std::complex<DOUBLE> t464;
  std::complex<DOUBLE> t467;
  std::complex<DOUBLE> t468;
  std::complex<DOUBLE> t473;
  std::complex<DOUBLE> t476;
  std::complex<DOUBLE> t479;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t484;
  std::complex<DOUBLE> t485;
  std::complex<DOUBLE> t489;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t501;
  std::complex<DOUBLE> t502;
  std::complex<DOUBLE> t515;
  std::complex<DOUBLE> t516;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t520;
  std::complex<DOUBLE> t521;
  std::complex<DOUBLE> t523;
  std::complex<DOUBLE> t524;
  std::complex<DOUBLE> t525;
  std::complex<DOUBLE> t529;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t530;
  std::complex<DOUBLE> t534;
  std::complex<DOUBLE> t536;
  std::complex<DOUBLE> t537;
  std::complex<DOUBLE> t538;
  std::complex<DOUBLE> t539;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t543;
  std::complex<DOUBLE> t548;
  std::complex<DOUBLE> t549;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t554;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t563;
  std::complex<DOUBLE> t570;
  std::complex<DOUBLE> t571;
  std::complex<DOUBLE> t578;
  std::complex<DOUBLE> t585;
  std::complex<DOUBLE> t586;
  std::complex<DOUBLE> t589;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t601;
  std::complex<DOUBLE> t602;
  std::complex<DOUBLE> t603;
  std::complex<DOUBLE> t604;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t612;
  std::complex<DOUBLE> t625;
  std::complex<DOUBLE> t626;
  std::complex<DOUBLE> t628;
  std::complex<DOUBLE> t632;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t641;
  std::complex<DOUBLE> t644;
  std::complex<DOUBLE> t649;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t660;
  std::complex<DOUBLE> t661;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t673;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t685;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t704;
  std::complex<DOUBLE> t707;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t734;
  std::complex<DOUBLE> t735;
  std::complex<DOUBLE> t736;
  std::complex<DOUBLE> t754;
  std::complex<DOUBLE> t755;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t760;
  std::complex<DOUBLE> t765;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t772;
  std::complex<DOUBLE> t775;
  std::complex<DOUBLE> t782;
  std::complex<DOUBLE> t784;
  std::complex<DOUBLE> t786;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t800;
  std::complex<DOUBLE> t801;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t813;
  std::complex<DOUBLE> t823;
  std::complex<DOUBLE> t829;
  std::complex<DOUBLE> t830;
  std::complex<DOUBLE> t839;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t844;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t857;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t864;
  std::complex<DOUBLE> t865;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t877;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t884;
  std::complex<DOUBLE> t885;
  std::complex<DOUBLE> t886;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t902;
  std::complex<DOUBLE> t904;
  std::complex<DOUBLE> t913;
  std::complex<DOUBLE> t922;
  std::complex<DOUBLE> t928;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t932;
  std::complex<DOUBLE> t933;
  std::complex<DOUBLE> t938;
  std::complex<DOUBLE> t942;
  std::complex<DOUBLE> t945;
  std::complex<DOUBLE> t951;
  std::complex<DOUBLE> t954;
  std::complex<DOUBLE> t955;
  std::complex<DOUBLE> t977;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t980;
  std::complex<DOUBLE> t981;
  std::complex<DOUBLE> t982;
  std::complex<DOUBLE> t986;
  std::complex<DOUBLE> t987;
  std::complex<DOUBLE> t99;
  {
    t2 = IAB(p2,p4+p3,p2);
    t3 = p4+p3+p2;
    t4 = S(t3);
    K1f = p4+p3+p2-t2*t4*p2;
    t7 = A(K1f,p4);
    t8 = A(p1,p2);
    t9 = t7*t8;
    t10 = A(p2,p3);
    t11 = t10*t10;
    t12 = B(K1f,p4);
    t13 = t11*t12;
    t14 = t9*t13;
    t15 = B(K1f,p6);
    t16 = t15*t15;
    t17 = t16*t15;
    t18 = B(p1,p2);
    t19 = t17*t18;
    t20 = B(p2,p6);
    t21 = IA(K1f,p2);
    t22 = t20*t21;
    t23 = IA(p2,p4);
    t24 = t23*t23;
    t25 = t22*t24;
    t28 = IA(p3,p4);
    t29 = IB(K1f,p1);
    t30 = t29*t29;
    t31 = t28*t30;
    t32 = IB(K1f,p5);
    t33 = IB(p1,p6);
    t34 = t32*t33;
    t35 = t31*t34;
    t36 = IB(p3,p4);
    t37 = IB(p5,p6);
    t38 = t36*t37;
    t39 = p6+p5;
    t40 = IAB(p1,t39,K1f);
    t41 = t2*t2;
    t43 = t40*t41*t4;
    t44 = t38*t43;
    t45 = t35*t44;
    t48 = B(p2,p5);
    t49 = t17*t48;
    t52 = t28*t29;
    t53 = t32*t32;
    t54 = t53*t33;
    t55 = t52*t54;
    t56 = t55*t44;
    t59 = t17*t20;
    t60 = t21*t24;
    t61 = t60*t28;
    t64 = t29*t32;
    t65 = t33*t36;
    t66 = t64*t65;
    t67 = AB(p1,t39,p2);
    t68 = t37*t67;
    t69 = t40*t40;
    t71 = t69*t41*t4;
    t72 = t68*t71;
    t73 = t66*t72;
    t76 = t20*t20;
    t77 = t16*t76;
    t80 = t52*t34;
    t81 = t80*t44;
    t84 = t16*t16;
    t85 = t11*t84;
    t86 = t9*t85;
    t87 = B(p2,p4);
    t88 = t18*t87;
    t93 = t87*t48;
    t98 = t87*t21;
    t99 = t24*t28;
    t100 = t98*t99;
    t104 = A(K1f,p5);
    t105 = A(p1,p5);
    t106 = t105*t105;
    t107 = t106*t106;
    t111 = IA(p1,p6);
    t112 = t12*t21*t111;
    t114 = IA(p2,p5);
    t115 = t114*t114;
    t116 = t23*t115;
    t118 = IA(p5,p6);
    t121 = IS(p6+p5+p1);
    t122 = t40*t121;
    t123 = t118*t36*t122;
    t127 = t8*t11;
    t128 = t12*t84;
    t129 = t127*t128;
    t130 = t18*t18;
    t133 = t23*t28;
    t136 = t30*t30;
    t139 = t37*t40;
    t140 = t41*t2;
    t141 = t4*t4;
    t142 = t140*t141;
    t143 = t139*t142;
    t148 = t21*t23;
    t151 = t30*t29;
    t152 = t28*t151;
    t156 = t38*t40*t140*t141;
    t161 = t130*t21*t133;
    t163 = t151*t32;
    t164 = t163*t65;
    t167 = t68*t69*t140*t141;
    t171 = EIGHT/THREE*t14*t19*t25*t45-EIGHT/THREE*t14*t49*t25*t56-EIGHT/THREE*t14*t59*t61*
t73+FOUR*t14*t77*t60*t81-FOUR/THREE*t86*t88*t60*t45+FOUR/THREE*t86*t93*t60*t56+FOUR/THREE
*t86*t100*t73+TWO*t104*t107*t11*t112*t116*t28*t123+TWO/THREE*t129*t130*t18*t21*
t133*t136*t32*t65*t143-TWO/THREE*t129*t130*t48*t148*t152*t54*t156-TWO/THREE*t129*
t161*t164*t167;
    t172 = t48*t48;
    t176 = t53*t32;
    t177 = t176*t33;
    t182 = t18*t48;
    t183 = t148*t28;
    t184 = t182*t183;
    t186 = t30*t53;
    t191 = t18*t21;
    t192 = t191*t133;
    t194 = t30*t32;
    t195 = t194*t65;
    t196 = t67*t67;
    t197 = t37*t196;
    t198 = t69*t40;
    t201 = t197*t198*t140*t141;
    t209 = t53*t53;
    t216 = t172*t21*t133;
    t218 = t29*t176;
    t219 = t218*t65;
    t223 = t48*t21;
    t224 = t223*t133;
    t226 = t29*t53;
    t227 = t226*t65;
    t231 = t148*t52;
    t233 = t34*t38;
    t235 = t69*t69;
    t241 = t8*t84;
    t242 = t87*t87;
    t243 = t242*t242;
    t245 = IA(p2,p3);
    t247 = IB(K1f,p2);
    t250 = IB(p2,p3);
    t252 = t36*t36;
    t253 = t252*t37;
    t254 = IS(t3);
    t255 = t40*t254;
    t262 = t252*t36;
    t263 = t34*t262;
    t265 = AB(K1f,p3+p2,p4);
    t266 = t265*t265;
    t267 = t37*t266;
    t268 = t40*t2;
    t269 = t267*t268;
    t273 = t107*t11;
    t275 = t21*t111;
    t276 = t23*t114;
    t277 = t275*t276;
    t279 = t28*t118;
    t282 = t69*t2;
    t283 = t121*t4;
    t295 = TWO/THREE*t129*t18*t172*t148*t31*t177*t156+TWO/THREE*t129*t184*t186*t65*
t167+TWO/THREE*t129*t192*t195*t201-TWO/THREE*t129*t172*t48*t21*t133*t29*t209*t65*
t143-TWO/THREE*t129*t216*t219*t167-TWO/THREE*t129*t224*t227*t201-TWO/THREE*t129*t231*
t233*t196*t67*t235*t142+TWO*t241*t243*t245*t29*t247*t34*t250*t253*t255+TWO/THREE*
t241*t87*t231*t263*t269-TWO*t273*t12*t277*t279*t36*t67*t282*t283+TWO*t273*t87*
t277*t279*t36*t268*t283;
    t301 = t114*t28;
    t308 = t12*t16;
    t309 = t127*t308;
    t313 = t35*t156;
    t316 = t12*t17;
    t317 = t127*t316;
    t318 = t48*t20;
    t319 = t318*t183;
    t321 = t227*t167;
    t324 = t22*t133;
    t329 = A(K1f,p1);
    t330 = t329*t7;
    t331 = t330*t13;
    t332 = t84*t18;
    t333 = t332*t60;
    t335 = t38*t268;
    t336 = t35*t335;
    t339 = t84*t48;
    t340 = t339*t60;
    t342 = t55*t335;
    t345 = t84*t21;
    t346 = t345*t99;
    t349 = t66*t68*t282;
    t352 = t59*t60;
    t354 = t80*t335;
    t358 = t84*t87;
    t359 = t358*t60;
    t363 = t10*t12;
    t366 = t33*t252;
    t367 = t64*t366;
    t368 = t37*t265;
    t370 = t367*t368*t255;
    t374 = t329*t104*t13;
    t375 = t339*t148;
    t379 = t65*t139*t2;
    t380 = t301*t226*t379;
    t383 = TWO*t107*t10*t12*t275*t23*t301*t118*t252*t265*t122+FOUR*t309*t18*t76*
t148*t313+EIGHT/THREE*t317*t319*t321+EIGHT/THREE*t317*t324*t66*t201-FOUR/THREE*t331*t333*
t336-TWO/THREE*t331*t340*t342-TWO/THREE*t331*t346*t349-FOUR/THREE*t331*t352*t354-FOUR/
THREE*t330*t11*t359*t354+TWO/THREE*t330*t363*t346*t370-TWO*t374*t375*t380;
    t386 = t67*t69;
    t387 = t386*t2;
    t389 = t80*t38*t387;
    t392 = t59*t148;
    t395 = t301*t64*t379;
    t398 = t329*t11;
    t399 = t398*t128;
    t401 = t41*t4;
    t402 = t139*t401;
    t409 = t55*t156;
    t413 = t76*t21*t133;
    t415 = t66*t167;
    t424 = t66*t143;
    t427 = t84*t130;
    t429 = t98*t133;
    t434 = t127*t332;
    t435 = t93*t148;
    t437 = t31*t54;
    t438 = t437*t156;
    t442 = t195*t167;
    t445 = t127*t358;
    t453 = A(K1f,p3);
    t454 = t453*t8;
    t455 = t454*t363;
    t458 = t195*t72;
    t461 = -TWO*t374*t345*t276*t389+FOUR*t374*t392*t395-TEN/THREE*t399*t161*t164*
t402-FOUR*t309*t48*t76*t148*t409-FOUR*t309*t413*t415+EIGHT/THREE*t127*t12*t15*t76*t20
*t21*t133*t424+TWO/THREE*t127*t427*t429*t164*t143-TWO/THREE*t434*t435*t438-TWO/THREE*
t434*t429*t442+TWO/THREE*t445*t216*t219*t143+TWO/THREE*t445*t224*t321+EIGHT/THREE*t455*
t332*t183*t458;
    t464 = t84*t172;
    t467 = t52*t177;
    t468 = t467*t44;
    t473 = t227*t72;
    t476 = t345*t133;
    t479 = t198*t41*t4;
    t481 = t66*t197*t479;
    t484 = t22*t23;
    t485 = t19*t484;
    t489 = t49*t484;
    t501 = t10*t84;
    t502 = t454*t501;
    t515 = t87*t20;
    t516 = t515*t148;
    t520 = -EIGHT/THREE*t455*t464*t148*t468-EIGHT/THREE*t455*t339*t183*t473-EIGHT/THREE*
t455*t476*t481-FOUR*EIGHT/THREE*t455*t485*t45+FOUR*EIGHT/THREE*t455*t489*t56+FOUR*EIGHT/THREE*t455*t59*
t183*t73-TWO*EIGHT*t455*t77*t148*t81+FOUR/THREE*t502*t88*t148*t45-FOUR/THREE*t502*t435*t56
-FOUR/THREE*t502*t429*t73+TWO*EIGHT/THREE*t454*t10*t17*t516*t81;
    t521 = t454*t128;
    t523 = t194*t366;
    t524 = t368*t268;
    t525 = t523*t524;
    t529 = t226*t366;
    t530 = t529*t524;
    t534 = t21*t245*t133;
    t536 = t33*t262;
    t537 = t64*t536;
    t538 = t267*t255;
    t539 = t537*t538;
    t543 = t196*t198;
    t548 = t398*t316;
    t549 = t18*t20;
    t554 = t318*t148;
    t563 = t66*t402;
    t570 = t17*t87;
    t571 = t127*t570;
    t578 = -EIGHT/THREE*t521*t192*t525+EIGHT/THREE*t521*t224*t530-EIGHT/THREE*t521*t534*
t539+TWO/THREE*t399*t231*t233*t543*t401-SEVEN*FOUR/THREE*t548*t549*t148*t45+FOUR/THREE*t548*
t554*t56+FOUR/THREE*t548*t324*t73-EIGHT*t398*t308*t413*t563+EIGHT/THREE*t127*t19*t516*
t313-EIGHT/THREE*t571*t554*t409-EIGHT/THREE*t571*t324*t415;
    t585 = t8*t10;
    t586 = t585*t128;
    t589 = t368*t43;
    t601 = t34*t253;
    t602 = t265*t67;
    t603 = t602*t71;
    t604 = t601*t603;
    t612 = t133*t29;
    t625 = t585*t316;
    t626 = t549*t183;
    t628 = t523*t589;
    t632 = t529*t589;
    t641 = t367*t589;
    t644 = FOUR*t127*t16*t87*t413*t424+TWO/THREE*t586*t161*t163*t366*t589-TWO/THREE*
t586*t184*t186*t366*t589-TWO/THREE*t586*t191*t133*t30*t604+TWO/THREE*t586*t216*t218
*t366*t589+TWO/THREE*t586*t223*t612*t54*t253*t603+TWO/THREE*t586*t231*t601*t265*
t196*t479+EIGHT/THREE*t625*t626*t628-EIGHT/THREE*t625*t319*t632-EIGHT/THREE*t625*t22*t612*
t604+FOUR*t585*t308*t413*t641;
    t649 = t585*t358;
    t660 = t8*t12;
    t661 = t660*t84;
    t673 = t252*t252;
    t685 = t152*t34;
    t704 = t437*t44;
    t707 = TWO/THREE*t585*t332*t429*t628-TWO/THREE*t649*t224*t632-TWO/THREE*t649*t231
*t604+EIGHT/THREE*t585*t570*t324*t641+TWO/THREE*t661*t192*t194*t536*t269-TWO/THREE*t661
*t224*t226*t536*t269+TWO/THREE*t661*t534*t64*t33*t673*t37*t266*t265*t255+EIGHT/THREE*
t317*t130*t20*t148*t685*t156-EIGHT/THREE*t317*t182*t484*t438-EIGHT/THREE*t317*t626*t442
+EIGHT/THREE*t317*t172*t20*t148*t467*t156+FOUR/THREE*t399*t182*t148*t704;
    t734 = t453*t7;
    t735 = t734*t585;
    t736 = t128*t60;
    t754 = t453*t104*t585;
    t755 = t223*t23;
    t760 = t148*t114;
    t765 = FOUR/THREE*t399*t192*t458+TWO/THREE*t399*t216*t219*t402+TWO/THREE*t399*t224
*t473-TWO/THREE*t661*t231*t34*t262*t37*t266*t67*t282+EIGHT/THREE*t660*t17*t324*t537*
t269+EIGHT/THREE*t735*t736*t389-FOUR*EIGHT/THREE*t735*t316*t22*t99*t64*t379+SEVEN*FOUR/THREE*t735*
t359*t354-EIGHT/THREE*t734*t660*t346*t370+EIGHT*t754*t128*t755*t380+EIGHT*t754*t128*
t760*t389;
    t772 = t685*t44;
    t775 = t332*t755;
    t782 = t104*t104;
    t784 = t782*t8*t13;
    t786 = t115*t28;
    t800 = t104*t8;
    t801 = t800*t13;
    t813 = t398*t358;
    t823 = -TWO*EIGHT*t754*t316*t484*t395-EIGHT/THREE*t455*t427*t148*t772+EIGHT/THREE*t455*
t775*t704+TWO/THREE*t14*t346*t481-TWO*t784*t375*t786*t226*t379-TWO*t784*t345*t116
*t389+FOUR*t784*t392*t786*t64*t379-TWO*t801*t775*t301*t186*t33*t44-FOUR/THREE*t398*
t332*t429*t195*t402-TWO/THREE*t813*t224*t227*t402-TWO/THREE*t813*t231*t233*t386*
t401;
    t829 = t329*t10;
    t830 = t829*t128;
    t839 = t601*t602*t282;
    t844 = t367*t524;
    t857 = t453*t453;
    t864 = t38*t255;
    t865 = t80*t864;
    t877 = t339*t760;
    t884 = -FOUR/THREE*t398*t570*t324*t563-FOUR/THREE*t830*t192*t525-TWO/THREE*t830*
t224*t530-TWO/THREE*t830*t231*t839-FOUR/THREE*t829*t316*t324*t844+TWO/THREE*t829*t84*
t429*t844+TWO/THREE*t329*t12*t84*t231*t263*t538-EIGHT/THREE*t857*t453*t8*t12*t345*
t245*t23*t865+FOUR*t857*t7*t8*t736*t865+FOUR*t801*t464*t760*t468+FOUR*t801*t877*
t52*t54*t36*t72;
    t885 = t276*t28;
    t886 = t345*t885;
    t902 = t59*t760;
    t904 = t34*t36;
    t913 = t800*t85;
    t922 = t11*t17;
    t928 = t800*t363;
    t932 = t253*t265*t40*t2;
    t933 = t55*t932;
    t938 = t367*t368*t387;
    t942 = t80*t932;
    t945 = t329*t453;
    t951 = FOUR*t801*t886*t481+FOUR*t801*t485*t301*t194*t33*t44-THREE*FOUR*t801*t489*
t301*t226*t33*t44-THREE*FOUR*t801*t902*t52*t904*t72+THREE*FOUR*t801*t77*t760*t81-TWO*t913*
t93*t760*t56-TWO*t913*t98*t885*t73+FOUR*t800*t922*t515*t760*t81-TWO*t928*t877*
t933-TWO*t928*t886*t938+FOUR*t928*t902*t942-EIGHT/THREE*t945*t7*t10*t736*t865;
    t954 = t945*t363;
    t955 = t332*t148;
    t977 = t7*t7;
    t980 = t24*t23;
    t981 = t21*t980;
    t982 = t128*t981;
    t986 = t857*t8;
    t987 = t986*t12;
    t1001 = TWO*EIGHT/THREE*t954*t955*t336+EIGHT/THREE*t954*t375*t342+EIGHT/THREE*t954*t476*
t349+TWO*EIGHT/THREE*t954*t392*t354+FOUR/THREE*t945*t10*t358*t148*t354-EIGHT/THREE*t945*t12*
t476*t370+TWO/THREE*t329*t977*t11*t982*t865+FOUR*t987*t955*t336-FOUR*t987*t375*t342
+FOUR*t986*t128*t534*t370-FOUR*t987*t476*t349;
    t1007 = t64*t33;
    t1039 = t9*t363;
    t1057 = TWO*EIGHT*t987*t392*t354-EIGHT*t986*t84*t429*t1007*t335-EIGHT/THREE*t453*t977*
t585*t982*t865-EIGHT/THREE*t735*t128*t191*t99*t194*t379+EIGHT/THREE*t735*t128*t223*t99*
t226*t379+FOUR*t329*t857*t12*t476*t1007*t864-TWO*EIGHT/THREE*t9*t922*t515*t60*t81+TWO/
THREE*t1039*t333*t35*t932-TWO/THREE*t1039*t340*t933-TWO/THREE*t1039*t346*t938+EIGHT/THREE
*t1039*t352*t942-FOUR/THREE*t9*t501*t100*t844;
    t1097 = t24*t24;
    t1103 = t977*t8;
    t1104 = t1103*t13;
    t1114 = t345*t980*t28;
    t1118 = TWO/THREE*t9*t12*t346*t539+TWO*t7*t107*t11*t112*t24*t114*t28*t123-TWO
*t801*t332*t760*t31*t904*t72+EIGHT/THREE*t521*t231*t839-FOUR*EIGHT/THREE*t454*t316*t324*
t844+FOUR/THREE*t454*t84*t429*t844-EIGHT*t453*t107*t10*t112*t885*t123+TWO/THREE*t977*
t7*t8*t11*t128*t21*t1097*t865+TWO/THREE*t1104*t332*t981*t336-TWO/THREE*t1104*t339*
t981*t342-TWO/THREE*t1104*t1114*t349;
    t1133 = t7*t104*t127;
    t1134 = t223*t24;
    t1173 = EIGHT/THREE*t1104*t59*t981*t354-TEN/THREE*t1103*t11*t358*t981*t354+TWO/
THREE*t1103*t363*t1114*t370-TWO*t1133*t128*t1134*t380-TWO*t1133*t128*t60*t114*
t389+FOUR*t1133*t316*t25*t395+TWO/THREE*t14*t427*t60*t772-TWO/THREE*t14*t332*t1134*
t704-TWO/THREE*t14*t332*t61*t458+TWO/THREE*t14*t464*t60*t468+TWO/THREE*t14*t339*t61*
t473+TWO/THREE*t445*t231*t233*t543*t142;
    return(t171+t295+t383+t461+t520+t578+t644+t707+t765+t823+t884+t951+t1001+
t1057+t1118+t1173);
  }
}


std::complex<DOUBLE> C2345mpmpmpbub(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  FourMomentum<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t154;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t273;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t282;
  std::complex<DOUBLE> t287;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t297;
  std::complex<DOUBLE> t299;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t300;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t310;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t321;
  std::complex<DOUBLE> t326;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t332;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t341;
  std::complex<DOUBLE> t343;
  std::complex<DOUBLE> t346;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t369;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t379;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t387;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t390;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t400;
  std::complex<DOUBLE> t402;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t414;
  std::complex<DOUBLE> t424;
  std::complex<DOUBLE> t426;
  std::complex<DOUBLE> t427;
  std::complex<DOUBLE> t439;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t452;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t481;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t97;
  {
    t1 = p5+p4;
    t2 = IAB(p3,t1,p3);
    t3 = p5+p4+p3;
    t4 = S(t3);
    K1f = p5+p4+p3-t2*t4*p3;
    t7 = A(K1f,p1);
    t8 = t7*t7;
    t9 = A(p1,p3);
    t10 = t9*t9;
    t12 = A(p3,p5);
    t13 = B(K1f,p2);
    t14 = t12*t13;
    t16 = B(K1f,p4);
    t17 = t16*t16;
    t18 = t17*t17;
    t19 = IA(p1,p2);
    t20 = t18*t19;
    t21 = IA(p1,p6);
    t22 = IA(p2,p3);
    t23 = t22*t21;
    t24 = t20*t23;
    t26 = IA(p3,p6);
    t27 = IB(K1f,p3);
    t28 = t27*t26;
    t29 = IB(K1f,p5);
    t30 = IB(p3,p4);
    t31 = t30*t29;
    t32 = t28*t31;
    t33 = IB(p4,p5);
    t34 = IAB(p3,t1,p2);
    t35 = t33*t34;
    t37 = IAB(p5,p4+p3,K1f);
    t38 = IS(t3);
    t39 = t37*t38;
    t40 = t35*t39;
    t41 = t32*t40;
    t44 = A(K1f,p2);
    t46 = t10*t9;
    t47 = t46*t12;
    t49 = t13*t18;
    t50 = t19*t21;
    t51 = t49*t50;
    t53 = t22*t22;
    t54 = t53*t26;
    t55 = t27*t29;
    t56 = t54*t55;
    t57 = t30*t33;
    t60 = t57*t34*t37*t38;
    t64 = A(K1f,p5);
    t71 = A(K1f,p6);
    t75 = t26*t26;
    t76 = t22*t75;
    t77 = t76*t55;
    t81 = t10*t10;
    t82 = t81*t12;
    t83 = t82*t49;
    t84 = B(p3,p5);
    t86 = t84*t19*t23;
    t88 = t29*t29;
    t90 = t28*t88*t30;
    t92 = AB(K1f,p6+p1,p2);
    t93 = t33*t92;
    t94 = t34*t34;
    t97 = t93*t94*t2*t37;
    t101 = B(p3,p6);
    t106 = IB(K1f,p6);
    t112 = t22*t26;
    t113 = t50*t112;
    t115 = t55*t57;
    t116 = t92*t92;
    t123 = t17*t16;
    t124 = t123*t13;
    t129 = t2*t2;
    t130 = t34*t129;
    t132 = t130*t37*t4;
    t138 = t123*t19*t23;
    t141 = t28*t29*t33;
    t142 = t92*t94;
    t143 = t2*t37;
    t144 = t142*t143;
    t150 = B(p3,p4);
    t154 = t141*t132;
    t157 = B(p2,p3);
    t159 = t82*t18*t157;
    t163 = t35*t129*t37*t4;
    t172 = t157*t19;
    t173 = t172*t23;
    t177 = FOUR*t8*t10*t14*t24*t41-EIGHT/THREE*t7*t44*t47*t51*t56*t60-EIGHT/THREE*t7*t64
*t46*t13*t24*t41+EIGHT/THREE*t7*t71*t47*t51*t77*t60+TWO/THREE*t83*t86*t90*t97+TWO*t83
*t101*t19*t23*t26*t55*t106*t30*t97+TWO/THREE*t83*t113*t115*t116*t94*t34*t39+FOUR/
THREE*t82*t124*t86*t28*t88*t33*t132+FOUR/THREE*t82*t13*t138*t141*t144-EIGHT*t82*t13*
t17*t150*t19*t23*t154+TWO/THREE*t159*t86*t90*t163+TWO/THREE*t159*t113*t115*t144+FOUR
/THREE*t82*t123*t173*t154;
    t178 = t12*t18;
    t179 = B(p2,p6);
    t180 = t179*t179;
    t181 = t180*t180;
    t185 = IB(p1,p2);
    t186 = IB(p1,p6);
    t190 = IS(p6+p2+p1);
    t196 = t64*t81;
    t197 = t196*t49;
    t199 = t35*t143;
    t204 = t142*t39;
    t216 = t32*t199;
    t219 = t71*t71;
    t229 = t71*t81;
    t230 = t12*t12;
    t231 = t230*t230;
    t233 = IA(K1f,p3);
    t235 = t13*t233*t50;
    t237 = IA(p3,p4);
    t238 = t22*t237;
    t240 = IA(p4,p5);
    t241 = t240*t34;
    t242 = t241*t39;
    t246 = A(p4,p5);
    t247 = t12*t246;
    t250 = t37*t37;
    t251 = t2*t250;
    t252 = t35*t251;
    t256 = t229*t14;
    t258 = t18*t84*t50;
    t260 = t27*t88;
    t264 = t57*t34*t2*t37;
    t270 = t18*t101*t50*t22;
    t272 = t75*t27;
    t273 = t29*t106;
    t279 = t272*t31;
    t282 = t93*t94*t37*t38;
    t287 = t124*t50;
    t297 = TWO*t178*t181*t55*t106*t185*t186*t30*t35*t37*t190-FOUR/THREE*t197*t86*
t90*t199+TWO/THREE*t197*t113*t115*t204+TWO*EIGHT/THREE*t196*t13*t138*t28*t29*t199+TWO/
THREE*t196*t18*t173*t216+TWO/THREE*t219*t81*t14*t24*t75*t26*t27*t31*t40+TWO*t229*
t231*t235*t238*t75*t242+FOUR/THREE*t229*t247*t51*t77*t252-TWO/THREE*t256*t258*t76*
t260*t264-TWO*t256*t270*t272*t273*t264-TWO/THREE*t256*t24*t279*t282-FOUR/THREE*t229*
t12*t287*t77*t199-TWO/THREE*t229*t178*t173*t279*t199;
    t299 = t81*t231;
    t300 = t299*t13;
    t310 = t233*t19;
    t321 = t238*t26;
    t326 = t246*t246;
    t332 = t112*t55;
    t341 = t82*t246*t13;
    t343 = t112*t260;
    t346 = t35*t129*t250*t4;
    t369 = t84*t84;
    t379 = t7*t46;
    t387 = t332*t252;
    t390 = t379*t14;
    t400 = -TWO*t300*t150*t233*t19*t23*t237*t26*t34*t251-TWO*t300*t310*t23*t237
*t26*t240*t204-TWO*t299*t157*t310*t21*t321*t241*t143-TEN/THREE*t82*t326*t13*t18*
t150*t50*t332*t35*t129*t250*t37*t4-FOUR/THREE*t341*t258*t343*t346-FOUR/THREE*t341*t24
*t141*t142*t251+SEVEN*FOUR/THREE*t341*t123*t150*t50*t332*t346-FOUR/THREE*t82*t246*t18*t173
*t141*t130*t250*t4+TWO/THREE*t83*t369*t19*t23*t28*t88*t29*t30*t163+EIGHT*t379*t231*
t235*t321*t242+TWO*EIGHT/THREE*t379*t247*t51*t387-EIGHT/THREE*t390*t258*t343*t264-FOUR*t390
*t270*t28*t273*t264;
    t402 = t32*t282;
    t414 = t44*t44;
    t424 = t81*t13;
    t426 = t21*t53;
    t427 = t20*t426;
    t439 = t44*t81;
    t452 = t439*t14;
    t481 = -EIGHT/THREE*t390*t24*t402-TWO*EIGHT/THREE*t379*t12*t287*t332*t199+FOUR/THREE*t379
*t178*t173*t216+TWO/THREE*t414*t81*t14*t20*t21*t53*t22*t41+TWO/THREE*t44*t64*t424*
t427*t41-TWO/THREE*t44*t71*t82*t51*t53*t75*t55*t60-TWO*t439*t231*t235*t53*t237*
t26*t242-FOUR/THREE*t439*t247*t51*t56*t252+TWO/THREE*t452*t258*t54*t260*t264+FOUR/THREE
*t439*t12*t287*t56*t199+TWO/THREE*t452*t427*t402-FOUR/THREE*t439*t178*t172*t426*t216
-TWO/THREE*t64*t71*t424*t24*t279*t40-FOUR/THREE*t196*t246*t51*t387;
    return(t177+t297+t400+t481);
  }
}


std::complex<DOUBLE> C212mpmpmpt1(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p2;
    t1 = A(p1,p2);
    t2 = B(p1,p2);
    S1 = -t2*t1;
    gg = AB(p1,K1f,p1);
    S2 = S(p1);
    t4 = AB(K1f,p1,K1f);
    t5 = gg*t4;
    t6 = AB(p1,p1,p1);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p1,p1,K1f);
    t19 = ONE/S1;
    t21 = AB(p1,p1,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C212mpmpmpt6(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  FourMomentum<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  FourMomentum<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t1 = A(p1,p2);
    t2 = B(p1,p2);
    S1 = -t2*t1;
    gg = AB(p1,p2,p1);
    S2 = S(p6);
    t4 = AB(p2,p6,p2);
    t5 = gg*t4;
    t6 = AB(p1,p6,p1);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p1,p6,p2);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p1,p6,p2);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p6);
    t27 = t26*t26;
    t28 = B(p2,p6);
    t29 = t28*t28;
    t32 = IA(p1,p2);
    t33 = t32*t32;
    t34 = t32*t33;
    t35 = IB(p1,p2);
    t36 = t35*t35;
    t37 = t36*t35;
    t39 = t27*t29*t28*t34*t37;
    t40 = IB(p3,p4);
    t41 = IB(p4,p5);
    t42 = t40*t41;
    t43 = p2+p1;
    t44 = IAB(p6,t43,p6);
    t45 = t44*t44;
    t46 = t45*t45;
    t47 = FbtAB(p1,p2,t43,p1,YY01);
    t48 = t47*t47;
    t52 = FbtAB(p1,p6,t43,p1,YY01);
    t53 = FbtAB(p3,p4,t43,p1,YY01);
    t54 = t53*t53;
    t56 = FbtAB(p5,p4,t43,p1,YY01);
    t57 = t56*t56;
    t58 = IFbtAB(p3,p3,t43,p1,YY01);
    t61 = IFbtAB(p5,p5,t43,p1,YY01);
    t62 = p4+p3;
    t63 = IFbtAA(p3,t62,p5,t43,p1,YY01);
    t64 = t61*t63;
    t66 = I/YY01;
    t67 = t64*t66;
    t71 = A(p3,p5);
    t72 = t71*t71;
    t73 = t72*t72;
    t76 = IA(p3,p4);
    t77 = IA(p4,p5);
    t80 = t73*t28*t33*t76*t77*t36;
    t82 = IS(p5+p4+p3);
    t83 = t45*t82;
    t92 = t28*t34*t37*t40;
    t93 = t41*t45;
    t95 = t93*t48*t52;
    t97 = t54*t57;
    t98 = FbtAB2(p1,p1,t43,p1,YY01,YY21);
    t101 = YY01*YY01;
    t108 = t28*t33;
    t110 = t108*t36*t40;
    t111 = t110*t95;
    t112 = FbtAB2(p3,p3,t43,p1,YY01,YY21);
    t113 = t58*t58;
    t119 = FbtAB2(p5,p5,t43,p1,YY01,YY21);
    t122 = t61*t61;
    t128 = FbtAA2(p3,t62,p5,t43,p1,YY01,YY21);
    t131 = t63*t63;
    t138 = FbtAB2(p5,p4,t43,p1,YY01,YY21);
    t145 = FbtAB2(p3,p4,t43,p1,YY01,YY21);
    t151 = t108*t36;
    t155 = FbtAB2(p1,p6,t43,p1,YY01,YY21);
    t164 = FbtAB2(p1,p2,t43,p1,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t176 = FbtAB(p1,p2,t43,p1,YY0);
    t177 = t176*t176;
    t181 = FbtAB(p1,p6,t43,p1,YY0);
    t182 = FbtAB(p3,p4,t43,p1,YY0);
    t183 = t182*t182;
    t185 = FbtAB(p5,p4,t43,p1,YY0);
    t186 = t185*t185;
    t187 = IFbtAB(p3,p3,t43,p1,YY0);
    t190 = IFbtAB(p5,p5,t43,p1,YY0);
    t191 = IFbtAA(p3,t62,p5,t43,p1,YY0);
    t192 = t190*t191;
    t194 = I/YY0;
    t195 = t192*t194;
    t206 = t93*t177*t181;
    t208 = t183*t186;
    t209 = FbtAB2(p1,p1,t43,p1,YY0,YY2);
    t212 = YY0*YY0;
    t219 = t110*t206;
    t220 = FbtAB2(p3,p3,t43,p1,YY0,YY2);
    t221 = t187*t187;
    t227 = FbtAB2(p5,p5,t43,p1,YY0,YY2);
    t230 = t190*t190;
    t236 = FbtAA2(p3,t62,p5,t43,p1,YY0,YY2);
    t239 = t191*t191;
    t246 = FbtAB2(p5,p4,t43,p1,YY0,YY2);
    t253 = FbtAB2(p3,p4,t43,p1,YY0,YY2);
    t262 = FbtAB2(p1,p6,t43,p1,YY0,YY2);
    t271 = FbtAB2(p1,p2,t43,p1,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t39*t42*t46*t48*t52*t54*t57*t58*t67-FOUR*t80*t83*t48*t52*t63*
t66-FOUR*t92*t95*t97*t98*t58*t64*I/t101+FOUR*t111*t97*t112*t113*t67+FOUR*t111*t97*
t119*t58*t122*t63*t66+FOUR*t111*t97*t128*t58*t61*t131*t66-EIGHT*t111*t54*t56*t138*
t58*t67-EIGHT*t111*t53*t57*t145*t58*t67-FOUR*t151*t42*t45*t48*t97*t155*t58*t67-EIGHT
*t110*t93*t47*t52*t97*t164*t58*t67+FOUR*EIGHT/THREE*t39*t42*t46*t177*t181*t183*t186*
t187*t195-FOUR*t80*t83*t177*t181*t191*t194-FOUR*t92*t206*t208*t209*t187*t192*I/
t212+FOUR*t219*t208*t220*t221*t195+FOUR*t219*t208*t227*t187*t230*t191*t194+FOUR*
t219*t208*t236*t187*t190*t239*t194-EIGHT*t219*t183*t185*t246*t187*t195-EIGHT*t219*
t182*t186*t253*t187*t195-FOUR*t151*t42*t45*t177*t208*t262*t187*t195-EIGHT*t110*t93
*t176*t181*t208*t271*t187*t195;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C212mpmpmpt56(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p6+p5;
    t1 = A(p1,p2);
    t2 = B(p1,p2);
    S1 = -t2*t1;
    gg = AB(p1,p2,p1);
    S2 = S(K2);
    t4 = AB(p2,K2,p2);
    t5 = gg*t4;
    t6 = AB(p1,K2,p1);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p1,K2,p2);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p1,K2,p2);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p2+p1;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p1+p2);
    t30 = S(p5+p6);
    t32 = -t28+t30*t29;
    t33 = ONE/t32;
    t34 = IA(p1,p2);
    t35 = t34*t34;
    t36 = t35*t34;
    t38 = IA(p3,p4);
    t39 = IA(p5,p6);
    t40 = t38*t39;
    t42 = IB(p1,p2);
    t43 = t42*t42;
    t44 = t43*t42;
    t45 = IB(p3,p4);
    t47 = IB(p5,p6);
    t48 = t47*t21;
    t50 = t33*t36*t40*t44*t45*t48;
    t51 = FbtAB(p1,p2,t26,p1,YY01);
    t52 = t51*t51;
    t53 = FbtAB(p3,p4,t26,p1,YY01);
    t54 = t53*t53;
    t55 = t52*t54;
    t56 = FbtAB(p5,p6,t26,p1,YY01);
    t57 = t56*t56;
    t58 = FbtAB2(p1,p1,t26,p1,YY01,YY21);
    t61 = IFbtAB(p3,p3,t26,p1,YY01);
    t62 = IFbtAB(p5,p5,t26,p1,YY01);
    t63 = t61*t62;
    t64 = YY01*YY01;
    t74 = t33*t35*t40*t43*t45*t48;
    t75 = FbtAB2(p3,p3,t26,p1,YY01,YY21);
    t78 = t61*t61;
    t81 = I/YY01;
    t85 = FbtAB2(p5,p5,t26,p1,YY01,YY21);
    t88 = t62*t62;
    t93 = FbtAB2(p5,p6,t26,p1,YY01,YY21);
    t96 = t63*t81;
    t101 = FbtAB2(p3,p4,t26,p1,YY01,YY21);
    t108 = FbtAB2(p1,p2,t26,p1,YY01,YY21);
    t114 = t32*t32;
    t121 = ONE/t114*t36*t38*t39*t44*t45*t47;
    t122 = t21*t21;
    t123 = t122*t21;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t136 = FbtAB(p1,p2,t26,p1,YY0);
    t137 = t136*t136;
    t138 = FbtAB(p3,p4,t26,p1,YY0);
    t139 = t138*t138;
    t140 = t137*t139;
    t141 = FbtAB(p5,p6,t26,p1,YY0);
    t142 = t141*t141;
    t143 = FbtAB2(p1,p1,t26,p1,YY0,YY2);
    t146 = IFbtAB(p3,p3,t26,p1,YY0);
    t147 = IFbtAB(p5,p5,t26,p1,YY0);
    t148 = t146*t147;
    t149 = YY0*YY0;
    t155 = FbtAB2(p3,p3,t26,p1,YY0,YY2);
    t158 = t146*t146;
    t161 = I/YY0;
    t165 = FbtAB2(p5,p5,t26,p1,YY0,YY2);
    t168 = t147*t147;
    t173 = FbtAB2(p5,p6,t26,p1,YY0,YY2);
    t176 = t148*t161;
    t181 = FbtAB2(p3,p4,t26,p1,YY0,YY2);
    t188 = FbtAB2(p1,p2,t26,p1,YY0,YY2);
    return(t50*t55*t57*t58*t63*I/t64/TWO-t74*t55*t57*t75*t78*t62*t81/TWO-t74*
t55*t57*t85*t61*t88*t81/TWO+t74*t55*t56*t93*t96+t74*t52*t53*t57*t101*t96+t74*
t51*t54*t57*t108*t96+t121*t123*t52*t54*t57*t96/THREE+t50*t140*t142*t143*t148*I/
t149/TWO-t74*t140*t142*t155*t158*t147*t161/TWO-t74*t140*t142*t165*t146*t168*
t161/TWO+t74*t140*t141*t173*t176+t74*t137*t138*t142*t181*t176+t74*t136*t139*
t142*t188*t176+t121*t123*t137*t139*t142*t176/THREE);
  }
}


std::complex<DOUBLE> C212mpmpmpt456(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p6+p5+p4;
    t1 = A(p1,p2);
    t2 = B(p1,p2);
    S1 = -t2*t1;
    gg = AB(p1,p2,p1);
    S2 = S(K2);
    t4 = AB(p2,K2,p2);
    t5 = gg*t4;
    t6 = AB(p1,K2,p1);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p1,K2,p2);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p1,K2,p2);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p3);
    t27 = t26*t26;
    t29 = B(p2,p3);
    t30 = t29*t29;
    t33 = IA(p1,p2);
    t34 = t33*t33;
    t36 = IA(p4,p5);
    t37 = t34*t33*t36;
    t39 = IA(p5,p6);
    t40 = IB(p1,p2);
    t41 = t40*t40;
    t43 = t39*t41*t40;
    t44 = IB(p1,p3);
    t45 = p2+p1;
    t46 = IAB(p3,t45,p3);
    t47 = t46*t46;
    t48 = t47*t47;
    t51 = t27*t26*t30*t29*t37*t43*t44*t48;
    t52 = FbtAB(p1,p2,t45,p1,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p3,p1,t45,p1,YY01);
    t56 = FbtAB(p5,p4,t45,p1,YY01);
    t57 = t56*t56;
    t58 = FbtAB(p5,p6,t45,p1,YY01);
    t59 = t58*t58;
    t62 = IFbtAB(p4,p4,t45,p1,YY01);
    t63 = IFbtAB(p6,p6,t45,p1,YY01);
    t64 = t62*t63;
    t65 = p5+p4;
    t66 = IFbtBB(p4,t65,p6,t45,p1,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = t26*t29;
    t75 = B(p4,p6);
    t76 = t75*t75;
    t77 = t76*t76;
    t80 = IB(p4,p5);
    t83 = t74*t77*t34*t41*t44*t80;
    t84 = IB(p5,p6);
    t85 = t84*t47;
    t86 = IS(K2);
    t95 = t74*t37;
    t96 = t44*t47;
    t97 = t96*t53;
    t100 = t54*t57;
    t101 = FbtAB2(p1,p1,t45,p1,YY01,YY21);
    t104 = YY01*YY01;
    t112 = t74*t34*t36;
    t113 = t39*t41;
    t115 = t112*t113*t97;
    t116 = FbtAB2(p4,p4,t45,p1,YY01,YY21);
    t119 = t62*t62;
    t125 = FbtAB2(p6,p6,t45,p1,YY01,YY21);
    t128 = t63*t63;
    t134 = FbtBB2(p4,t65,p6,t45,p1,YY01,YY21);
    t137 = t66*t66;
    t144 = FbtAB2(p5,p6,t45,p1,YY01,YY21);
    t151 = FbtAB2(p5,p4,t45,p1,YY01,YY21);
    t158 = t112*t113*t96;
    t160 = FbtAB2(p3,p1,t45,p1,YY01,YY21);
    t169 = FbtAB2(p1,p2,t45,p1,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t181 = FbtAB(p1,p2,t45,p1,YY0);
    t182 = t181*t181;
    t183 = FbtAB(p3,p1,t45,p1,YY0);
    t185 = FbtAB(p5,p4,t45,p1,YY0);
    t186 = t185*t185;
    t187 = FbtAB(p5,p6,t45,p1,YY0);
    t188 = t187*t187;
    t191 = IFbtAB(p4,p4,t45,p1,YY0);
    t192 = IFbtAB(p6,p6,t45,p1,YY0);
    t193 = t191*t192;
    t194 = IFbtBB(p4,t65,p6,t45,p1,YY0);
    t195 = t194*I;
    t196 = ONE/YY0;
    t197 = t195*t196;
    t198 = t193*t197;
    t210 = t96*t182;
    t213 = t183*t186;
    t214 = FbtAB2(p1,p1,t45,p1,YY0,YY2);
    t217 = YY0*YY0;
    t225 = t112*t113*t210;
    t226 = FbtAB2(p4,p4,t45,p1,YY0,YY2);
    t229 = t191*t191;
    t235 = FbtAB2(p6,p6,t45,p1,YY0,YY2);
    t238 = t192*t192;
    t244 = FbtBB2(p4,t65,p6,t45,p1,YY0,YY2);
    t247 = t194*t194;
    t254 = FbtAB2(p5,p6,t45,p1,YY0,YY2);
    t261 = FbtAB2(p5,p4,t45,p1,YY0,YY2);
    t268 = FbtAB2(p3,p1,t45,p1,YY0,YY2);
    t277 = FbtAB2(p1,p2,t45,p1,YY0,YY2);
    ans = -FOUR*EIGHT/THREE*t51*t53*t54*t57*t59*t70+FOUR*t83*t85*t86*t53*t54*t66*I*t68+
FOUR*t95*t43*t97*t100*t59*t101*t64*t67/t104-FOUR*t115*t100*t59*t116*t119*t63*t69
-FOUR*t115*t100*t59*t125*t62*t128*t69-FOUR*t115*t100*t59*t134*t64*t137*I*t68+EIGHT*
t115*t100*t58*t144*t70+EIGHT*t115*t54*t56*t59*t151*t70+FOUR*t158*t53*t57*t59*t160*
t70+EIGHT*t112*t113*t96*t52*t100*t59*t169*t70-FOUR*EIGHT/THREE*t51*t182*t183*t186*t188*
t198+FOUR*t83*t85*t86*t182*t183*t194*I*t196+FOUR*t95*t43*t210*t213*t188*t214*t193
*t195/t217-FOUR*t225*t213*t188*t226*t229*t192*t197-FOUR*t225*t213*t188*t235*t191*
t238*t197-FOUR*t225*t213*t188*t244*t193*t247*I*t196+EIGHT*t225*t213*t187*t254*t198
+EIGHT*t225*t183*t185*t188*t261*t198+FOUR*t158*t182*t186*t188*t268*t198+EIGHT*t112*
t113*t96*t181*t213*t188*t277*t198;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C223mpmpmpt2(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p3;
    t1 = A(p2,p3);
    t2 = B(p2,p3);
    S1 = -t2*t1;
    gg = AB(p2,K1f,p2);
    S2 = S(p2);
    t4 = AB(K1f,p2,K1f);
    t5 = gg*t4;
    t6 = AB(p2,p2,p2);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p2,p2,K1f);
    t19 = ONE/S1;
    t21 = AB(p2,p2,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C223mpmpmpt1(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t1 = A(p2,p3);
    t2 = B(p2,p3);
    S1 = -t2*t1;
    gg = AB(p2,p3,p2);
    S2 = S(p1);
    t4 = AB(p3,p1,p3);
    t5 = gg*t4;
    t6 = AB(p2,p1,p2);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p2,p1,p3);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p2,p1,p3);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p2);
    t27 = t26*t26;
    t29 = B(p1,p3);
    t30 = t29*t29;
    t33 = IA(p2,p3);
    t34 = t33*t33;
    t36 = IA(p4,p5);
    t37 = t34*t33*t36;
    t39 = IA(p5,p6);
    t40 = IB(p1,p2);
    t41 = t39*t40;
    t42 = IB(p2,p3);
    t43 = t42*t42;
    t44 = t43*t42;
    t45 = p3+p2;
    t46 = IAB(p1,t45,p1);
    t47 = t46*t46;
    t48 = t47*t47;
    t51 = t27*t26*t30*t29*t37*t41*t44*t48;
    t52 = FbtAB(p1,p2,t45,p2,YY01);
    t53 = FbtAB(p3,p2,t45,p2,YY01);
    t54 = t53*t53;
    t56 = FbtAB(p5,p4,t45,p2,YY01);
    t57 = t56*t56;
    t58 = FbtAB(p5,p6,t45,p2,YY01);
    t59 = t58*t58;
    t62 = IFbtAB(p4,p4,t45,p2,YY01);
    t63 = IFbtAB(p6,p6,t45,p2,YY01);
    t64 = t62*t63;
    t65 = p5+p4;
    t66 = IFbtBB(p4,t65,p6,t45,p2,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = t26*t29;
    t75 = B(p4,p6);
    t76 = t75*t75;
    t77 = t76*t76;
    t80 = IB(p4,p5);
    t83 = t74*t77*t34*t40*t43*t80;
    t84 = IB(p5,p6);
    t85 = t84*t47;
    t87 = IS(p6+p5+p4);
    t96 = t74*t37;
    t97 = t44*t47;
    t101 = t54*t57;
    t102 = FbtAB2(p2,p2,t45,p2,YY01,YY21);
    t105 = YY01*YY01;
    t113 = t74*t34*t36;
    t114 = t43*t47;
    t117 = t113*t41*t114*t52;
    t118 = FbtAB2(p4,p4,t45,p2,YY01,YY21);
    t121 = t62*t62;
    t127 = FbtAB2(p6,p6,t45,p2,YY01,YY21);
    t130 = t63*t63;
    t136 = FbtBB2(p4,t65,p6,t45,p2,YY01,YY21);
    t139 = t66*t66;
    t146 = FbtAB2(p5,p6,t45,p2,YY01,YY21);
    t153 = FbtAB2(p5,p4,t45,p2,YY01,YY21);
    t160 = FbtAB2(p3,p2,t45,p2,YY01,YY21);
    t167 = t113*t41*t114;
    t168 = FbtAB2(p1,p2,t45,p2,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t180 = FbtAB(p1,p2,t45,p2,YY0);
    t181 = FbtAB(p3,p2,t45,p2,YY0);
    t182 = t181*t181;
    t184 = FbtAB(p5,p4,t45,p2,YY0);
    t185 = t184*t184;
    t186 = FbtAB(p5,p6,t45,p2,YY0);
    t187 = t186*t186;
    t190 = IFbtAB(p4,p4,t45,p2,YY0);
    t191 = IFbtAB(p6,p6,t45,p2,YY0);
    t192 = t190*t191;
    t193 = IFbtBB(p4,t65,p6,t45,p2,YY0);
    t194 = t193*I;
    t195 = ONE/YY0;
    t196 = t194*t195;
    t197 = t192*t196;
    t212 = t182*t185;
    t213 = FbtAB2(p2,p2,t45,p2,YY0,YY2);
    t216 = YY0*YY0;
    t225 = t113*t41*t114*t180;
    t226 = FbtAB2(p4,p4,t45,p2,YY0,YY2);
    t229 = t190*t190;
    t235 = FbtAB2(p6,p6,t45,p2,YY0,YY2);
    t238 = t191*t191;
    t244 = FbtBB2(p4,t65,p6,t45,p2,YY0,YY2);
    t247 = t193*t193;
    t254 = FbtAB2(p5,p6,t45,p2,YY0,YY2);
    t261 = FbtAB2(p5,p4,t45,p2,YY0,YY2);
    t268 = FbtAB2(p3,p2,t45,p2,YY0,YY2);
    t274 = FbtAB2(p1,p2,t45,p2,YY0,YY2);
    ans = -FOUR*EIGHT/THREE*t51*t52*t54*t57*t59*t70+FOUR*t83*t85*t87*t52*t54*t66*I*t68+
FOUR*t96*t41*t97*t52*t101*t59*t102*t64*t67/t105-FOUR*t117*t101*t59*t118*t121*t63*
t69-FOUR*t117*t101*t59*t127*t62*t130*t69-FOUR*t117*t101*t59*t136*t64*t139*I*t68+
EIGHT*t117*t101*t58*t146*t70+EIGHT*t117*t54*t56*t59*t153*t70+EIGHT*t117*t53*t57*t59*
t160*t70+FOUR*t167*t101*t59*t168*t70-FOUR*EIGHT/THREE*t51*t180*t182*t185*t187*t197+FOUR*
t83*t85*t87*t180*t182*t193*I*t195+FOUR*t96*t41*t97*t180*t212*t187*t213*t192*t194
/t216-FOUR*t225*t212*t187*t226*t229*t191*t196-FOUR*t225*t212*t187*t235*t190*t238*
t196-FOUR*t225*t212*t187*t244*t192*t247*I*t195+EIGHT*t225*t212*t186*t254*t197+EIGHT*
t225*t182*t184*t187*t261*t197+EIGHT*t225*t181*t185*t187*t268*t197+FOUR*t167*t212*
t187*t274*t197;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C223mpmpmpt61(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t170;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p6+p1;
    t1 = A(p2,p3);
    t2 = B(p2,p3);
    S1 = -t2*t1;
    gg = AB(p2,p3,p2);
    S2 = S(K2);
    t4 = AB(p3,K2,p3);
    t5 = gg*t4;
    t6 = AB(p2,K2,p2);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p2,K2,p3);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p2,K2,p3);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p3+p2;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p1+p6);
    t30 = S(p2+p3);
    t32 = -t28+t29*t30;
    t34 = IA(p1,p6);
    t35 = ONE/t32*t34;
    t36 = IA(p2,p3);
    t37 = t36*t36;
    t38 = t37*t36;
    t39 = IA(p4,p5);
    t42 = IB(p1,p6);
    t43 = IB(p2,p3);
    t44 = t43*t43;
    t45 = t44*t43;
    t47 = IB(p4,p5);
    t49 = AB(p2,p5+p4,p3);
    t50 = t47*t49;
    t52 = t35*t38*t39*t42*t45*t50;
    t53 = FbtAB(p1,p6,t26,p2,YY01);
    t54 = t53*t53;
    t55 = FbtAB(p3,p2,t26,p2,YY01);
    t56 = t55*t55;
    t57 = t54*t56;
    t58 = FbtAB(p5,p4,t26,p2,YY01);
    t59 = t58*t58;
    t60 = FbtAB2(p2,p2,t26,p2,YY01,YY21);
    t63 = IFbtAB(p4,p4,t26,p2,YY01);
    t64 = IFbtAB(p6,p6,t26,p2,YY01);
    t65 = t63*t64;
    t66 = YY01*YY01;
    t76 = t35*t37*t39*t42*t44*t50;
    t77 = FbtAB2(p4,p4,t26,p2,YY01,YY21);
    t80 = t63*t63;
    t83 = I/YY01;
    t87 = FbtAB2(p6,p6,t26,p2,YY01,YY21);
    t90 = t64*t64;
    t95 = FbtAB2(p5,p4,t26,p2,YY01,YY21);
    t98 = t65*t83;
    t103 = FbtAB2(p3,p2,t26,p2,YY01,YY21);
    t110 = FbtAB2(p1,p6,t26,p2,YY01,YY21);
    t116 = t32*t32;
    t123 = ONE/t116*t34*t38*t39*t42*t45*t47;
    t124 = t49*t49;
    t125 = t124*t49;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t138 = FbtAB(p1,p6,t26,p2,YY0);
    t139 = t138*t138;
    t140 = FbtAB(p3,p2,t26,p2,YY0);
    t141 = t140*t140;
    t142 = t139*t141;
    t143 = FbtAB(p5,p4,t26,p2,YY0);
    t144 = t143*t143;
    t145 = FbtAB2(p2,p2,t26,p2,YY0,YY2);
    t148 = IFbtAB(p4,p4,t26,p2,YY0);
    t149 = IFbtAB(p6,p6,t26,p2,YY0);
    t150 = t148*t149;
    t151 = YY0*YY0;
    t157 = FbtAB2(p4,p4,t26,p2,YY0,YY2);
    t160 = t148*t148;
    t163 = I/YY0;
    t167 = FbtAB2(p6,p6,t26,p2,YY0,YY2);
    t170 = t149*t149;
    t175 = FbtAB2(p5,p4,t26,p2,YY0,YY2);
    t178 = t150*t163;
    t183 = FbtAB2(p3,p2,t26,p2,YY0,YY2);
    t190 = FbtAB2(p1,p6,t26,p2,YY0,YY2);
    return(-t52*t57*t59*t60*t65*I/t66/TWO+t76*t57*t59*t77*t80*t64*t83/TWO+t76*
t57*t59*t87*t63*t90*t83/TWO-t76*t57*t58*t95*t98-t76*t54*t55*t59*t103*t98-t76*
t53*t56*t59*t110*t98-t123*t125*t54*t56*t59*t98/THREE-t52*t142*t144*t145*t150*I/
t151/TWO+t76*t142*t144*t157*t160*t149*t163/TWO+t76*t142*t144*t167*t148*t170*
t163/TWO-t76*t142*t143*t175*t178-t76*t139*t140*t144*t183*t178-t76*t138*t141*
t144*t190*t178-t123*t125*t139*t141*t144*t178/THREE);
  }
}


std::complex<DOUBLE> C223mpmpmpt561(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  FourMomentum<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  FourMomentum<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p6+p5+p1;
    t1 = A(p2,p3);
    t2 = B(p2,p3);
    S1 = -t2*t1;
    gg = AB(p2,p3,p2);
    S2 = S(K2);
    t4 = AB(p3,K2,p3);
    t5 = gg*t4;
    t6 = AB(p2,K2,p2);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p2,K2,p3);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p2,K2,p3);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p5);
    t27 = t26*t26;
    t28 = t27*t27;
    t29 = B(p3,p4);
    t31 = IA(p1,p6);
    t33 = IA(p2,p3);
    t34 = t33*t33;
    t35 = IA(p5,p6);
    t37 = IB(p2,p3);
    t38 = t37*t37;
    t40 = t28*t29*t31*t34*t35*t38;
    t41 = p3+p2;
    t42 = IAB(p4,t41,p4);
    t43 = t42*t42;
    t44 = IS(K2);
    t45 = t44*t43;
    t46 = FbtAB(p2,p4,t41,p2,YY01);
    t48 = FbtAB(p3,p2,t41,p2,YY01);
    t49 = t48*t48;
    t50 = p6+p5;
    t51 = IFbtAA(p5,t50,p1,t41,p2,YY01);
    t54 = I/YY01;
    t59 = A(p2,p4);
    t60 = t59*t59;
    t61 = t29*t29;
    t64 = t34*t33;
    t65 = IB(p1,p6);
    t67 = t60*t61*t29*t64*t65;
    t68 = t38*t37;
    t69 = IB(p5,p6);
    t70 = t68*t69;
    t71 = t43*t43;
    t72 = FbtAB(p1,p6,t41,p2,YY01);
    t73 = t72*t72;
    t78 = FbtAB(p5,p6,t41,p2,YY01);
    t79 = t78*t78;
    t80 = IFbtAB(p1,p1,t41,p2,YY01);
    t83 = IFbtAB(p5,p5,t41,p2,YY01);
    t84 = t83*t51;
    t85 = t84*t54;
    t91 = t29*t64*t65*t68;
    t92 = t69*t43;
    t94 = t92*t73*t46;
    t96 = t49*t79;
    t97 = FbtAB2(p2,p2,t41,p2,YY01,YY21);
    t100 = YY01*YY01;
    t107 = t29*t34;
    t109 = t107*t65*t38;
    t110 = t109*t94;
    t111 = FbtAB2(p1,p1,t41,p2,YY01,YY21);
    t112 = t80*t80;
    t118 = FbtAB2(p5,p5,t41,p2,YY01,YY21);
    t121 = t83*t83;
    t127 = FbtAA2(p5,t50,p1,t41,p2,YY01,YY21);
    t130 = t51*t51;
    t137 = FbtAB2(p5,p6,t41,p2,YY01,YY21);
    t144 = FbtAB2(p3,p2,t41,p2,YY01,YY21);
    t150 = t107*t65;
    t151 = t38*t69;
    t155 = FbtAB2(p2,p4,t41,p2,YY01,YY21);
    t164 = FbtAB2(p1,p6,t41,p2,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t176 = FbtAB(p2,p4,t41,p2,YY0);
    t178 = FbtAB(p3,p2,t41,p2,YY0);
    t179 = t178*t178;
    t180 = IFbtAA(p5,t50,p1,t41,p2,YY0);
    t183 = I/YY0;
    t188 = FbtAB(p1,p6,t41,p2,YY0);
    t189 = t188*t188;
    t194 = FbtAB(p5,p6,t41,p2,YY0);
    t195 = t194*t194;
    t196 = IFbtAB(p1,p1,t41,p2,YY0);
    t199 = IFbtAB(p5,p5,t41,p2,YY0);
    t200 = t199*t180;
    t201 = t200*t183;
    t206 = t92*t189*t176;
    t208 = t179*t195;
    t209 = FbtAB2(p2,p2,t41,p2,YY0,YY2);
    t212 = YY0*YY0;
    t219 = t109*t206;
    t220 = FbtAB2(p1,p1,t41,p2,YY0,YY2);
    t221 = t196*t196;
    t227 = FbtAB2(p5,p5,t41,p2,YY0,YY2);
    t230 = t199*t199;
    t236 = FbtAA2(p5,t50,p1,t41,p2,YY0,YY2);
    t239 = t180*t180;
    t246 = FbtAB2(p5,p6,t41,p2,YY0,YY2);
    t253 = FbtAB2(p3,p2,t41,p2,YY0,YY2);
    t262 = FbtAB2(p2,p4,t41,p2,YY0,YY2);
    t271 = FbtAB2(p1,p6,t41,p2,YY0,YY2);
    ans = -FOUR*t40*t45*t46*t49*t51*t54+FOUR*EIGHT/THREE*t67*t70*t71*t73*t46*t49*t79*t80
*t85-FOUR*t91*t94*t96*t97*t80*t84*I/t100+FOUR*t110*t96*t111*t112*t85+FOUR*t110*t96
*t118*t80*t121*t51*t54+FOUR*t110*t96*t127*t80*t83*t130*t54-EIGHT*t110*t49*t78*t137
*t80*t85-EIGHT*t110*t48*t79*t144*t80*t85-FOUR*t150*t151*t43*t73*t96*t155*t80*t85
-EIGHT*t109*t92*t72*t46*t96*t164*t80*t85-FOUR*t40*t45*t176*t179*t180*t183+FOUR*EIGHT/THREE
*t67*t70*t71*t189*t176*t179*t195*t196*t201-FOUR*t91*t206*t208*t209*t196*t200*I/
t212+FOUR*t219*t208*t220*t221*t201+FOUR*t219*t208*t227*t196*t230*t180*t183+FOUR*
t219*t208*t236*t196*t199*t239*t183-EIGHT*t219*t179*t194*t246*t196*t201-EIGHT*t219*
t178*t195*t253*t196*t201-FOUR*t150*t151*t43*t189*t208*t262*t196*t201-EIGHT*t109*
t92*t188*t176*t208*t271*t196*t201;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C234mpmpmpt3(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p4;
    t1 = A(p3,p4);
    t2 = B(p3,p4);
    S1 = -t2*t1;
    gg = AB(p3,K1f,p3);
    S2 = S(p3);
    t4 = AB(K1f,p3,K1f);
    t5 = gg*t4;
    t6 = AB(p3,p3,p3);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p3,p3,K1f);
    t19 = ONE/S1;
    t21 = AB(p3,p3,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C234mpmpmpt2(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t247;
  std::complex<DOUBLE> t254;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t272;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  FourMomentum<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t1 = A(p3,p4);
    t2 = B(p3,p4);
    S1 = -t2*t1;
    gg = AB(p3,p4,p3);
    S2 = S(p2);
    t4 = AB(p4,p2,p4);
    t5 = gg*t4;
    t6 = AB(p3,p2,p3);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p3,p2,p4);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p3,p2,p4);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p5);
    t27 = t26*t26;
    t28 = t27*t27;
    t29 = B(p2,p4);
    t31 = IA(p1,p6);
    t33 = IA(p3,p4);
    t34 = t33*t33;
    t35 = IA(p5,p6);
    t37 = IB(p3,p4);
    t38 = t37*t37;
    t40 = t28*t29*t31*t34*t35*t38;
    t41 = p4+p3;
    t42 = IAB(p2,t41,p2);
    t43 = t42*t42;
    t45 = IS(p6+p5+p1);
    t46 = t43*t45;
    t47 = FbtAB(p3,p2,t41,p3,YY01);
    t49 = FbtAB(p3,p4,t41,p3,YY01);
    t50 = t49*t49;
    t51 = p6+p5;
    t52 = IFbtAA(p5,t51,p1,t41,p3,YY01);
    t55 = I/YY01;
    t60 = A(p2,p3);
    t61 = t60*t60;
    t62 = t29*t29;
    t65 = t34*t33;
    t66 = IB(p1,p6);
    t68 = t61*t62*t29*t65*t66;
    t69 = t38*t37;
    t70 = IB(p5,p6);
    t71 = t69*t70;
    t72 = t43*t43;
    t73 = FbtAB(p1,p6,t41,p3,YY01);
    t74 = t73*t73;
    t79 = FbtAB(p5,p6,t41,p3,YY01);
    t80 = t79*t79;
    t81 = IFbtAB(p1,p1,t41,p3,YY01);
    t84 = IFbtAB(p5,p5,t41,p3,YY01);
    t85 = t84*t52;
    t86 = t85*t55;
    t92 = t29*t65*t66*t69;
    t93 = t70*t43;
    t95 = t93*t74*t47;
    t97 = t50*t80;
    t98 = FbtAB2(p3,p3,t41,p3,YY01,YY21);
    t101 = YY01*YY01;
    t108 = t29*t34;
    t110 = t108*t66*t38;
    t111 = t110*t95;
    t112 = FbtAB2(p1,p1,t41,p3,YY01,YY21);
    t113 = t81*t81;
    t119 = FbtAB2(p5,p5,t41,p3,YY01,YY21);
    t122 = t84*t84;
    t128 = FbtAA2(p5,t51,p1,t41,p3,YY01,YY21);
    t131 = t52*t52;
    t138 = FbtAB2(p5,p6,t41,p3,YY01,YY21);
    t145 = FbtAB2(p3,p4,t41,p3,YY01,YY21);
    t151 = t108*t66;
    t152 = t38*t70;
    t156 = FbtAB2(p3,p2,t41,p3,YY01,YY21);
    t165 = FbtAB2(p1,p6,t41,p3,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t177 = FbtAB(p3,p2,t41,p3,YY0);
    t179 = FbtAB(p3,p4,t41,p3,YY0);
    t180 = t179*t179;
    t181 = IFbtAA(p5,t51,p1,t41,p3,YY0);
    t184 = I/YY0;
    t189 = FbtAB(p1,p6,t41,p3,YY0);
    t190 = t189*t189;
    t195 = FbtAB(p5,p6,t41,p3,YY0);
    t196 = t195*t195;
    t197 = IFbtAB(p1,p1,t41,p3,YY0);
    t200 = IFbtAB(p5,p5,t41,p3,YY0);
    t201 = t200*t181;
    t202 = t201*t184;
    t207 = t93*t190*t177;
    t209 = t180*t196;
    t210 = FbtAB2(p3,p3,t41,p3,YY0,YY2);
    t213 = YY0*YY0;
    t220 = t110*t207;
    t221 = FbtAB2(p1,p1,t41,p3,YY0,YY2);
    t222 = t197*t197;
    t228 = FbtAB2(p5,p5,t41,p3,YY0,YY2);
    t231 = t200*t200;
    t237 = FbtAA2(p5,t51,p1,t41,p3,YY0,YY2);
    t240 = t181*t181;
    t247 = FbtAB2(p5,p6,t41,p3,YY0,YY2);
    t254 = FbtAB2(p3,p4,t41,p3,YY0,YY2);
    t263 = FbtAB2(p3,p2,t41,p3,YY0,YY2);
    t272 = FbtAB2(p1,p6,t41,p3,YY0,YY2);
    ans = -FOUR*t40*t46*t47*t50*t52*t55+FOUR*EIGHT/THREE*t68*t71*t72*t74*t47*t50*t80*t81
*t86-FOUR*t92*t95*t97*t98*t81*t85*I/t101+FOUR*t111*t97*t112*t113*t86+FOUR*t111*t97
*t119*t81*t122*t52*t55+FOUR*t111*t97*t128*t81*t84*t131*t55-EIGHT*t111*t50*t79*t138
*t81*t86-EIGHT*t111*t49*t80*t145*t81*t86-FOUR*t151*t152*t43*t74*t97*t156*t81*t86
-EIGHT*t110*t93*t73*t47*t97*t165*t81*t86-FOUR*t40*t46*t177*t180*t181*t184+FOUR*EIGHT/THREE
*t68*t71*t72*t190*t177*t180*t196*t197*t202-FOUR*t92*t207*t209*t210*t197*t201*I/
t213+FOUR*t220*t209*t221*t222*t202+FOUR*t220*t209*t228*t197*t231*t181*t184+FOUR*
t220*t209*t237*t197*t200*t240*t184-EIGHT*t220*t180*t195*t247*t197*t202-EIGHT*t220*
t179*t196*t254*t197*t202-FOUR*t151*t152*t43*t190*t209*t263*t197*t202-EIGHT*t110*
t93*t189*t177*t209*t272*t197*t202;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C234mpmpmpt12(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p2+p1;
    t1 = A(p3,p4);
    t2 = B(p3,p4);
    S1 = -t2*t1;
    gg = AB(p3,p4,p3);
    S2 = S(K2);
    t4 = AB(p4,K2,p4);
    t5 = gg*t4;
    t6 = AB(p3,K2,p3);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p3,K2,p4);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p3,K2,p4);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p4+p3;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p1+p2);
    t30 = S(p3+p4);
    t32 = -t28+t29*t30;
    t34 = IA(p1,p2);
    t35 = ONE/t32*t34;
    t36 = IA(p3,p4);
    t37 = t36*t36;
    t38 = t37*t36;
    t39 = IA(p5,p6);
    t42 = IB(p1,p2);
    t43 = IB(p3,p4);
    t44 = t43*t43;
    t45 = t44*t43;
    t47 = IB(p5,p6);
    t48 = t47*t21;
    t50 = t35*t38*t39*t42*t45*t48;
    t51 = FbtAB(p1,p2,t26,p3,YY01);
    t52 = t51*t51;
    t53 = FbtAB(p3,p4,t26,p3,YY01);
    t54 = t53*t53;
    t55 = t52*t54;
    t56 = FbtAB(p5,p6,t26,p3,YY01);
    t57 = t56*t56;
    t58 = FbtAB2(p3,p3,t26,p3,YY01,YY21);
    t61 = IFbtAB(p1,p1,t26,p3,YY01);
    t62 = IFbtAB(p5,p5,t26,p3,YY01);
    t63 = t61*t62;
    t64 = YY01*YY01;
    t74 = t35*t37*t39*t42*t44*t48;
    t75 = FbtAB2(p1,p1,t26,p3,YY01,YY21);
    t78 = t61*t61;
    t81 = I/YY01;
    t85 = FbtAB2(p5,p5,t26,p3,YY01,YY21);
    t88 = t62*t62;
    t93 = FbtAB2(p5,p6,t26,p3,YY01,YY21);
    t96 = t63*t81;
    t101 = FbtAB2(p3,p4,t26,p3,YY01,YY21);
    t108 = FbtAB2(p1,p2,t26,p3,YY01,YY21);
    t114 = t32*t32;
    t121 = ONE/t114*t34*t38*t39*t42*t45*t47;
    t122 = t21*t21;
    t123 = t122*t21;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t136 = FbtAB(p1,p2,t26,p3,YY0);
    t137 = t136*t136;
    t138 = FbtAB(p3,p4,t26,p3,YY0);
    t139 = t138*t138;
    t140 = t137*t139;
    t141 = FbtAB(p5,p6,t26,p3,YY0);
    t142 = t141*t141;
    t143 = FbtAB2(p3,p3,t26,p3,YY0,YY2);
    t146 = IFbtAB(p1,p1,t26,p3,YY0);
    t147 = IFbtAB(p5,p5,t26,p3,YY0);
    t148 = t146*t147;
    t149 = YY0*YY0;
    t155 = FbtAB2(p1,p1,t26,p3,YY0,YY2);
    t158 = t146*t146;
    t161 = I/YY0;
    t165 = FbtAB2(p5,p5,t26,p3,YY0,YY2);
    t168 = t147*t147;
    t173 = FbtAB2(p5,p6,t26,p3,YY0,YY2);
    t176 = t148*t161;
    t181 = FbtAB2(p3,p4,t26,p3,YY0,YY2);
    t188 = FbtAB2(p1,p2,t26,p3,YY0,YY2);
    return(t50*t55*t57*t58*t63*I/t64/TWO-t74*t55*t57*t75*t78*t62*t81/TWO-t74*
t55*t57*t85*t61*t88*t81/TWO+t74*t55*t56*t93*t96+t74*t52*t53*t57*t101*t96+t74*
t51*t54*t57*t108*t96+t121*t123*t52*t54*t57*t96/THREE+t50*t140*t142*t143*t148*I/
t149/TWO-t74*t140*t142*t155*t158*t147*t161/TWO-t74*t140*t142*t165*t146*t168*
t161/TWO+t74*t140*t141*t173*t176+t74*t137*t138*t142*t181*t176+t74*t136*t139*
t142*t188*t176+t121*t123*t137*t139*t142*t176/THREE);
  }
}


std::complex<DOUBLE> C234mpmpmpt612(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t167;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t275;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p6+p2+p1;
    t1 = A(p3,p4);
    t2 = B(p3,p4);
    S1 = -t2*t1;
    gg = AB(p3,p4,p3);
    S2 = S(K2);
    t4 = AB(p4,K2,p4);
    t5 = gg*t4;
    t6 = AB(p3,K2,p3);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p3,K2,p4);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p3,K2,p4);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p3,p5);
    t27 = t26*t26;
    t29 = B(p4,p5);
    t30 = t29*t29;
    t33 = IA(p1,p2);
    t34 = IA(p1,p6);
    t35 = t34*t33;
    t37 = IA(p3,p4);
    t38 = t37*t37;
    t40 = IB(p3,p4);
    t41 = t40*t40;
    t43 = t38*t37*t41*t40;
    t44 = IB(p3,p5);
    t45 = p4+p3;
    t46 = IAB(p5,t45,p5);
    t47 = t46*t46;
    t48 = t47*t47;
    t51 = t27*t26*t30*t29*t35*t43*t44*t48;
    t52 = FbtAB(p1,p2,t45,p3,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p1,p6,t45,p3,YY01);
    t55 = t54*t54;
    t56 = t53*t55;
    t57 = FbtAB(p3,p4,t45,p3,YY01);
    t58 = t57*t57;
    t59 = FbtAB(p5,p3,t45,p3,YY01);
    t62 = IFbtAB(p2,p2,t45,p3,YY01);
    t63 = IFbtAB(p6,p6,t45,p3,YY01);
    t64 = t62*t63;
    t65 = p6+p1;
    t66 = IFbtBB(p6,t65,p2,t45,p3,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = B(p2,p6);
    t75 = t74*t74;
    t76 = t75*t75;
    t79 = IB(p1,p2);
    t81 = IB(p1,p6);
    t84 = t26*t76*t29*t38*t79*t81*t41;
    t85 = t44*t47;
    t86 = IS(K2);
    t96 = t26*t29*t35;
    t97 = t85*t53;
    t100 = t55*t58;
    t101 = FbtAB2(p3,p3,t45,p3,YY01,YY21);
    t104 = YY01*YY01;
    t111 = t38*t41;
    t113 = t96*t111*t97;
    t114 = FbtAB2(p2,p2,t45,p3,YY01,YY21);
    t117 = t62*t62;
    t123 = FbtAB2(p6,p6,t45,p3,YY01,YY21);
    t126 = t63*t63;
    t132 = FbtBB2(p6,t65,p2,t45,p3,YY01,YY21);
    t135 = t66*t66;
    t143 = t96*t111*t85;
    t144 = FbtAB2(p5,p3,t45,p3,YY01,YY21);
    t151 = FbtAB2(p3,p4,t45,p3,YY01,YY21);
    t158 = FbtAB2(p1,p6,t45,p3,YY01,YY21);
    t167 = FbtAB2(p1,p2,t45,p3,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t179 = FbtAB(p1,p2,t45,p3,YY0);
    t180 = t179*t179;
    t181 = FbtAB(p1,p6,t45,p3,YY0);
    t182 = t181*t181;
    t183 = t180*t182;
    t184 = FbtAB(p3,p4,t45,p3,YY0);
    t185 = t184*t184;
    t186 = FbtAB(p5,p3,t45,p3,YY0);
    t189 = IFbtAB(p2,p2,t45,p3,YY0);
    t190 = IFbtAB(p6,p6,t45,p3,YY0);
    t191 = t189*t190;
    t192 = IFbtBB(p6,t65,p2,t45,p3,YY0);
    t193 = t192*I;
    t194 = ONE/YY0;
    t195 = t193*t194;
    t196 = t191*t195;
    t208 = t85*t180;
    t211 = t182*t185;
    t212 = FbtAB2(p3,p3,t45,p3,YY0,YY2);
    t215 = YY0*YY0;
    t223 = t96*t111*t208;
    t224 = FbtAB2(p2,p2,t45,p3,YY0,YY2);
    t227 = t189*t189;
    t233 = FbtAB2(p6,p6,t45,p3,YY0,YY2);
    t236 = t190*t190;
    t242 = FbtBB2(p6,t65,p2,t45,p3,YY0,YY2);
    t245 = t192*t192;
    t252 = FbtAB2(p5,p3,t45,p3,YY0,YY2);
    t259 = FbtAB2(p3,p4,t45,p3,YY0,YY2);
    t266 = FbtAB2(p1,p6,t45,p3,YY0,YY2);
    t275 = FbtAB2(p1,p2,t45,p3,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t51*t56*t58*t59*t70-FOUR*t84*t85*t86*t58*t59*t66*I*t68-FOUR*
t96*t43*t97*t100*t59*t101*t64*t67/t104+FOUR*t113*t100*t59*t114*t117*t63*t69+FOUR*
t113*t100*t59*t123*t62*t126*t69+FOUR*t113*t100*t59*t132*t64*t135*I*t68-FOUR*t143*
t56*t58*t144*t70-EIGHT*t113*t55*t57*t59*t151*t70-EIGHT*t113*t54*t58*t59*t158*t70
-EIGHT*t96*t111*t85*t52*t100*t59*t167*t70+FOUR*EIGHT/THREE*t51*t183*t185*t186*t196-FOUR*
t84*t85*t86*t185*t186*t192*I*t194-FOUR*t96*t43*t208*t211*t186*t212*t191*t193/
t215+FOUR*t223*t211*t186*t224*t227*t190*t195+FOUR*t223*t211*t186*t233*t189*t236*
t195+FOUR*t223*t211*t186*t242*t191*t245*I*t194-FOUR*t143*t183*t185*t252*t196-EIGHT*
t223*t182*t184*t186*t259*t196-EIGHT*t223*t181*t185*t186*t266*t196-EIGHT*t96*t111*
t85*t179*t211*t186*t275*t196;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C245mpmpmpt4(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p5;
    t1 = A(p4,p5);
    t2 = B(p4,p5);
    S1 = -t2*t1;
    gg = AB(p4,K1f,p4);
    S2 = S(p4);
    t4 = AB(K1f,p4,K1f);
    t5 = gg*t4;
    t6 = AB(p4,p4,p4);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p4,p4,K1f);
    t19 = ONE/S1;
    t21 = AB(p4,p4,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C245mpmpmpt3(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t153;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t255;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t268;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t1 = A(p4,p5);
    t2 = B(p4,p5);
    S1 = -t2*t1;
    gg = AB(p4,p5,p4);
    S2 = S(p3);
    t4 = AB(p5,p3,p5);
    t5 = gg*t4;
    t6 = AB(p4,p3,p4);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p4,p3,p5);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p4,p3,p5);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p3,p4);
    t27 = t26*t26;
    t29 = B(p3,p5);
    t30 = t29*t29;
    t33 = IA(p1,p2);
    t34 = IA(p1,p6);
    t35 = t34*t33;
    t37 = IA(p4,p5);
    t38 = t37*t37;
    t40 = IB(p3,p4);
    t41 = t38*t37*t40;
    t42 = IB(p4,p5);
    t43 = t42*t42;
    t44 = t43*t42;
    t45 = p5+p4;
    t46 = IAB(p3,t45,p3);
    t47 = t46*t46;
    t48 = t47*t47;
    t51 = t27*t26*t30*t29*t35*t41*t44*t48;
    t52 = FbtAB(p1,p2,t45,p4,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p1,p6,t45,p4,YY01);
    t55 = t54*t54;
    t56 = t53*t55;
    t57 = FbtAB(p3,p4,t45,p4,YY01);
    t58 = FbtAB(p5,p4,t45,p4,YY01);
    t59 = t58*t58;
    t62 = IFbtAB(p2,p2,t45,p4,YY01);
    t63 = IFbtAB(p6,p6,t45,p4,YY01);
    t64 = t62*t63;
    t65 = p6+p1;
    t66 = IFbtBB(p6,t65,p2,t45,p4,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = B(p2,p6);
    t75 = t74*t74;
    t76 = t75*t75;
    t79 = IB(p1,p2);
    t81 = IB(p1,p6);
    t84 = t26*t76*t29*t38*t79*t81*t40;
    t85 = t43*t47;
    t87 = IS(p6+p2+p1);
    t97 = t26*t29*t35;
    t98 = t44*t47;
    t102 = t55*t57;
    t103 = FbtAB2(p4,p4,t45,p4,YY01,YY21);
    t106 = YY01*YY01;
    t113 = t38*t40;
    t116 = t97*t113*t85*t53;
    t117 = FbtAB2(p2,p2,t45,p4,YY01,YY21);
    t120 = t62*t62;
    t126 = FbtAB2(p6,p6,t45,p4,YY01,YY21);
    t129 = t63*t63;
    t135 = FbtBB2(p6,t65,p2,t45,p4,YY01,YY21);
    t138 = t66*t66;
    t145 = FbtAB2(p5,p4,t45,p4,YY01,YY21);
    t152 = t97*t113*t85;
    t153 = FbtAB2(p3,p4,t45,p4,YY01,YY21);
    t160 = FbtAB2(p1,p6,t45,p4,YY01,YY21);
    t169 = FbtAB2(p1,p2,t45,p4,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t181 = FbtAB(p1,p2,t45,p4,YY0);
    t182 = t181*t181;
    t183 = FbtAB(p1,p6,t45,p4,YY0);
    t184 = t183*t183;
    t185 = t182*t184;
    t186 = FbtAB(p3,p4,t45,p4,YY0);
    t187 = FbtAB(p5,p4,t45,p4,YY0);
    t188 = t187*t187;
    t191 = IFbtAB(p2,p2,t45,p4,YY0);
    t192 = IFbtAB(p6,p6,t45,p4,YY0);
    t193 = t191*t192;
    t194 = IFbtBB(p6,t65,p2,t45,p4,YY0);
    t195 = t194*I;
    t196 = ONE/YY0;
    t197 = t195*t196;
    t198 = t193*t197;
    t213 = t184*t186;
    t214 = FbtAB2(p4,p4,t45,p4,YY0,YY2);
    t217 = YY0*YY0;
    t226 = t97*t113*t85*t182;
    t227 = FbtAB2(p2,p2,t45,p4,YY0,YY2);
    t230 = t191*t191;
    t236 = FbtAB2(p6,p6,t45,p4,YY0,YY2);
    t239 = t192*t192;
    t245 = FbtBB2(p6,t65,p2,t45,p4,YY0,YY2);
    t248 = t194*t194;
    t255 = FbtAB2(p5,p4,t45,p4,YY0,YY2);
    t261 = FbtAB2(p3,p4,t45,p4,YY0,YY2);
    t268 = FbtAB2(p1,p6,t45,p4,YY0,YY2);
    t277 = FbtAB2(p1,p2,t45,p4,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t51*t56*t57*t59*t70-FOUR*t84*t85*t87*t57*t59*t66*I*t68-FOUR*
t97*t41*t98*t53*t102*t59*t103*t64*t67/t106+FOUR*t116*t102*t59*t117*t120*t63*t69+
FOUR*t116*t102*t59*t126*t62*t129*t69+FOUR*t116*t102*t59*t135*t64*t138*I*t68-EIGHT*
t116*t102*t58*t145*t70-FOUR*t152*t56*t59*t153*t70-EIGHT*t116*t54*t57*t59*t160*t70
-EIGHT*t97*t113*t85*t52*t102*t59*t169*t70+FOUR*EIGHT/THREE*t51*t185*t186*t188*t198-FOUR*
t84*t85*t87*t186*t188*t194*I*t196-FOUR*t97*t41*t98*t182*t213*t188*t214*t193*t195
/t217+FOUR*t226*t213*t188*t227*t230*t192*t197+FOUR*t226*t213*t188*t236*t191*t239*
t197+FOUR*t226*t213*t188*t245*t193*t248*I*t196-EIGHT*t226*t213*t187*t255*t198-FOUR*
t152*t185*t188*t261*t198-EIGHT*t226*t183*t186*t188*t268*t198-EIGHT*t97*t113*t85*
t181*t213*t188*t277*t198;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C245mpmpmpt23(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p3+p2;
    t1 = A(p4,p5);
    t2 = B(p4,p5);
    S1 = -t2*t1;
    gg = AB(p4,p5,p4);
    S2 = S(K2);
    t4 = AB(p5,K2,p5);
    t5 = gg*t4;
    t6 = AB(p4,K2,p4);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p4,K2,p5);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p4,K2,p5);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p5+p4;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p2+p3);
    t30 = S(p4+p5);
    t32 = -t28+t29*t30;
    t34 = IA(p1,p6);
    t35 = ONE/t32*t34;
    t36 = IA(p2,p3);
    t37 = IA(p4,p5);
    t38 = t37*t37;
    t39 = t38*t37;
    t42 = IB(p1,p6);
    t43 = IB(p2,p3);
    t44 = t43*t42;
    t45 = IB(p4,p5);
    t46 = t45*t45;
    t47 = t46*t45;
    t50 = t35*t36*t39*t44*t47*t21;
    t51 = FbtAB(p1,p6,t26,p4,YY01);
    t52 = t51*t51;
    t53 = FbtAB(p3,p2,t26,p4,YY01);
    t54 = t53*t53;
    t55 = t52*t54;
    t56 = FbtAB(p5,p4,t26,p4,YY01);
    t57 = t56*t56;
    t58 = FbtAB2(p4,p4,t26,p4,YY01,YY21);
    t61 = IFbtAB(p2,p2,t26,p4,YY01);
    t62 = IFbtAB(p6,p6,t26,p4,YY01);
    t63 = t61*t62;
    t64 = YY01*YY01;
    t74 = t35*t36*t38*t44*t46*t21;
    t75 = FbtAB2(p2,p2,t26,p4,YY01,YY21);
    t78 = t61*t61;
    t81 = I/YY01;
    t85 = FbtAB2(p6,p6,t26,p4,YY01,YY21);
    t88 = t62*t62;
    t93 = FbtAB2(p5,p4,t26,p4,YY01,YY21);
    t96 = t63*t81;
    t101 = FbtAB2(p3,p2,t26,p4,YY01,YY21);
    t108 = FbtAB2(p1,p6,t26,p4,YY01,YY21);
    t114 = t32*t32;
    t121 = ONE/t114*t34*t36*t39*t42*t43*t47;
    t122 = t21*t21;
    t123 = t122*t21;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t136 = FbtAB(p1,p6,t26,p4,YY0);
    t137 = t136*t136;
    t138 = FbtAB(p3,p2,t26,p4,YY0);
    t139 = t138*t138;
    t140 = t137*t139;
    t141 = FbtAB(p5,p4,t26,p4,YY0);
    t142 = t141*t141;
    t143 = FbtAB2(p4,p4,t26,p4,YY0,YY2);
    t146 = IFbtAB(p2,p2,t26,p4,YY0);
    t147 = IFbtAB(p6,p6,t26,p4,YY0);
    t148 = t146*t147;
    t149 = YY0*YY0;
    t155 = FbtAB2(p2,p2,t26,p4,YY0,YY2);
    t158 = t146*t146;
    t161 = I/YY0;
    t165 = FbtAB2(p6,p6,t26,p4,YY0,YY2);
    t168 = t147*t147;
    t173 = FbtAB2(p5,p4,t26,p4,YY0,YY2);
    t176 = t148*t161;
    t181 = FbtAB2(p3,p2,t26,p4,YY0,YY2);
    t188 = FbtAB2(p1,p6,t26,p4,YY0,YY2);
    return(t50*t55*t57*t58*t63*I/t64/TWO-t74*t55*t57*t75*t78*t62*t81/TWO-t74*
t55*t57*t85*t61*t88*t81/TWO+t74*t55*t56*t93*t96+t74*t52*t53*t57*t101*t96+t74*
t51*t54*t57*t108*t96+t121*t123*t52*t54*t57*t96/THREE+t50*t140*t142*t143*t148*I/
t149/TWO-t74*t140*t142*t155*t158*t147*t161/TWO-t74*t140*t142*t165*t146*t168*
t161/TWO+t74*t140*t141*t173*t176+t74*t137*t138*t142*t181*t176+t74*t136*t139*
t142*t188*t176+t121*t123*t137*t139*t142*t176/THREE);
  }
}


std::complex<DOUBLE> C245mpmpmpt123(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t261;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t279;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  FourMomentum<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  FourMomentum<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p3+p2+p1;
    t1 = A(p4,p5);
    t2 = B(p4,p5);
    S1 = -t2*t1;
    gg = AB(p4,p5,p4);
    S2 = S(K2);
    t4 = AB(p5,K2,p5);
    t5 = gg*t4;
    t6 = AB(p4,K2,p4);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p4,K2,p5);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p4,K2,p5);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p3);
    t27 = t26*t26;
    t28 = t27*t27;
    t29 = B(p5,p6);
    t31 = IA(p1,p2);
    t33 = IA(p2,p3);
    t34 = IA(p4,p5);
    t35 = t34*t34;
    t37 = IB(p4,p5);
    t38 = t37*t37;
    t40 = t28*t29*t31*t33*t35*t38;
    t41 = p5+p4;
    t42 = IAB(p6,t41,p6);
    t43 = t42*t42;
    t44 = IS(K2);
    t45 = t44*t43;
    t46 = FbtAB(p4,p6,t41,p4,YY01);
    t48 = FbtAB(p5,p4,t41,p4,YY01);
    t49 = t48*t48;
    t50 = p2+p1;
    t51 = IFbtAA(p1,t50,p3,t41,p4,YY01);
    t54 = I/YY01;
    t59 = A(p4,p6);
    t60 = t59*t59;
    t61 = t29*t29;
    t64 = t35*t34;
    t65 = IB(p1,p2);
    t67 = t60*t61*t29*t64*t65;
    t68 = IB(p2,p3);
    t69 = t38*t37;
    t70 = t68*t69;
    t71 = t43*t43;
    t72 = FbtAB(p1,p2,t41,p4,YY01);
    t73 = t72*t72;
    t77 = FbtAB(p3,p2,t41,p4,YY01);
    t78 = t77*t77;
    t80 = IFbtAB(p1,p1,t41,p4,YY01);
    t83 = IFbtAB(p3,p3,t41,p4,YY01);
    t84 = t83*t51;
    t85 = t84*t54;
    t90 = t65*t68;
    t91 = t29*t64*t90;
    t92 = t69*t43;
    t93 = t73*t78;
    t96 = t46*t49;
    t97 = FbtAB2(p4,p4,t41,p4,YY01,YY21);
    t100 = YY01*YY01;
    t107 = t29*t35;
    t108 = t107*t90;
    t109 = t38*t43;
    t111 = t108*t109*t93;
    t112 = FbtAB2(p1,p1,t41,p4,YY01,YY21);
    t113 = t80*t80;
    t119 = FbtAB2(p3,p3,t41,p4,YY01,YY21);
    t122 = t83*t83;
    t128 = FbtAA2(p1,t50,p3,t41,p4,YY01,YY21);
    t131 = t51*t51;
    t138 = FbtAB2(p5,p4,t41,p4,YY01,YY21);
    t144 = t107*t65;
    t145 = t68*t38;
    t150 = FbtAB2(p4,p6,t41,p4,YY01,YY21);
    t159 = FbtAB2(p3,p2,t41,p4,YY01,YY21);
    t168 = FbtAB2(p1,p2,t41,p4,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t180 = FbtAB(p4,p6,t41,p4,YY0);
    t182 = FbtAB(p5,p4,t41,p4,YY0);
    t183 = t182*t182;
    t184 = IFbtAA(p1,t50,p3,t41,p4,YY0);
    t187 = I/YY0;
    t192 = FbtAB(p1,p2,t41,p4,YY0);
    t193 = t192*t192;
    t197 = FbtAB(p3,p2,t41,p4,YY0);
    t198 = t197*t197;
    t200 = IFbtAB(p1,p1,t41,p4,YY0);
    t203 = IFbtAB(p3,p3,t41,p4,YY0);
    t204 = t203*t184;
    t205 = t204*t187;
    t209 = t193*t198;
    t212 = t180*t183;
    t213 = FbtAB2(p4,p4,t41,p4,YY0,YY2);
    t216 = YY0*YY0;
    t224 = t108*t109*t209;
    t225 = FbtAB2(p1,p1,t41,p4,YY0,YY2);
    t226 = t200*t200;
    t232 = FbtAB2(p3,p3,t41,p4,YY0,YY2);
    t235 = t203*t203;
    t241 = FbtAA2(p1,t50,p3,t41,p4,YY0,YY2);
    t244 = t184*t184;
    t251 = FbtAB2(p5,p4,t41,p4,YY0,YY2);
    t261 = FbtAB2(p4,p6,t41,p4,YY0,YY2);
    t270 = FbtAB2(p3,p2,t41,p4,YY0,YY2);
    t279 = FbtAB2(p1,p2,t41,p4,YY0,YY2);
    ans = FOUR*t40*t45*t46*t49*t51*t54-FOUR*EIGHT/THREE*t67*t70*t71*t73*t78*t46*t49*t80*
t85+FOUR*t91*t92*t93*t96*t97*t80*t84*I/t100-FOUR*t111*t96*t112*t113*t85-FOUR*t111*
t96*t119*t80*t122*t51*t54-FOUR*t111*t96*t128*t80*t83*t131*t54+EIGHT*t111*t46*t48*
t138*t80*t85+FOUR*t144*t145*t43*t73*t78*t49*t150*t80*t85+EIGHT*t108*t109*t73*t77*
t96*t159*t80*t85+EIGHT*t108*t109*t72*t78*t96*t168*t80*t85+FOUR*t40*t45*t180*t183*
t184*t187-FOUR*EIGHT/THREE*t67*t70*t71*t193*t198*t180*t183*t200*t205+FOUR*t91*t92*t209*
t212*t213*t200*t204*I/t216-FOUR*t224*t212*t225*t226*t205-FOUR*t224*t212*t232*t200
*t235*t184*t187-FOUR*t224*t212*t241*t200*t203*t244*t187+EIGHT*t224*t180*t182*t251*
t200*t205+FOUR*t144*t145*t43*t193*t198*t183*t261*t200*t205+EIGHT*t108*t109*t193*
t197*t212*t270*t200*t205+EIGHT*t108*t109*t192*t198*t212*t279*t200*t205;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C256mpmpmpt5(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p6;
    t1 = A(p5,p6);
    t2 = B(p5,p6);
    S1 = -t2*t1;
    gg = AB(p5,K1f,p5);
    S2 = S(p5);
    t4 = AB(K1f,p5,K1f);
    t5 = gg*t4;
    t6 = AB(p5,p5,p5);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p5,p5,K1f);
    t19 = ONE/S1;
    t21 = AB(p5,p5,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C256mpmpmpt4(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t206;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t233;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t242;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t262;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t271;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t280;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  FourMomentum<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t1 = A(p5,p6);
    t2 = B(p5,p6);
    S1 = -t2*t1;
    gg = AB(p5,p6,p5);
    S2 = S(p4);
    t4 = AB(p6,p4,p6);
    t5 = gg*t4;
    t6 = AB(p5,p4,p5);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p5,p4,p6);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p5,p4,p6);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p3);
    t27 = t26*t26;
    t28 = t27*t27;
    t29 = B(p4,p6);
    t31 = IA(p1,p2);
    t33 = IA(p2,p3);
    t34 = IA(p5,p6);
    t35 = t34*t34;
    t37 = IB(p5,p6);
    t38 = t37*t37;
    t40 = t28*t29*t31*t33*t35*t38;
    t41 = p6+p5;
    t42 = IAB(p4,t41,p4);
    t43 = t42*t42;
    t45 = IS(p3+p2+p1);
    t46 = t43*t45;
    t47 = FbtAB(p5,p4,t41,p5,YY01);
    t49 = FbtAB(p5,p6,t41,p5,YY01);
    t50 = t49*t49;
    t51 = p2+p1;
    t52 = IFbtAA(p1,t51,p3,t41,p5,YY01);
    t55 = I/YY01;
    t60 = A(p4,p5);
    t61 = t60*t60;
    t62 = t29*t29;
    t65 = t35*t34;
    t66 = IB(p1,p2);
    t68 = t61*t62*t29*t65*t66;
    t69 = IB(p2,p3);
    t70 = t38*t37;
    t71 = t69*t70;
    t72 = t43*t43;
    t73 = FbtAB(p1,p2,t41,p5,YY01);
    t74 = t73*t73;
    t78 = FbtAB(p3,p2,t41,p5,YY01);
    t79 = t78*t78;
    t81 = IFbtAB(p1,p1,t41,p5,YY01);
    t84 = IFbtAB(p3,p3,t41,p5,YY01);
    t85 = t84*t52;
    t86 = t85*t55;
    t91 = t66*t69;
    t92 = t29*t65*t91;
    t93 = t70*t43;
    t94 = t74*t79;
    t97 = t47*t50;
    t98 = FbtAB2(p5,p5,t41,p5,YY01,YY21);
    t101 = YY01*YY01;
    t108 = t29*t35;
    t109 = t108*t91;
    t110 = t38*t43;
    t112 = t109*t110*t94;
    t113 = FbtAB2(p1,p1,t41,p5,YY01,YY21);
    t114 = t81*t81;
    t120 = FbtAB2(p3,p3,t41,p5,YY01,YY21);
    t123 = t84*t84;
    t129 = FbtAA2(p1,t51,p3,t41,p5,YY01,YY21);
    t132 = t52*t52;
    t139 = FbtAB2(p5,p6,t41,p5,YY01,YY21);
    t145 = t108*t66;
    t146 = t38*t69;
    t151 = FbtAB2(p5,p4,t41,p5,YY01,YY21);
    t160 = FbtAB2(p3,p2,t41,p5,YY01,YY21);
    t169 = FbtAB2(p1,p2,t41,p5,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t181 = FbtAB(p5,p4,t41,p5,YY0);
    t183 = FbtAB(p5,p6,t41,p5,YY0);
    t184 = t183*t183;
    t185 = IFbtAA(p1,t51,p3,t41,p5,YY0);
    t188 = I/YY0;
    t193 = FbtAB(p1,p2,t41,p5,YY0);
    t194 = t193*t193;
    t198 = FbtAB(p3,p2,t41,p5,YY0);
    t199 = t198*t198;
    t201 = IFbtAB(p1,p1,t41,p5,YY0);
    t204 = IFbtAB(p3,p3,t41,p5,YY0);
    t205 = t204*t185;
    t206 = t205*t188;
    t210 = t194*t199;
    t213 = t181*t184;
    t214 = FbtAB2(p5,p5,t41,p5,YY0,YY2);
    t217 = YY0*YY0;
    t225 = t109*t110*t210;
    t226 = FbtAB2(p1,p1,t41,p5,YY0,YY2);
    t227 = t201*t201;
    t233 = FbtAB2(p3,p3,t41,p5,YY0,YY2);
    t236 = t204*t204;
    t242 = FbtAA2(p1,t51,p3,t41,p5,YY0,YY2);
    t245 = t185*t185;
    t252 = FbtAB2(p5,p6,t41,p5,YY0,YY2);
    t262 = FbtAB2(p5,p4,t41,p5,YY0,YY2);
    t271 = FbtAB2(p3,p2,t41,p5,YY0,YY2);
    t280 = FbtAB2(p1,p2,t41,p5,YY0,YY2);
    ans = FOUR*t40*t46*t47*t50*t52*t55-FOUR*EIGHT/THREE*t68*t71*t72*t74*t79*t47*t50*t81*
t86+FOUR*t92*t93*t94*t97*t98*t81*t85*I/t101-FOUR*t112*t97*t113*t114*t86-FOUR*t112*
t97*t120*t81*t123*t52*t55-FOUR*t112*t97*t129*t81*t84*t132*t55+EIGHT*t112*t47*t49*
t139*t81*t86+FOUR*t145*t146*t43*t74*t79*t50*t151*t81*t86+EIGHT*t109*t110*t74*t78*
t97*t160*t81*t86+EIGHT*t109*t110*t73*t79*t97*t169*t81*t86+FOUR*t40*t46*t181*t184*
t185*t188-FOUR*EIGHT/THREE*t68*t71*t72*t194*t199*t181*t184*t201*t206+FOUR*t92*t93*t210*
t213*t214*t201*t205*I/t217-FOUR*t225*t213*t226*t227*t206-FOUR*t225*t213*t233*t201
*t236*t185*t188-FOUR*t225*t213*t242*t201*t204*t245*t188+EIGHT*t225*t181*t183*t252*
t201*t206+FOUR*t145*t146*t43*t194*t199*t184*t262*t201*t206+EIGHT*t109*t110*t194*
t198*t213*t271*t201*t206+EIGHT*t109*t110*t193*t199*t213*t280*t201*t206;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C256mpmpmpt34(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p4+p3;
    t1 = A(p5,p6);
    t2 = B(p5,p6);
    S1 = -t2*t1;
    gg = AB(p5,p6,p5);
    S2 = S(K2);
    t4 = AB(p6,K2,p6);
    t5 = gg*t4;
    t6 = AB(p5,K2,p5);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p5,K2,p6);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p5,K2,p6);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p6+p5;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p3+p4);
    t30 = S(p5+p6);
    t32 = -t28+t29*t30;
    t34 = IA(p1,p2);
    t35 = ONE/t32*t34;
    t36 = IA(p3,p4);
    t37 = IA(p5,p6);
    t38 = t37*t37;
    t39 = t38*t37;
    t42 = IB(p1,p2);
    t43 = IB(p3,p4);
    t44 = t43*t42;
    t45 = IB(p5,p6);
    t46 = t45*t45;
    t47 = t46*t45;
    t50 = t35*t36*t39*t44*t47*t21;
    t51 = FbtAB(p1,p2,t26,p5,YY01);
    t52 = t51*t51;
    t53 = FbtAB(p3,p4,t26,p5,YY01);
    t54 = t53*t53;
    t55 = t52*t54;
    t56 = FbtAB(p5,p6,t26,p5,YY01);
    t57 = t56*t56;
    t58 = FbtAB2(p5,p5,t26,p5,YY01,YY21);
    t61 = IFbtAB(p1,p1,t26,p5,YY01);
    t62 = IFbtAB(p3,p3,t26,p5,YY01);
    t63 = t61*t62;
    t64 = YY01*YY01;
    t74 = t35*t36*t38*t44*t46*t21;
    t75 = FbtAB2(p1,p1,t26,p5,YY01,YY21);
    t78 = t61*t61;
    t81 = I/YY01;
    t85 = FbtAB2(p3,p3,t26,p5,YY01,YY21);
    t88 = t62*t62;
    t93 = FbtAB2(p5,p6,t26,p5,YY01,YY21);
    t96 = t63*t81;
    t101 = FbtAB2(p3,p4,t26,p5,YY01,YY21);
    t108 = FbtAB2(p1,p2,t26,p5,YY01,YY21);
    t114 = t32*t32;
    t121 = ONE/t114*t34*t36*t39*t42*t43*t47;
    t122 = t21*t21;
    t123 = t122*t21;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t136 = FbtAB(p1,p2,t26,p5,YY0);
    t137 = t136*t136;
    t138 = FbtAB(p3,p4,t26,p5,YY0);
    t139 = t138*t138;
    t140 = t137*t139;
    t141 = FbtAB(p5,p6,t26,p5,YY0);
    t142 = t141*t141;
    t143 = FbtAB2(p5,p5,t26,p5,YY0,YY2);
    t146 = IFbtAB(p1,p1,t26,p5,YY0);
    t147 = IFbtAB(p3,p3,t26,p5,YY0);
    t148 = t146*t147;
    t149 = YY0*YY0;
    t155 = FbtAB2(p1,p1,t26,p5,YY0,YY2);
    t158 = t146*t146;
    t161 = I/YY0;
    t165 = FbtAB2(p3,p3,t26,p5,YY0,YY2);
    t168 = t147*t147;
    t173 = FbtAB2(p5,p6,t26,p5,YY0,YY2);
    t176 = t148*t161;
    t181 = FbtAB2(p3,p4,t26,p5,YY0,YY2);
    t188 = FbtAB2(p1,p2,t26,p5,YY0,YY2);
    return(t50*t55*t57*t58*t63*I/t64/TWO-t74*t55*t57*t75*t78*t62*t81/TWO-t74*
t55*t57*t85*t61*t88*t81/TWO+t74*t55*t56*t93*t96+t74*t52*t53*t57*t101*t96+t74*
t51*t54*t57*t108*t96+t121*t123*t52*t54*t57*t96/THREE+t50*t140*t142*t143*t148*I/
t149/TWO-t74*t140*t142*t155*t158*t147*t161/TWO-t74*t140*t142*t165*t146*t168*
t161/TWO+t74*t140*t141*t173*t176+t74*t137*t138*t142*t181*t176+t74*t136*t139*
t142*t188*t176+t121*t123*t137*t139*t142*t176/THREE);
  }
}


std::complex<DOUBLE> C256mpmpmpt234(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t117;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t264;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p4+p3+p2;
    t1 = A(p5,p6);
    t2 = B(p5,p6);
    S1 = -t2*t1;
    gg = AB(p5,p6,p5);
    S2 = S(K2);
    t4 = AB(p6,K2,p6);
    t5 = gg*t4;
    t6 = AB(p5,K2,p5);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p5,K2,p6);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p5,K2,p6);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p1,p5);
    t27 = t26*t26;
    t29 = B(p1,p6);
    t30 = t29*t29;
    t33 = IA(p2,p3);
    t34 = IA(p3,p4);
    t35 = t34*t33;
    t37 = IA(p5,p6);
    t38 = t37*t37;
    t40 = IB(p1,p5);
    t41 = t38*t37*t40;
    t42 = IB(p5,p6);
    t43 = t42*t42;
    t44 = t43*t42;
    t45 = p6+p5;
    t46 = IAB(p1,t45,p1);
    t47 = t46*t46;
    t48 = t47*t47;
    t51 = t27*t26*t30*t29*t35*t41*t44*t48;
    t52 = FbtAB(p1,p5,t45,p5,YY01);
    t53 = FbtAB(p3,p2,t45,p5,YY01);
    t54 = t53*t53;
    t56 = FbtAB(p3,p4,t45,p5,YY01);
    t57 = t56*t56;
    t58 = FbtAB(p5,p6,t45,p5,YY01);
    t59 = t58*t58;
    t62 = IFbtAB(p2,p2,t45,p5,YY01);
    t63 = IFbtAB(p4,p4,t45,p5,YY01);
    t64 = t62*t63;
    t65 = p3+p2;
    t66 = IFbtBB(p2,t65,p4,t45,p5,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = t26*t29;
    t75 = B(p2,p4);
    t76 = t75*t75;
    t77 = t76*t76;
    t79 = t38*t40;
    t80 = IB(p2,p3);
    t81 = IB(p3,p4);
    t84 = t74*t77*t79*t80*t81;
    t85 = t43*t47;
    t86 = IS(K2);
    t95 = t74*t35;
    t96 = t44*t47;
    t100 = t54*t57;
    t101 = FbtAB2(p5,p5,t45,p5,YY01,YY21);
    t104 = YY01*YY01;
    t113 = t95*t79*t85*t52;
    t114 = FbtAB2(p2,p2,t45,p5,YY01,YY21);
    t117 = t62*t62;
    t123 = FbtAB2(p4,p4,t45,p5,YY01,YY21);
    t126 = t63*t63;
    t132 = FbtBB2(p2,t65,p4,t45,p5,YY01,YY21);
    t135 = t66*t66;
    t142 = FbtAB2(p5,p6,t45,p5,YY01,YY21);
    t149 = FbtAB2(p3,p4,t45,p5,YY01,YY21);
    t156 = FbtAB2(p3,p2,t45,p5,YY01,YY21);
    t163 = t95*t79*t85;
    t164 = FbtAB2(p1,p5,t45,p5,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t176 = FbtAB(p1,p5,t45,p5,YY0);
    t177 = FbtAB(p3,p2,t45,p5,YY0);
    t178 = t177*t177;
    t180 = FbtAB(p3,p4,t45,p5,YY0);
    t181 = t180*t180;
    t182 = FbtAB(p5,p6,t45,p5,YY0);
    t183 = t182*t182;
    t186 = IFbtAB(p2,p2,t45,p5,YY0);
    t187 = IFbtAB(p4,p4,t45,p5,YY0);
    t188 = t186*t187;
    t189 = IFbtBB(p2,t65,p4,t45,p5,YY0);
    t190 = t189*I;
    t191 = ONE/YY0;
    t192 = t190*t191;
    t193 = t188*t192;
    t208 = t178*t181;
    t209 = FbtAB2(p5,p5,t45,p5,YY0,YY2);
    t212 = YY0*YY0;
    t221 = t95*t79*t85*t176;
    t222 = FbtAB2(p2,p2,t45,p5,YY0,YY2);
    t225 = t186*t186;
    t231 = FbtAB2(p4,p4,t45,p5,YY0,YY2);
    t234 = t187*t187;
    t240 = FbtBB2(p2,t65,p4,t45,p5,YY0,YY2);
    t243 = t189*t189;
    t250 = FbtAB2(p5,p6,t45,p5,YY0,YY2);
    t257 = FbtAB2(p3,p4,t45,p5,YY0,YY2);
    t264 = FbtAB2(p3,p2,t45,p5,YY0,YY2);
    t270 = FbtAB2(p1,p5,t45,p5,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t51*t52*t54*t57*t59*t70-FOUR*t84*t85*t86*t52*t59*t66*I*t68
-FOUR*t95*t41*t96*t52*t100*t59*t101*t64*t67/t104+FOUR*t113*t100*t59*t114*t117*t63
*t69+FOUR*t113*t100*t59*t123*t62*t126*t69+FOUR*t113*t100*t59*t132*t64*t135*I*t68
-EIGHT*t113*t100*t58*t142*t70-EIGHT*t113*t54*t56*t59*t149*t70-EIGHT*t113*t53*t57*t59*
t156*t70-FOUR*t163*t100*t59*t164*t70+FOUR*EIGHT/THREE*t51*t176*t178*t181*t183*t193-FOUR*
t84*t85*t86*t176*t183*t189*I*t191-FOUR*t95*t41*t96*t176*t208*t183*t209*t188*t190
/t212+FOUR*t221*t208*t183*t222*t225*t187*t192+FOUR*t221*t208*t183*t231*t186*t234*
t192+FOUR*t221*t208*t183*t240*t188*t243*I*t191-EIGHT*t221*t208*t182*t250*t193-EIGHT*
t221*t178*t180*t183*t257*t193-EIGHT*t221*t177*t181*t183*t264*t193-FOUR*t163*t208*
t183*t270*t193;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C261mpmpmpt6(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<std::complex<DOUBLE> > K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K1f = -p6;
    t1 = A(p1,p6);
    t2 = B(p1,p6);
    S1 = -t2*t1;
    gg = AB(p6,K1f,p6);
    S2 = S(p6);
    t4 = AB(K1f,p6,K1f);
    t5 = gg*t4;
    t6 = AB(p6,p6,p6);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p6,p6,K1f);
    t19 = ONE/S1;
    t21 = AB(p6,p6,K1f);
    t24 = t14/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C261mpmpmpt5(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  DOUBLE xsign;
  {
//    std::cout << "C2_61:t5" << std::endl;
    K1f = p1;
//    std::cout << "K1f=" << K1f << std::endl;
    K2 = p5;
    S1 = -A(p1,p6)*B(p1,p6);
    gg = AB(p6,K1f,p6);
    S2 = S(K2);
    alpha = pow(gg*AB(K1f,K2,K1f)+S1*AB(p6,K2,p6),TWO)-FOUR*S1*S2*gg*gg;
//    std::cout << "alpha=" << alpha << std::endl;
    xsign = ONE;
    YY0 = (gg*AB(K1f,K2,K1f)-S1*AB(p6,K2,p6)+xsign*sqrt(alpha))*IAB(p6,K2,K1f)/S1/TWO;
    YY2 = -xsign*AB(p6,K2,K1f)*sqrt(alpha)/alpha;

//    std::cout << "yy0+=" << YY0 << std::endl;
//    std::cout << "test=" << 
//	FOUR*EIGHT/THREE*
//	pow(A(p5,p6),THREE)*pow(B(p1,p5),THREE)*pow(IA(p1,p6),THREE)
//	*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),THREE)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),FOUR)
//	*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)
//	*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)
//	*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)
//	FbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)
//      << std::endl;

    ans = FOUR*EIGHT/THREE*pow(A(p5,p6),THREE)*pow(B(p1,p5),THREE)*pow(IA(p1,p6),THREE)*IA(p2
,p3)*IA(p3,p4)*pow(IB(p1,p6),THREE)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),FOUR)*pow(FbtAB
(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6
+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4
,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-FOUR*A(p5,p6)*B(p1,p5)*
pow(B(p2,p4),FOUR)*pow(IA(p1,p6),TWO)*pow(IB(p1,p6),TWO)*IB(p2,p3)*IB(p3,p4)*IB(
p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*IS(p4+p3+p2)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO
)*FbtAB(p5,p6,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-FOUR*A(p5,p6)
*B(p1,p5)*pow(IA(p1,p6),THREE)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),THREE)*IB(p5,p6)*
pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6
+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*
FbtAB2(p6,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6
,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/(YY0*YY0)+FOUR*A(p5,p6)*B(p1,p5)*pow(IA
(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,
p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*
pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p2,p2,p6+p1
,p6,YY0,YY2)*pow(IFbtAB(p2,p2,p6+p1,p6,YY0),TWO)*IFbtAB(p4,p4,p6+p1,p6,YY0)*
IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0+FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)
*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow
(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3
,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p4,p4,p6+p1,p6,YY0,YY2)
*IFbtAB(p2,p2,p6+p1,p6,YY0)*pow(IFbtAB(p4,p4,p6+p1,p6,YY0),TWO)*IFbtBB(p2,p3+p2
,p4,p6+p1,p6,YY0)*I/YY0+FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(
p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,
p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,
YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtBB2(p2,p3+p2,p4,p6+p1,p6,YY0,YY2)*IFbtAB
(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*pow(IFbtBB(p2,p3+p2,p4,p6+p1,p6
,YY0),TWO)*I/YY0-FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*
pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6
,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO
)*FbtAB2(p5,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,
p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-EIGHT*A(p5,p6)*B(p1,p5)*pow(IA(p1,
p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),
TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*
FbtAB(p3,p4,p6+p1,p6,YY0)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p3,p4,p6+p1,p6,YY0,
YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,
p6+p1,p6,YY0)*I/YY0-EIGHT*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4
)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,
p6,YY0),TWO)*FbtAB(p3,p2,p6+p1,p6,YY0)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB
(p5,p6,p6+p1,p6,YY0)*FbtAB2(p3,p2,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*
IFbtAB(p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-EIGHT*A(p5,p6)*
B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*
pow(IAB(p5,p6+p1,p5),TWO)*FbtAB(p1,p6,p6+p1,p6,YY0)*pow(FbtAB(p3,p2,p6+p1,p6,
YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(
p1,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*
IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0;
    xsign = -ONE;
    YY0 = (gg*AB(K1f,K2,K1f)-S1*AB(p6,K2,p6)+xsign*sqrt(alpha))*IAB(p6,K2,K1f)/
S1/TWO;
    YY2 = -xsign*AB(p6,K2,K1f)*sqrt(alpha)/alpha;
    ans += FOUR*EIGHT/THREE*pow(A(p5,p6),THREE)*pow(B(p1,p5),THREE)*pow(IA(p1,p6),THREE)*IA(
p2,p3)*IA(p3,p4)*pow(IB(p1,p6),THREE)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),FOUR)*pow(
FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,
p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*IFbtAB(p2,p2,p6+p1,p6,YY0)*
IFbtAB(p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-FOUR*A(p5,p6)*
B(p1,p5)*pow(B(p2,p4),FOUR)*pow(IA(p1,p6),TWO)*pow(IB(p1,p6),TWO)*IB(p2,p3)*IB(
p3,p4)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*IS(p4+p3+p2)*pow(FbtAB(p1,p6,p6+p1,
p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0
-FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),THREE)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),THREE
)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(
FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,
p6+p1,p6,YY0)*FbtAB2(p6,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(
p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/(YY0*YY0)+FOUR*A(p5,p6)*B
(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow
(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1
,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*
FbtAB2(p2,p2,p6+p1,p6,YY0,YY2)*pow(IFbtAB(p2,p2,p6+p1,p6,YY0),TWO)*IFbtAB(p4,p4
,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0+FOUR*A(p5,p6)*B(p1,p5)*pow
(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+
p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),
TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p4,p4,
p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*pow(IFbtAB(p4,p4,p6+p1,p6,YY0),TWO
)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0+FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),
TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)
*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(
FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtBB2(p2,p3+p2,p4,p6+
p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*pow(IFbtBB
(p2,p3+p2,p4,p6+p1,p6,YY0),TWO)*I/YY0-FOUR*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)*
IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow(
FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,
p4,p6+p1,p6,YY0),TWO)*FbtAB2(p5,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)
*IFbtAB(p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-EIGHT*A(p5,p6)
*B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*
pow(IAB(p5,p6+p1,p5),TWO)*pow(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p2,p6
+p1,p6,YY0),TWO)*FbtAB(p3,p4,p6+p1,p6,YY0)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p3,
p4,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*
IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0-EIGHT*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)
*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*pow
(FbtAB(p1,p6,p6+p1,p6,YY0),TWO)*FbtAB(p3,p2,p6+p1,p6,YY0)*pow(FbtAB(p3,p4,p6+p1
,p6,YY0),TWO)*FbtAB(p5,p6,p6+p1,p6,YY0)*FbtAB2(p3,p2,p6+p1,p6,YY0,YY2)*IFbtAB(
p2,p2,p6+p1,p6,YY0)*IFbtAB(p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)
*I/YY0-EIGHT*A(p5,p6)*B(p1,p5)*pow(IA(p1,p6),TWO)*IA(p2,p3)*IA(p3,p4)*pow(IB(p1,
p6),TWO)*IB(p5,p6)*pow(IAB(p5,p6+p1,p5),TWO)*FbtAB(p1,p6,p6+p1,p6,YY0)*pow(
FbtAB(p3,p2,p6+p1,p6,YY0),TWO)*pow(FbtAB(p3,p4,p6+p1,p6,YY0),TWO)*FbtAB(p5,p6,
p6+p1,p6,YY0)*FbtAB2(p1,p6,p6+p1,p6,YY0,YY2)*IFbtAB(p2,p2,p6+p1,p6,YY0)*IFbtAB(
p4,p4,p6+p1,p6,YY0)*IFbtBB(p2,p3+p2,p4,p6+p1,p6,YY0)*I/YY0;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C261mpmpmpt45(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t146;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t161;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  FourMomentum<DOUBLE> t26;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t93;
  std::complex<DOUBLE> t96;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p5+p4;
    t1 = A(p1,p6);
    t2 = B(p1,p6);
    S1 = -t2*t1;
    gg = AB(p6,p1,p6);
    S2 = S(K2);
    t4 = AB(p1,K2,p1);
    t5 = gg*t4;
    t6 = AB(p6,K2,p6);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p6,K2,p1);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p6,K2,p1);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = p6+p1;
    t27 = dot(t26,K2);
    t28 = t27*t27;
    t29 = S(p1+p6);
    t30 = S(p4+p5);
    t32 = -t28+t29*t30;
    t33 = ONE/t32;
    t34 = IA(p1,p6);
    t35 = t34*t34;
    t36 = t35*t34;
    t38 = IA(p2,p3);
    t39 = IA(p4,p5);
    t40 = t38*t39;
    t42 = IB(p1,p6);
    t43 = t42*t42;
    t44 = t43*t42;
    t45 = IB(p2,p3);
    t47 = IB(p4,p5);
    t48 = t47*t21;
    t50 = t33*t36*t40*t44*t45*t48;
    t51 = FbtAB(p1,p6,t26,p6,YY01);
    t52 = t51*t51;
    t53 = FbtAB(p3,p2,t26,p6,YY01);
    t54 = t53*t53;
    t55 = t52*t54;
    t56 = FbtAB(p5,p4,t26,p6,YY01);
    t57 = t56*t56;
    t58 = FbtAB2(p6,p6,t26,p6,YY01,YY21);
    t61 = IFbtAB(p2,p2,t26,p6,YY01);
    t62 = IFbtAB(p4,p4,t26,p6,YY01);
    t63 = t61*t62;
    t64 = YY01*YY01;
    t74 = t33*t35*t40*t43*t45*t48;
    t75 = FbtAB2(p2,p2,t26,p6,YY01,YY21);
    t78 = t61*t61;
    t81 = I/YY01;
    t85 = FbtAB2(p4,p4,t26,p6,YY01,YY21);
    t88 = t62*t62;
    t93 = FbtAB2(p5,p4,t26,p6,YY01,YY21);
    t96 = t63*t81;
    t101 = FbtAB2(p3,p2,t26,p6,YY01,YY21);
    t108 = FbtAB2(p1,p6,t26,p6,YY01,YY21);
    t114 = t32*t32;
    t121 = ONE/t114*t36*t38*t39*t44*t45*t47;
    t122 = t21*t21;
    t123 = t122*t21;
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t136 = FbtAB(p1,p6,t26,p6,YY0);
    t137 = t136*t136;
    t138 = FbtAB(p3,p2,t26,p6,YY0);
    t139 = t138*t138;
    t140 = t137*t139;
    t141 = FbtAB(p5,p4,t26,p6,YY0);
    t142 = t141*t141;
    t143 = FbtAB2(p6,p6,t26,p6,YY0,YY2);
    t146 = IFbtAB(p2,p2,t26,p6,YY0);
    t147 = IFbtAB(p4,p4,t26,p6,YY0);
    t148 = t146*t147;
    t149 = YY0*YY0;
    t155 = FbtAB2(p2,p2,t26,p6,YY0,YY2);
    t158 = t146*t146;
    t161 = I/YY0;
    t165 = FbtAB2(p4,p4,t26,p6,YY0,YY2);
    t168 = t147*t147;
    t173 = FbtAB2(p5,p4,t26,p6,YY0,YY2);
    t176 = t148*t161;
    t181 = FbtAB2(p3,p2,t26,p6,YY0,YY2);
    t188 = FbtAB2(p1,p6,t26,p6,YY0,YY2);
    return(t50*t55*t57*t58*t63*I/t64/TWO-t74*t55*t57*t75*t78*t62*t81/TWO-t74*
t55*t57*t85*t61*t88*t81/TWO+t74*t55*t56*t93*t96+t74*t52*t53*t57*t101*t96+t74*
t51*t54*t57*t108*t96+t121*t123*t52*t54*t57*t96/THREE+t50*t140*t142*t143*t148*I/
t149/TWO-t74*t140*t142*t155*t158*t147*t161/TWO-t74*t140*t142*t165*t146*t168*
t161/TWO+t74*t140*t141*t173*t176+t74*t137*t138*t142*t181*t176+t74*t136*t139*
t142*t188*t176+t121*t123*t137*t139*t142*t176/THREE);
  }
}


std::complex<DOUBLE> C261mpmpmpt345(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t111;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t140;
  std::complex<DOUBLE> t147;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t213;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t221;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t240;
  std::complex<DOUBLE> t249;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  FourMomentum<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  FourMomentum<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t91;
  std::complex<DOUBLE> t92;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    K2 = p5+p4+p3;
    t1 = A(p1,p6);
    t2 = B(p1,p6);
    S1 = -t2*t1;
    gg = AB(p6,p1,p6);
    S2 = S(K2);
    t4 = AB(p1,K2,p1);
    t5 = gg*t4;
    t6 = AB(p6,K2,p6);
    t7 = S1*t6;
    t9 = pow(t5+t7,TWO);
    t11 = gg*gg;
    alpha = t9-FOUR*S1*S2*t11;
    xsign1 = ONE;
    t14 = sqrt(alpha);
    t17 = IAB(p6,K2,p1);
    t19 = ONE/S1;
    YY01 = (t5-t7+xsign1*t14)*t17*t19/TWO;
    t21 = AB(p6,K2,p1);
    t24 = t14/alpha;
    YY21 = -xsign1*t21*t24;
    t26 = A(p2,p6);
    t27 = t26*t26;
    t28 = B(p1,p2);
    t29 = t28*t28;
    t32 = IA(p1,p6);
    t33 = t32*t32;
    t34 = t33*t32;
    t35 = IB(p1,p6);
    t36 = t35*t35;
    t37 = t36*t35;
    t39 = t27*t29*t28*t34*t37;
    t40 = IB(p3,p4);
    t41 = IB(p4,p5);
    t42 = t40*t41;
    t43 = p6+p1;
    t44 = IAB(p2,t43,p2);
    t45 = t44*t44;
    t46 = t45*t45;
    t47 = FbtAB(p1,p6,t43,p6,YY01);
    t48 = t47*t47;
    t52 = FbtAB(p3,p4,t43,p6,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p5,p4,t43,p6,YY01);
    t55 = t54*t54;
    t56 = t53*t55;
    t57 = FbtAB(p6,p2,t43,p6,YY01);
    t58 = IFbtAB(p3,p3,t43,p6,YY01);
    t61 = IFbtAB(p5,p5,t43,p6,YY01);
    t62 = p4+p3;
    t63 = IFbtAA(p3,t62,p5,t43,p6,YY01);
    t64 = t61*t63;
    t66 = I/YY01;
    t67 = t64*t66;
    t71 = A(p3,p5);
    t72 = t71*t71;
    t73 = t72*t72;
    t76 = IA(p3,p4);
    t77 = IA(p4,p5);
    t80 = t73*t28*t33*t76*t77*t36;
    t81 = IS(K2);
    t82 = t45*t81;
    t91 = t28*t34*t37*t40;
    t92 = t41*t45;
    t94 = t92*t48*t53;
    t96 = t55*t57;
    t97 = FbtAB2(p6,p6,t43,p6,YY01,YY21);
    t100 = YY01*YY01;
    t107 = t28*t33;
    t109 = t107*t36*t40;
    t110 = t109*t94;
    t111 = FbtAB2(p3,p3,t43,p6,YY01,YY21);
    t112 = t58*t58;
    t118 = FbtAB2(p5,p5,t43,p6,YY01,YY21);
    t121 = t61*t61;
    t127 = FbtAA2(p3,t62,p5,t43,p6,YY01,YY21);
    t130 = t63*t63;
    t136 = t107*t36;
    t140 = FbtAB2(p6,p2,t43,p6,YY01,YY21);
    t147 = FbtAB2(p5,p4,t43,p6,YY01,YY21);
    t156 = FbtAB2(p3,p4,t43,p6,YY01,YY21);
    t165 = FbtAB2(p1,p6,t43,p6,YY01,YY21);
    xsign = -ONE;
    YY0 = (t5-t7+xsign*t14)*t17*t19/TWO;
    YY2 = -xsign*t21*t24;
    t177 = FbtAB(p1,p6,t43,p6,YY0);
    t178 = t177*t177;
    t182 = FbtAB(p3,p4,t43,p6,YY0);
    t183 = t182*t182;
    t184 = FbtAB(p5,p4,t43,p6,YY0);
    t185 = t184*t184;
    t186 = t183*t185;
    t187 = FbtAB(p6,p2,t43,p6,YY0);
    t188 = IFbtAB(p3,p3,t43,p6,YY0);
    t191 = IFbtAB(p5,p5,t43,p6,YY0);
    t192 = IFbtAA(p3,t62,p5,t43,p6,YY0);
    t193 = t191*t192;
    t195 = I/YY0;
    t196 = t193*t195;
    t207 = t92*t178*t183;
    t209 = t185*t187;
    t210 = FbtAB2(p6,p6,t43,p6,YY0,YY2);
    t213 = YY0*YY0;
    t220 = t109*t207;
    t221 = FbtAB2(p3,p3,t43,p6,YY0,YY2);
    t222 = t188*t188;
    t228 = FbtAB2(p5,p5,t43,p6,YY0,YY2);
    t231 = t191*t191;
    t237 = FbtAA2(p3,t62,p5,t43,p6,YY0,YY2);
    t240 = t192*t192;
    t249 = FbtAB2(p6,p2,t43,p6,YY0,YY2);
    t256 = FbtAB2(p5,p4,t43,p6,YY0,YY2);
    t265 = FbtAB2(p3,p4,t43,p6,YY0,YY2);
    t274 = FbtAB2(p1,p6,t43,p6,YY0,YY2);
    ans = -FOUR*EIGHT/THREE*t39*t42*t46*t48*t56*t57*t58*t67+FOUR*t80*t82*t48*t57*t63*t66
+FOUR*t91*t94*t96*t97*t58*t64*I/t100-FOUR*t110*t96*t111*t112*t67-FOUR*t110*t96*
t118*t58*t121*t63*t66-FOUR*t110*t96*t127*t58*t61*t130*t66+FOUR*t136*t42*t45*t48*
t56*t140*t58*t67+EIGHT*t110*t54*t57*t147*t58*t67+EIGHT*t109*t92*t48*t52*t96*t156*
t58*t67+EIGHT*t109*t92*t47*t53*t96*t165*t58*t67-FOUR*EIGHT/THREE*t39*t42*t46*t178*t186*
t187*t188*t196+FOUR*t80*t82*t178*t187*t192*t195+FOUR*t91*t207*t209*t210*t188*t193
*I/t213-FOUR*t220*t209*t221*t222*t196-FOUR*t220*t209*t228*t188*t231*t192*t195-FOUR
*t220*t209*t237*t188*t191*t240*t195+FOUR*t136*t42*t45*t178*t186*t249*t188*t196+
EIGHT*t220*t184*t187*t256*t188*t196+EIGHT*t109*t92*t178*t182*t209*t265*t188*t196+
EIGHT*t109*t92*t177*t183*t209*t274*t188*t196;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2123mpmpmpt1(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p1,p3+p2,p1);
    t4 = S(p3+p2+p1);
    //K1f = -p3-p2-p1+t2*t4*p1;
    K1f = -p3-p2-p1+S(p1+p2+p3)/TWO/dot(p1,p2+p3)*p1;
    S1 = t4;
    gg = AB(p1,K1f,p1);
    S2 = S(p1);
    t7 = AB(K1f,p1,K1f);
    t8 = gg*t7;
    t9 = AB(p1,p1,p1);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p1,p1,K1f);
    t22 = ONE/S1;
    t24 = AB(p1,p1,K1f);
    t27 = t17/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C2123mpmpmpt12(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p1,p3+p2,p1);
    t4 = S(p3+p2+p1);
    //K1f = -p3-p2-p1+t2*t4*p1;
    K1f = -p3-p2-p1+S(p1+p2+p3)/TWO/dot(p1,p2+p3)*p1;
    K2 = p2+p1;
    S1 = t4;
    gg = AB(p1,K1f,p1);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p1,K2,p1);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p1,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p1,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p1,p2);
    t30 = t29*t29;
    t31 = B(K1f,p2);
    t32 = t31*t31;
    t35 = IA(p4,p5);
    t36 = IA(p5,p6);
    t39 = IB(p1,p2);
    t40 = IB(p1,p3);
    t42 = t2*t2;
    t43 = t42*t42;
    t44 = S1*S1;
    t46 = p6+p5+p4;
    t47 = IDelta(t46,K2);
    t48 = t47*t47;
    t51 = t30*t32*t31*t35*t36*t39*t40*t43*t44*t48;
    t52 = FbtAB(p1,p2,t46,p1,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p3,p1,t46,p1,YY01);
    t56 = FbtAB(p5,p4,t46,p1,YY01);
    t57 = t56*t56;
    t58 = FbtAB(p5,p6,t46,p1,YY01);
    t59 = t58*t58;
    t62 = IFbtAB(p4,p4,t46,p1,YY01);
    t63 = IFbtAB(p6,p6,t46,p1,YY01);
    t64 = t62*t63;
    t65 = p5+p4;
    t66 = IFbtBB(p4,t65,p6,t46,p1,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = B(p4,p6);
    t75 = t74*t74;
    t76 = t75*t75;
    t79 = IB(p4,p5);
    t81 = IB(p5,p6);
    t84 = t31*t76*t39*t40*t79*t81*t42;
    t85 = IS(t46);
    t86 = t85*S1;
    t96 = t31*t35*t36*t39;
    t98 = t40*t42*t2;
    t99 = S1*t47;
    t100 = t99*t53;
    t103 = t54*t57;
    t104 = FbtAB2(p1,p1,t46,p1,YY01,YY21);
    t107 = YY01*YY01;
    t113 = t40*t42;
    t115 = t96*t113*t100;
    t116 = FbtAB2(p4,p4,t46,p1,YY01,YY21);
    t119 = t62*t62;
    t124 = FbtAB2(p6,p6,t46,p1,YY01,YY21);
    t127 = t63*t63;
    t132 = FbtBB2(p4,t65,p6,t46,p1,YY01,YY21);
    t135 = t66*t66;
    t141 = FbtAB2(p5,p6,t46,p1,YY01,YY21);
    t148 = FbtAB2(p5,p4,t46,p1,YY01,YY21);
    t155 = t96*t113*t99;
    t157 = FbtAB2(p3,p1,t46,p1,YY01,YY21);
    t165 = FbtAB2(p1,p2,t46,p1,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t177 = FbtAB(p1,p2,t46,p1,YY0);
    t178 = t177*t177;
    t179 = FbtAB(p3,p1,t46,p1,YY0);
    t181 = FbtAB(p5,p4,t46,p1,YY0);
    t182 = t181*t181;
    t183 = FbtAB(p5,p6,t46,p1,YY0);
    t184 = t183*t183;
    t187 = IFbtAB(p4,p4,t46,p1,YY0);
    t188 = IFbtAB(p6,p6,t46,p1,YY0);
    t189 = t187*t188;
    t190 = IFbtBB(p4,t65,p6,t46,p1,YY0);
    t191 = t190*I;
    t192 = ONE/YY0;
    t193 = t191*t192;
    t194 = t189*t193;
    t205 = t99*t178;
    t208 = t179*t182;
    t209 = FbtAB2(p1,p1,t46,p1,YY0,YY2);
    t212 = YY0*YY0;
    t219 = t96*t113*t205;
    t220 = FbtAB2(p4,p4,t46,p1,YY0,YY2);
    t223 = t187*t187;
    t228 = FbtAB2(p6,p6,t46,p1,YY0,YY2);
    t231 = t188*t188;
    t236 = FbtBB2(p4,t65,p6,t46,p1,YY0,YY2);
    t239 = t190*t190;
    t245 = FbtAB2(p5,p6,t46,p1,YY0,YY2);
    t252 = FbtAB2(p5,p4,t46,p1,YY0,YY2);
    t259 = FbtAB2(p3,p1,t46,p1,YY0,YY2);
    t267 = FbtAB2(p1,p2,t46,p1,YY0,YY2);
    ans = -TWO/THREE*t51*t53*t54*t57*t59*t70+t84*t86*t47*t53*t54*t66*I*t68+t96*
t98*t100*t103*t59*t104*t64*t67/t107-t115*t103*t59*t116*t119*t63*t69-t115*t103*
t59*t124*t62*t127*t69-t115*t103*t59*t132*t64*t135*I*t68+TWO*t115*t103*t58*t141*
t70+TWO*t115*t54*t56*t59*t148*t70+t155*t53*t57*t59*t157*t70+TWO*t96*t113*t99*
t52*t103*t59*t165*t70-TWO/THREE*t51*t178*t179*t182*t184*t194+t84*t86*t47*t178*
t179*t190*I*t192+t96*t98*t205*t208*t184*t209*t189*t191/t212-t219*t208*t184*t220
*t223*t188*t193-t219*t208*t184*t228*t187*t231*t193-t219*t208*t184*t236*t189*
t239*I*t192+TWO*t219*t208*t183*t245*t194+TWO*t219*t179*t181*t184*t252*t194+t155
*t178*t182*t184*t259*t194+TWO*t96*t113*t99*t177*t208*t184*t267*t194;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2123mpmpmpt6(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t133;
  std::complex<DOUBLE> t136;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t143;
  std::complex<DOUBLE> t150;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t168;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t210;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t214;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t232;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t241;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t265;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t274;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p1,p3+p2,p1);
    t3 = p3+p2+p1;
    t4 = S(t3);
    //K1f = p3+p2+p1-t2*t4*p1;
    K1f = p3+p2+p1-S(p1+p2+p3)/TWO/dot(p1,p2+p3)*p1;
    S1 = t4;
    gg = AB(p1,K1f,p1);
    S2 = S(p6);
    t7 = AB(K1f,p6,K1f);
    t8 = gg*t7;
    t9 = AB(p1,p6,p1);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p1,p6,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p1,p6,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p1,p3);
    t30 = t29*t29;
    t31 = t30*t30;
    t32 = B(K1f,p6);
    t34 = IA(p1,p2);
    t36 = IA(p2,p3);
    t37 = IA(p4,p5);
    t39 = IB(p4,p5);
    t41 = t31*t32*t34*t36*t37*t39;
    t43 = IAB(p6,p5+p4,p6);
    t44 = t43*t43;
    t45 = t2*t44;
    t46 = FbtAB(p1,p6,t3,p1,YY01);
    t48 = FbtAB(p5,p4,t3,p1,YY01);
    t49 = t48*t48;
    t50 = IFbtAB(p4,p4,t3,p1,YY01);
    t52 = p2+p1;
    t53 = IFbtAA(p1,t52,p3,t3,p1,YY01);
    t54 = t53*I;
    t59 = A(p1,p6);
    t60 = t59*t59;
    t61 = t32*t32;
    t64 = IB(p1,p2);
    t67 = IB(p2,p3);
    t69 = t2*t2;
    t70 = t69*t69;
    t71 = t44*t44;
    t73 = S1*S1;
    t76 = t60*t61*t32*t37*t64*t67*t39*t70*t71*t73;
    t77 = FbtAB(p1,p2,t3,p1,YY01);
    t78 = t77*t77;
    t80 = FbtAB(p3,p2,t3,p1,YY01);
    t81 = t80*t80;
    t84 = IFbtAB(p3,p3,t3,p1,YY01);
    t85 = t84*t50;
    t86 = ONE/YY01;
    t87 = t54*t86;
    t88 = t85*t87;
    t94 = t32*t37*t64*t67;
    t96 = t39*t69*t2;
    t97 = t44*S1;
    t98 = t97*t78;
    t101 = t46*t81;
    t102 = FbtAB2(p1,p1,t3,p1,YY01,YY21);
    t105 = YY01*YY01;
    t112 = t39*t69;
    t114 = t94*t112*t98;
    t115 = FbtAB2(p3,p3,t3,p1,YY01,YY21);
    t118 = t84*t84;
    t124 = FbtAB2(p4,p4,t3,p1,YY01,YY21);
    t127 = t50*t50;
    t133 = FbtAA2(p1,t52,p3,t3,p1,YY01,YY21);
    t136 = t53*t53;
    t143 = FbtAB2(p5,p4,t3,p1,YY01,YY21);
    t150 = FbtAB2(p3,p2,t3,p1,YY01,YY21);
    t157 = t94*t112*t97;
    t159 = FbtAB2(p1,p6,t3,p1,YY01,YY21);
    t168 = FbtAB2(p1,p2,t3,p1,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t180 = FbtAB(p1,p6,t3,p1,YY0);
    t182 = FbtAB(p5,p4,t3,p1,YY0);
    t183 = t182*t182;
    t184 = IFbtAB(p4,p4,t3,p1,YY0);
    t186 = IFbtAA(p1,t52,p3,t3,p1,YY0);
    t187 = t186*I;
    t192 = FbtAB(p1,p2,t3,p1,YY0);
    t193 = t192*t192;
    t195 = FbtAB(p3,p2,t3,p1,YY0);
    t196 = t195*t195;
    t199 = IFbtAB(p3,p3,t3,p1,YY0);
    t200 = t199*t184;
    t201 = ONE/YY0;
    t202 = t187*t201;
    t203 = t200*t202;
    t207 = t97*t193;
    t210 = t180*t196;
    t211 = FbtAB2(p1,p1,t3,p1,YY0,YY2);
    t214 = YY0*YY0;
    t222 = t94*t112*t207;
    t223 = FbtAB2(p3,p3,t3,p1,YY0,YY2);
    t226 = t199*t199;
    t232 = FbtAB2(p4,p4,t3,p1,YY0,YY2);
    t235 = t184*t184;
    t241 = FbtAA2(p1,t52,p3,t3,p1,YY0,YY2);
    t244 = t186*t186;
    t251 = FbtAB2(p5,p4,t3,p1,YY0,YY2);
    t258 = FbtAB2(p3,p2,t3,p1,YY0,YY2);
    t265 = FbtAB2(p1,p6,t3,p1,YY0,YY2);
    t274 = FbtAB2(p1,p2,t3,p1,YY0,YY2);
    ans = FOUR*t41*t45*t46*t49*t50*t54+FOUR*EIGHT/THREE*t76*t78*t46*t81*t49*t88-FOUR*t94*
t96*t98*t101*t49*t102*t85*t54/t105-FOUR*t114*t101*t49*t115*t118*t50*t87-FOUR*t114
*t101*t49*t124*t84*t127*t87-FOUR*t114*t101*t49*t133*t85*t136*I*t86+EIGHT*t114*t101
*t48*t143*t88+EIGHT*t114*t46*t80*t49*t150*t88+FOUR*t157*t78*t81*t49*t159*t88+EIGHT*
t94*t112*t97*t77*t101*t49*t168*t88+FOUR*t41*t45*t180*t183*t184*t187+FOUR*EIGHT/THREE*t76
*t193*t180*t196*t183*t203-FOUR*t94*t96*t207*t210*t183*t211*t200*t187/t214-FOUR*
t222*t210*t183*t223*t226*t184*t202-FOUR*t222*t210*t183*t232*t199*t235*t202-FOUR*
t222*t210*t183*t241*t200*t244*I*t201+EIGHT*t222*t210*t182*t251*t203+EIGHT*t222*t180
*t195*t183*t258*t203+FOUR*t157*t193*t196*t183*t265*t203+EIGHT*t94*t112*t97*t192*
t210*t183*t274*t203;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2123mpmpmpt56(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t108;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t121;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t130;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t173;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t251;
  std::complex<DOUBLE> t258;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p1,p3+p2,p1);
    t3 = p3+p2+p1;
    t4 = S(t3);
    //K1f = p3+p2+p1-t2*t4*p1;
    K1f = p3+p2+p1-S(p1+p2+p3)/TWO/dot(p1,p2+p3)*p1;
    K2 = p6+p5;
    S1 = t4;
    gg = AB(p1,K1f,p1);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p1,K2,p1);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p1,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p1,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p1,p3);
    t30 = t29*t29;
    t31 = t30*t30;
    t32 = IA(p1,p2);
    t34 = IA(p1,p4);
    t36 = IA(p2,p3);
    t37 = IA(p5,p6);
    t39 = IB(p5,p6);
    t40 = t39*t24;
    t42 = t31*t32*t34*t36*t37*t40;
    t43 = IAB(p4,K2,p4);
    t44 = t43*t43;
    t45 = t2*t44;
    t46 = FbtAB(p1,p4,t3,p1,YY01);
    t48 = FbtAB(p5,p6,t3,p1,YY01);
    t49 = t48*t48;
    t50 = IFbtAB(p5,p5,t3,p1,YY01);
    t52 = p2+p1;
    t53 = IFbtAA(p1,t52,p3,t3,p1,YY01);
    t54 = t53*I;
    t60 = IB(p1,p2);
    t61 = IB(p2,p3);
    t63 = t34*t37*t60*t61;
    t64 = t24*t24;
    t67 = t2*t2;
    t68 = t67*t67;
    t69 = t44*t44;
    t71 = S1*S1;
    t74 = t63*t39*t64*t24*t68*t69*t71;
    t75 = FbtAB(p1,p2,t3,p1,YY01);
    t76 = t75*t75;
    t77 = t76*t46;
    t78 = FbtAB(p3,p2,t3,p1,YY01);
    t79 = t78*t78;
    t80 = t79*t49;
    t82 = IFbtAB(p3,p3,t3,p1,YY01);
    t83 = t82*t50;
    t84 = ONE/YY01;
    t85 = t54*t84;
    t86 = t83*t85;
    t94 = t63*t40*t67*t2*t44*S1;
    t95 = FbtAB2(p1,p1,t3,p1,YY01,YY21);
    t98 = YY01*YY01;
    t108 = t63*t40*t67*t44*S1;
    t109 = FbtAB2(p3,p3,t3,p1,YY01,YY21);
    t112 = t82*t82;
    t118 = FbtAB2(p5,p5,t3,p1,YY01,YY21);
    t121 = t50*t50;
    t127 = FbtAA2(p1,t52,p3,t3,p1,YY01,YY21);
    t130 = t53*t53;
    t138 = FbtAB2(p5,p6,t3,p1,YY01,YY21);
    t145 = FbtAB2(p3,p2,t3,p1,YY01,YY21);
    t152 = FbtAB2(p1,p4,t3,p1,YY01,YY21);
    t159 = FbtAB2(p1,p2,t3,p1,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t171 = FbtAB(p1,p4,t3,p1,YY0);
    t173 = FbtAB(p5,p6,t3,p1,YY0);
    t174 = t173*t173;
    t175 = IFbtAB(p5,p5,t3,p1,YY0);
    t177 = IFbtAA(p1,t52,p3,t3,p1,YY0);
    t178 = t177*I;
    t183 = FbtAB(p1,p2,t3,p1,YY0);
    t184 = t183*t183;
    t185 = t184*t171;
    t186 = FbtAB(p3,p2,t3,p1,YY0);
    t187 = t186*t186;
    t188 = t187*t174;
    t190 = IFbtAB(p3,p3,t3,p1,YY0);
    t191 = t190*t175;
    t192 = ONE/YY0;
    t193 = t178*t192;
    t194 = t191*t193;
    t198 = FbtAB2(p1,p1,t3,p1,YY0,YY2);
    t201 = YY0*YY0;
    t208 = FbtAB2(p3,p3,t3,p1,YY0,YY2);
    t211 = t190*t190;
    t217 = FbtAB2(p5,p5,t3,p1,YY0,YY2);
    t220 = t175*t175;
    t226 = FbtAA2(p1,t52,p3,t3,p1,YY0,YY2);
    t229 = t177*t177;
    t237 = FbtAB2(p5,p6,t3,p1,YY0,YY2);
    t244 = FbtAB2(p3,p2,t3,p1,YY0,YY2);
    t251 = FbtAB2(p1,p4,t3,p1,YY0,YY2);
    t258 = FbtAB2(p1,p2,t3,p1,YY0,YY2);
    ans = -FOUR*t42*t45*t46*t49*t50*t54-FOUR*EIGHT/THREE*t74*t77*t80*t86+FOUR*t94*t77*t80
*t95*t83*t54/t98+FOUR*t108*t77*t80*t109*t112*t50*t85+FOUR*t108*t77*t80*t118*t82*
t121*t85+FOUR*t108*t77*t80*t127*t83*t130*I*t84-EIGHT*t108*t77*t79*t48*t138*t86-EIGHT
*t108*t77*t78*t49*t145*t86-FOUR*t108*t76*t79*t49*t152*t86-EIGHT*t108*t75*t46*t80*
t159*t86-FOUR*t42*t45*t171*t174*t175*t178-FOUR*EIGHT/THREE*t74*t185*t188*t194+FOUR*t94*
t185*t188*t198*t191*t178/t201+FOUR*t108*t185*t188*t208*t211*t175*t193+FOUR*t108*
t185*t188*t217*t190*t220*t193+FOUR*t108*t185*t188*t226*t191*t229*I*t192-EIGHT*t108
*t185*t187*t173*t237*t194-EIGHT*t108*t185*t186*t174*t244*t194-FOUR*t108*t184*t187*
t174*t251*t194-EIGHT*t108*t183*t171*t188*t258*t194;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2234mpmpmpt2(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p2,p4+p3,p2);
    t4 = S(p4+p3+p2);
    K1f = -p4-p3-p2+S(p2+p3+p4)/TWO/dot(p2,p3+p4)*p2;
    S1 = t4;
    gg = AB(p2,K1f,p2);
    S2 = S(p2);
    t7 = AB(K1f,p2,K1f);
    t8 = gg*t7;
    t9 = AB(p2,p2,p2);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p2,p2,K1f);
    t22 = ONE/S1;
    t24 = AB(p2,p2,K1f);
    t27 = t17/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C2234mpmpmpt23(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t124;
  std::complex<DOUBLE> t127;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t135;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t157;
  std::complex<DOUBLE> t165;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t205;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t209;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t223;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t231;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t252;
  std::complex<DOUBLE> t259;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t43;
  FourMomentum<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  FourMomentum<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t89;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p2,p4+p3,p2);
    t4 = S(p4+p3+p2);
    K1f = -p4-p3-p2+S(p2+p3+p4)/TWO/dot(p2,p3+p4)*p2;
    K2 = p3+p2;
    S1 = t4;
    gg = AB(p2,K1f,p2);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p2,K2,p2);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p2,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p2,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p1,p5);
    t30 = t29*t29;
    t31 = t30*t30;
    t32 = B(K1f,p3);
    t34 = IA(p1,p6);
    t36 = IA(p2,p4);
    t37 = IA(p5,p6);
    t39 = IB(p2,p3);
    t40 = t2*t2;
    t43 = t31*t32*t34*t36*t37*t39*t40;
    t44 = p6+p5+p1;
    t45 = IS(t44);
    t46 = t45*S1;
    t47 = IDelta(t44,K2);
    t48 = FbtAB(p2,p4,t44,p2,YY01);
    t51 = FbtAB(p3,p2,t44,p2,YY01);
    t52 = t51*t51;
    t53 = p6+p5;
    t54 = IFbtAA(p5,t53,p1,t44,p2,YY01);
    t56 = ONE/YY01;
    t61 = A(p2,p3);
    t62 = t61*t61;
    t63 = t32*t32;
    t66 = IB(p1,p6);
    t69 = IB(p5,p6);
    t71 = t40*t40;
    t72 = S1*S1;
    t74 = t47*t47;
    t77 = t62*t63*t32*t36*t66*t39*t69*t71*t72*t74;
    t78 = FbtAB(p1,p6,t44,p2,YY01);
    t79 = t78*t78;
    t81 = FbtAB(p5,p6,t44,p2,YY01);
    t82 = t81*t81;
    t85 = IFbtAB(p1,p1,t44,p2,YY01);
    t86 = IFbtAB(p5,p5,t44,p2,YY01);
    t87 = t85*t86;
    t88 = t54*I;
    t89 = t88*t56;
    t90 = t87*t89;
    t96 = t32*t36*t66*t39;
    t98 = t69*t40*t2;
    t99 = S1*t47;
    t100 = t99*t79;
    t103 = t48*t52;
    t104 = FbtAB2(p2,p2,t44,p2,YY01,YY21);
    t107 = YY01*YY01;
    t113 = t69*t40;
    t115 = t96*t113*t100;
    t116 = FbtAB2(p1,p1,t44,p2,YY01,YY21);
    t119 = t85*t85;
    t124 = FbtAB2(p5,p5,t44,p2,YY01,YY21);
    t127 = t86*t86;
    t132 = FbtAA2(p5,t53,p1,t44,p2,YY01,YY21);
    t135 = t54*t54;
    t141 = FbtAB2(p5,p6,t44,p2,YY01,YY21);
    t148 = FbtAB2(p3,p2,t44,p2,YY01,YY21);
    t155 = t96*t113*t99;
    t157 = FbtAB2(p2,p4,t44,p2,YY01,YY21);
    t165 = FbtAB2(p1,p6,t44,p2,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t177 = FbtAB(p2,p4,t44,p2,YY0);
    t180 = FbtAB(p3,p2,t44,p2,YY0);
    t181 = t180*t180;
    t182 = IFbtAA(p5,t53,p1,t44,p2,YY0);
    t184 = ONE/YY0;
    t189 = FbtAB(p1,p6,t44,p2,YY0);
    t190 = t189*t189;
    t192 = FbtAB(p5,p6,t44,p2,YY0);
    t193 = t192*t192;
    t196 = IFbtAB(p1,p1,t44,p2,YY0);
    t197 = IFbtAB(p5,p5,t44,p2,YY0);
    t198 = t196*t197;
    t199 = t182*I;
    t200 = t199*t184;
    t201 = t198*t200;
    t205 = t99*t190;
    t208 = t177*t181;
    t209 = FbtAB2(p2,p2,t44,p2,YY0,YY2);
    t212 = YY0*YY0;
    t219 = t96*t113*t205;
    t220 = FbtAB2(p1,p1,t44,p2,YY0,YY2);
    t223 = t196*t196;
    t228 = FbtAB2(p5,p5,t44,p2,YY0,YY2);
    t231 = t197*t197;
    t236 = FbtAA2(p5,t53,p1,t44,p2,YY0,YY2);
    t239 = t182*t182;
    t245 = FbtAB2(p5,p6,t44,p2,YY0,YY2);
    t252 = FbtAB2(p3,p2,t44,p2,YY0,YY2);
    t259 = FbtAB2(p2,p4,t44,p2,YY0,YY2);
    t267 = FbtAB2(p1,p6,t44,p2,YY0,YY2);
    ans = -t43*t46*t47*t48*t52*t54*I*t56+TWO/THREE*t77*t79*t48*t52*t82*t90-t96*
t98*t100*t103*t82*t104*t87*t88/t107+t115*t103*t82*t116*t119*t86*t89+t115*t103*
t82*t124*t85*t127*t89+t115*t103*t82*t132*t87*t135*I*t56-TWO*t115*t103*t81*t141*
t90-TWO*t115*t48*t51*t82*t148*t90-t155*t79*t52*t82*t157*t90-TWO*t96*t113*t99*
t78*t103*t82*t165*t90-t43*t46*t47*t177*t181*t182*I*t184+TWO/THREE*t77*t190*t177*
t181*t193*t201-t96*t98*t205*t208*t193*t209*t198*t199/t212+t219*t208*t193*t220*
t223*t197*t200+t219*t208*t193*t228*t196*t231*t200+t219*t208*t193*t236*t198*t239
*I*t184-TWO*t219*t208*t192*t245*t201-TWO*t219*t177*t180*t193*t252*t201-t155*
t190*t181*t193*t259*t201-TWO*t96*t113*t99*t189*t208*t193*t267*t201;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2234mpmpmpt1(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t106;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t120;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t129;
  std::complex<DOUBLE> t132;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t149;
  std::complex<DOUBLE> t156;
  std::complex<DOUBLE> t163;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t199;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t200;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t202;
  std::complex<DOUBLE> t203;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t216;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t239;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t245;
  std::complex<DOUBLE> t248;
  std::complex<DOUBLE> t256;
  std::complex<DOUBLE> t263;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t270;
  std::complex<DOUBLE> t277;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  FourMomentum<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t87;
  std::complex<DOUBLE> t88;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t97;
  std::complex<DOUBLE> t98;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p2,p4+p3,p2);
    t3 = p4+p3+p2;
    t4 = S(t3);
    //K1f = p4+p3+p2-t2*t4*p2;
    K1f = p4+p3+p2-S(p2+p3+p4)/TWO/dot(p2,p3+p4)*p2;
    S1 = t4;
    gg = AB(p2,K1f,p2);
    S2 = S(p1);
    t7 = AB(K1f,p1,K1f);
    t8 = gg*t7;
    t9 = AB(p2,p1,p2);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p2,p1,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p2,p1,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p1,p2);
    t30 = t29*t29;
    t32 = B(K1f,p1);
    t33 = t32*t32;
    t36 = IA(p2,p3);
    t37 = IA(p3,p4);
    t38 = t36*t37;
    t40 = IA(p5,p6);
    t41 = IB(p1,p2);
    t42 = t40*t41;
    t43 = IB(p5,p6);
    t45 = IAB(p1,p6+p5,p1);
    t46 = t45*t45;
    t47 = t46*t46;
    t49 = t2*t2;
    t50 = t49*t49;
    t53 = t30*t29*t33*t32*t38*t42*t43*t47*t50;
    t54 = S1*S1;
    t55 = FbtAB(p1,p2,t3,p2,YY01);
    t57 = FbtAB(p3,p2,t3,p2,YY01);
    t58 = t57*t57;
    t59 = FbtAB(p3,p4,t3,p2,YY01);
    t60 = t59*t59;
    t62 = FbtAB(p5,p6,t3,p2,YY01);
    t63 = t62*t62;
    t66 = IFbtAB(p4,p4,t3,p2,YY01);
    t67 = IFbtAB(p5,p5,t3,p2,YY01);
    t68 = t66*t67;
    t69 = p3+p2;
    t70 = IFbtBB(p2,t69,p4,t3,p2,YY01);
    t71 = t70*I;
    t72 = ONE/YY01;
    t73 = t71*t72;
    t74 = t68*t73;
    t78 = t29*t32;
    t79 = B(p2,p4);
    t80 = t79*t79;
    t81 = t80*t80;
    t83 = IB(p2,p3);
    t84 = IB(p3,p4);
    t87 = t78*t81*t42*t83*t84;
    t88 = t43*t46;
    t97 = t78*t38*t40;
    t98 = t41*t43;
    t103 = t97*t98*t46*t49*t2*S1;
    t104 = t55*t58;
    t105 = t60*t63;
    t106 = FbtAB2(p2,p2,t3,p2,YY01,YY21);
    t109 = YY01*YY01;
    t119 = t97*t98*t46*t49*S1;
    t120 = FbtAB2(p4,p4,t3,p2,YY01,YY21);
    t123 = t66*t66;
    t129 = FbtAB2(p5,p5,t3,p2,YY01,YY21);
    t132 = t67*t67;
    t138 = FbtBB2(p2,t69,p4,t3,p2,YY01,YY21);
    t141 = t70*t70;
    t149 = FbtAB2(p5,p6,t3,p2,YY01,YY21);
    t156 = FbtAB2(p3,p4,t3,p2,YY01,YY21);
    t163 = FbtAB2(p3,p2,t3,p2,YY01,YY21);
    t172 = t78*t38*t42*t88*t49;
    t174 = FbtAB2(p1,p2,t3,p2,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t186 = FbtAB(p1,p2,t3,p2,YY0);
    t188 = FbtAB(p3,p2,t3,p2,YY0);
    t189 = t188*t188;
    t190 = FbtAB(p3,p4,t3,p2,YY0);
    t191 = t190*t190;
    t193 = FbtAB(p5,p6,t3,p2,YY0);
    t194 = t193*t193;
    t197 = IFbtAB(p4,p4,t3,p2,YY0);
    t198 = IFbtAB(p5,p5,t3,p2,YY0);
    t199 = t197*t198;
    t200 = IFbtBB(p2,t69,p4,t3,p2,YY0);
    t201 = t200*I;
    t202 = ONE/YY0;
    t203 = t201*t202;
    t204 = t199*t203;
    t215 = t186*t189;
    t216 = t191*t194;
    t217 = FbtAB2(p2,p2,t3,p2,YY0,YY2);
    t220 = YY0*YY0;
    t227 = FbtAB2(p4,p4,t3,p2,YY0,YY2);
    t230 = t197*t197;
    t236 = FbtAB2(p5,p5,t3,p2,YY0,YY2);
    t239 = t198*t198;
    t245 = FbtBB2(p2,t69,p4,t3,p2,YY0,YY2);
    t248 = t200*t200;
    t256 = FbtAB2(p5,p6,t3,p2,YY0,YY2);
    t263 = FbtAB2(p3,p4,t3,p2,YY0,YY2);
    t270 = FbtAB2(p3,p2,t3,p2,YY0,YY2);
    t277 = FbtAB2(p1,p2,t3,p2,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t53*t54*t55*t58*t60*t63*t74+FOUR*t87*t88*t2*t55*t63*t67*t71
-FOUR*t103*t104*t105*t106*t68*t71/t109-FOUR*t119*t104*t105*t120*t123*t67*t73-FOUR*
t119*t104*t105*t129*t66*t132*t73-FOUR*t119*t104*t105*t138*t68*t141*I*t72+EIGHT*
t119*t104*t60*t62*t149*t74+EIGHT*t119*t104*t59*t63*t156*t74+EIGHT*t119*t55*t57*t105
*t163*t74+FOUR*t172*S1*t58*t105*t174*t74+FOUR*EIGHT/THREE*t53*t54*t186*t189*t191*t194*
t204+FOUR*t87*t88*t2*t186*t194*t198*t201-FOUR*t103*t215*t216*t217*t199*t201/t220
-FOUR*t119*t215*t216*t227*t230*t198*t203-FOUR*t119*t215*t216*t236*t197*t239*t203
-FOUR*t119*t215*t216*t245*t199*t248*I*t202+EIGHT*t119*t215*t191*t193*t256*t204+EIGHT
*t119*t215*t190*t194*t263*t204+EIGHT*t119*t186*t188*t216*t270*t204+FOUR*t172*S1*
t189*t216*t277*t204;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2234mpmpmpt61(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t138;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t236;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t99;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p2,p4+p3,p2);
    t3 = p4+p3+p2;
    t4 = S(t3);
    //K1f = p4+p3+p2-t2*t4*p2;
    K1f = p4+p3+p2-S(p2+p3+p4)/TWO/dot(p2,p3+p4)*p2;
    K2 = p6+p1;
    S1 = t4;
    gg = AB(p2,K1f,p2);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p2,K2,p2);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p2,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p2,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = B(p2,p4);
    t30 = t29*t29;
    t31 = t30*t30;
    t32 = IA(p1,p6);
    t34 = IB(p1,p6);
    t36 = IB(p2,p3);
    t37 = IB(p2,p5);
    t39 = IB(p3,p4);
    t42 = t31*t32*t34*t36*t37*t39*t24;
    t43 = IAB(p5,K2,p5);
    t44 = t43*t43;
    t45 = t2*t44;
    t46 = FbtAB(p1,p6,t3,p2,YY01);
    t47 = t46*t46;
    t49 = FbtAB(p5,p2,t3,p2,YY01);
    t50 = IFbtAB(p6,p6,t3,p2,YY01);
    t52 = p3+p2;
    t53 = IFbtBB(p2,t52,p4,t3,p2,YY01);
    t54 = t53*I;
    t59 = IA(p2,p3);
    t61 = IA(p3,p4);
    t63 = t32*t59*t61*t34;
    t64 = t24*t24;
    t67 = t2*t2;
    t68 = t67*t67;
    t69 = t44*t44;
    t71 = S1*S1;
    t74 = t63*t37*t64*t24*t68*t69*t71;
    t75 = FbtAB(p3,p2,t3,p2,YY01);
    t76 = t75*t75;
    t77 = t47*t76;
    t78 = FbtAB(p3,p4,t3,p2,YY01);
    t79 = t78*t78;
    t80 = t79*t49;
    t82 = IFbtAB(p4,p4,t3,p2,YY01);
    t83 = t82*t50;
    t84 = ONE/YY01;
    t85 = t54*t84;
    t86 = t83*t85;
    t90 = t37*t24;
    t95 = t63*t90*t67*t2*t44*S1;
    t96 = FbtAB2(p2,p2,t3,p2,YY01,YY21);
    t99 = YY01*YY01;
    t109 = t63*t90*t67*t44*S1;
    t110 = FbtAB2(p4,p4,t3,p2,YY01,YY21);
    t113 = t82*t82;
    t119 = FbtAB2(p6,p6,t3,p2,YY01,YY21);
    t122 = t50*t50;
    t128 = FbtBB2(p2,t52,p4,t3,p2,YY01,YY21);
    t131 = t53*t53;
    t138 = FbtAB2(p5,p2,t3,p2,YY01,YY21);
    t145 = FbtAB2(p3,p4,t3,p2,YY01,YY21);
    t152 = FbtAB2(p3,p2,t3,p2,YY01,YY21);
    t159 = FbtAB2(p1,p6,t3,p2,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t171 = FbtAB(p1,p6,t3,p2,YY0);
    t172 = t171*t171;
    t174 = FbtAB(p5,p2,t3,p2,YY0);
    t175 = IFbtAB(p6,p6,t3,p2,YY0);
    t177 = IFbtBB(p2,t52,p4,t3,p2,YY0);
    t178 = t177*I;
    t183 = FbtAB(p3,p2,t3,p2,YY0);
    t184 = t183*t183;
    t185 = t172*t184;
    t186 = FbtAB(p3,p4,t3,p2,YY0);
    t187 = t186*t186;
    t188 = t187*t174;
    t190 = IFbtAB(p4,p4,t3,p2,YY0);
    t191 = t190*t175;
    t192 = ONE/YY0;
    t193 = t178*t192;
    t194 = t191*t193;
    t198 = FbtAB2(p2,p2,t3,p2,YY0,YY2);
    t201 = YY0*YY0;
    t208 = FbtAB2(p4,p4,t3,p2,YY0,YY2);
    t211 = t190*t190;
    t217 = FbtAB2(p6,p6,t3,p2,YY0,YY2);
    t220 = t175*t175;
    t226 = FbtBB2(p2,t52,p4,t3,p2,YY0,YY2);
    t229 = t177*t177;
    t236 = FbtAB2(p5,p2,t3,p2,YY0,YY2);
    t243 = FbtAB2(p3,p4,t3,p2,YY0,YY2);
    t250 = FbtAB2(p3,p2,t3,p2,YY0,YY2);
    t257 = FbtAB2(p1,p6,t3,p2,YY0,YY2);
    ans = -FOUR*t42*t45*t47*t49*t50*t54-FOUR*EIGHT/THREE*t74*t77*t80*t86+FOUR*t95*t77*t80
*t96*t83*t54/t99+FOUR*t109*t77*t80*t110*t113*t50*t85+FOUR*t109*t77*t80*t119*t82*
t122*t85+FOUR*t109*t77*t80*t128*t83*t131*I*t84-FOUR*t109*t77*t79*t138*t86-EIGHT*
t109*t77*t78*t49*t145*t86-EIGHT*t109*t47*t75*t80*t152*t86-EIGHT*t109*t46*t76*t80*
t159*t86-FOUR*t42*t45*t172*t174*t175*t178-FOUR*EIGHT/THREE*t74*t185*t188*t194+FOUR*t95*
t185*t188*t198*t191*t178/t201+FOUR*t109*t185*t188*t208*t211*t175*t193+FOUR*t109*
t185*t188*t217*t190*t220*t193+FOUR*t109*t185*t188*t226*t191*t229*I*t192-FOUR*t109
*t185*t187*t236*t194-EIGHT*t109*t185*t186*t174*t243*t194-EIGHT*t109*t172*t183*t188*
t250*t194-EIGHT*t109*t171*t184*t188*t257*t194;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2345mpmpmpt3(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p3,p5+p4,p3);
    t4 = S(p5+p4+p3);
    //K1f = -p5-p4-p3+t2*t4*p3;
    K1f = -p3-p4-p5+S(p3+p4+p5)/TWO/dot(p3,p4+p5)*p3;
    S1 = t4;
    gg = AB(p3,K1f,p3);
    S2 = S(p3);
    t7 = AB(K1f,p3,K1f);
    t8 = gg*t7;
    t9 = AB(p3,p3,p3);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p3,p3,K1f);
    t22 = ONE/S1;
    t24 = AB(p3,p3,K1f);
    t27 = t17/alpha;
    xsign = -ONE;
    return(ZERO);
  }
}


std::complex<DOUBLE> C2345mpmpmpt34(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t100;
  std::complex<DOUBLE> t103;
  std::complex<DOUBLE> t104;
  std::complex<DOUBLE> t107;
  std::complex<DOUBLE> t114;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t118;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t123;
  std::complex<DOUBLE> t126;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t141;
  std::complex<DOUBLE> t142;
  std::complex<DOUBLE> t148;
  std::complex<DOUBLE> t155;
  std::complex<DOUBLE> t164;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t176;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t179;
  std::complex<DOUBLE> t180;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t189;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t204;
  std::complex<DOUBLE> t207;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t218;
  std::complex<DOUBLE> t219;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t222;
  std::complex<DOUBLE> t227;
  std::complex<DOUBLE> t230;
  std::complex<DOUBLE> t235;
  std::complex<DOUBLE> t238;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t244;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t266;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  FourMomentum<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  FourMomentum<DOUBLE> t65;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t98;
  std::complex<DOUBLE> t99;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p3,p5+p4,p3);
    t4 = S(p5+p4+p3);
    //K1f = -p5-p4-p3+t2*t4*p3;
    K1f = -p3-p4-p5+S(p3+p4+p5)/TWO/dot(p3,p4+p5)*p3;
    K2 = p4+p3;
    S1 = t4;
    gg = AB(p3,K1f,p3);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p3,K2,p3);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p3,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p3,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p3,p4);
    t30 = t29*t29;
    t31 = B(K1f,p4);
    t32 = t31*t31;
    t35 = IA(p1,p2);
    t36 = IA(p1,p6);
    t39 = IB(p3,p4);
    t40 = IB(p3,p5);
    t42 = t2*t2;
    t43 = t42*t42;
    t44 = S1*S1;
    t46 = p6+p2+p1;
    t47 = IDelta(t46,K2);
    t48 = t47*t47;
    t51 = t30*t32*t31*t35*t36*t39*t40*t43*t44*t48;
    t52 = FbtAB(p1,p2,t46,p3,YY01);
    t53 = t52*t52;
    t54 = FbtAB(p1,p6,t46,p3,YY01);
    t55 = t54*t54;
    t56 = t53*t55;
    t57 = FbtAB(p3,p4,t46,p3,YY01);
    t58 = t57*t57;
    t59 = FbtAB(p5,p3,t46,p3,YY01);
    t62 = IFbtAB(p2,p2,t46,p3,YY01);
    t63 = IFbtAB(p6,p6,t46,p3,YY01);
    t64 = t62*t63;
    t65 = p6+p1;
    t66 = IFbtBB(p6,t65,p2,t46,p3,YY01);
    t67 = t66*I;
    t68 = ONE/YY01;
    t69 = t67*t68;
    t70 = t64*t69;
    t74 = B(p2,p6);
    t75 = t74*t74;
    t76 = t75*t75;
    t78 = IB(p1,p2);
    t80 = IB(p1,p6);
    t82 = t40*t42;
    t84 = t31*t76*t78*t80*t39*t82;
    t85 = IS(t46);
    t86 = t85*S1;
    t96 = t31*t35*t36*t39;
    t98 = t40*t42*t2;
    t99 = S1*t47;
    t100 = t99*t53;
    t103 = t55*t58;
    t104 = FbtAB2(p3,p3,t46,p3,YY01,YY21);
    t107 = YY01*YY01;
    t114 = t96*t82*t100;
    t115 = FbtAB2(p2,p2,t46,p3,YY01,YY21);
    t118 = t62*t62;
    t123 = FbtAB2(p6,p6,t46,p3,YY01,YY21);
    t126 = t63*t63;
    t131 = FbtBB2(p6,t65,p2,t46,p3,YY01,YY21);
    t134 = t66*t66;
    t141 = t96*t82*t99;
    t142 = FbtAB2(p5,p3,t46,p3,YY01,YY21);
    t148 = FbtAB2(p3,p4,t46,p3,YY01,YY21);
    t155 = FbtAB2(p1,p6,t46,p3,YY01,YY21);
    t164 = FbtAB2(p1,p2,t46,p3,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t176 = FbtAB(p1,p2,t46,p3,YY0);
    t177 = t176*t176;
    t178 = FbtAB(p1,p6,t46,p3,YY0);
    t179 = t178*t178;
    t180 = t177*t179;
    t181 = FbtAB(p3,p4,t46,p3,YY0);
    t182 = t181*t181;
    t183 = FbtAB(p5,p3,t46,p3,YY0);
    t186 = IFbtAB(p2,p2,t46,p3,YY0);
    t187 = IFbtAB(p6,p6,t46,p3,YY0);
    t188 = t186*t187;
    t189 = IFbtBB(p6,t65,p2,t46,p3,YY0);
    t190 = t189*I;
    t191 = ONE/YY0;
    t192 = t190*t191;
    t193 = t188*t192;
    t204 = t99*t177;
    t207 = t179*t182;
    t208 = FbtAB2(p3,p3,t46,p3,YY0,YY2);
    t211 = YY0*YY0;
    t218 = t96*t82*t204;
    t219 = FbtAB2(p2,p2,t46,p3,YY0,YY2);
    t222 = t186*t186;
    t227 = FbtAB2(p6,p6,t46,p3,YY0,YY2);
    t230 = t187*t187;
    t235 = FbtBB2(p6,t65,p2,t46,p3,YY0,YY2);
    t238 = t189*t189;
    t244 = FbtAB2(p5,p3,t46,p3,YY0,YY2);
    t250 = FbtAB2(p3,p4,t46,p3,YY0,YY2);
    t257 = FbtAB2(p1,p6,t46,p3,YY0,YY2);
    t266 = FbtAB2(p1,p2,t46,p3,YY0,YY2);
    ans = TWO/THREE*t51*t56*t58*t59*t70-t84*t86*t47*t58*t59*t66*I*t68-t96*t98*
t100*t103*t59*t104*t64*t67/t107+t114*t103*t59*t115*t118*t63*t69+t114*t103*t59*
t123*t62*t126*t69+t114*t103*t59*t131*t64*t134*I*t68-t141*t56*t58*t142*t70-TWO*
t114*t55*t57*t59*t148*t70-TWO*t114*t54*t58*t59*t155*t70-TWO*t96*t82*t99*t52*
t103*t59*t164*t70+TWO/THREE*t51*t180*t182*t183*t193-t84*t86*t47*t182*t183*t189*I*
t191-t96*t98*t204*t207*t183*t208*t188*t190/t211+t218*t207*t183*t219*t222*t187*
t192+t218*t207*t183*t227*t186*t230*t192+t218*t207*t183*t235*t188*t238*I*t191-
t141*t180*t182*t244*t193-TWO*t218*t179*t181*t183*t250*t193-TWO*t218*t178*t182*
t183*t257*t193-TWO*t96*t82*t99*t176*t207*t183*t266*t193;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2345mpmpmpt2(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t101;
  std::complex<DOUBLE> t102;
  std::complex<DOUBLE> t105;
  std::complex<DOUBLE> t112;
  std::complex<DOUBLE> t115;
  std::complex<DOUBLE> t116;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t125;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t134;
  std::complex<DOUBLE> t137;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t144;
  std::complex<DOUBLE> t151;
  std::complex<DOUBLE> t158;
  std::complex<DOUBLE> t160;
  std::complex<DOUBLE> t169;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t181;
  std::complex<DOUBLE> t182;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t195;
  std::complex<DOUBLE> t196;
  std::complex<DOUBLE> t197;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t212;
  std::complex<DOUBLE> t215;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t224;
  std::complex<DOUBLE> t225;
  std::complex<DOUBLE> t228;
  std::complex<DOUBLE> t234;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t246;
  std::complex<DOUBLE> t253;
  std::complex<DOUBLE> t260;
  std::complex<DOUBLE> t267;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t276;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t57;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t65;
  FourMomentum<DOUBLE> t66;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t81;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t94;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t97;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p3,p5+p4,p3);
    t3 = p5+p4+p3;
    t4 = S(t3);
    //K1f = p5+p4+p3-t2*t4*p3;
    K1f = p3+p4+p5-S(p3+p4+p5)/TWO/dot(p3,p4+p5)*p3;
    S1 = t4;
    gg = AB(p3,K1f,p3);
    S2 = S(p2);
    t7 = AB(K1f,p2,K1f);
    t8 = gg*t7;
    t9 = AB(p3,p2,p3);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p3,p2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p3,p2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p2,p3);
    t30 = t29*t29;
    t31 = B(K1f,p2);
    t32 = t31*t31;
    t35 = IA(p1,p6);
    t36 = IB(p1,p6);
    t39 = IB(p3,p4);
    t40 = IB(p4,p5);
    t43 = IAB(p2,p6+p1,p2);
    t44 = t43*t43;
    t45 = t44*t44;
    t46 = t2*t2;
    t47 = t46*t46;
    t49 = S1*S1;
    t52 = t30*t32*t31*t35*t36*t39*t40*t45*t47*t49;
    t53 = FbtAB(p1,p6,t3,p3,YY01);
    t54 = t53*t53;
    t55 = FbtAB(p3,p2,t3,p3,YY01);
    t57 = FbtAB(p3,p4,t3,p3,YY01);
    t58 = t57*t57;
    t59 = FbtAB(p5,p4,t3,p3,YY01);
    t60 = t59*t59;
    t63 = IFbtAB(p5,p5,t3,p3,YY01);
    t64 = IFbtAB(p6,p6,t3,p3,YY01);
    t65 = t63*t64;
    t66 = p4+p3;
    t67 = IFbtAA(p3,t66,p5,t3,p3,YY01);
    t68 = t67*I;
    t69 = ONE/YY01;
    t70 = t68*t69;
    t71 = t65*t70;
    t75 = A(p3,p5);
    t76 = t75*t75;
    t77 = t76*t76;
    t80 = IA(p3,p4);
    t81 = IA(p4,p5);
    t84 = t77*t31*t35*t80*t81*t36;
    t85 = t2*t44;
    t94 = t31*t35*t36*t39;
    t95 = t40*t44;
    t97 = t46*t2*S1;
    t101 = t55*t58;
    t102 = FbtAB2(p3,p3,t3,p3,YY01,YY21);
    t105 = YY01*YY01;
    t112 = t46*S1;
    t115 = t94*t95*t112*t54;
    t116 = FbtAB2(p5,p5,t3,p3,YY01,YY21);
    t119 = t63*t63;
    t125 = FbtAB2(p6,p6,t3,p3,YY01,YY21);
    t128 = t64*t64;
    t134 = FbtAA2(p3,t66,p5,t3,p3,YY01,YY21);
    t137 = t67*t67;
    t144 = FbtAB2(p5,p4,t3,p3,YY01,YY21);
    t151 = FbtAB2(p3,p4,t3,p3,YY01,YY21);
    t158 = t94*t95*t112;
    t160 = FbtAB2(p3,p2,t3,p3,YY01,YY21);
    t169 = FbtAB2(p1,p6,t3,p3,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t181 = FbtAB(p1,p6,t3,p3,YY0);
    t182 = t181*t181;
    t183 = FbtAB(p3,p2,t3,p3,YY0);
    t185 = FbtAB(p3,p4,t3,p3,YY0);
    t186 = t185*t185;
    t187 = FbtAB(p5,p4,t3,p3,YY0);
    t188 = t187*t187;
    t191 = IFbtAB(p5,p5,t3,p3,YY0);
    t192 = IFbtAB(p6,p6,t3,p3,YY0);
    t193 = t191*t192;
    t194 = IFbtAA(p3,t66,p5,t3,p3,YY0);
    t195 = t194*I;
    t196 = ONE/YY0;
    t197 = t195*t196;
    t198 = t193*t197;
    t211 = t183*t186;
    t212 = FbtAB2(p3,p3,t3,p3,YY0,YY2);
    t215 = YY0*YY0;
    t224 = t94*t95*t112*t182;
    t225 = FbtAB2(p5,p5,t3,p3,YY0,YY2);
    t228 = t191*t191;
    t234 = FbtAB2(p6,p6,t3,p3,YY0,YY2);
    t237 = t192*t192;
    t243 = FbtAA2(p3,t66,p5,t3,p3,YY0,YY2);
    t246 = t194*t194;
    t253 = FbtAB2(p5,p4,t3,p3,YY0,YY2);
    t260 = FbtAB2(p3,p4,t3,p3,YY0,YY2);
    t267 = FbtAB2(p3,p2,t3,p3,YY0,YY2);
    t276 = FbtAB2(p1,p6,t3,p3,YY0,YY2);
    ans = FOUR*EIGHT/THREE*t52*t54*t55*t58*t60*t71+FOUR*t84*t85*t54*t55*t64*t68-FOUR*t94*
t95*t97*t54*t101*t60*t102*t65*t68/t105-FOUR*t115*t101*t60*t116*t119*t64*t70-FOUR*
t115*t101*t60*t125*t63*t128*t70-FOUR*t115*t101*t60*t134*t65*t137*I*t69+EIGHT*t115*
t101*t59*t144*t71+EIGHT*t115*t55*t57*t60*t151*t71+FOUR*t158*t54*t58*t60*t160*t71+
EIGHT*t94*t95*t112*t53*t101*t60*t169*t71+FOUR*EIGHT/THREE*t52*t182*t183*t186*t188*t198+
FOUR*t84*t85*t182*t183*t192*t195-FOUR*t94*t95*t97*t182*t211*t188*t212*t193*t195/
t215-FOUR*t224*t211*t188*t225*t228*t192*t197-FOUR*t224*t211*t188*t234*t191*t237*
t197-FOUR*t224*t211*t188*t243*t193*t246*I*t196+EIGHT*t224*t211*t187*t253*t198+EIGHT*
t224*t183*t185*t188*t260*t198+FOUR*t158*t182*t186*t188*t267*t198+EIGHT*t94*t95*
t112*t181*t211*t188*t276*t198;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C2345mpmpmpt12(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  FourMomentum<DOUBLE> K1f;
  FourMomentum<DOUBLE> K2;
  std::complex<DOUBLE> S1;
  std::complex<DOUBLE> S2;
  std::complex<DOUBLE> YY0;
  std::complex<DOUBLE> YY01;
  std::complex<DOUBLE> YY2;
  std::complex<DOUBLE> YY21;
  std::complex<DOUBLE> alpha;
  std::complex<DOUBLE> ans;
  std::complex<DOUBLE> gg;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t109;
  std::complex<DOUBLE> t110;
  std::complex<DOUBLE> t113;
  std::complex<DOUBLE> t119;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t122;
  std::complex<DOUBLE> t128;
  std::complex<DOUBLE> t131;
  std::complex<DOUBLE> t139;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t145;
  std::complex<DOUBLE> t152;
  std::complex<DOUBLE> t159;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t171;
  std::complex<DOUBLE> t172;
  std::complex<DOUBLE> t174;
  std::complex<DOUBLE> t175;
  std::complex<DOUBLE> t177;
  std::complex<DOUBLE> t178;
  std::complex<DOUBLE> t183;
  std::complex<DOUBLE> t184;
  std::complex<DOUBLE> t185;
  std::complex<DOUBLE> t186;
  std::complex<DOUBLE> t187;
  std::complex<DOUBLE> t188;
  std::complex<DOUBLE> t190;
  std::complex<DOUBLE> t191;
  std::complex<DOUBLE> t192;
  std::complex<DOUBLE> t193;
  std::complex<DOUBLE> t194;
  std::complex<DOUBLE> t198;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t201;
  std::complex<DOUBLE> t208;
  std::complex<DOUBLE> t211;
  std::complex<DOUBLE> t217;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t220;
  std::complex<DOUBLE> t226;
  std::complex<DOUBLE> t229;
  std::complex<DOUBLE> t237;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t243;
  std::complex<DOUBLE> t250;
  std::complex<DOUBLE> t257;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  FourMomentum<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t50;
  FourMomentum<DOUBLE> t52;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t63;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t67;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  std::complex<DOUBLE> t82;
  std::complex<DOUBLE> t83;
  std::complex<DOUBLE> t84;
  std::complex<DOUBLE> t85;
  std::complex<DOUBLE> t86;
  std::complex<DOUBLE> t9;
  std::complex<DOUBLE> t90;
  std::complex<DOUBLE> t95;
  std::complex<DOUBLE> t96;
  std::complex<DOUBLE> t99;
  DOUBLE xsign;
  DOUBLE xsign1;
  {
    t2 = IAB(p3,p5+p4,p3);
    t3 = p5+p4+p3;
    t4 = S(t3);
    //K1f = p5+p4+p3-t2*t4*p3;
    K1f = p3+p4+p5-S(p3+p4+p5)/TWO/dot(p3,p4+p5)*p3;
    K2 = p2+p1;
    S1 = t4;
    gg = AB(p3,K1f,p3);
    S2 = S(K2);
    t7 = AB(K1f,K2,K1f);
    t8 = gg*t7;
    t9 = AB(p3,K2,p3);
    t10 = S1*t9;
    t12 = pow(t8+t10,TWO);
    t14 = gg*gg;
    alpha = t12-FOUR*S1*S2*t14;
    xsign1 = ONE;
    t17 = sqrt(alpha);
    t20 = IAB(p3,K2,K1f);
    t22 = ONE/S1;
    YY01 = (t8-t10+xsign1*t17)*t20*t22/TWO;
    t24 = AB(p3,K2,K1f);
    t27 = t17/alpha;
    YY21 = -xsign1*t24*t27;
    t29 = A(p3,p5);
    t30 = t29*t29;
    t31 = t30*t30;
    t32 = IA(p1,p2);
    t34 = IA(p3,p4);
    t36 = IA(p3,p6);
    t37 = IA(p4,p5);
    t39 = IB(p1,p2);
    t42 = t31*t32*t34*t36*t37*t39*t24;
    t43 = IAB(p6,K2,p6);
    t44 = t43*t43;
    t45 = t2*t44;
    t46 = FbtAB(p1,p2,t3,p3,YY01);
    t47 = t46*t46;
    t49 = FbtAB(p3,p6,t3,p3,YY01);
    t50 = IFbtAB(p1,p1,t3,p3,YY01);
    t52 = p4+p3;
    t53 = IFbtAA(p3,t52,p5,t3,p3,YY01);
    t54 = t53*I;
    t60 = IB(p3,p4);
    t62 = t32*t36*t39*t60;
    t63 = IB(p4,p5);
    t64 = t24*t24;
    t67 = t2*t2;
    t68 = t67*t67;
    t69 = t44*t44;
    t71 = S1*S1;
    t74 = t62*t63*t64*t24*t68*t69*t71;
    t75 = FbtAB(p3,p4,t3,p3,YY01);
    t76 = t75*t75;
    t77 = t47*t76;
    t78 = FbtAB(p5,p4,t3,p3,YY01);
    t79 = t78*t78;
    t80 = t79*t49;
    t82 = IFbtAB(p5,p5,t3,p3,YY01);
    t83 = t82*t50;
    t84 = ONE/YY01;
    t85 = t54*t84;
    t86 = t83*t85;
    t90 = t63*t24;
    t95 = t62*t90*t67*t2*t44*S1;
    t96 = FbtAB2(p3,p3,t3,p3,YY01,YY21);
    t99 = YY01*YY01;
    t109 = t62*t90*t67*t44*S1;
    t110 = FbtAB2(p1,p1,t3,p3,YY01,YY21);
    t113 = t50*t50;
    t119 = FbtAB2(p5,p5,t3,p3,YY01,YY21);
    t122 = t82*t82;
    t128 = FbtAA2(p3,t52,p5,t3,p3,YY01,YY21);
    t131 = t53*t53;
    t139 = FbtAB2(p5,p4,t3,p3,YY01,YY21);
    t145 = FbtAB2(p3,p6,t3,p3,YY01,YY21);
    t152 = FbtAB2(p3,p4,t3,p3,YY01,YY21);
    t159 = FbtAB2(p1,p2,t3,p3,YY01,YY21);
    xsign = -ONE;
    YY0 = (t8-t10+xsign*t17)*t20*t22/TWO;
    YY2 = -xsign*t24*t27;
    t171 = FbtAB(p1,p2,t3,p3,YY0);
    t172 = t171*t171;
    t174 = FbtAB(p3,p6,t3,p3,YY0);
    t175 = IFbtAB(p1,p1,t3,p3,YY0);
    t177 = IFbtAA(p3,t52,p5,t3,p3,YY0);
    t178 = t177*I;
    t183 = FbtAB(p3,p4,t3,p3,YY0);
    t184 = t183*t183;
    t185 = t172*t184;
    t186 = FbtAB(p5,p4,t3,p3,YY0);
    t187 = t186*t186;
    t188 = t187*t174;
    t190 = IFbtAB(p5,p5,t3,p3,YY0);
    t191 = t190*t175;
    t192 = ONE/YY0;
    t193 = t178*t192;
    t194 = t191*t193;
    t198 = FbtAB2(p3,p3,t3,p3,YY0,YY2);
    t201 = YY0*YY0;
    t208 = FbtAB2(p1,p1,t3,p3,YY0,YY2);
    t211 = t175*t175;
    t217 = FbtAB2(p5,p5,t3,p3,YY0,YY2);
    t220 = t190*t190;
    t226 = FbtAA2(p3,t52,p5,t3,p3,YY0,YY2);
    t229 = t177*t177;
    t237 = FbtAB2(p5,p4,t3,p3,YY0,YY2);
    t243 = FbtAB2(p3,p6,t3,p3,YY0,YY2);
    t250 = FbtAB2(p3,p4,t3,p3,YY0,YY2);
    t257 = FbtAB2(p1,p2,t3,p3,YY0,YY2);
    ans = -FOUR*t42*t45*t47*t49*t50*t54-FOUR*EIGHT/THREE*t74*t77*t80*t86+FOUR*t95*t77*t80
*t96*t83*t54/t99+FOUR*t109*t77*t80*t110*t113*t82*t85+FOUR*t109*t77*t80*t119*t50*
t122*t85+FOUR*t109*t77*t80*t128*t83*t131*I*t84-EIGHT*t109*t77*t49*t78*t139*t86-FOUR
*t109*t77*t79*t145*t86-EIGHT*t109*t47*t75*t80*t152*t86-EIGHT*t109*t46*t76*t80*t159*
t86-FOUR*t42*t45*t172*t174*t175*t178-FOUR*EIGHT/THREE*t74*t185*t188*t194+FOUR*t95*t185*
t188*t198*t191*t178/t201+FOUR*t109*t185*t188*t208*t211*t190*t193+FOUR*t109*t185*
t188*t217*t175*t220*t193+FOUR*t109*t185*t188*t226*t191*t229*I*t192-EIGHT*t109*t185
*t174*t186*t237*t194-FOUR*t109*t185*t187*t243*t194-EIGHT*t109*t172*t183*t188*t250*
t194-EIGHT*t109*t171*t184*t188*t257*t194;
    return(ans/TWO);
  }
}


std::complex<DOUBLE> C212mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C212mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C212mpmpmpt456(p1,p2,p3,p4,p5,p6);
    t4 = C212mpmpmpt1(p1,p2,p3,p4,p5,p6);
    t5 = C212mpmpmpt56(p1,p2,p3,p4,p5,p6);
    t6 = C212mpmpmpt6(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C223mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C223mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C223mpmpmpt1(p1,p2,p3,p4,p5,p6);
    t4 = C223mpmpmpt561(p1,p2,p3,p4,p5,p6);
    t5 = C223mpmpmpt2(p1,p2,p3,p4,p5,p6);
    t6 = C223mpmpmpt61(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C234mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C234mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C234mpmpmpt2(p1,p2,p3,p4,p5,p6);
    t4 = C234mpmpmpt612(p1,p2,p3,p4,p5,p6);
    t5 = C234mpmpmpt3(p1,p2,p3,p4,p5,p6);
    t6 = C234mpmpmpt12(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C245mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C245mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C245mpmpmpt4(p1,p2,p3,p4,p5,p6);
    t4 = C245mpmpmpt3(p1,p2,p3,p4,p5,p6);
    t5 = C245mpmpmpt123(p1,p2,p3,p4,p5,p6);
    t6 = C245mpmpmpt23(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C256mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C256mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C256mpmpmpt4(p1,p2,p3,p4,p5,p6);
    t4 = C256mpmpmpt5(p1,p2,p3,p4,p5,p6);
    t5 = C256mpmpmpt234(p1,p2,p3,p4,p5,p6);
    t6 = C256mpmpmpt34(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C261mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C261mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C261mpmpmpt45(p1,p2,p3,p4,p5,p6);
    t4 = C261mpmpmpt5(p1,p2,p3,p4,p5,p6);
    t5 = C261mpmpmpt6(p1,p2,p3,p4,p5,p6);
    t6 = C261mpmpmpt345(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C2123mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C2123mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C2123mpmpmpt1(p1,p2,p3,p4,p5,p6);
    t4 = C2123mpmpmpt56(p1,p2,p3,p4,p5,p6);
    t5 = C2123mpmpmpt6(p1,p2,p3,p4,p5,p6);
    t6 = C2123mpmpmpt12(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C2234mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C2234mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C2234mpmpmpt1(p1,p2,p3,p4,p5,p6);
    t4 = C2234mpmpmpt2(p1,p2,p3,p4,p5,p6);
    t5 = C2234mpmpmpt61(p1,p2,p3,p4,p5,p6);
    t6 = C2234mpmpmpt23(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> C2345mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  {
    t1 = C2345mpmpmpbub(p1,p2,p3,p4,p5,p6);
    t3 = C2345mpmpmpt2(p1,p2,p3,p4,p5,p6);
    t4 = C2345mpmpmpt3(p1,p2,p3,p4,p5,p6);
    t5 = C2345mpmpmpt12(p1,p2,p3,p4,p5,p6);
    t6 = C2345mpmpmpt34(p1,p2,p3,p4,p5,p6);
    return(-t1*I-t3-t4-t5-t6);
  }
}


std::complex<DOUBLE> R6mpmpmp(FourMomentum<DOUBLE> p1,FourMomentum<DOUBLE> p2,FourMomentum<DOUBLE> p3,FourMomentum<DOUBLE> p4,FourMomentum<DOUBLE> p5,
FourMomentum<DOUBLE> p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t71;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t73;
  std::complex<DOUBLE> t74;
  std::complex<DOUBLE> t75;
  std::complex<DOUBLE> t76;
  std::complex<DOUBLE> t77;
  std::complex<DOUBLE> t78;
  std::complex<DOUBLE> t79;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t80;
  {
    t1 = S(p1+p2+p3);
    t2 = C2123mpmpmp(p1,p2,p3,p4,p5,p6);
    t4 = S(p1+p6);
    t5 = C261mpmpmp(p1,p2,p3,p4,p5,p6);
    t7 = S(p2+p3);
    t8 = C223mpmpmp(p1,p2,p3,p4,p5,p6);
    t10 = S(p2+p3+p4);
    t11 = C2234mpmpmp(p1,p2,p3,p4,p5,p6);
    t13 = S(p3+p4);
    t14 = C234mpmpmp(p1,p2,p3,p4,p5,p6);
    t16 = S(p3+p4+p5);
    t17 = C2345mpmpmp(p1,p2,p3,p4,p5,p6);
    t19 = S(p4+p5);
    t20 = C245mpmpmp(p1,p2,p3,p4,p5,p6);
    t22 = S(p5+p6);
    t23 = C256mpmpmp(p1,p2,p3,p4,p5,p6);
    t25 = S(p1+p2);
    t26 = C212mpmpmp(p1,p2,p3,p4,p5,p6);
    t28 = C4123mpmpmp(p1,p2,p3,p4,p5,p6);
    t29 = C4234mpmpmp(p1,p2,p3,p4,p5,p6);
    t30 = -t1*t2-t4*t5-t7*t8-t10*t11-t13*t14-t16*t17-t19*t20-t22*t23-t25*t26-
t28-t29;
    t32 = C4345mpmpmp(p1,p2,p3,p4,p5,p6);
    t33 = C4456mpmpmp(p1,p2,p3,p4,p5,p6);
    t34 = C4561mpmpmp(p1,p2,p3,p4,p5,p6);
    t35 = C4612mpmpmp(p1,p2,p3,p4,p5,p6);
    t36 = C412x34mpmpmp(p1,p2,p3,p4,p5,p6);
    t37 = C423x45mpmpmp(p1,p2,p3,p4,p5,p6);
    t38 = C434x56mpmpmp(p1,p2,p3,p4,p5,p6);
    t39 = C445x61mpmpmp(p1,p2,p3,p4,p5,p6);
    t40 = C456x12mpmpmp(p1,p2,p3,p4,p5,p6);
    t41 = C461x23mpmpmp(p1,p2,p3,p4,p5,p6);
    t42 = C412x45mpmpmp(p1,p2,p3,p4,p5,p6);
    t43 = -t32-t33-t34-t35-t36-t37-t38-t39-t40-t41-t42;
    t46 = C423x56mpmpmp(p1,p2,p3,p4,p5,p6);
    t48 = C434x61mpmpmp(p1,p2,p3,p4,p5,p6);
    t50 = C31234mpmpmp(p1,p2,p3,p4,p5,p6);
    t52 = C32345mpmpmp(p1,p2,p3,p4,p5,p6);
    t54 = C33456mpmpmp(p1,p2,p3,p4,p5,p6);
    t56 = C34561mpmpmp(p1,p2,p3,p4,p5,p6);
    t58 = C35612mpmpmp(p1,p2,p3,p4,p5,p6);
    t60 = C36123mpmpmp(p1,p2,p3,p4,p5,p6);
    t62 = C312x345mpmpmp(p1,p2,p3,p4,p5,p6);
    t64 = C323x456mpmpmp(p1,p2,p3,p4,p5,p6);
    t66 = C334x561mpmpmp(p1,p2,p3,p4,p5,p6);
    t68 = -t46/SIX-t48/SIX-t50/TWO-t52/TWO-t54/TWO-t56/TWO-t58/TWO-t60/TWO-t62/
TWO-t64/TWO-t66/TWO;
    t69 = C345x612mpmpmp(p1,p2,p3,p4,p5,p6);
    t70 = C356x123mpmpmp(p1,p2,p3,p4,p5,p6);
    t71 = C361x234mpmpmp(p1,p2,p3,p4,p5,p6);
    t72 = C3123x45mpmpmp(p1,p2,p3,p4,p5,p6);
    t73 = C3234x56mpmpmp(p1,p2,p3,p4,p5,p6);
    t74 = C3345x61mpmpmp(p1,p2,p3,p4,p5,p6);
    t75 = C3456x12mpmpmp(p1,p2,p3,p4,p5,p6);
    t76 = C3561x23mpmpmp(p1,p2,p3,p4,p5,p6);
    t77 = C3612x34mpmpmp(p1,p2,p3,p4,p5,p6);
    t78 = C312x34x56mpmpmp(p1,p2,p3,p4,p5,p6);
    t79 = C323x45x61mpmpmp(p1,p2,p3,p4,p5,p6);
    t80 = -t69-t70-t71-t72-t73-t74-t75-t76-t77-t78-t79;
    return(t30/SIX+t43/SIX+t68+t80/TWO);
  }
}
